// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "DefenceBreaker.h"
#include "Basics.h"

#include "body_hold_ball2008.h"

#include <rcsc/player/interception.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include <math.h>

#define maxX 52.5
#define maxY 34
#define maxDC 5
#define RATESTEPCOUNT 10

using namespace std;
using namespace rcsc;

int DefenceBreaker::finalCycle = -1000;

DefenceBreaker::DefenceBreaker()
{
	mDelay = 2.f;
	players = "";
	tmpOpenPlayers = "";
	maxPossibleDelay = 8.f;
	oppKickable = 3.f;
	oppCatchable = 3.6;
	for ( int i =0 ; i < 11; i++)
		playerDelay[i] = 0;
	maxDistanceToOffsideLine = 15;
	isPassSetThisCycle = false;
	isPassPossible = false;
}

DefenceBreaker::~DefenceBreaker()
{
}

void DefenceBreaker::reset()
{
	finalCycle = -1000;
}

void DefenceBreaker::setMaxDistanceToOffsideLine(float max)
{
	maxDistanceToOffsideLine = max;
}

void DefenceBreaker::setMaxPossibleDelay(rcsc::PlayerAgent *agent)
{
	const WorldModel & wm = agent->world();
	maxPossibleDelay = 8.f;
	Interception intercept(wm.ball().pos() , 0 , 0);
	// Calculating Opponents Min Reach Cycle
	float minReachCycle = 1000;
	const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
		it != end;
		++it )
	{
		Vector2D hisPos = (*it)->pos();
		float hisReachCycle = (float)intercept.getReachCycle( hisPos , NULL , NULL , (*it)->posCount() , 1 , 1);
//		dlog.addText( Logger::PASS , "player %d at ( %f , %f ) -> Reaching in %f " , (*it)->unum() , hisPos.x, hisPos.y , hisReachCycle);
		if ( hisReachCycle < minReachCycle )
			minReachCycle = hisReachCycle;
	}
	minReachCycle += 2;
	maxPossibleDelay = min( 8.f , minReachCycle);
	maxPossibleDelay = 8.f;
}

float DefenceBreaker::getMaxDelay(rcsc::PlayerAgent *agent)
{
	setMaxPossibleDelay(agent);
	const WorldModel & wm = agent->world();
	float min = maxPossibleDelay + 1.f;
//	dlog.addText ( Logger::PASS , "Maxdelay	=	%f" , maxPossibleDelay );
//	dlog.addText ( Logger::PASS , "OffsideLine	=	%f" , wm.offsideLineX() );
	const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
		it != end;
		++it )
	{
		int hisUnum = (*it)->unum();
		if ( hisUnum > 0 )
			if ( Basics::isInPlayers( hisUnum , players ) )
			{
				Vector2D hisPos = (*it)->pos();
				Vector2D ballPos = hisPos;
				ballPos.x = wm.offsideLineX();
				const Vector2D * hisVel = NULL;
				const AngleDeg * hisAngle = NULL;
				if ( (*it)->velValid() )
					hisVel = & ( (*it)->vel() );
				if ( (*it)->faceValid() )
					hisAngle = & ( (*it)->body() );			
				Interception intercept( ballPos , 0.f , 0.f );
				float hisReachCycle = (float)(intercept.getReachCycle( hisPos , hisVel , hisAngle , (*it)->posCount() , 0.9 , 1 ));
//				dlog.addText ( Logger::PASS , "Number	%d" , hisUnum );
//				dlog.addText ( Logger::PASS , "hisReachCycle	=	%f" , hisReachCycle );
				hisReachCycle = fabsf(hisReachCycle);
				playerDelay[hisUnum - 1] = roundf(hisReachCycle) + 1.f;
				if ( playerDelay[hisUnum - 1] == 0 )
					playerDelay[hisUnum - 1]++;
				if ( hisReachCycle < min )
					min = hisReachCycle;
				if ( hisReachCycle <= maxPossibleDelay )
				{
					if ( hisUnum <= 9 )
						tmpOpenPlayers += (char)( hisUnum + 48 );
					if ( hisUnum == 10 )
						tmpOpenPlayers += "A";
					if ( hisUnum == 11 )
						tmpOpenPlayers += "B";
				}
			}
	}
	return min;
}

void DefenceBreaker::setReceiverPlayers(string openPlayers)
{
	players = openPlayers;
}

void DefenceBreaker::setOppControlBuff(float kickable, float catchable)
{
	oppKickable = kickable;
	oppCatchable = catchable;
}

bool DefenceBreaker::isAvailable(PlayerAgent *agent)
{
//	dlog.addText ( Logger::PASS , "DFB::ISAvailable()");
//	dlog.flush();
	const WorldModel & wm = agent->world();
//	dlog.addText ( Logger::PASS , "Ball ( %f , %f )" , wm.ball().pos().x , wm.ball().pos().y );
//	float currentDelay = getMaxDelay(agent);
	Vector2D myPos = wm.self().pos();
//	if ( currentDelay > maxPossibleDelay )
//		return false;
	float oppBuffs[11];
	for ( int i = 0; i < 11; i++ )
		oppBuffs[i] = oppKickable;
	float playersOffsideLine = 0;
	{
		int oppGoalieNum = wm.opponentGoalieUnum();
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int unum = (*it)->unum();
			if ( ( hisPos.x > playersOffsideLine ) and ( unum > 0 ) and ( unum != oppGoalieNum) )
				playersOffsideLine = hisPos.x;
			if ( unum > 0 )
			{
				float hisDist = myPos.dist(hisPos);
				if ( hisDist <= 2.f * maxDistanceToOffsideLine / 3.f )
					oppBuffs[unum - 1] = oppKickable / 2.f;
				else if ( hisDist <= 4.f * maxDistanceToOffsideLine / 3.f )
					oppBuffs[unum - 1] = oppKickable;
				else if ( hisDist <= 2 * maxDistanceToOffsideLine )
					oppBuffs[unum - 1] = 3.f * oppKickable / 2.f;
				else
					oppBuffs[unum - 1] = 2.f * oppKickable;
//				dlog.addText ( Logger::PASS , "player %d has %f buff" , unum , oppBuffs[unum - 1] );
			}
		}
	}
	if ( myPos.x >= playersOffsideLine )
	{
//		dlog.addText ( Logger::PASS , "I am In!!" );
		return false;
	}
	playersOffsideLine = wm.offsideLineX();
	float foreLine = min( 36.f , playersOffsideLine + 20 );
	if ( playersOffsideLine - myPos.x >= maxDistanceToOffsideLine )
	{
//		dlog.addText ( Logger::PASS , "Too Far!!" );
		return false;
	}
	if ( playersOffsideLine >= 36.f )
	{
//		dlog.addText ( Logger::PASS , "They are In!!" );
		return false;
	}
	const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
		it != end;
		++it )
	{
		int hisUnum = (*it)->unum();
		if ( hisUnum > 0 )
			if ( Basics::isInPlayers( hisUnum , players ) )
			{
				Vector2D hisPos = (*it)->pos();
				if ( hisPos.x > playersOffsideLine )
					continue;
				if ( playersOffsideLine - hisPos.x <= maxDistanceToOffsideLine / 3.f )
				{
					if ( hisUnum <= 9 )
						tmpOpenPlayers += (char)( hisUnum + 48 );
					if ( hisUnum == 10 )
						tmpOpenPlayers += "A";
					if ( hisUnum == 11 )
						tmpOpenPlayers += "B";
				}
			}
	}

//	dlog.addText ( Logger::PASS , "tmpOpenPlayers	'%s'" , tmpOpenPlayers.c_str() );
//	dlog.flush();
	if ( tmpOpenPlayers == "" )
		return false;
	dfbPass.reconstruct();
	dfbPass.setForwardMovementParameters();
	dfbPass.setReceiverYChangeParameters( 1 , 10 , true);
	dfbPass.setOpenPlayers(tmpOpenPlayers);
	dfbPass.setSafetyFactors( 0.9 , 1.2 , 0.85 , 0 , 10 , oppKickable , oppCatchable , 0 , 0 , 30 );
	dfbPass.setDelayFactors( 0, 1 , 1 , min ( 5.f , (float) ( playersOffsideLine - myPos.x ) ) );
	for ( int i = 0; i < 11; i++)
		dfbPass.setOppPlayerControlBuff ( i + 1 , oppBuffs[i] );
	dfbPass.setMaxDirectionCount(maxDC);
	dfbPass.setInsidePolygonPass( true );
	dfbPass.targetPolygon().addVertex ( Vector2D ( playersOffsideLine , -30 ) );
	dfbPass.targetPolygon().addVertex ( Vector2D ( playersOffsideLine , 30 ) );
	dfbPass.targetPolygon().addVertex ( Vector2D ( foreLine , 30 ) );
	dfbPass.targetPolygon().addVertex ( Vector2D ( foreLine , -30 ) );
	float value = dfbPass.getValue(agent , false , true , -90 , 90 , true , RATESTEPCOUNT);
	isPassSetThisCycle = true;
	if ( value > 0.f )
	{
		isPassPossible = true;
		finalCycle = floor( wm.time().cycle() + maxPossibleDelay);
		return true;
	}
//	dlog.addText ( Logger::PASS , "No PASS!!" );
	isPassPossible = false;
	return false;
}

void DefenceBreaker::execute(PlayerAgent *agent)
{
	if ( shouldPass ( agent ) )
		doKickBall(agent);
	else if ( shouldHoldBall ( agent ) )
		doHoldBall(agent);
}

bool DefenceBreaker::shouldHoldBall(PlayerAgent *agent)
{
	if ( agent->world().time().cycle() <= finalCycle )
		return true;
	return false;
}

bool DefenceBreaker::shouldPass(PlayerAgent *agent)
{
//	dlog.addText ( Logger::PASS , "DFB::ShouldPass()");
//	dlog.flush();
	if ( isPassSetThisCycle )
		return isPassPossible;
	const WorldModel & wm = agent->world();
	float oppBuffs[11];
	Vector2D myPos = wm.self().pos();
//		dlog.addText ( Logger::PASS , "NO PASS BEFORE!!" );
	for ( int i = 0; i < 11; i++ )
		oppBuffs[i] = oppKickable;
	float playersOffsideLine = 0;
	{
		int oppGoalieNum = -1;
		if ( wm.getOpponentGoalie() )
			oppGoalieNum = wm.getOpponentGoalie()->unum();
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int unum = (*it)->unum();
			if ( ( hisPos.x > playersOffsideLine ) and ( unum > 0 ) and ( unum != oppGoalieNum) )
				playersOffsideLine = hisPos.x;
			if ( unum > 0 )
			{
				float hisDist = myPos.dist(hisPos);
				if ( hisDist <= 2.f * maxDistanceToOffsideLine / 3.f )
					oppBuffs[unum - 1] = oppKickable / 2.f;
				else if ( hisDist <= 4.f * maxDistanceToOffsideLine / 3.f )
					oppBuffs[unum - 1] = oppKickable;
				else if ( hisDist <= 2 * maxDistanceToOffsideLine )
					oppBuffs[unum - 1] = 3.f * oppKickable / 2.f;
				else
					oppBuffs[unum - 1] = 2.f * oppKickable;
//					dlog.addText ( Logger::PASS , "Opp %d with Buff %f" , unum , oppBuffs[unum - 1] );
			}
		}
	}
	if ( wm.self().pos().x >= playersOffsideLine )
	{
		reset();
		return false;
	}
	playersOffsideLine = wm.offsideLineX();
	float foreLine = min( 36.f , playersOffsideLine + 20 );
	string currentlyReceiverPlayer;
	const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
		it != end;
		++it )
		{
			int hisUnum = (*it)->unum();
			if ( hisUnum > 0 )
			if ( Basics::isInPlayers( hisUnum , players ) )
			{
				Vector2D hisPos = (*it)->pos();
				if ( hisPos.x > playersOffsideLine )
					continue;
				if ( playersOffsideLine - hisPos.x <= maxDistanceToOffsideLine / 3.f )
				{
					if ( hisUnum <= 9 )
						currentlyReceiverPlayer += (char)( hisUnum + 48 );
					if ( hisUnum == 10 )
						currentlyReceiverPlayer += "A";
					if ( hisUnum == 11 )
						currentlyReceiverPlayer += "B";
				}
			}
		}
	if ( currentlyReceiverPlayer.length() == 0 )
		return false;
//	dlog.addText ( Logger::PASS , "CurrentlyReceivingPlayers	'%s'" , currentlyReceiverPlayer.c_str() );
//	dlog.flush();
	dfbPass.reconstruct();
	dfbPass.setForwardMovementParameters();
	dfbPass.setReceiverYChangeParameters( 1 , 10 , true);
	dfbPass.setOpenPlayers(currentlyReceiverPlayer);
	dfbPass.setSafetyFactors( 0.9 , 1.2 , 0.85 , 0 , 10 , oppKickable , oppCatchable , 0 , 0 , 30 );
	for ( int i = 0; i < 11; i++)
		dfbPass.setOppPlayerControlBuff ( i + 1 , oppBuffs[i] );
	dfbPass.setDelayFactors ( 0 , 1 , 1 , min( 5.f , (float) (playersOffsideLine - myPos.x ) ) );
	dfbPass.setMaxDirectionCount(maxDC);
	dfbPass.setInsidePolygonPass( true );
	dfbPass.targetPolygon().addVertex ( Vector2D ( playersOffsideLine , -30 ) );
	dfbPass.targetPolygon().addVertex ( Vector2D ( playersOffsideLine , 30 ) );
	dfbPass.targetPolygon().addVertex ( Vector2D ( foreLine , 30 ) );
	dfbPass.targetPolygon().addVertex ( Vector2D ( foreLine , -30 ) );
	float value = dfbPass.getValue(agent , false , true , -90 , 90 , true , RATESTEPCOUNT);
	if ( value > 0.f )
		return true;
//	dlog.addText ( Logger::PASS , "No PASS!!!!!!!!!!!!" );
	return false;
}

void DefenceBreaker::doHoldBall(PlayerAgent *agent)
{
	Body_HoldBall2008( true , Vector2D( 52.5 , 0.0 ) , Vector2D( 52.5 , 0.0 ) ).execute(agent);
	agent->addSayMessage(new  WaitRequestMessage() );
//	cerr << "RUN!!!" << endl;
}

void DefenceBreaker::doKickBall(PlayerAgent *agent)
{
	dlog.addText ( Logger::PASS , "DFB::Ball Passed!!" );
//	cerr << "PASSED!!" << endl;
	dfbPass.execute( agent , true , 1);
}

bool DefenceBreaker::areWeInTheMiddle(PlayerAgent *agent)
{
	if ( agent->world().time().cycle() <= finalCycle )
		return true;
	return false;
}
