// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "GlobalPositioning2.h"
#include <Share.h>

#include <rcsc/common/logger.h>

using namespace rcsc;
using namespace std;

int GlobalPositioning2::isSet = 0;
PosConf GlobalPositioning2::posConf;
int GlobalPositioning2::lastCycle = -1;
Vector2D GlobalPositioning2::lastPos = Vector2D::INVALIDATED;

GlobalPositioning2::GlobalPositioning2(string players, unsigned style): M_LineDistance(10)
{
	rcsc::dlog.addText( Logger::PASS,
   	  		                "OffensePositioning");

	if (!isSet || posConf.style != isSet)
	{
		posConf = PosConf(CONFIG, style);
		isSet = true;
	}
	setPlayers(players);
}

GlobalPositioning2::GlobalPositioning2(string main, string back, unsigned style): mainPlayers(main), backPlayers(back), allPlayers(main + back), M_LineDistance(10)
{
	rcsc::dlog.addText( Logger::PASS,
   	  		                "OffensePositioning constructor 2");

	if (!isSet || posConf.style != isSet)
	{
		posConf = PosConf(CONFIG, style);
		isSet = true;
	}
}

void GlobalPositioning2::setLineDistance(double linedist)
{
	M_LineDistance = linedist;
}

Vector2D
GlobalPositioning2::getHomePos(const WorldModel *wm)
{
	Vector2D HomePos = getPlayerPos(wm, wm->self().unum());;
	if (wm->self().pos().x > wm->offsideLineX() - 2.5)
	    HomePos.y = wm->self().pos().y;

	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "I'm going to Vector2D(%f, %f)",HomePos.x, HomePos.y);
	dlog.addText( Logger::DRIBBLE, "last pos = (%f, %f)", lastPos.x, lastPos.y);

	return HomePos;
}

Vector2D
GlobalPositioning2::getPlayerPos(const WorldModel *wm, int unum)
{
	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "My UniNum is %d and I want find player %d Pos",wm->self().unum(),unum);

	findPlayerPos(wm, unum);
	Vector2D PlayerPos = Vector2D(SortedPos[unum].x, SortedPos[unum].y);
	return PlayerPos;
}

Rect2D
GlobalPositioning2::getHomeArea(const WorldModel *wm)
{
	return getPlayerArea(wm, wm->self().unum());
}

Rect2D
GlobalPositioning2::getPlayerArea(const WorldModel *wm, int unum)
{
	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "My UniNum is %d and I want find player %d Area",wm->self().unum(),unum);

	findAllAreas(wm);

	return SortedAreas[unum];
}

void 
GlobalPositioning2::determinePosLines(const WorldModel *wm)
{
	//rcsc::dlog.addText( Logger::PASS,
   	//  		                "determinePosLines lastCycle = %d now = %d", lastCycle, wm->time().cycle());

	if (lastCycle < wm->time().cycle())
	{
		lastCycle = wm->time().cycle();
		Vector2D ballPos = receivePos;
		double offsideLine = wm->offsideLineX();
		dlog.addText(Logger::DRIBBLE, "Offside Line X = %f, receivePos = (%f, %f)", offsideLine, ballPos.x, ballPos.y);
		double Middle, PlayerSize, Size, Extra = 0.0;
		//Finding main line
		PosConf::Weight mainweight = posConf.getMainWeight(ballPos.x);
		mainLine.x = ((mainweight.OffsideWeight * offsideLine) + (mainweight.BallXWeight * ballPos.x)) / mainweight.Sum();
		if (wm->interceptTable()->isOurTeamBallPossessor())
		    mainLine.x += 5.0;
		dlog.addText(Logger::DRIBBLE, "Our Team Ball Possessor %d", wm->interceptTable()->isOurTeamBallPossessor());
		mainLine.x = min(mainLine.x, offsideLine - 2.5);
		mainLine.x = min(mainLine.x, 46.0);
		mainLine.x = max(mainLine.x, -10.0);
		dlog.addText(Logger::PASS, "main line = %f", mainLine.x);
		const PlayerObject * goalie = wm->getOpponentGoalie();
		Vector2D goaliePos(0, 0);
		if (goalie)
			goaliePos = goalie->pos();
		Middle = ((posConf.getWeight(posConf.Mid_BallYWeight, ballPos.x) * ballPos.y) + (posConf.getWeight(posConf.Mid_GoalieYWeight, ballPos.x) * goaliePos.y));
		dlog.addText(Logger::DRIBBLE, "Middle = %f", Middle);

		PlayerSize = (25 - (posConf.getWeight(posConf.Size_MainXWeight, ballPos.x) * mainLine.x)) * (1 - posConf.getWeight(posConf.Size_BallAbsYWeight, ballPos.x) * ballPos.absY());
		dlog.addText(Logger::DRIBBLE, "PlayerSize = %f", PlayerSize);
		Size = ((getMainCount() - 1) * PlayerSize * (1 - getMainCount() * 0.05));
		Size = max(Size, getMainCount() * 9.0);
		Size = min(Size, 60.0);
		dlog.addText(Logger::DRIBBLE, "Size = %f", Size);
		mainLine.minY = Middle - Size / 2.f;
		if (mainLine.minY < -30)
		{
			Extra = -mainLine.minY - 30;
			mainLine.minY = -30;
		}
		mainLine.maxY = Middle + Size / 2.f + Extra;
		if (mainLine.maxY > 30)
		{
			mainLine.minY -= mainLine.maxY - 30;
			mainLine.maxY = 30;
		}
		dlog.addText(Logger::PASS, "main line max Y = %f min Y = %f", mainLine.maxY, mainLine.minY);
		//Finding back line
		if (getBackCount())
		{
			PlayerSize = (20 - (posConf.getWeight(posConf.Size_MainXWeight, ballPos.x) * backLine.x)) * (1 - posConf.getWeight(posConf.Size_BallAbsYWeight, ballPos.x) * ballPos.absY());
			PosConf::Weight backweight = posConf.getBackWeight(ballPos.x);
			backLine.x = ((backweight.OffsideWeight * offsideLine) + (backweight.BallXWeight * ballPos.x)) / backweight.Sum();
			backLine.x = min(backLine.x, offsideLine - 10.0);
			backLine.x = min(backLine.x, 35.0);
			backLine.x = min(backLine.x, mainLine.x - 10);
			backLine.x = max(backLine.x, 0.0);
			dlog.addText(Logger::PASS, "back line = %f", backLine.x);
		
			Size = ((getBackCount() - 1) * (PlayerSize * 1.2) * (1 - getBackCount() * 0.05));
			Size = max(Size, getMainCount() * 11.0);
			Size = min(Size, getMainCount() * 17.0);
		
			backLine.minY = Middle / 2.f - Size / 2.f;	
			if (backLine.minY < -30)
			{
				Extra = -backLine.minY - 30;
				backLine.minY = -30;
			}
			backLine.maxY = Middle / 2.f + Size / 2.f + Extra;
			if (backLine.maxY > 30)
			{
				backLine.minY -= backLine.maxY - 30;
				backLine.maxY = 30;
			}
			dlog.addText(Logger::PASS, "back line max Y = %f min Y = %f", backLine.maxY, backLine.minY);
		}
	}		
}

void
GlobalPositioning2::findAllAreas(const rcsc::WorldModel *wm)
{
	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "findAllAreas");

	detectReceiveInfo(wm);
	determinePosLines(wm);
	double mainX, backX = 0.0;
	Vector2D ballpos(receivePos);
	double N1, N2 = 0.0, n1, n2 = 0.0;
	if(getMainCount() == 4)
	{
		N1 = (abs(mainLine.maxY) + abs(mainLine.minY)) / 3.5;
		n1 = (abs(mainLine.maxY) + abs(mainLine.minY)) / 4.f;
	}
	else
	{
		N1 = (abs(mainLine.maxY) + abs(mainLine.minY)) / 2.8;
		n1 = (abs(mainLine.maxY) + abs(mainLine.minY)) / 3.f;
	}

	if(getBackCount() != 0 )
	{
		double k;
		mainX = (mainLine.x - backLine.x) / 1.8;
		backX = (mainLine.x - backLine.x) / 1.8;
		
		n2 = (abs(backLine.maxY) + abs(backLine.minY)) / getBackCount();
		k = backPlayers.length() * (4.f / 5.f);
		N2 = (abs(backLine.maxY) + abs(backLine.minY)) / k;
	}
	else
		mainX = 8 - abs(ballpos.x / 6.f );

	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "N1 = %f, n1 = %f, mainX = %f"
       	      		        ,N1, n1, mainX);

	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "getMainCount = %d"
       	      		        , getMainCount());

	for(unsigned i = 0; i < getMainCount(); i++)
	{
		rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "->finding Area num %d (Areas num start from 0 to MainCount)"
       	      		        ,i );

		Vector2D left_top, right_bottom;
		double upY,downY;
		Rect2D baghali;
//		upY = mainLine.maxY - N1 * i;
//		downY = upY - n1;

		upY = mainLine.maxY - (i * n1);
//		if( getMainCount() == 4 and i == 2)
//		    upY += n / 3;
		downY = upY - n1;

		rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "---->downY = %f and mainLine.x = %f"
           	      		        , downY, mainLine.x);

//		downY -= i * (n1 - N1);

		rcsc::dlog.addText( Logger::DRIBBLE,
							"---->upY = %f and downY = %f"
							, upY, downY);

		if( getMainCount() == 4 and i == 1 and receivePos.y > (mainLine.maxY + mainLine.minY) / 2 )
		{
		    rcsc::dlog.addText( Logger::DRIBBLE,
   		  	                "yeki aghab vaise 1");
		    left_top = Vector2D(mainLine.x - mainX * 2, upY);
		    right_bottom = Vector2D(mainLine.x , downY);
		}
		else if( getMainCount() == 4 and i == 2 and receivePos.y <  (mainLine.maxY + mainLine.minY) / 2  )
		{
		    rcsc::dlog.addText( Logger::DRIBBLE,
   		  	                "yeki aghab vaise 2");
		    left_top = Vector2D(mainLine.x - mainX * 2, upY);
		    right_bottom = Vector2D(mainLine.x , downY);
		}
		else
		{
		    left_top = Vector2D(mainLine.x - mainX, upY);
		    right_bottom = Vector2D(mainLine.x + mainX, downY);
		}
		if( getMainCount() == 4 and (i == 1 or i == 2) and wm->ball().pos().x > 43)
		{
		    left_top = Vector2D(left_top.x - 3, left_top.y);
		    right_bottom = Vector2D(right_bottom.x - 3, right_bottom.y);
		}

		baghali = Rect2D(left_top, right_bottom);

		Areas.push_back(Rect2D(left_top, right_bottom));
		rcsc::dlog.addText( Logger::DRIBBLE,
							"---->left_top point is %f, %f and right_bottom is %f, %f"
							, left_top.x, left_top.y, right_bottom.x, right_bottom.y);
		rcsc::dlog.addText( Logger::DRIBBLE,
							"---->mese adam az to vector left_top point is %f, %f and right_bottom is %f, %f"
							, baghali.minX(), baghali.minY(), baghali.maxX(), baghali.maxY());
	}

	if(getBackCount() != 0)
	{
		rcsc::dlog.addText( Logger::DRIBBLE,
    	  		                "getBackCount != 0 and getBackCount() = %d"
	   	      		        , getBackCount());
		rcsc::dlog.addText( Logger::DRIBBLE,
    	  		                "backLine.x = %f       and     backLine.maxY = %f       and     backLine.minY = %f"
	   	      		        , backLine.x, backLine.maxY, backLine.minY);

		for (unsigned j = 0; j < getBackCount(); j++)
		{
			rcsc::dlog.addText( Logger::DRIBBLE,
   		  		                "->finding Area num %d (Areas num start from 0 to BackCount)"
       	    				        ,j );

	        	Vector2D left_top, right_bottom;
	    		double upY,downY;
    			Rect2D baghali;
//			upY = mainLine.maxY - N1 * i;
//			downY = upY - n1;

			upY = backLine.maxY - (j * n2);
			downY = upY - n2;

			rcsc::dlog.addText( Logger::DRIBBLE,
   		  		                "---->downY = %f and backLine.x = %f"
           		      		        , downY, backLine.x);

//			downY -= i * (n1 - N1);

			rcsc::dlog.addText( Logger::DRIBBLE,
								"---->upY = %f and downY = %f"
								, upY, downY);

			left_top = Vector2D(backLine.x - backX, upY);
			right_bottom = Vector2D(backLine.x + backX, downY);

			baghali = Rect2D(left_top, right_bottom);

			Areas.push_back(Rect2D(left_top, right_bottom));

			rcsc::dlog.addText( Logger::DRIBBLE,
								"---->upY - downY = %f"
								, upY - downY);
			rcsc::dlog.addText( Logger::DRIBBLE,
								"---->left_top point is %f, %f and right_bottom is %f, %f"
								, left_top.x, left_top.y, right_bottom.x, right_bottom.y);
			rcsc::dlog.addText( Logger::DRIBBLE,
								"---->mese adam az to vector left_top point is %f, %f and right_bottom is %f, %f"
								, baghali.minX(), baghali.minY(), baghali.maxX(), baghali.maxY());

		}
	}

	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "Areas Found ***"
       	      		        ,__FILE__, __LINE__ );

	setAreasToPlayers(wm);
}

void
GlobalPositioning2::setAreasToPlayers(const WorldModel *wm)
{
    dlog.addText(Logger::DRIBBLE , "flag-2");
    vector <Vector2D> hisPosMain;
    Vector2D pos;
    int i = 0;
    vector <int> hisUNumMain;
    bool isAreaDeleted;
    dlog.addText(Logger::DRIBBLE, "flag-1");
    vector <Rect2D>::iterator rend = Areas.end();
    dlog.addText(Logger::DRIBBLE, "FLAG0");
    if (isInPlayers(receiver, allPlayers))
	for (vector <Rect2D>::iterator it = Areas.begin();it != rend;it++)
	{
	    dlog.addText(Logger::DRIBBLE, "receiver = %d", receiver);
	    if ((*it).contains(receivePos))
	    {
		dlog.addText(Logger::DRIBBLE, "FLAG2");
		vector <Rect2D>::iterator next = it;
		next++;
		if (next != Areas.end() && (*next).contains(receivePos))
		    if ((*it).center().dist(receivePos) > (*next).center().dist(receivePos))
		    {
			Areas.erase(next);
			isAreaDeleted = true;
			break;
		    }
		    else
		    {
			Areas.erase(it);
			isAreaDeleted = true;
			break;
		    }
		else
		{	
		    Areas.erase(it);
		    isAreaDeleted = true;
		    break;
		}
	    }
	}

    for ( unsigned q = 0;
	    q < getMainCount();
	    q++ )
    {
	dlog.addText(Logger::DRIBBLE, "FLAG3");
    	int NUM;
	if (mainPlayers[q] == 'A')
	    NUM = 10;
	else
	if (mainPlayers[q] == 'B')
	    NUM = 11;
	else
	    NUM = mainPlayers[q] - 48;
	rcsc::dlog.addText( Logger::DRIBBLE,
                "now check player %c = %d", mainPlayers[q]
	        , NUM);

	if (!(NUM == receiver and isAreaDeleted))
	{
	    dlog.addText(Logger::DRIBBLE, "INIF");
	    if (wm->teammate(NUM))
		hisPosMain.push_back(wm->teammate(NUM)->pos());
	    else
		hisPosMain.push_back(Vector2D(0,0));
	    hisUNumMain.push_back(NUM);
	    
	    
	    rcsc::dlog.addText( Logger::DRIBBLE,
   		                "-->i = %d and hisUNumMain[i] = %d and his y is %f"
    	      		        ,i, hisUNumMain[i], hisPosMain[i].y);
	    i += 1;
	}
    }
    for ( unsigned q = 0;
	    q < getBackCount();
	    q++ )
    {
    	int NUM;
	if (backPlayers[q] == 'A')
	    NUM = 10;
	else
	if (backPlayers[q] == 'B')
	    NUM = 11;
	else
	    NUM = backPlayers[q] - 48;
	rcsc::dlog.addText( Logger::DRIBBLE,
                "now check player %d"
	        , NUM);
	if (!(NUM == receiver and isAreaDeleted))
	{
	    dlog.addText(Logger::DRIBBLE, "Man zendam!");
	    if (wm->teammate(NUM))
		hisPosMain.push_back(wm->teammate(NUM)->pos());
	    else
		hisPosMain.push_back(Vector2D(0,0));
	    hisUNumMain.push_back(NUM);

	    rcsc::dlog.addText( Logger::DRIBBLE,
   		                "-->i = %d and hisUNumBack[i] = %d and his y is %f"
    	      		        ,i, hisUNumMain[i], hisPosMain[i].y);
	    i += 1;
	}
    }
    
    for(int a = 0; a < i; a++)
    {
	dlog.addText(Logger::DRIBBLE, "UniNum = %d in area ", hisUNumMain[a]);
	dlog.addRect(Logger::DRIBBLE, Areas[a], 0, 0, 0);
	SortedAreas[hisUNumMain[a]] = Areas[a];
    }
}

void
GlobalPositioning2::setAreas2Players(const WorldModel *wm)
{
	rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "setAreas2Players"
       	      		        ,__FILE__, __LINE__ );

//	if (mainPlayers.length() == 3) FOR TEST
	{

		Vector2D hisPosMain[3], pos;
		int hisUNumMain[3], i = 0, num;

		const PlayerPtrCont::const_iterator end = wm->teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm->teammatesFromSelf().begin();
			it != end;
			++it )
		{
			int NUM = (*it)->unum();

			rcsc::dlog.addText( Logger::DRIBBLE,
   	  			                "now check player %d"
       	    	  		        , NUM);

		 	if (isInPlayers(NUM, mainPlayers) and NUM != receiver)
			{
				hisPosMain[i] = (*it)->pos();
				hisUNumMain[i] = (*it)->unum();

				rcsc::dlog.addText( Logger::DRIBBLE,
   		  				                "-->i = %d and hisUNumMain[i] = %d and his y is %f"
    		   		      		        ,i, hisUNumMain[i], hisPosMain[i].y);

				i += 1;
			}
		}

		rcsc::dlog.addText( Logger::DRIBBLE,
  			                "now check player %d"
   	    	  		        , wm->self().unum());

		if (isInPlayers(wm->self().unum(), mainPlayers) and wm->self().unum() != receiver)
		{
			hisPosMain[i] = wm->self().pos();
			hisUNumMain[i] = wm->self().unum();
			rcsc::dlog.addText( Logger::DRIBBLE,
	  				                "-->i = %d and hisUNumMain[i] = %d and his y is %f"
  		   		      		        ,i, hisUNumMain[i], hisPosMain[i].y);

			i += 1;
		}

		rcsc::dlog.addText( Logger::DRIBBLE,
   	  		                "We go to sort Players(in main line) by their Y");

		for(int j = 0; j < i - 1; j++)
			for(int k = j; k < i; k++)
			{
				if(hisPosMain[j].y > hisPosMain[k].y/* and (hisPosMain[j].y - hisPosMain[k].y) >  6*/)
				{
					pos = hisPosMain[j];
					hisPosMain[j] = hisPosMain[k];
					hisPosMain[k] = pos;
					
					num = hisUNumMain[j];
					hisUNumMain[j] = hisUNumMain[k];
					hisUNumMain[k] = num;
				}
				if((hisPosMain[j].y - hisPosMain[k].y) <=  6 and hisUNumMain[j] < hisUNumMain[k])
				{
					pos = hisPosMain[j];
					hisPosMain[j] = hisPosMain[k];
					hisPosMain[k] = pos;
					
					num = hisUNumMain[j];
					hisUNumMain[j] = hisUNumMain[k];
					hisUNumMain[k] = num;
				}
			}

		int t = 0;
		for(int a = 0; a < i; t++)
		{
			dlog.addText(Logger::DRIBBLE,"first line in for -----> a = %d    and   t = %d",a,t);
			if( isForReceiver(t) )
				continue;

			rcsc::dlog.addText( Logger::DRIBBLE,
   		  		                "SortedAreas[%d] = Areas[%d]",hisUNumMain[a], t);

			if(hisUNumMain[a] < -100)
				hisUNumMain[a] = 10;

			rcsc::dlog.addText( Logger::DRIBBLE,
   		  		                "After check 10 ::: >>>> SortedAreas[%d] = Areas[%d]",hisUNumMain[a], t);

			SortedAreas[hisUNumMain[a]] = Areas[t];
			a++;
		}

/**********************/

		int hisUNumBack[3];
		Vector2D hisPosBack[3];
		i = 0;

//		const PlayerPtrCont::const_iterator end = wm->teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm->teammatesFromSelf().begin();
			it != end;
			++it )
		{
			int NUM = (*it)->unum();
			if (isInPlayers(NUM, backPlayers) and NUM != receiver)
			{
				hisPosBack[i] = (*it)->pos();
				hisUNumBack[i] = (*it)->unum();
				i += 1;
			}
		}

		for(int j = 0; j <= i; j++)
			for(int k = j; k < i+1; k++)
			{
				if(hisPosBack[j].y > hisPosBack[k].y)
				{
					pos = hisPosBack[j];
					hisPosBack[j] = hisPosBack[k];
					hisPosBack[k] = pos;
					
					num = hisUNumBack[j];
					hisUNumBack[j] = hisUNumBack[k];
					hisUNumBack[k] = num;
				}
			}

		t = 0;
		for(int a = 0; a < i; t++)
		{
			if( isForReceiver(t) )
				continue;
			SortedAreas[hisUNumBack[a]] = Areas[t];
			a += 1;
		}

	}
/*	else
	{
		Vector2D hisPos[4], pos;
		int hisUNum[4], i = 0, num;

		const PlayerPtrCont::const_iterator end = wm->teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm->teammatesFromSelf().begin();
			it != end;
			++it )
		{
			int NUM = (*it)->unum();
			if (isInPlayers(NUM, allPlayers) and NUM != receiver)
			{
				hisPos[i] = (*it)->pos();
				hisUNum[i] = (*it)->unum();
				i += 1;
			}
		}

		for(int j = 0; j <= i; j++)
			for(int k = j; k < i+1; k++)
			{
				if(hisPos[j].y > hisPos[k].y)
				{
					pos = hisPos[j];
					hisPos[j] = hisPos[k];
					hisPos[k] = pos;
					
					num = hisUNum[j];
					hisUNum[j] = hisUNum[k];
					hisUNum[k] = num;
				}
			}

		int t = 0;
		for(int a = 0; a < i; t++)
		{
			if( isForReceiver(t) )
				continue;
			SortedAreas[hisUNum[a]] = Areas[t];
			a += 1;
		}

	}
*/
}

bool
GlobalPositioning2::isForReceiver(unsigned a)
{
    if(!isInPlayers(receiver, mainPlayers))
    {
	dlog.addText(Logger::DRIBBLE,"receiver is not in thsi line! return false");
	return false;
    }
    unsigned /*length = Areas.size(), */b = 100, aa = 0;

    dlog.addText(Logger::DRIBBLE,"Reveicer is %d and he receive ball in %f, %f", receiver, receivePos.x, receivePos.y);
    dlog.addText(Logger::DRIBBLE,"we want to check area[%d]", a);
//    if(getMainCount() == length or (Areas[a].center().x - backLine.x > Areas[a].center().x - mainLine.x and getBackCount() != 0 ))
    {
    	for(unsigned i = 0; i <= getMainCount(); i++)
    	{
    		if(abs(Areas[i].center().y - receivePos.y) < b)
    		{
    			b = abs(Areas[i].center().y - receivePos.y);
    			aa = i;
    		}
    	}
    }
/*    else
    {
    	for(unsigned i = 3; i <= length; i++)
    	{
    		if(abs(Areas[i].center().y - receivePos.y) < b)
    		{
    			b = abs(Areas[i].center().y - receivePos.y);
    			aa = i;
    		}
    	}
    }*/

    dlog.addText(Logger::DRIBBLE,"receiveArea is %d", aa);
    if (aa == a)
    	return true;

    dlog.addText(Logger::DRIBBLE,"Area[%d] is not receiveArea ", a);
    return false;
}

void
GlobalPositioning2::findPlayerPos(const WorldModel *wm, int num)
{
	findAllAreas(wm);

	Vector2D Pos;
	Pos = SortedAreas[num].center();

        SortedPos[num] = Pos;

	rcsc::dlog.addText( Logger::DRIBBLE,
	  		                "SortedPos[%d] = %f, %f", num, Pos.x, Pos.y );
	rcsc::dlog.addText( Logger::DRIBBLE,
	  		                "--->SortedAreas[%d] information : left = %f, minX = %f", num, SortedAreas[num].left(), SortedAreas[num].minX() );
	if (SortedPos[num].dist(lastPos) < 1.0)
	    SortedPos[num] = lastPos;
	lastPos = SortedPos[num];
}

void
GlobalPositioning2::detectReceiveInfo(const WorldModel *wm)
{
	if (wm->interceptTable()->teammateReachCycle() < 8)
	{
		receivePos = getBallPosAfterCycles(wm->ball(),wm->interceptTable()->teammateReachCycle(),rcsc::ServerParam::i().ballDecay());
		receiver = wm->interceptTable()->fastestTeammate() ? wm->interceptTable()->fastestTeammate()->unum() : 0;
		if (!isInField(receivePos))
		{
			receivePos = wm->ball().pos();
			receiver = 0;
		}
		return;
	}
	receivePos = wm->ball().pos();
	receiver = 0;
}

unsigned 
GlobalPositioning2::setPlayers(string players)
{
	allPlayers = players;
	mainPlayers = "   ";
	backPlayers = "";
	mainPlayers[0] = players[0];
	mainPlayers[1] = players[1];
	mainPlayers[2] = players[2];
	if (players.size() == 4)
	{
	    mainPlayers += " ";
	    mainPlayers[3] = players[3];
	}
	else 
	if (players.size() > 4)
	{
	    backPlayers += " ";
	    backPlayers += " ";
	    backPlayers[0] = players[3];
	    backPlayers[1] = players[4];
	    if (players.size() == 6)
	    {
		backPlayers += " ";
		backPlayers[2] = players[5];
	    }
	}
	dlog.addText(Logger::PASS, "Main Count = %d", getMainCount());
	dlog.addText(Logger::PASS, "Back Count = %d", getBackCount());
	return getMainCount();
}

unsigned 
GlobalPositioning2::getMainCount() const
{
	return mainPlayers.length();
}

unsigned 
GlobalPositioning2::getBackCount() const
{
	return backPlayers.length();
}

PosConf::PosConf()
{
	style = 0;
}

PosConf::PosConf(ConfigList & config, unsigned _style)
{
	dlog.addText(Logger::PASS, "PosConf creating %d", _style);
	
	style = _style;
	
	mainWeights.clear();
	backWeights.clear();
	Mid_BallYWeight.clear();
	Mid_GoalieYWeight.clear();
	Size_MainXWeight.clear();
	Size_BallAbsYWeight.clear();
	
	string addr = "OffensePositioning ";
	addr[addr.length() - 1] = (char)(_style + 48);
	for (unsigned i = 0;i < config[addr]["Config"].size();i++)
	{
		Expression exp = config[addr]["Config"][i];
		Weight w;
		w.From          = exp["From"][0][1].AsDouble();
		w.To            = exp["To"]  [0][1].AsDouble();
		w.OffsideWeight = exp["Main"][0]["Offside"][0][1].AsDouble();
		w.BallXWeight   = exp["Main"][0]["BallX"]  [0][1].AsDouble();
		mainWeights.push_back(w);
		w.From          = exp["From"][0][1].AsDouble();
		w.To            = exp["To"]  [0][1].AsDouble();
		w.OffsideWeight = exp["Back"][0]["Offside"][0][1].AsDouble();
		w.BallXWeight   = exp["Back"][0]["BallX"]  [0][1].AsDouble();
		backWeights.push_back(w);
		LocalWeight lw;
		lw.From = exp["From"][0][1].AsDouble();
		lw.To = exp["To"][0][1].AsDouble();
		lw.Value = exp["Middle"][0]["BallY"][0][1].AsDouble();
		Mid_BallYWeight.push_back(lw);
		lw.Value = exp["Middle"][0]["GoalieY"][0][1].AsDouble();
		Mid_GoalieYWeight.push_back(lw);
		lw.Value = exp["Size"][0]["MainX"][0][1].AsDouble();
		Size_MainXWeight.push_back(lw);
		lw.Value = exp["Size"][0]["BallAbsY"][0][1].AsDouble();
		Size_BallAbsYWeight.push_back(lw);
	}
}

PosConf::Weight
PosConf::getMainWeight(double ballx)
{
	for (unsigned i = 0;i < mainWeights.size();i++)
		if ((mainWeights[i].From < ballx) && (mainWeights[i].To > ballx))
		    return mainWeights[i];
	return mainWeights[0]; // No weight found!!!
}

PosConf::Weight
PosConf::getBackWeight(double ballx)
{
	for (unsigned i = 0;i < backWeights.size();i++)
		if ((backWeights[i].From < ballx) && (backWeights[i].To > ballx))
		    return backWeights[i];
	return mainWeights[0]; // No weight found!!!
}

double
PosConf::getWeight(vector<PosConf::LocalWeight> local, double ballx)
{
	for (unsigned i = 0;i < local.size();i++)
		if ((local[i].From < ballx) && (local[i].To > ballx))
		    return local[i].Value;
	return local[0].Value; // No weight found!!!
}

