// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef _GLOBAL_POSITIONING2_H
#define _GLOBAL_POSITIONING2_H

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/polygon_2d.h>
#include <rcsc/player/player_agent.h>

#include <map>
#include <vector>
#include <cstring>

#include "PositioningConfig.h"

#include <rcsc/player/ball_object.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/common/server_param.h>

#include <Config.h>
#include <Share.h>
#include <Basics.h>

using namespace std;
using namespace rcsc;
using namespace Basics;

struct PosConf
{
	struct Weight
	{
	    double From;
	    double To;
	    double OffsideWeight;
	    double BallXWeight;
	    double Sum()
	    {
			return BallXWeight + OffsideWeight;
	    }
	};
	
	struct LocalWeight
	{
		double From;
		double To;
		double Value;
	};
	
	int style;
	
	vector <Weight> mainWeights, backWeights;
	// Middle pos factors
	vector <LocalWeight> Mid_BallYWeight;
	vector <LocalWeight> Mid_GoalieYWeight;
	// Size factors
	vector <LocalWeight> Size_MainXWeight;
	vector <LocalWeight> Size_BallAbsYWeight;
	
	double getWeight(vector<LocalWeight> local, double ballx);
	
	PosConf();
	PosConf(ConfigList & config, unsigned _style);
	Weight getMainWeight(double ballx);
	Weight getBackWeight(double ballx);
};

class GlobalPositioning2
{
private:
	struct PosLine
	{
		double x, maxY, minY;
	};
	
	static int positionStyle;
	static std::vector<PositioningConfig> posConfs;
	static Vector2D lastPos;	
	static PosConf posConf;
	static int isSet;
	static int lastCycle;
	
	string mainPlayers, backPlayers, allPlayers;
	PosLine backLine, mainLine;
	vector <Rect2D> Areas;
	vector <Vector2D> points;
	map <int, Rect2D> SortedAreas;
	map <int, Vector2D> SortedPos;
	int receiver;
	Vector2D receivePos;
	double M_LineDistance;

public:
	GlobalPositioning2(string players = "AB9", unsigned style = 1);
	GlobalPositioning2(string mainplayers = "AB9", string backplayers = "867", unsigned style = 1);
	
	Vector2D getHomePos(const WorldModel *wm);
	Rect2D getHomeArea(const WorldModel *wm);
	Vector2D getPlayerPos(const WorldModel *wm, int unum);
	Rect2D getPlayerArea(const WorldModel *wm, int unum);

	unsigned getMainCount() const;
	unsigned getBackCount() const;
	
	unsigned setPlayers(string players); //return number of main players	
	void setLineDistance(double linedist);

private:
	void detectReceiveInfo(const WorldModel *wm);
	void determinePosLines(const WorldModel *wm);
	void findAllAreas(const WorldModel *wm);
	void setAreas2Players(const WorldModel *wm);
	void setAreasToPlayers(const WorldModel *wm);
	void findPlayerPos(const WorldModel *wm, int num);
	bool isForReceiver(unsigned a);

};

#endif //Global Positioning
