// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <HelliBASH09Pass.h>
#include <Basics.h>

#include <math.h>

#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>
#include <rcsc/action/intention_kick.h>


#include <rcsc/player/interception.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include "body_hold_ball2008.h"

#define maxX 51.5
#define maxY 33
#define MAXDIST 5

using namespace std;
using namespace rcsc;

HelliBASH09Pass::HelliBASH09Pass()
{
	for (uint i = 0 ; i < 11 ; i++)
		oppControlBuff[i] = 0.9;

	mustOneKick = false;
	
	insidePolygonPass = false;
	mTargetPolygon.assign();
	
	distToGoalWeight = 0;
	checkDistToGoal = false;
	maxDistanceToGoal = 110.f;
	
	forwardMovementWeight = 0;
	checkForwardMovement = false;
	maxForwardMovement = 105.f;

	receiverMovementWeight = 0;
	checkReceiverMovement = false;
	maxReceiverMovement = 110.f;
	
	receiverYChangeWeight = 0;
	checkReceiverYChange = false;
	maxReceiverYChange = 34.f;

	receiverAbilityWeight = 0;
	checkReceiverAbility = false;
	maxOpponents = 11;
	
	insidePolygonPass = false;
	bestAngle = bestVelocity = 0.f;
	bestReceiver = 0;
	for (map<float, map<float, bool> >::iterator it = isChecked.begin();
		it != isChecked.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}
	for ( int i = 0; i < 11; i++ )
		isOpen[i] = true;
		
	oppDelay = 1;
	oppKickable = 0.9;
	oppCatchable = 2;
	oppMaxDash = 1;
	tmmDelay = 1;
	tmmKickable = 0.9;
	tmmCatchable = 1.2;
	tmmMaxDash = 1;
	oppMaxPosCount = tmmMaxPosCount = 30;
	maxDirectionCount = 4;
	tmmMaxDistanceDelay = oppMaxDistanceDelay = 40;
}

void HelliBASH09Pass::reconstruct()
{
	insidePolygonPass = false;
	mTargetPolygon.assign();
	
	bestAngle = bestVelocity = 0.f;
	bestReceiver = 0;
	for (map<float, map<float, bool> >::iterator it = isChecked.begin();
		it != isChecked.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}
}

HelliBASH09Pass::~HelliBASH09Pass()
{
}

int HelliBASH09Pass::calculateMyReachCycle(rcsc::PlayerAgent *agent,float rate,float direction)
{
	float nextRate = rate * ServerParam::i().ballDecay();
	float distInFirstCycle = (rate + nextRate) / 2;
	const WorldModel & wm = agent->world();
	Vector2D ballrPos = Vector2D::polar2vector(distInFirstCycle , direction);
	Vector2D nextBallPos = wm.ball().pos() + ballrPos;
	Interception intercept(nextBallPos,nextRate,direction);
	float nextReachCycle = intercept.getReachCycle( wm.self().pos() , NULL , NULL , 0 , tmmKickable , tmmMaxDash );
	nextReachCycle += 1.f;
	return ceil(nextReachCycle);
}

passStats HelliBASH09Pass::getDirectionVelocityPassStats(PlayerAgent *agent,float rate,float direction , bool estimateHalfs)
{
	if ( isChecked[direction][rate] )
	{
		return checkedStats[direction][rate];
	}
	const WorldModel & wm = agent->world();
	Vector2D ballPos = wm.ball().pos();
	Vector2D tmmBallPos = Basics::getBallPosAfterCycles( ballPos , rate , direction , (float)tmmDelay , ServerParam::i().ballDecay());
	Vector2D oppBallPos = Basics::getBallPosAfterCycles( ballPos , rate , direction , (float)oppDelay , ServerParam::i().ballDecay());
    Interception normIntercept( ballPos , rate , direction);
    Interception tmmIntercept(tmmBallPos,rate * pow ( ServerParam::i().ballDecay() , tmmDelay ) ,direction);
    Interception oppIntercept(oppBallPos,rate * pow ( ServerParam::i().ballDecay() , oppDelay ) ,direction);
    int minReachCycle = calculateMyReachCycle(agent,rate,direction);
	Vector2D theReceiverPos = Vector2D( -1000 , -1000 );
	int fastest = -10;
	bool isTeamMate = false;
	passStats thisDirVelStats;
	// Filling wrong stats
		thisDirVelStats.receivePoint = Vector2D ( -1000 , -1000);
		thisDirVelStats.receiver = -10;
		thisDirVelStats.isTeamMateReceiver = false;
		thisDirVelStats.receiverPos = Vector2D ( -1000 , -1000);
		thisDirVelStats.reachCycle = 10000;
	
	// Checking TeamMates
	{
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( ( hisUNum > 0 ) and ( (*it)->posCount() <= tmmMaxPosCount ) )
			{
				const Vector2D * hisVel = NULL;
				const AngleDeg * hisAngle = NULL;
				if ( (*it)->velValid() and ( tmmMaxDash > 0 ) and ( (*it)->posCount() <= 3 ) )
					hisVel = & ( (*it)->vel() );
				if ( (*it)->faceValid() )
					hisAngle = & ( (*it)->body() );
				int hisReachCycle;
				if ( hisPos.dist ( ballPos ) < tmmMaxDistanceDelay )
				{
//					dlog.addText ( Logger::PASS , "Number %d delayed!!!" , hisUNum );
					hisReachCycle = ceil( tmmIntercept.getReachCycleGreedly( hisPos , hisVel , hisAngle , (*it)->posCount() , tmmKickable , tmmMaxDash , minReachCycle ) );
					hisReachCycle += tmmDelay;
				}
				else hisReachCycle = ceil ( normIntercept.getReachCycleGreedly ( hisPos , hisVel , hisAngle , (*it)->posCount() , tmmKickable , tmmMaxDash , minReachCycle ) );
				if ( hisReachCycle < minReachCycle )
				{
					theReceiverPos = hisPos;
					minReachCycle = hisReachCycle;
					fastest = hisUNum;
					if ( hisPos.x <= wm.offsideLineX() + 1.f )
						isTeamMate = true;
					else isTeamMate = false;
				}
			}
		}
	}

	// Checking Opponents
	{
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( ( ( hisUNum > 0 ) or ( estimateHalfs ) ) and ( (*it)->posCount() <= oppMaxPosCount ) )
			{
				float oppBuf = oppKickable;
				if ( hisUNum > 0 )
					oppBuf = oppControlBuff[hisUNum - 1];
				if ( hisUNum == wm.opponentGoalieUnum() )
					oppBuf = oppCatchable;
				int hisReachCycle;
				if ( ( hisPos.dist ( ballPos ) < oppMaxDistanceDelay ) and ( hisUNum != wm.opponentGoalieUnum() ) )
				{
					hisReachCycle = ceil( oppIntercept.getReachCycleGreedly( hisPos , NULL , NULL , (*it)->posCount() , oppBuf , oppMaxDash , minReachCycle + 2 ) );
					hisReachCycle += oppDelay;
				}
				else hisReachCycle = ceil( normIntercept.getReachCycleGreedly( hisPos , NULL , NULL , (*it)->posCount() , oppBuf , oppMaxDash , minReachCycle + 2) );
				if ( hisReachCycle <= minReachCycle )
				{
					theReceiverPos = hisPos;
					minReachCycle = hisReachCycle;
					fastest = hisUNum;
					isTeamMate = false;
				}
			}
		}
	}
	// Filling right stats
	if ( minReachCycle <= 25 )
	{
		thisDirVelStats.receiver = fastest;
		thisDirVelStats.isTeamMateReceiver = isTeamMate;
		thisDirVelStats.receiverPos = theReceiverPos;
		thisDirVelStats.reachCycle = minReachCycle;
		thisDirVelStats.receivePoint = Basics::getBallPosAfterCycles(ballPos,rate,direction,minReachCycle,ServerParam::i().ballDecay());
	}
	
	checkedStats[direction][rate] = thisDirVelStats;
	isChecked[direction][rate] = true;
	return thisDirVelStats;
}

float HelliBASH09Pass::getValue(PlayerAgent *agent , bool oneKick , bool estimateHalfs , float minAngle , float maxAngle , bool checkDist , float rateStepCount )
{
	checkDist = false;
	bool toffing = false;
	float bestValue = 0.f;
	mustOneKick = oneKick;
	const WorldModel & wm = agent->world();
	if ( not agent->world().self().isKickable() )
		return false;
	bestAngle = bestVelocity = -1.f;
	bestReceiver = 0;
/*	{
	const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
		it != end;
		++it )
		{
			Vector2D hisPos = (*it)->pos();
//			int hisUNum = (*it)->unum();
//			dlog.addText ( Logger::PASS , "seeing TMM %d in ( %f , %f ) with confidence %d" , hisUNum , hisPos.x , hisPos.y , (*it)->posCount());
//			if ( (*it)->velValid() )
//				dlog.addText ( Logger::PASS , "\tHis Vel is ( %f , %f )" , (*it)->vel().x , (*it)->vel().y);
//			if ( (*it)->faceValid() )
//				dlog.addText ( Logger::PASS , "\tHis Face To %f " , (*it)->body().degree());
		}
	}
	{
	const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
		it != end;
		++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			dlog.addText ( Logger::PASS , "seeing OPP %d in ( %f , %f ) with confidence %d" , hisUNum , hisPos.x , hisPos.y , (*it)->posCount());
			if ( (*it)->velValid() )
				dlog.addText ( Logger::PASS , "\tHis Vel is ( %f , %f )" , (*it)->vel().x , (*it)->vel().y);
			if ( (*it)->faceValid() )
				dlog.addText ( Logger::PASS , "\tHis Face To %f " , (*it)->body().degree());
			dlog.addText ( Logger::PASS , "\tHis Distance To Ball	%f" , hisPos.dist( wm.ball().pos() ) );
		}
	}
*/	for ( float direction = minAngle; direction <= maxAngle; direction += 10.f)
	{
//		dlog.addText ( Logger::PASS , "Direction	%f	:	%d" , direction , wm.dirCount ( direction ) );
		int sum,max,ave;
		wm.dirRangeCount( direction,30.0,&max,&sum,&ave);
//		dlog.addText ( Logger::PASS , "\t\tMax = %d\tSum =%d\tAve = %d" , max , sum , ave );
		
		if ( wm.dirCount( direction ) > maxDirectionCount )
			continue;
		if ( ave > maxDirectionCount ) // <---------- Check this in the game
			continue;
		float maxRate = 3.f;
		if ( mustOneKick )
		{
			maxRate = Body_KickOneStep(Vector2D(1000 , 0) ,1000).get_max_possible_vel(direction,1000,wm.ball().vel()).r();
		}
		float rateStep = maxRate / rateStepCount;
		if ( rateStep < 0.1 )
			continue;
		for ( float rate = rateStep; rate <= maxRate ; rate += rateStep)
		{
			passStats currentStat = getDirectionVelocityPassStats(agent,rate,direction,estimateHalfs);
//			dlog.addText ( Logger::PASS , "Direction %f -> Rate %f ====> Player %d->%d recieving in ( %f , %f ) after %d" , direction , rate , currentStat.isTeamMateReceiver , 
//										currentStat.receiver , currentStat.receivePoint.x , currentStat.receivePoint.y , currentStat.reachCycle);
//			dlog.flush();
			if ( not currentStat.isTeamMateReceiver )
			{
//				dlog.addText ( Logger::PASS , "Not TeamMate!!!");
				continue;
			}
			if ( currentStat.receiver < 0 )
			{
//				dlog.addText ( Logger::PASS , "Not Complete!!!");
				continue;
			}
			if ( ( insidePolygonPass ) and ( not mTargetPolygon.contains( currentStat.receivePoint ) ) )
			{
//				dlog.addText ( Logger::PASS , "Not In Polygon!!!");
				continue;
			}
			if ( not isOpen[ currentStat.receiver - 1] )
			{
//				dlog.addText ( Logger::PASS , "Not Open!!!");
				continue;
			}
			if ( ( abs(currentStat.receivePoint.x) > maxX ) or ( abs(currentStat.receivePoint.y) > maxY ) )
			{
//				dlog.addText ( Logger::PASS , "OUT!!!");
				continue;
			}
			if ( currentStat.receivePoint.x > wm.offsideLineX() + 1.f )
				continue;
			float curVal = 1.f;
			float ballMovement = currentStat.receivePoint.dist( wm.ball().pos() );
			if ( ballMovement <= 3.f )
				continue;
			float currentRate = rate;
			if ( checkDist )
			{
				float preVel = rate - rateStep;
//				dlog.addText ( Logger::PASS , "PreVel	%f" , preVel );
				bool shallIIgnore = true;
				if ( preVel >= 0.f )
				{
					passStats preStat = checkedStats[direction][preVel];
					if ( not preStat.isTeamMateReceiver )
						shallIIgnore = false;
					if ( currentStat.receiver < 0 )
						shallIIgnore = false;
					if ( ( insidePolygonPass ) and ( not mTargetPolygon.contains( preStat.receivePoint ) ) ) 
						shallIIgnore = false;
					if ( not isOpen[ preStat.receiver - 1] )
						shallIIgnore = false;
					if ( ( abs(preStat.receivePoint.x) > maxX ) or ( abs(preStat.receivePoint.y) > maxY ) )
						shallIIgnore = false;
				}
				if ( ballMovement <= 10 )
				{
					if ( ( rate >= 2.f ) and ( shallIIgnore ) )
						continue;
				}
				else if ( ballMovement <= 15 )
				{
					if ( ( rate >= 2.5 ) and ( shallIIgnore ) )
						continue;
				}
				else if ( ballMovement <= 20 )
				{
					if ( ( rate >= 3.f ) and ( shallIIgnore ) )
						continue;
				}
			}

			int weightSum = distToGoalWeight + forwardMovementWeight + receiverMovementWeight + receiverYChangeWeight + receiverAbilityWeight;
			float forwardMovement = currentStat.receivePoint.x - wm.ball().pos().x;
			if ( ( ( forwardMovement < 0 ) or ( forwardMovement > maxForwardMovement ) ) and ( checkForwardMovement ) )
			{
//				dlog.addText ( Logger::PASS , "FORWARD PROBLEM!!!");
				continue;
			}
			float distToGoal = currentStat.receivePoint.dist( Vector2D(52.5,0) );
			if ( ( distToGoal > maxDistanceToGoal ) and ( checkDistToGoal ) )
			{
//				dlog.addText ( Logger::PASS , "DISTTOGOAL PROBLEM!!!");
				continue;
			}
			float playerMovement = (currentStat.receivePoint - currentStat.receiverPos).r();
			if ( ( playerMovement > maxReceiverMovement ) and ( checkReceiverMovement ) )
			{
//				dlog.addText ( Logger::PASS , "PlayerMovement PROBLEM!!!");
				continue;
			}
			float receiverYChange = abs( (currentStat.receivePoint - currentStat.receiverPos).y );
			if ( ( receiverYChange > maxReceiverYChange ) and ( checkReceiverYChange ) )
			{
//				dlog.addText ( Logger::PASS , "ReceiverYChange PROBLEM!!!");
				continue;
			}
			int oppsNear = 0;
			const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
			for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
				it != end;
				++it )
			{
				Vector2D hisPos = (*it)->pos();
				int hisUNum = (*it)->unum();
				if ( ( hisUNum > 0 ) or ( estimateHalfs ) )
				{
					if ( currentStat.receivePoint.dist(hisPos) <= MAXDIST )
						oppsNear++;
				}
			}
			if ( ( oppsNear > maxOpponents ) and ( checkReceiverAbility ) )
			{
//				dlog.addText ( Logger::PASS , "MAX Opponents Problem");
				continue;
			}
			if ( weightSum > 0 )
			{
				curVal = ( forwardMovementWeight * ( forwardMovement / maxForwardMovement ) )
							+ ( distToGoalWeight * ( ( maxDistanceToGoal - distToGoal ) / maxDistanceToGoal ) )
							+ ( receiverMovementWeight * ( ( maxReceiverMovement - playerMovement ) / maxReceiverMovement ) )
							+ ( receiverYChangeWeight * ( ( maxReceiverYChange - receiverYChange ) / maxReceiverYChange ) )
							+ ( receiverAbilityWeight * ( ( maxOpponents - oppsNear ) / maxOpponents ) );
				curVal /= weightSum;
			}
			if ( curVal > bestValue )
			{
//				dlog.addText ( Logger::PASS , "CHANGED!!!" );
				toffing = false;
				bestValue = curVal;
				bestVelocity = currentRate;
				bestAngle = direction;
				bestReceiver = currentStat.receiver;
				receivePoint = currentStat.receivePoint;
			}
		}
	}
	dlog.addText ( Logger::PASS , "Receiver	=	%d",bestReceiver);
	dlog.addText ( Logger::PASS , "receivePoint	=	( %f , %f )" , receivePoint.x , receivePoint.y );
	dlog.addText ( Logger::PASS , "Value	=	%f" , bestValue);
	return bestValue;
}

void HelliBASH09Pass::setDistToGoalParameters(int weight,float maxDist,bool checkIt)
{
	distToGoalWeight = weight;
	checkDistToGoal = ( checkIt or weight );
	maxDistanceToGoal = maxDist;
}

void HelliBASH09Pass::setForwardMovementParameters(int weight,float maxMove,bool checkIt)
{
	forwardMovementWeight = weight;
	checkForwardMovement = ( checkIt or weight );
	maxForwardMovement = maxMove;
}

void HelliBASH09Pass::setReceiverMovementParameters(int weight,float maxMove,bool checkIt)
{
	receiverMovementWeight = weight;
	checkReceiverMovement = ( checkIt or weight );
	maxReceiverMovement = maxMove;
}

void HelliBASH09Pass::setReceiverYChangeParameters(int weight,float maxChange, bool checkIt)
{
	receiverYChangeWeight = weight;
	checkReceiverYChange = (checkIt or weight);
	maxReceiverYChange = maxChange;
}

void HelliBASH09Pass::setReceiverAbilityParameters(int weight,int max, bool checkIt)
{
	receiverAbilityWeight = weight;
	checkReceiverAbility = (checkIt or weight);
	maxOpponents = max;
}

void HelliBASH09Pass::setSafetyFactors(float _tmmKickable,float _tmmCatchable,float _tmmMaxDash,int _tmmDelay,int _tmmMaxPosCount,float _oppKickable,float _oppCatchable,float _oppMaxDash,int _oppDelay,int _oppMaxPosCount)
{
	for (map<float, map<float, bool> >::iterator it = isChecked.begin();
		it != isChecked.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}
	oppDelay = _oppDelay;
	oppKickable = _oppKickable;
	oppCatchable = _oppCatchable;
	oppMaxDash = _oppMaxDash;
	oppMaxPosCount = _oppMaxPosCount;
	tmmDelay = _tmmDelay;
	tmmKickable = _tmmKickable;
	tmmCatchable = _tmmCatchable;
	tmmMaxDash = _tmmMaxDash;
	tmmMaxPosCount = _tmmMaxPosCount;
	for (uint i = 0 ; i < 11 ; i++)
		oppControlBuff[i] = _oppKickable;
}

void HelliBASH09Pass::setOpenPlayers(string players)
{
	for ( int i = 0; i < 11; i++ )
		isOpen[i] = false;
	for ( uint i = 0; i < players.size(); i++)
	{
		if ( players[i] == 'A' )
			isOpen[9] = true;
		else if ( players[i] == 'B' )
			isOpen[10] = true;
		else
			isOpen[players[i] - '1' ] = true;
	}
}

void HelliBASH09Pass::execute(PlayerAgent *agent,bool isEnforced,int delay)
{
	if ( agent->world().self().pos().dist( receivePoint ) <= 5.f )
	{
		if ( agent->world().self().pos().x <= 36 )
		{
//			dlog.addText ( Logger::PASS , "Tofe Nazdik" );
			receivePoint = checkedStats[bestAngle][bestVelocity].receiverPos;
			bestVelocity = 1.f;
		}
	}
	else if ( agent->world().self().pos().dist( receivePoint ) <= 15.f )
	{
//		dlog.addText ( Logger::PASS , "Tofe Dur" );
		receivePoint = receivePoint + checkedStats[bestAngle][bestVelocity].receiverPos;
		receivePoint = receivePoint / 2.f;
		bestVelocity = min ( bestVelocity , 2.f );
	}
	if ( ( mustOneKick ) or ( delay == 0 ) )
	{
//		cerr << "ONe Kick!!!" << endl;
		Body_KickOneStep(receivePoint,bestVelocity).execute(agent);
	}
	else
	{
//		dlog.addText ( Logger::PASS , "MultiStepKick	: Velocity %f , isEnforced %d , delay %d" , bestVelocity , isEnforced , delay);
		if ( delay == 1 )
		{
			Body_KickTwoStep(receivePoint,bestVelocity,isEnforced).execute(agent);
//			dlog.addText(Logger::PASS, "TwoKick");
		}
		else
			Body_KickMultiStep(receivePoint,bestVelocity,isEnforced).execute(agent);
		agent->setIntention( new IntentionKick( receivePoint , bestVelocity , delay + 1 , isEnforced ,agent->world().time() ) );
	}
	sayPass(agent);
}

void HelliBASH09Pass::sayPass(rcsc::PlayerAgent *agent)
{
	Vector2D target_buf = receivePoint - agent->world().self().pos();
	target_buf.setLength( 1.0 );

	agent->addSayMessage( new PassMessage( bestReceiver , receivePoint /*+ target_buf */, agent->effector().queuedNextBallPos() , agent->effector().queuedNextBallVel() ) );
}

Polygon2D & HelliBASH09Pass::targetPolygon()
{
	return mTargetPolygon;
}

void HelliBASH09Pass::setInsidePolygonPass(bool b)
{
	insidePolygonPass = b;
}

int HelliBASH09Pass::receiver()
{
	return bestReceiver;
}

passStats HelliBASH09Pass::getBestPassStats()
{
	return checkedStats[bestAngle][bestVelocity];
}

Vector2D HelliBASH09Pass::hisReceivePoint()
{
	return receivePoint;
}

void HelliBASH09Pass::setMaxDirectionCount(float _maxDirectionCount)
{
	maxDirectionCount = _maxDirectionCount;
}

void HelliBASH09Pass::setDelayFactors( int _tmmDelay , float _tmmMaxDistanceDelay , int _oppDelay , float _oppMaxDistanceDelay)
{
	for (map<float, map<float, bool> >::iterator it = isChecked.begin();
		it != isChecked.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}
	tmmDelay = _tmmDelay;
	tmmMaxDistanceDelay = _tmmMaxDistanceDelay;
	oppDelay = _oppDelay;
	oppMaxDistanceDelay = _oppMaxDistanceDelay;
}

void HelliBASH09Pass::setOppPlayerControlBuff ( int unum , float controlBuff )
{
	oppControlBuff[unum - 1] = controlBuff;
}

