// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef _HELLIBASH_09_PASS_H
#define _HELLIBASH_09_PASS_H

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/polygon_2d.h>
#include <rcsc/player/player_agent.h>

#include <map>
#include <string>

struct passStats
{
	rcsc::Vector2D receivePoint;
	int receiver;
	bool isTeamMateReceiver;
	rcsc::Vector2D receiverPos;
	int reachCycle;
};

class HelliBASH09Pass
{
private:
	bool mustOneKick;
	float bestAngle;
	float bestVelocity;
	int bestReceiver;
	rcsc::Vector2D receivePoint;

	// Factors
	bool insidePolygonPass;
	rcsc::Polygon2D mTargetPolygon;

	int distToGoalWeight;
	bool checkDistToGoal;
	float maxDistanceToGoal;

	int forwardMovementWeight;
	bool checkForwardMovement;
	float maxForwardMovement;
	
	int receiverMovementWeight;
	bool checkReceiverMovement;
	float maxReceiverMovement;
	
	int receiverYChangeWeight;
	bool checkReceiverYChange;
	float maxReceiverYChange;
	
	int receiverAbilityWeight;
	bool checkReceiverAbility;
	int maxOpponents;

	// Safety Factors
	float oppKickable;
	float oppCatchable;
	float oppMaxDash;
	int oppMaxPosCount;
	float tmmKickable;
	float tmmCatchable;
	float tmmMaxDash;
	int tmmMaxPosCount;
	int maxDirectionCount;
	float oppControlBuff[11];
	
		// Delay Factors
		int tmmDelay;
		float tmmMaxDistanceDelay;
		int oppDelay;
		float oppMaxDistanceDelay;

	
	bool isOpen[11];
	
	int calculateMyReachCycle(rcsc::PlayerAgent *agent,float rate,float direction);
	
	passStats getDirectionVelocityPassStats(rcsc::PlayerAgent *agent,float rate,float direction,bool estimateHalfs);
	
	std::map<float , std::map<float,passStats> > checkedStats;
	std::map<float , std::map<float,bool> > isChecked;
	
public:
	HelliBASH09Pass();
	void reconstruct();
	~HelliBASH09Pass();
	rcsc::Polygon2D & targetPolygon();
	void setInsidePolygonPass(bool b);
	void setDistToGoalParameters(int weight = 1,float maxDist = 110.f , bool checkIt = true);
	void setForwardMovementParameters(int weight = 1,float maxMove = 105 , bool checkIt = true);
	void setReceiverMovementParameters(int weight = 1,float maxMove = 110 , bool checkIt = true);
	void setReceiverYChangeParameters(int weight = 1,float maxChange = 68 , bool checkIt = true);
	void setReceiverAbilityParameters(int weight = 1,int max = 11 , bool checkIt = true);
	void setOpenPlayers(std::string players);
	void setSafetyFactors(float _tmmKickable,float _tmmCatchable,float _tmmMaxDash,int _tmmDelay,int _tmmMaxPosCount,float _oppKickable,float _oppCatchable,float _oppMaxDash,int _oppDelay,int _oppMaxPosCount);
	void setDelayFactors( int _tmmDelay , float _tmmMaxDistanceDelay , int _oppDelay , float _oppMaxDistanceDelay);
	void setMaxDirectionCount(float _maxDirectionCount);
	float getValue(rcsc::PlayerAgent *agent, bool oneKick = false , bool estimateHalfs = false , float minAngle = -180 , float maxAngle = 180 , bool checkDist = true , float rateStepCount = 6);
	void execute(rcsc::PlayerAgent *agent,bool isEnforced = false ,int delay = 2);
	void sayPass(rcsc::PlayerAgent *agent);
	int receiver();
	passStats getBestPassStats();
	rcsc::Vector2D hisReceivePoint();

	void setOppPlayerControlBuff ( int unum , float controlBuff );
};

#endif // _HELLIBASH_09_PASS_H
