// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <HelliBASH09SRP.h>
#include <Basics.h>

#include <math.h>
#include <iostream>

#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/intention_kick.h>


#include <rcsc/player/interception.h>
#include <rcsc/player/world_model.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#define maxX 51.5
#define maxY 33

using namespace rcsc;
using namespace std;

int HelliBASH09SRP::lastKickCycle = -1000;

HelliBASH09SRP::HelliBASH09SRP(Vector2D target):mTarget(target) , mForbiddenPolygon()
{
	othersKickable = 1.f;
	othersMaxDash = 1.f;
	myKickable = 1.f;
	myMaxDash = 1.f;
	hasForbiddenPolygon = false;
	maxForwardMovement = 10.f;
	maxVisionDelay = 4;
}

HelliBASH09SRP::HelliBASH09SRP()
{
	mTarget = Vector2D(0,0);
	othersKickable = 1.f;
	othersMaxDash = 1.f;
	myKickable = 1.f;
	myMaxDash = 1.f;
	hasForbiddenPolygon = false;
	maxForwardMovement = 10.f;
	maxVisionDelay = 4;
}

void HelliBASH09SRP::reconstruct(Vector2D target)
{
	mTarget = target;
	mForbiddenPolygon.assign();
	othersKickable = 1.f;
	othersMaxDash = 1.f;
	myKickable = 1.f;
	myMaxDash = 1.f;
	hasForbiddenPolygon = false;
	maxForwardMovement = 10.f;
	maxVisionDelay = 2;
}

HelliBASH09SRP::~HelliBASH09SRP()
{
}

float HelliBASH09SRP::getCycleBeforeExit(PlayerAgent *agent,float curRate)
{
	const WorldModel & wm = agent->world();
	Vector2D ballPos = wm.ball().pos();
	float ballVel = curRate;
	float direction = ( mTarget - wm.self().pos() ).th().degree();
	int cycleBeforeExit = -1;
	while ( ( wm.self().pos().dist(ballPos) <= 1.f ) and ( cycleBeforeExit <= 7 ))
	{
		cycleBeforeExit++;
		float midVel = (ballVel + ballVel * ServerParam::i().ballDecay() ) / 2;
		ballPos = ballPos + Vector2D::polar2vector(midVel,direction);
		ballVel *= ServerParam::i().ballDecay();
	}
	if ( cycleBeforeExit == -1 )
		cycleBeforeExit = 10000;
	return (float)(cycleBeforeExit + 1);
}

float HelliBASH09SRP::getOthersMinReachCycle(PlayerAgent *agent,float curRate)
{
	const WorldModel & wm = agent->world();
	Vector2D ballPos = wm.ball().pos();
	float direction = ( mTarget - wm.self().pos() ).th().degree();
	Interception intercept(ballPos,curRate,direction);
	float minReachCycle = 1000;
	// Checking TeamMates
	{
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( hisUNum > 0 )
			{
				double hisReachCycle = intercept.getReachCycle( hisPos , NULL , NULL , (*it)->posCount() , othersKickable , othersMaxDash );
				if ( hisReachCycle < minReachCycle )
				{
					minReachCycle = hisReachCycle;
				}
			}
		}
	}
	// Checking TeamMates
	{
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( hisUNum > 0 )
			{
				double hisReachCycle = intercept.getReachCycle( hisPos , NULL , NULL , (*it)->posCount() , othersKickable , othersMaxDash );
				if ( hisReachCycle < minReachCycle )
				{
					minReachCycle = hisReachCycle;
				}
			}
		}
	}
	return minReachCycle;

}

float HelliBASH09SRP::getMyReachCycle(PlayerAgent *agent,float curRate)
{
	float cyclesToExit = getCycleBeforeExit(agent,curRate);
	if ( ( cyclesToExit < 0.f ) or ( cyclesToExit > 5.f ) )
		return -1.f;
	float startRate = curRate * pow ( (float)ServerParam::i().ballDecay() , cyclesToExit );
	const WorldModel & wm = agent->world();
	float direction = ( mTarget - wm.self().pos() ).th().degree();
	Vector2D ballPos = Basics::getBallPosAfterCycles( wm.ball().pos() , curRate , direction , cyclesToExit , ServerParam::i().ballDecay());
	Interception interception(ballPos , startRate , direction);
	float myReachCycle = interception.getReachCycle ( wm.self().pos() , &(wm.self().vel()) , &(wm.self().body()) , 0 , myKickable , myMaxDash );
	return (myReachCycle + (float)cyclesToExit);
}

float HelliBASH09SRP::isGood(PlayerAgent *agent,float curRate)
{
	float others = getOthersMinReachCycle(agent,curRate);
	float myReachCycle = getMyReachCycle(agent,curRate);
	if ( others > myReachCycle )
		return myReachCycle;
	else return -1.f;
}

void HelliBASH09SRP::setHasForbiddenPolygon(bool b)
{
	hasForbiddenPolygon = b;
}

void HelliBASH09SRP::setMaxForwardMovement(float max)
{
	maxForwardMovement = max;
}

void HelliBASH09SRP::setMaxVisionDelay(float max)
{
	maxVisionDelay = max;
}

rcsc::Polygon2D & HelliBASH09SRP::forbiddenPolygon()
{
	return mForbiddenPolygon;
}

void HelliBASH09SRP::setSafetyValues(float _myKickable,float _myMaxDash,float _othersKickable,float _othersMaxDash)
{
	myKickable = _myKickable;
	myMaxDash = _myMaxDash;
	othersKickable = _othersKickable;
	othersMaxDash = _othersMaxDash;
}

bool HelliBASH09SRP::getValue(rcsc::PlayerAgent *agent)
{
	bestRate = 0.f;
	ballMovement = 0.f;
	const WorldModel & wm = agent->world();
	if ( wm.self().stamina() < 2000.f )
		return false;
	float direction = ( mTarget - wm.self().pos() ).th().degree();
	if ( wm.dirCount( direction ) > maxVisionDelay )
			return false;
	float maxRate = Body_KickOneStep(Vector2D(1000 , 0) ,1000).get_max_possible_vel(direction,1000,wm.ball().vel()).r();	
	float rateStep = maxRate / 10.f;
	if ( rateStep <= 0.2 )
		return false;
	for ( float f = rateStep; f <= maxRate; f += rateStep)
	{
		dlog.flush();
		float good = isGood(agent , f);
		if ( good < 0 )
			continue;
		if ( f > bestRate )
		{
			Vector2D reunionPoint = Basics::getBallPosAfterCycles( wm.ball().pos() , f , direction , good , ServerParam::i().ballDecay());
			if ( ( abs( reunionPoint.y ) > maxY ) or ( abs( reunionPoint.x ) > maxX ) )
				continue;
			dlog.addText ( Logger::PASS , "Rate	=	%f -> reunionPoint	=	( %f , %f )" , f , reunionPoint.x , reunionPoint.y );
			if ( hasForbiddenPolygon )
				if ( mForbiddenPolygon.contains( reunionPoint ) )
					continue;
			float bm = wm.ball().pos().dist(reunionPoint);
			if ( bm > maxForwardMovement )
				continue;
			ballMovement = bm;
			bestRate = f;
		}
	}
	dlog.addText ( Logger::PASS , "Best Rate	=	%f" , bestRate );
	if ( bestRate > 0 )
		return true;
	return false;
}

float HelliBASH09SRP::getBallMovement()
{
	return ballMovement;
}

void HelliBASH09SRP::execute(rcsc::PlayerAgent *agent)
{
	dlog.addText ( Logger::PASS , "HelliBASH09SRP::execute()" );
	Body_KickOneStep(mTarget , bestRate).execute(agent);
	lastKickCycle = agent->world().time().cycle();
	dlog.addText ( Logger::PASS , "lastKickCycle	=	%d " , lastKickCycle );
}

bool HelliBASH09SRP::shouldIntercept(PlayerAgent *agent)
{
	int cycle = agent->world().time().cycle();
	dlog.addText ( Logger::PASS , "delta Cycle	=	%d" , cycle - lastKickCycle );
	if ( ( cycle > lastKickCycle ) and ( cycle - lastKickCycle <= 2 ) )
		return true;
	else
		return false;
}

void HelliBASH09SRP::reset()
{
	lastKickCycle = -1000;
}
