// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <Intercept.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/body_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_scan_field.h>


#include <strategy.h>
#include <Basics.h>
#include <iostream>
#include <vector>
#include <Share.h>

//#define INTERCEPT_LOG
#define EXTRA_THREAD 0.2

using namespace Basics;
using namespace std;
using namespace rcsc;

Intercept::InterceptModel Intercept::finalIntercept(rcsc::PlayerAgent * agent, bool useMyBall, float moveError, int maxCycles, int playerDelay, int updateCycle)
{
	maxSideDashCycle = 10;
	Vector2D selfPos = agent->world().self().inertiaPoint(updateCycle)
	, selfVel = agent->world().self().vel();
	for (int i = 0;i < updateCycle;i++)
	    selfVel *= agent->world().self().playerType().playerDecay();
	const WorldModel & wm = agent->world();
	BallObject calculatingBall = wm.ball();
	if (useMyBall)
		calculatingBall = myBall;
	minTurnThread = 0.f;
	if (selfPos.dist(calculatingBall.pos()) > 30.f)
		minTurnThread = 2;

	Vector2D interceptPoint = calculatingBall.pos();
	const Vector2D & inertiaFinalPoint = calculatingBall.inertiaFinalPoint();

	vector<InterceptModel> intercepts;
//	InterceptModel finalInterceptModel(9999, 9999, Vector2D(-70, -70), false, 9999, false, 9999);
	InterceptModel finalInterceptModel(9999, 9999, Vector2D(-70, -70), false, 9999, false, false, false, 9999);
	int cycle = 0;
	Vector2D curBallPos = calculatingBall.pos();
	while (true)
	{
		Vector2D curBallPos = calculatingBall.inertiaPoint(cycle);
		dlog.addText(Logger::INTERCEPT, "cycle %d pos(%f, %f), dist = %f", cycle, curBallPos.x, curBallPos.y, curBallPos.dist(wm.self().pos()));
		
		if (curBallPos.dist(inertiaFinalPoint) < 0.1) // if ball stopped
		{
			dlog.addText(Logger::INTERCEPT, "Special!!!");
			int arriveCycle = getArriveCycle(agent, curBallPos, maxCycles, updateCycle);
			if (arriveCycle == -1)
			    arriveCycle = maxCycles + 1; // :)
			intercepts.push_back(InterceptModel(arriveCycle, cycle, curBallPos, isSideDash, dirSideDash, isTurnMove, isMinusDash, isPlusDash, turnSimAngle));
			break;
		}
		int arriveCycle = (playerDelay >= cycle) ? -1 : getArriveCycle(agent, curBallPos, cycle - playerDelay, updateCycle);
		
		if (arriveCycle != -1)
		{
		    if (SRP)
				return InterceptModel(arriveCycle, cycle, curBallPos, isSideDash, dirSideDash, isTurnMove, isMinusDash, isPlusDash, turnSimAngle);
		    else
				intercepts.push_back(InterceptModel(arriveCycle, cycle, curBallPos, isSideDash, dirSideDash, isTurnMove, isMinusDash, isPlusDash, turnSimAngle));
		}
		cycle++;
		if (cycle > maxCycles)
		    break;
	}

	int minReachCycle = 9999;
	for (vector<InterceptModel>::iterator it = intercepts.begin(); it != intercepts.end(); it++)
	{
		if (it->getBallReachCycle() < minReachCycle)
		{
			minReachCycle = it->getReachCycle();
			finalInterceptModel = (*it);
		}
	}

	return finalInterceptModel;
}

/*Intercept::InterceptModel Intercept::finalIntercept(rcsc::PlayerAgent * agent, PlayerObject * player, bool useMyBall, float moveError, int maxCycles, int playerDelay)
{
	maxSideDashCycle = 10;
	const WorldModel & wm = agent->world();
	BallObject calculatingBall = wm.ball();
	if (useMyBall)
		calculatingBall = myBall;
	minTurnThread = 0.f;
	if (player->pos().dist(calculatingBall.pos()) > 30.f)
		minTurnThread = 2;
	InterceptModel finalInterceptModel(9999, 9999, Vector2D(-70, -70), false, 9999, false, false, false, 9999);
	Vector2D interceptPoint = calculatingBall.pos();
	const Vector2D & inertiaFinalPoint = calculatingBall.inertiaFinalPoint();

	vector<InterceptModel> intercepts;
	int cycle = 0;
	Vector2D curBallPos = calculatingBall.pos();
	while (true)
	{
		Vector2D curBallPos = calculatingBall.inertiaPoint(cycle);
		if (curBallPos.dist(inertiaFinalPoint) < 0.1) // if ball stopped
			break;

		int arriveCycle = (playerDelay >= cycle) ? -1 : getArriveCycle(agent, curBallPos, cycle);
		if (arriveCycle != -1)
		{
		    if (SRP)
				return InterceptModel(arriveCycle, cycle, curBallPos, isSideDash, dirSideDash, isTurnMove, isMinusDash, isPlusDash, turnSimAngle);
		    else
				intercepts.push_back(InterceptModel(arriveCycle, cycle, curBallPos, isSideDash, dirSideDash, isTurnMove, isMinusDash, isPlusDash, turnSimAngle));
		}
		cycle++;
		if (cycle > maxCycles)
		    break;
	}

	int minReachCycle = 9999;
	for (vector<InterceptModel>::iterator it = intercepts.begin(); it != intercepts.end(); it++)
	{
		if (it->getBallReachCycle() < minReachCycle)
		{
			minReachCycle = it->getReachCycle();
			finalInterceptModel = (*it);
		}
	}

	return finalInterceptModel;
}
*/

void Intercept::execute(rcsc::PlayerAgent * agent , bool useMyBall , float moveError)
{

	Share::isInterceptExecuted = true;
// I AM HERE
	dlog.addText(Logger::ROLE, "Intercept::execute");
	maxSideDashCycle = 10;
	const WorldModel & wm = agent->world();
	BallObject calculatingBall = wm.ball();
	if (useMyBall)
		calculatingBall = myBall;
	minTurnThread = 0.f;
	if (wm.self().pos().dist(calculatingBall.pos()) > 30.f)
		minTurnThread = 2;
//#ifdef INTERCEPT_LOG
	dlog.addText(Logger::INTERCEPT, "Ball -> (%f, %f)", calculatingBall.pos().x, calculatingBall.pos().y);
	dlog.addText(Logger::INTERCEPT, "Vel  -> (%f, %f)", calculatingBall.vel().x, calculatingBall.vel().y);
//#endif
	Vector2D interceptPoint = calculatingBall.pos();
	const Vector2D & inertiaFinalPoint = calculatingBall.inertiaFinalPoint();
//	dlog.addText ( Logger::ROLE , "myPos ( %f , %f )" , wm.self().pos().x , wm.self().pos().y );

	vector<InterceptModel> intercepts;
	InterceptModel finalInterceptModel(9999, 9999, Vector2D(-70, -70), false, 9999, false, false, false, 9999);
	int cycle = 0;
	while (true)
	{
		Vector2D curBallPos;// = calculatingBall.inertiaPoint(cycle);
		if ( cycle >= ballDelay )
			curBallPos = calculatingBall.inertiaPoint(cycle - ballDelay);
		else
			curBallPos = calculatingBall.inertiaPoint(0);
//		if (cycle < 15)
//			dlog.addText(Logger::ROLE, "SimBallPos.x(%.2f).y(%.2f)cycle(%d)ballDelay(%d)", curBallPos.x, curBallPos.y, cycle - ballDelay , ballDelay);
		if (curBallPos.dist(inertiaFinalPoint) < 0.1 or (curBallPos.absX() > 55 or curBallPos.absY() > 37)) // if ball stopped
			break;

		int arriveCycle = getArriveCycle(agent, curBallPos, cycle, 0);
		if (arriveCycle != -1)
			intercepts.push_back(InterceptModel(arriveCycle, cycle, curBallPos, isSideDash, dirSideDash, isTurnMove, isMinusDash, isPlusDash, turnSimAngle));
		cycle++;
//		curBallPos += ballVel();
//		ballVel *= ballDecay;
	}

	int minReachCycle = 9999;
	for (vector<InterceptModel>::iterator it = intercepts.begin(); it != intercepts.end(); it++)
	{
		if (it->getBallReachCycle() < minReachCycle)
		{
			minReachCycle = it->getReachCycle();
			finalInterceptModel = (*it);
		}
	}

	interceptPoint = finalInterceptModel.getBallPos();
//#ifdef INTERCEPT_LOG
	dlog.addText(Logger::ROLE, "INTERCEPT: Ball Vell %.2f", calculatingBall.vel().r());
	dlog.addText(Logger::ROLE, "INTERCEPT: FinalInterceptPoint=(%.2f, %.2f) FinalInterceptTime(%d) FinalInterceptBallTime(%d)"
		, interceptPoint.x, interceptPoint.y, finalInterceptModel.getReachCycle(), finalInterceptModel.getBallReachCycle());
//#endif
	if (intercepts.size() == 0)
	{
#ifdef INTERCEPT_LOG
		dlog.addText(Logger::ROLE, "INTERCEPT GOTO INERTIA POINT");
#endif
		double dashPower = 100;
		/*if (wm.self().stamina() < 1500 )
			dashPower = 25;*/
		dashPower = wm.self().getSafetyDashPower(dashPower);
		if (!Body_GoToPoint(calculatingBall.inertiaPoint(cycle + 1)/*inertiaFinalPoint*/, moveError , dashPower).execute(agent))
			Body_TurnToBall().execute(agent);
	}
	else if (finalInterceptModel.isSideDash())
	{
#ifdef INTERCEPT_LOG
		dlog.addText(Logger::ROLE, "SIDE DASH INTERCEPT");
#endif
		double dashPower = 100;
		Vector2D targetRel = interceptPoint - wm.self().pos();
		if (wm.self().stamina() < 2000 /* ServerParam::i().recoverDecThrValue() */)
			dashPower = 25;
		else if (targetRel.r() < 2)
		{
			int cycle = finalInterceptModel.getBallReachCycle() - finalInterceptModel.getReachCycle();
			for (int curDashPower = -100; curDashPower <= 100; curDashPower += 25)
			{
				Vector2D bodyPos = wm.self().pos(), bodyVel = wm.self().vel();
				for (int i = 0; i < cycle; i++)
				{
					Vector2D dashAccelerate;
					double sideDashRate = (fabs(finalInterceptModel.getDirSideDash()) > 90.0 ? ServerParam::i().backDashRate() - ((ServerParam::i().backDashRate() - ServerParam::i().sideDashRate())
					    * (1.0 - (fabs(finalInterceptModel.getDirSideDash()) - 90.0) / 90.0))
					    : ServerParam::i().sideDashRate() + ((1.0 - ServerParam::i().sideDashRate())
					    * (1.0 - fabs(finalInterceptModel.getDirSideDash()) / 90.0)));
					sideDashRate = min(max(0.0, sideDashRate), 1.0);

					dashAccelerate.setPolar(curDashPower * wm.self().dashRate() * sideDashRate, finalInterceptModel.getDirSideDash());
					bodyVel += dashAccelerate;
					if (bodyVel.r() > wm.self().playerType().playerSpeedMax())
						bodyVel.setPolar(wm.self().playerType().playerSpeedMax(), bodyVel.th());
					bodyPos += bodyVel;
					bodyVel *= wm.self().playerType().playerDecay();
				}
				double control = wm.self().kickableArea();
				if (wm.self().goalie())
				    control = wm.self().catchableArea();
				if (bodyPos.dist(interceptPoint) < control - 0.055)
					dashPower = curDashPower;
			}
		}
		

#ifdef INTERCEPT_LOG
		/*dlog.addText(Logger::INTERCEPT, "1>");
		dlog.flush();*/
		dlog.addText(Logger::INTERCEPT, "dashpower = %f, safety = %f stamina = %f", dashPower, wm.self().getSafetyDashPower(100), wm.self().stamina());
		/*dlog.addText(Logger::INTERCEPT, "1<");
		dlog.flush();*/
#endif
		agent->doDash(dashPower, finalInterceptModel.getDirSideDash());
	}
	else if (finalInterceptModel.isMinusDash() or finalInterceptModel.isPlusDash())
	{
#ifdef INTERCEPT_LOG
		dlog.addText(Logger::ROLE, "MINUS OR PLUS DASH INTERCEPT");
#endif
		AngleDeg bodyAngle = wm.self().body();
		Vector2D targetRel = interceptPoint - wm.self().pos();
		Vector2D myPoint = wm.self().playerType().inertiaPoint(wm.self().pos(), wm.self().vel(), 1);
		AngleDeg turnAngle = targetRel.th() - bodyAngle;
		double dashPower = 100;
		/*if (wm.self().stamina() < 1500 )
			dashPower = 25;*/
		/*dlog.addText(Logger::INTERCEPT, "2>");
		dlog.flush();*/
		dashPower = wm.self().getSafetyDashPower(dashPower);
		/*dlog.addText(Logger::INTERCEPT, "2<");
		dlog.flush();*/
		if (finalInterceptModel.isMinusDash())
			dashPower = -100;
#ifdef INTERCEPT_LOG
		dlog.addText(Logger::ROLE, "M: %d, %d", finalInterceptModel.isMinusDash(), finalInterceptModel.isPlusDash());
#endif
		if (targetRel.r() < 2) // for accurate intercept
		{
			double min = wm.self().kickableArea() - 0.055;
			if (wm.self().goalie())
			    min = wm.self().catchableArea() - 0.055;
			int cycle = finalInterceptModel.getBallReachCycle() - finalInterceptModel.getReachCycle();
			for (int curDashPower = -100; curDashPower <= 100; curDashPower += 25)
			{
				Vector2D bodyPos = wm.self().pos(), bodyVel = wm.self().vel();
				for (int i = 0; i < cycle; i++)
				{
					Vector2D dashAccelerate;
					dashAccelerate.setPolar(curDashPower * wm.self().dashRate(), bodyAngle);
					bodyVel += dashAccelerate;
					if (bodyVel.r() > wm.self().playerType().playerSpeedMax())
						bodyVel.setPolar(wm.self().playerType().playerSpeedMax(), bodyVel.th());
					bodyPos += bodyVel;
					bodyVel *= wm.self().playerType().playerDecay();
				}
				if (bodyPos.dist(interceptPoint) < min)
				{
					min = bodyPos.dist(interceptPoint);
					dashPower = curDashPower;
				}
			}
		}
		/*if (not finalInterceptModel.getReachCycle() and not SRP_EXE)
		{
		    dashPower = 0;
#ifdef INTERCEPT_LOG
		    dlog.addText(Logger::INTERCEPT, "Already reached!");
#endif		    
		}*/
#ifdef INTERCEPT_LOG
		/*dlog.addText(Logger::INTERCEPT, "3>");
		dlog.flush();*/
		dlog.addText(Logger::INTERCEPT, "dashpower = %f, safety = %f stamina = %f", dashPower, wm.self().getSafetyDashPower(100), wm.self().stamina());
		/*dlog.addText(Logger::INTERCEPT, "3<");
		dlog.flush();*/
#endif
		agent->doDash(dashPower);
	}
	else
	{
#ifdef INTERCEPT_LOG
		dlog.addText(Logger::ROLE, "NON SIDE DASH INTERCEPT");
#endif
		float turnThread = 180.0; // if already reached, turn is not required.
		float targetDist = interceptPoint.dist(wm.self().pos());
		double control = wm.self().kickableArea();
		if (wm.self().goalie())
			control = wm.self().catchableArea();
		if (control - 0.055 < targetDist)
			turnThread = AngleDeg::asin_deg((control - 0.055) / targetDist);
		turnThread = max(minTurnThread, turnThread);

		AngleDeg bodyAngle = wm.self().body();
		Vector2D targetRel = interceptPoint - wm.self().pos();
		Vector2D myPoint = wm.self().playerType().inertiaPoint(wm.self().pos(), wm.self().vel(), 1);
		AngleDeg turnAngle = targetRel.th() - bodyAngle;
		if (finalInterceptModel.isTurnMove())
		{
			double dashPower = 100;
			/*if (wm.self().stamina() < 1500 )
				dashPower = 25;*/
			/*dlog.addText(Logger::INTERCEPT, "4>");
			dlog.flush();*/
			if (wm.self().stamina() < 2500)
			{
				dashPower = wm.self().getSafetyDashPower(dashPower);
				/*dlog.addText(Logger::INTERCEPT, "4<");
				dlog.flush();*/
			}
			else if (targetRel.r() < 2) // for accurate intercept
			{
				double min = wm.self().kickableArea() - 0.055;
				if (wm.self().goalie())
				    min = wm.self().catchableArea() - 0.055;
				int cycle = finalInterceptModel.getBallReachCycle() - finalInterceptModel.getReachCycle();
				for (int curDashPower = -100; curDashPower <= 100; curDashPower += 25)
				{
					Vector2D bodyPos = wm.self().pos(), bodyVel = wm.self().vel();
					for (int i = 0; i < cycle; i++)
					{
						Vector2D dashAccelerate;
						dashAccelerate.setPolar(curDashPower * wm.self().dashRate(), bodyAngle);
						bodyVel += dashAccelerate;
						if (bodyVel.r() > wm.self().playerType().playerSpeedMax())
							bodyVel.setPolar(wm.self().playerType().playerSpeedMax(), bodyVel.th());
						bodyPos += bodyVel;
						bodyVel *= wm.self().playerType().playerDecay();
					}
					if (bodyPos.dist(interceptPoint) < min)
					{
						min = bodyPos.dist(interceptPoint);
						dashPower = curDashPower;
					}
				}
			}
#ifdef INTERCEPT_LOG
			dlog.addText(Logger::ROLE, "Dash Power:%.2f", dashPower);
			dlog.addText(Logger::INTERCEPT, "dashpower = %f, safety = %f stamina = %f", dashPower, wm.self().getSafetyDashPower(100), wm.self().stamina());
#endif
			agent->doDash(dashPower);
		}
		else
		{
#ifdef INTERCEPT_LOG
			dlog.addText(Logger::ROLE, "Turn Angle:%.2f", finalInterceptModel.getTurnAngle());
#endif
			agent->doTurn(finalInterceptModel.getTurnAngle(), false);
		}
	}
	if (finalInterceptModel.getBallReachCycle() > 3 or agent->world().ball().posCount() > 3)
	    agent->setNeckAction(new Neck_TurnToBall());
	else
	    if ( agent->world().gameMode().isPenaltyKickMode() and agent->world().getOpponentGoalie() )
		agent->setNeckAction(new Neck_TurnToBallAndPlayer( agent->world().getOpponentGoalie() ));           
	    else
		agent->setNeckAction(new Neck_ScanField());
}

int Intercept::getArriveCycle(const PlayerAgent *agent, const Vector2D & pos, int maxCycle, int updateCycle)
{
//	maxCycle += ballDelay;
	Vector2D selfPos = agent->world().self().inertiaPoint(updateCycle)
	, selfVel = agent->world().self().vel();
	for (int i = 0;i < updateCycle;i++)
	    selfVel *= agent->world().self().playerType().playerDecay();
	dlog.addText(Logger::INTERCEPT, "Pos(%f, %f), Vel(%f, %f), SelfPos(%f, %f), SelfVel(%f, %f)", agent->world().self().pos().x, agent->world().self().pos().y, agent->world().self().vel().x, agent->world().self().vel().y, selfPos.x, selfPos.y, selfVel.x, selfVel.y);
	dlog.addText(Logger::INTERCEPT, "getArriveCycle pos(%f, %f) - cycle(%d)", pos.x, pos.y, maxCycle);
	int turnCycle = 0, /*sideDashReachCycle = 0, */dashReachCycle = 0, minusDashReachCycle = 0, plusDashReachCycle = 0;
	int reachCycle = 0;
	const SelfObject & self = agent->world().self();
	isMinusDash = false;
	isSideDash = false;
	isPlusDash = false;

	// Calculate turnCycle for NonSideIntercept
	double distThread = self.goalie() ? self.catchableArea() - 0.055 : self.kickableArea() - 0.055; // kickableDist
	if (SRP_EXE)
		distThread += EXTRA_THREAD;
	// calculate turn moment threshold
	float turnThread = 180.0; // if already reached, turn is not required.
	float targetDist = pos.dist(selfPos);
	if (distThread < targetDist)
		turnThread = AngleDeg::asin_deg(distThread / targetDist);

	double bodyAngle = self.body().degree();
	Vector2D targetRel = pos - selfPos;
	Vector2D myPoint = self.playerType().inertiaPoint(selfPos, selfVel, 1);

	int maxTurnCycle = 99;
	for (int i = 0; i < maxTurnCycle; i++)
	{
		Vector2D bodyPos = selfPos, bodyVel = selfVel;
		Vector2D tmpPos(0, 0), tmpVel(0, 0);
		double bodyStamina = self.stamina();
		int turnInterceptTime = 0;
		double dashDist = 0;
		while (true)
		{
			if (pos.dist(bodyPos) < dashDist + distThread)
				break;

			// One cycle past
			if (turnInterceptTime > i - 1)
			{
				double dashPower = 100;
				/*if (bodyStamina < 1500)
					dashPower = 25;*/
				/*dlog.addText(Logger::INTERCEPT, "5>");
				dlog.flush();*/
				dashPower = self.getSafetyDashPower(dashPower, bodyStamina);
				/*dlog.addText(Logger::INTERCEPT, "5<");
				dlog.flush();*/
				Vector2D dashAccelerate;
				dashAccelerate.setPolar(dashPower * self.dashRate(), 0);
				tmpVel += dashAccelerate;
				if (tmpVel.r() > self.playerType().playerSpeedMax())
					tmpVel.setPolar(self.playerType().playerSpeedMax(), tmpVel.th().degree());
				tmpPos += tmpVel;
				tmpVel *= self.playerType().playerDecay();
				dashDist = tmpPos.x;
			}
			turnInterceptTime++;

			bodyPos += bodyVel;
			bodyVel *= self.playerType().playerDecay();
			bodyStamina += self.playerType().staminaIncMax();
		}

		turnCycle = 0;
		bodyAngle = self.body().degree();
		Vector2D r = selfVel;

		while (true)
		{
			Vector2D targetRel = pos - bodyPos;
			float turnAngle = AngleDeg(targetRel.th().degree() - bodyAngle).degree();

			if (abs(turnAngle) < turnThread)
				break;

			turnAngle = turnAngle * (1 + self.playerType().inertiaMoment() * r.r());
			if (abs(turnAngle) > ServerParam::i().maxMoment())
				turnAngle = AngleDeg((turnAngle / abs(turnAngle)) * ServerParam::i().maxMoment()).degree();
			if (turnCycle == 0)
			{
#ifdef INTERCEPT_LOG
				//if (i < 20) dlog.addText(Logger::INTERCEPT, "I Must Turn %f and Body = %f", turnAngle , bodyAngle + turnAngle);
#endif
				turnSimAngle = turnAngle;
			}
			turnAngle = turnAngle / (1 + self.playerType().inertiaMoment() * r.r());
			r *= self.playerType().playerDecay();
			bodyAngle += turnAngle;

			turnCycle++;
		}

		if (turnCycle <= i)
		{
			isTurnMove = (i == 0);
			break;
		}
	}
	dashReachCycle = turnCycle;
	Vector2D bodyPos = selfPos, bodyVel = selfVel;
#ifdef INTERCEPT_LOG
	//dlog.addText(Logger::INTERCEPT, "BoDy = (%f, %f), pOs = (%f, %f)", bodyPos.x, bodyPos.y, pos.x, pos.y);
#endif
	double bodyStamina = self.stamina();
	for (int i = 0; i < turnCycle; i++)
	{
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();
		bodyStamina += self.playerType().staminaIncMax();
	}
	if (bodyStamina > ServerParam::i().staminaMax())
		bodyStamina = ServerParam::i().staminaMax();
	// TODO: talk with pooria about this while
	while (dashReachCycle < maxCycle)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
			break;
		//dlog.addText(Logger::INTERCEPT, "dist = %f", bodyPos.dist(pos));
		double dashPower = 100;
		/*if (bodyStamina < 1500)
			dashPower = 25;*/
		/*dlog.addText(Logger::INTERCEPT, "6>");
		dlog.flush();*/
		dashPower = self.getSafetyDashPower(dashPower);
		/*dlog.addText(Logger::INTERCEPT, "6<");
		dlog.flush();*/
		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * self.dashRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > self.playerType().playerSpeedMax())
			bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();
		bodyStamina -= dashPower - self.playerType().staminaIncMax();
		dashReachCycle++;
	}
	if (dashReachCycle >= maxCycle)
		dashReachCycle = -1;

//	Minus Dash Calculate
	bodyPos = selfPos;
	bodyVel = selfVel;
	bodyStamina = self.stamina();

	minusDashReachCycle = 0;
	bool ahmaghane = false;
	bodyAngle = self.body().degree();
	while (minusDashReachCycle <= maxCycle/* && bodyStamina > 2500 */&& minusDashReachCycle < 7/*ServerParam::i().recoverDecThrValue()*/)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
		{
			if (minusDashReachCycle != 0)
			    ahmaghane = true;
			break;
		}

		double dashPower = self.getSafetyDashPower(-100,bodyStamina);

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * self.dashRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > self.playerType().playerSpeedMax())
			bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();
		bodyStamina -= -2 * dashPower - self.playerType().staminaIncMax();
		minusDashReachCycle++;
	}
	if (!ahmaghane)
		minusDashReachCycle = -1;

//	Plus Dash Calculate
	bodyPos = selfPos;
	bodyVel = selfVel;
	bodyStamina = self.stamina();

	plusDashReachCycle = 0;
	ahmaghane = false;
/*	if (bodyStamina < 2500)
	    dlog.addText(Logger::INTERCEPT, "CAUTION");*/
	while (plusDashReachCycle <= maxCycle /*&& bodyStamina > 2500*/ && plusDashReachCycle < 50/*ServerParam::i().recoverDecThrValue()*/)
	{
#ifdef INTERCEPT_LOG
		dlog.addText(Logger::INTERCEPT,"cycle  = %d, dist = %f, distThread = %f,stamina = %f, bodyPos(%f, %f)", plusDashReachCycle, bodyPos.dist(pos), distThread , bodyStamina, bodyPos.x, bodyPos.y);
#endif
		if (bodyPos.dist(pos) < distThread) // we reached the pos
		{
			if (plusDashReachCycle != 0)
			    ahmaghane = true;
			break;
		}

		double dashPower = self.getSafetyDashPower(100, bodyStamina);
		
		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * self.dashRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > self.playerType().playerSpeedMax())
			bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
#ifdef INTERCEPT_LOG
		dlog.addText(Logger::INTERCEPT,"bodyVel = %f, dashAccelerate = %f, Rate = %f", bodyVel.r(), dashAccelerate.r(), self.dashRate());
#endif
		bodyVel *= self.playerType().playerDecay();
		bodyStamina -= dashPower - self.playerType().staminaIncMax();
		plusDashReachCycle++;
	}
	if (!ahmaghane)
		plusDashReachCycle = -1;

//	Side Dash Calculation
/*	bodyPos = selfPos;
	bodyVel = selfVel;
	bodyStamina = self.stamina();
	sideDashReachCycle = 0;
	ahmaghane = false;
	while (sideDashReachCycle < min(maxCycle, maxSideDashCycle))
	{
		double sideDashAngle = ((pos - bodyPos).th().degree() - self.body()).degree();
		sideDashAngle = ServerParam::i().dashAngleStep() * rint( sideDashAngle / ServerParam::i().dashAngleStep() );
		double sideDashRate = (fabs(sideDashAngle) > 90.0 ? ServerParam::i().backDashRate() - ((ServerParam::i().backDashRate() - ServerParam::i().sideDashRate())
				* (1.0 - (fabs(sideDashAngle) - 90.0) / 90.0))
				: ServerParam::i().sideDashRate() + ((1.0 - ServerParam::i().sideDashRate())
				* (1.0 - fabs(sideDashAngle) / 90.0)));
		sideDashRate = min(max(0.0, sideDashRate), 1.0);
		//dlog.addText(Logger::ROLE, "Side = %f", bodyPos.dist(pos));
		if (bodyPos.dist(pos) < distThread) // we reached the pos
		{
			ahmaghane = true;
			break;
		}
		
		double dashPower = 100;
		if (bodyStamina < 2000/ServerParam::i().recoverDecThrValue()/)
			dashPower = 25;
		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * self.dashRate() * sideDashRate, sideDashAngle + self.body().degree());
		bodyVel += dashAccelerate;
		if (bodyVel.r() > self.playerType().playerSpeedMax())
			bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();

		double powerNeed = (abs(sideDashAngle) > 90.0 ? dashPower * -2.0 : dashPower);
		powerNeed = min(powerNeed, bodyStamina + self.playerType().extraStamina());
		bodyStamina = max(0.0, bodyStamina - powerNeed) + self.playerType().staminaIncMax();

		sideDashReachCycle++;
	}
	if (sideDashReachCycle >= min(maxCycle, maxSideDashCycle))
		sideDashReachCycle = -1;

	if (!ahmaghane)
		sideDashReachCycle = -1;*/
		
	reachCycle = dashReachCycle;
#ifdef INTERCEPT_LOG
	dlog.addText(Logger::INTERCEPT, "Turn = %d", turnCycle);
	dlog.addText(Logger::INTERCEPT, "Reach = %d", dashReachCycle);
	dlog.addText(Logger::INTERCEPT, "Minus = %d", minusDashReachCycle);
	//dlog.addText(Logger::INTERCEPT, "Side = %d", sideDashReachCycle);
#endif
	if (minusDashReachCycle != -1 && (minusDashReachCycle < reachCycle or reachCycle == -1))
	{
		isMinusDash = true;
		reachCycle = minusDashReachCycle;
	}
#ifdef INTERCEPT_LOG
	dlog.addText(Logger::INTERCEPT, "Plus = %d", plusDashReachCycle);
#endif
	if (plusDashReachCycle != -1 && ((plusDashReachCycle < reachCycle or reachCycle == -1) /*or (SRP_EXE and plusDashReachCycle <= reachCycle + 2)*/))
	{
		/*if (SRP_EXE)
		    dlog.addText(Logger::ROLE, "plus = %d, reach = %d", plusDashReachCycle, reachCycle);*/
		isMinusDash = false;
		isPlusDash = true;
		reachCycle = plusDashReachCycle;
	}
	/*if (maxCycle < 9)
	{
	    dlog.addText(Logger::ROLE, "SSSALAM");
	}*/
	/*if (sideDashReachCycle != -1 and (reachCycle > sideDashReachCycle && sideDashReachCycle != -1) and (not SRP_EXE))
	{
		isSideDash = true;
		dirSideDash = ((pos - selfPos).th() - self.body()).degree();
		reachCycle = sideDashReachCycle;
		dlog.addText(Logger::INTERCEPT, "SiDe %d", sideDashReachCycle);
	}
	else*/
		isSideDash = false;

/*
if (maxCycle < 15)
{
dlog.addText(Logger::ROLE, "RRRRRRRRRRRRRRRRRRRRRRRRRReachCycle: %d", reachCycle);
}
*/	return reachCycle;
}

int Intercept::getArriveCycle(const PlayerAgent *agent, const Vector2D & pos, int maxCycle, rcsc::PlayerObject * player)
{
//	maxCycle += ballDelay;
	int turnCycle = 0, /*sideDashReachCycle = -1,*/ dashReachCycle = 0, minusDashReachCycle = 0, plusDashReachCycle = 0;
	int reachCycle = 0;
	//const SelfObject & self = agent->world().self();
	isMinusDash = false;
	isSideDash = false;
	isPlusDash = false;

	// Calculate turnCycle for NonSideIntercept
	double distThread = player->playerTypePtr()->kickableArea() - 0.055; // kickableDist
	// calculate turn moment threshold
	float turnThread = 180.0; // if already reached, turn is not required.
	float targetDist = pos.dist(player->pos());
	if (distThread < targetDist)
		turnThread = AngleDeg::asin_deg(distThread / targetDist);

	double bodyAngle = player->body().degree();
	Vector2D targetRel = pos - player->pos();
	Vector2D myPoint = player->playerTypePtr()->inertiaPoint(player->pos(), player->vel(), 1);

	int maxTurnCycle = 99;
	for (int i = 0; i < maxTurnCycle; i++)
	{
		Vector2D bodyPos = player->pos(), bodyVel = player->vel();
		Vector2D tmpPos(0, 0), tmpVel(0, 0);
		double bodyStamina = 8000;
		int turnInterceptTime = 0;
		double dashDist = 0;
		while (true)
		{
			if (pos.dist(bodyPos) < dashDist + distThread)
				break;

			// One cycle past
			if (turnInterceptTime > i - 1)
			{
				double dashPower = 100;
				if (bodyStamina < 2500)
					dashPower = 25;
				//dashPower = self.getSafetyDashPower(dashPower);
				
				Vector2D dashAccelerate;
				dashAccelerate.setPolar(dashPower * player->playerTypePtr()->dashPowerRate(), 0);
				tmpVel += dashAccelerate;
				if (tmpVel.r() > player->playerTypePtr()->playerSpeedMax())
					tmpVel.setPolar(player->playerTypePtr()->playerSpeedMax(), tmpVel.th().degree());
				tmpPos += tmpVel;
				tmpVel *= player->playerTypePtr()->playerDecay();
				dashDist = tmpPos.x;
			}
			turnInterceptTime++;

			bodyPos += bodyVel;
			bodyVel *= player->playerTypePtr()->playerDecay();
			bodyStamina += player->playerTypePtr()->staminaIncMax();
		}

		turnCycle = 0;
		bodyAngle = player->body().degree();
		Vector2D r = player->vel();

		while (true)
		{
			Vector2D targetRel = pos - bodyPos;
			float turnAngle = AngleDeg(targetRel.th().degree() - bodyAngle).degree();

			if (abs(turnAngle) < turnThread)
				break;

			turnAngle = turnAngle * (1 + player->playerTypePtr()->inertiaMoment() * r.r());
			if (abs(turnAngle) > ServerParam::i().maxMoment())
				turnAngle = AngleDeg((turnAngle / abs(turnAngle)) * ServerParam::i().maxMoment()).degree();
			if (turnCycle == 0)
				turnSimAngle = turnAngle;
			turnAngle = turnAngle / (1 + player->playerTypePtr()->inertiaMoment() * r.r());
			r *= player->playerTypePtr()->playerDecay();
			bodyAngle += turnAngle;

			turnCycle++;
		}

		if (turnCycle <= i)
		{
			isTurnMove = (i == 0);
			break;
		}
	}

	dashReachCycle = turnCycle;
	Vector2D bodyPos = player->pos(), bodyVel = player->vel();
	double bodyStamina = 8000;
	for (int i = 0; i < turnCycle; i++)
	{
		bodyPos += bodyVel;
		bodyVel *= player->playerTypePtr()->playerDecay();
		bodyStamina += player->playerTypePtr()->staminaIncMax();
	}
	if (bodyStamina > ServerParam::i().staminaMax())
		bodyStamina = ServerParam::i().staminaMax();

	while (dashReachCycle < maxCycle)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
			break;

		double dashPower = 100;
		if (bodyStamina < 2500)
			dashPower = 25;
		//dashPower = self.getSafetyDashPower(dashPower);

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * player->playerTypePtr()->dashPowerRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > player->playerTypePtr()->playerSpeedMax())
			bodyVel.setPolar(player->playerTypePtr()->playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= player->playerTypePtr()->playerDecay();
		bodyStamina -= dashPower - player->playerTypePtr()->staminaIncMax();
		dashReachCycle++;
	}
	if (dashReachCycle >= maxCycle)
		dashReachCycle = -1;

//	Minus Dash Calculate
	bodyPos = player->pos();
	bodyVel = player->vel();
	bodyStamina = 8000;

	minusDashReachCycle = 0;
	bool ahmaghane = false;
	bodyAngle = player->body().degree();
	while (minusDashReachCycle < maxCycle && bodyStamina > 2500 && minusDashReachCycle < 7/*ServerParam::i().recoverDecThrValue()*/)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
		{
			ahmaghane = true;
			break;
		}

		double dashPower = -100;

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * player->playerTypePtr()->dashPowerRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > player->playerTypePtr()->playerSpeedMax())
			bodyVel.setPolar(player->playerTypePtr()->playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= player->playerTypePtr()->playerDecay();
		bodyStamina -= -2 * dashPower - player->playerTypePtr()->staminaIncMax();
		minusDashReachCycle++;
	}
	if (!ahmaghane)
		minusDashReachCycle = -1;

//	Plus Dash Calculate
	bodyPos = player->pos();
	bodyVel = player->vel();
	bodyStamina = 8000;

	plusDashReachCycle = 0;
	ahmaghane = false;
	while (plusDashReachCycle < maxCycle && bodyStamina > 2500 && plusDashReachCycle < 50/*ServerParam::i().recoverDecThrValue()*/)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
		{
			ahmaghane = true;
			break;
		}

		double dashPower = 100;

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * player->playerTypePtr()->dashPowerRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > player->playerTypePtr()->playerSpeedMax())
			bodyVel.setPolar(player->playerTypePtr()->playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= player->playerTypePtr()->playerDecay();
		bodyStamina -= dashPower - player->playerTypePtr()->staminaIncMax();
		plusDashReachCycle++;
	}
	if (!ahmaghane)
		plusDashReachCycle = -1;

//	Side Dash Calculation
/*	bodyPos = player->pos();
	bodyVel = player->vel();
	bodyStamina = player->stamina();
	sideDashReachCycle = 0;
	while (sideDashReachCycle < min(maxCycle, maxSideDashCycle))
	{
		double sideDashAngle = ((pos - bodyPos).th().degree() - player->body()).degree();
		double sideDashRate = (fabs(sideDashAngle) > 90.0 ? ServerParam::i().backDashRate() - ((ServerParam::i().backDashRate() - ServerParam::i().sideDashRate())
				* (1.0 - (fabs(sideDashAngle) - 90.0) / 90.0))
				: ServerParam::i().sideDashRate() + ((1.0 - ServerParam::i().sideDashRate())
				* (1.0 - fabs(sideDashAngle) / 90.0)));
		sideDashRate = min(max(0.0, sideDashRate), 1.0);

		if (bodyPos.dist(pos) < distThread) // we reached the pos
			break;

		double dashPower = 100;
		if (bodyStamina < 2000/ServerParam::i().recoverDecThrValue()/)
			dashPower = 25;
		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * player->playerTypePtr()->dashPowerRate() * sideDashRate, sideDashAngle + player->body().degree());
		bodyVel += dashAccelerate;
		if (bodyVel.r() > player->playerTypePtr()->playerSpeedMax())
			bodyVel.setPolar(player->playerTypePtr()->playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= player->playerTypePtr()->playerDecay();

		double powerNeed = (abs(sideDashAngle) > 90.0 ? dashPower * -2.0 : dashPower);
		powerNeed = min(powerNeed, bodyStamina + player->playerTypePtr()->extraStamina());
		bodyStamina = max(0.0, bodyStamina - powerNeed) + player->playerTypePtr()->staminaIncMax();

		sideDashReachCycle++;
	}
	if (sideDashReachCycle >= min(maxCycle, maxSideDashCycle))
		sideDashReachCycle = -1;
*/
	reachCycle = dashReachCycle;
//#ifdef INTERCEPT_LOG
	dlog.addText(Logger::INTERCEPT, "reach = %d", dashReachCycle);
	dlog.addText(Logger::INTERCEPT, "Minus = %d", minusDashReachCycle);
//#endif
	if (minusDashReachCycle != -1 && (minusDashReachCycle < reachCycle or reachCycle == -1))
	{
		isMinusDash = true;
		reachCycle = minusDashReachCycle;
	}
//#ifdef INTERCEPT_LOG
	dlog.addText(Logger::INTERCEPT, "Plus = %d", plusDashReachCycle);
//#endif
	if ((plusDashReachCycle != -1 && (plusDashReachCycle < reachCycle or reachCycle == -1)) or (SRP_EXE and plusDashReachCycle <= reachCycle + 2))
	{
		isMinusDash = false;
		isPlusDash = true;
		reachCycle = plusDashReachCycle;
/*if (maxCycle < 9)
{
dlog.addText(Logger::ROLE, "SSSALAM");
}
*/	}
/*	if (reachCycle > sideDashReachCycle && sideDashReachCycle != -1)
	{
		isSideDash = true;
		dirSideDash = ((pos - player->pos()).th() - player->body()).degree();
		reachCycle = sideDashReachCycle;
	}
	else
		isSideDash = false;
*/
/*
if (maxCycle < 15)
{
dlog.addText(Logger::ROLE, "RRRRRRRRRRRRRRRRRRRRRRRRRReachCycle: %d", reachCycle);
}
*/	return reachCycle;
}

