// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef __INTERCEPT_H
#define __INTERCEPT_H

#include <rcsc/geom/vector_2d.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/ball_object.h>

class Intercept
{
private:
	float minTurnThread;
	float dirSideDash;
	bool isSideDash;
	bool isTurnMove;
	bool isMinusDash;
	bool isPlusDash;
	int maxSideDashCycle;
	int getArriveCycle(const rcsc::PlayerAgent *agent, const rcsc::Vector2D & pos, int maxCycle, int updateCycle);
	int getArriveCycle(const rcsc::PlayerAgent *agent, const rcsc::Vector2D & pos, int maxCycle, rcsc::PlayerObject * player);
	double turnSimAngle;
	rcsc::BallObject myBall;
	int ballDelay;
	bool SRP, SRP_EXE;

public:
	Intercept(bool _SRP = false, bool _SRP_EXE = false)
	{
		SRP = _SRP;
		SRP_EXE = _SRP_EXE;
		ballDelay = 0;
	}

	~Intercept()
	{
	}

	void execute(rcsc::PlayerAgent * agent, bool useMyBall = false, float moveError = 1.f);

	rcsc::BallObject& ball()
	{
		return myBall;
	}
	
	void setBallDelay(int delay)
	{
		ballDelay = delay;
	}

	class InterceptModel
	{
	private:
		double turnAngle;
		bool turnMove;
		bool sideDash;
		bool minusDash;
		bool plusDash;
		double dirSideDash;
		int reachCycle;
		int ballReachCycle;
		rcsc::Vector2D ballPos;

	public:
		InterceptModel()
		{
			sideDash = false;
			reachCycle = -1;
			ballReachCycle = -1;
		}

		InterceptModel(int reachCycle, int ballReachCycle, rcsc::Vector2D ballPos, bool sideDash, double dirSideDash, bool turnMove, bool minusDash,
				bool plusDash, double turnAngle)
		{
			this->sideDash = sideDash;
			this->dirSideDash = dirSideDash;
			this->reachCycle = reachCycle;
			this->ballReachCycle = ballReachCycle;
			this->ballPos = ballPos;
			this->turnMove = turnMove;
			this->turnAngle = turnAngle;
			this->minusDash = minusDash;
			this->plusDash = plusDash;
		}

		bool isTurnMove()
		{
			return turnMove;
		}

		double getTurnAngle()
		{
			return turnAngle;
		}

		bool isSideDash()
		{
			return sideDash;
		}

		bool isMinusDash()
		{
			return minusDash;
		}

		bool isPlusDash()
		{
			return plusDash;
		}

		double getDirSideDash()
		{
			return dirSideDash;
		}

		int getReachCycle()
		{
			return reachCycle;
		}

		int getBallReachCycle()
		{
			return ballReachCycle;
		}

		rcsc::Vector2D getBallPos()
		{
			return ballPos;
		}
	};

	public:
		InterceptModel finalIntercept(rcsc::PlayerAgent* agent, bool useMyBall = false, float moveError = 1.f, int maxCycles = 500, int playerDelay = 0, int updateCycle = 0);
		//InterceptModel finalIntercept(rcsc::PlayerAgent* agent, rcsc::PlayerObject * player, bool useMyBall = false, float moveError = 1.f, int maxCycles = 500, int playerDelay = -1);
};

#endif // __INTERCEPT_H
