// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "PositioningConfig.h"
#include "Basics.h"

#include <string>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include <rcsc/common/logger.h>


using namespace std;
using namespace rcsc;
using namespace Basics;

PositioningConfig::PositioningConfig(Expression & passConf)
{
	offsideLineWeight = passConf["OffsideLineWeight"][0][1].AsFloat();
	ballXWeight = passConf["BallXWeight"][0][1].AsFloat();
	minX = passConf["MinX"][0][1].AsFloat();
	maxX = passConf["MaxX"][0][1].AsFloat();
	midLineDif = passConf["MidLineDif"][0][1].AsFloat();
	QBLineDif = passConf["QBLineDif"][0][1].AsFloat();
	for ( int i = 0; i < 5; i++ )
	{
		Expression area = passConf["Area"][i];
		areas[i].minY = area["MinY"][0][1].AsFloat();
		areas[i].maxY = area["MaxY"][0][1].AsFloat();
		areas[i].minPosY = area["MinPosY"][0][1].AsFloat();
		areas[i].maxPosY = area["MaxPosY"][0][1].AsFloat();
	}
}

Vector2D PositioningConfig::getAreaPosition(int area , Vector2D ballPos, float offsideLine , int ballArea , bool isMid)
{
	Vector2D pos;
	pos.x = ( offsideLineWeight * offsideLine )+( ballXWeight * ballPos.x )/( offsideLineWeight + ballXWeight );

	pos.x = min(offsideLine - 2.0, pos.x);

	if ( isMid )
		pos.x -= midLineDif;
	float downDelta = - areas[ballArea].minY + ballPos.y;
	float completeDelta = areas[ballArea].maxY - areas[ballArea].minY;
	float areaDelta = areas[area].maxPosY - areas[area].minPosY;
	float distToMin = ( areaDelta * downDelta ) / completeDelta;
	pos.y = areas[area].minPosY + distToMin;

	return pos;
}

Vector2D PositioningConfig::getDefensivePositioning(const rcsc::WorldModel *wm)
{	
//	if (isInPlayers(wm->self().unum(), "9AB"))
//		return Offense::defensivePositioning(wm->self().unum(), wm);
//	else
//		return Middle::defensivePositioning(wm);

}

Vector2D PositioningConfig::getOffensivePositioning(const rcsc::WorldModel *wm)
{
	float offsideLine = wm->offsideLineX();
	int reachCycle = wm->interceptTable()->teammateReachCycle();
	Vector2D ballPos = Basics::getBallPosAfterCycles(wm->ball().pos(),wm->ball().vel().r(),wm->ball().vel().r(),(float)reachCycle,ServerParam::i().ballDecay());
	if ( reachCycle >= 8 )
		ballPos = wm->ball().pos();
	int receiver = 50;
	if ( wm->interceptTable()->fastestTeammate() )
		receiver = wm->interceptTable()->fastestTeammate()->unum();
//	dlog.addText ( Logger::PASS , "Fastest	=	%d" , receiver );
	if ( ballPos.y > 34 )
		ballPos.y = 34;
	if ( ballPos.y < -34 )
		ballPos.y = -34;
	if ( ( ballPos.x > maxX ) or ( ballPos.x < minX ) )
		return Vector2D(-1000,-1000);
	int ballArea = 0;
	while ( ( ballPos.y > areas[ballArea].minY ) and ( ballArea < 5 ) )
		ballArea++;
	ballArea--;
	bool isMiddle = Basics::isInPlayers( wm->self().unum() , "678" );
	if ( wm->self().unum() == 6 )
	{
		Vector2D qbPos = getAreaPosition( 2 , ballPos, offsideLine , ballArea);
		qbPos.x -= QBLineDif;
		qbPos.y = wm->ball().pos().y / 3;
		return qbPos;
	}
	if ( wm->self().unum() == receiver )
		return getAreaPosition( ballArea , ballPos, offsideLine , ballArea , isMiddle);
	string players = "97B8A";
	int playerPointer , areaPointer;
	playerPointer = areaPointer = 0;
	string poses = "QQQQQ";
	while ( ( playerPointer < 5 ) and ( areaPointer < 5 ) )
	{
		string currentPlayer = " ";
		currentPlayer[0] = players[playerPointer];
		if ( ( areaPointer == ballArea ) and ( Basics::isInPlayers ( receiver , players ) ) )
		{
			areaPointer++;
			continue;
		}
		if ( Basics::isInPlayers( receiver , currentPlayer) )
		{
			playerPointer++;
			continue;
		}
		poses[areaPointer] = players[playerPointer];
		areaPointer++;
		playerPointer++;
	}
//	dlog.addText ( Logger::PASS , "Players	=	%s" , poses.c_str());
	for ( int i = 0; i < 5; i++ )
	{
		string thisPart = " ";
		thisPart[0] = poses[i];
		if ( thisPart == "Q" )
			continue;
		if ( Basics::isInPlayers( wm->self().unum() , thisPart ) )
		{
			Vector2D myPos = getAreaPosition( i , ballPos, offsideLine , ballArea , isMiddle);
			dlog.addText (Logger::PASS,  "myPos	=	(%f , %f)" , myPos.x , myPos.y );

			if (wm->self().pos().x > wm->offsideLineX() - 2)
			{
				myPos.x = wm->offsideLineX() - 10;
				myPos.y = wm->self().pos().y;
			}

			return myPos;
		}
	}
	return Vector2D(-1000,-1000);
}

Vector2D PositioningConfig::getOffensivePositioning(float offsideLine , const rcsc::BallObject & ball , int unum )
{
	Vector2D ballPos = ball.pos();
	if ( ( ballPos.x > maxX ) or ( ballPos.x < minX ) )
		return Vector2D(-1000,-1000);
	int ballArea = 0;
	while ( ( ballPos.y > areas[ballArea].minY ) and ( ballArea < 5 ) )
		ballArea++;
	ballArea--;
	bool isMiddle = Basics::isInPlayers( unum , "678" );
	if ( unum == 6 )
	{
		Vector2D qbPos = getAreaPosition( 2 , ballPos, offsideLine , ballArea);
		qbPos.x -= QBLineDif;
		return qbPos;
	}
	string poses = "97B8A";
	for ( int i = 0; i < 5; i++ )
	{
		string thisPart = " ";
		thisPart[0] = poses[i];
		if ( thisPart == "Q" )
			continue;
		if ( Basics::isInPlayers( unum , thisPart ) )
		{
			Vector2D myPos = getAreaPosition( i , ballPos, offsideLine , ballArea , isMiddle);
			dlog.addText (Logger::PASS,  "myPos	=	(%f , %f)" , myPos.x , myPos.y );
			return myPos;
		}
	}
	return Vector2D(-1000, -1000);
}

Vector2D PositioningConfig::getHomePos(const rcsc::WorldModel *wm)
{
	return getOffensivePositioning(wm);
}
