// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <SideDashGoToPoint.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <cmath>
#include <Share.h>

using namespace std;
using namespace rcsc;


SideDashGoToPoint::SideDashGoToPoint(const Vector2D& target, double error, double power):
	mTarget( target ),
	mError( error ),
	mPower( power )
{
}

bool SideDashGoToPoint::execute(rcsc::PlayerAgent * agent)
{
	if ( agent->world().self().pos().dist( mTarget ) < mError )
	{
		return false;
		dlog.addText(Logger::PASS, "SideDashGoToPoint::execute() --> In Position ( )");
	}

	const double _sideDashRate = 0.25;
	const double _backDashRate = 0.50;

	const WorldModel* wm = &agent->world();

	// Checking dash angle
	double sideDashAngle = ((mTarget - wm->self().pos()).th().degree() - wm->self().body()).degree();
	if (abs(sideDashAngle - Share::lastSideDashAngle) < 15.0 and Share::lastSideDashCycle == wm->time().cycle() - 1)
		sideDashAngle = Share::lastSideDashAngle;
	else
	{
		Share::lastSideDashAngle = sideDashAngle;
		Share::lastSideDashCycle = wm->time().cycle();
	}

	// Checking dash power
	double sideDashRate = (fabs(sideDashAngle) > 90.0 ? _backDashRate - ((_backDashRate - _sideDashRate)
			* (1.0 - (fabs(sideDashAngle) - 90.0) / 90.0))
			: _sideDashRate + ((1.0 - _sideDashRate)
			* (1.0 - fabs(sideDashAngle) / 90.0)));
	sideDashRate = min(max(0.0, sideDashRate), 1.0);


	double nearestDist = 1000;
	double _dashPower = 0;

	for (double dashPower = +100 ; dashPower >= -100 ; dashPower -= 10)
	{
		int bodyStamina = wm->self().stamina();
		Vector2D bodyVel = wm->self().vel();
		Vector2D bodyPos = wm->self().pos();

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * wm->self().dashRate() * sideDashRate, sideDashAngle + wm->self().body().degree());
		bodyVel += dashAccelerate;
		if (bodyVel.r() > wm->self().playerType().playerSpeedMax())
			bodyVel.setPolar(wm->self().playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= wm->self().playerType().playerDecay();

		double powerNeed = (abs(sideDashAngle) > 90.0 ? dashPower * -2.0 : dashPower);
		powerNeed = min(powerNeed, bodyStamina + wm->self().playerType().extraStamina());
		bodyStamina = max(0.0, bodyStamina - powerNeed) + wm->self().playerType().staminaIncMax();

		if (bodyPos.dist(mTarget) < nearestDist)
		{
			nearestDist = bodyPos.dist(mTarget);
			_dashPower = dashPower;
		}
	}

	if (wm->self().stamina() < 2000/*ServerParam::i().recoverDecThrValue()*/)
		_dashPower = 25;

	mPower = _dashPower;
	return agent->doDash( mPower, sideDashAngle );
}
