// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <Tackle.h>

#include <rcsc/common/logger.h>

#include <bhv_basic_tackle.h>
#include <bhv_danger_area_tackle.h>
#include <bhv_deflecting_tackle.h>

#include <FastIC.h>
#include <Intercept.h>

using namespace rcsc;

bool Tackle::execute(rcsc::PlayerAgent* agent)
{
	Vector2D selfReachPoint = Intercept().finalIntercept(agent).getBallPos();
	Vector2D fastestTeammateReachPoint = agent->world().ball().inertiaPoint(FastIC::getFastestTeammateReachCycle());

	dlog.addText( Logger::CLEAR, "selfReachPoint is %f, %f", selfReachPoint.x, selfReachPoint.y);
	dlog.addText( Logger::CLEAR, "self().tackleProbability() == %f", agent->world().self().tackleProbability());
	if (FastIC::isSelfFastestPlayer() and selfReachPoint.absX() < 52.0 and selfReachPoint.absY() < 34.0)
	{
		dlog.addText(Logger::CLEAR, "TACKLE: self is fastest [FALSE]");
		return false;
	}

	if (Bhv_DangerAreaTackle().execute(agent))
	{
		dlog.addText(Logger::CLEAR, "TACKLE: danger area tackle [TRUE]");
		return true;
	}
	else if (Bhv_BasicTackle(0.85, 50).execute(agent))
	{
		dlog.addText(Logger::CLEAR, "TACKLE: basic tackle [TRUE]");
		return true;
	}

	Line2D line(agent->world().ball().pos(), selfReachPoint);
	Vector2D bottom_left(-52.5, 34), top_left(-52.5, -34);
	if (selfReachPoint == Vector2D(-70, -70))
	{
	    //dlog.addText(Logger::BLOCK, "nemiresam be toop %f", agent->world().ball().vel().th().degree());
	    line.assign(agent->world().ball().pos(), agent->world().ball().vel().th());
	}
	Vector2D noghte = line.intersection( Line2D(bottom_left, top_left) );

	dlog.addText( Logger::BLOCK, "noghte is %f, %f", noghte.x, noghte.y);

	if
	(
		(selfReachPoint.x <  -52 and noghte.absY() < 8 and FastIC::isSelfFastestTeamMate())
		or
		(fastestTeammateReachPoint.x < -52 and noghte.absY() < 8 )
	)
	{
		dlog.addText( Logger::CLEAR, "Tackle nazanam rafte too goal");
		double min_prob = 0.1;
		if (Bhv_DangerAreaTackle(min_prob).execute(agent) )
		{
			dlog.addText( Logger::CLEAR, "Tackle: DangerAreaTackle  [TRUE]");
			return true;
		}
		else if (Bhv_BasicTackle(min_prob, 50).execute(agent))
		{
			dlog.addText( Logger::CLEAR, "Tackle: BasicTackle [TRUE]");
			return true;
		}
		else if (Bhv_DeflectingTackle(min_prob, true).execute(agent))
		{
			dlog.addText( Logger::CLEAR, "Tackle: DeflectingTackle [TRUE]");
			return true;
		}
	}

	dlog.addText(Logger::CLEAR, "TACKLE: no tackle available [FALSE]");
	return false;
}
