// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <HelliCoach.h>

#include <Basics.h>

#include <rcsc/coach/coach_audio_sensor.h>

#include <rcsc/coach/coach_command.h>
#include <rcsc/coach/coach_config.h>
#include <rcsc/common/basic_client.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/player_param.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/player_type.h>
#include <rcsc/common/audio_memory.h>
#include <rcsc/common/say_message_parser.h>
#include <rcsc/param/param_map.h>
#include <rcsc/param/cmd_line_parser.h>

#include <cstdio>
#include <vector>
#include <algorithm>
#include <sstream>
#include <iostream>
#include <functional>

#include "team_logo.xpm"

#include <WMatching.h>

using namespace std;
using namespace rcsc;
using namespace Basics;

std::string HelliCoach::centralSayMsg = "M000000000";

/////////////////////////////////////////////////////////

struct TypeStaminaIncComp
    : public std::binary_function< const rcsc::PlayerType *,
                                   const rcsc::PlayerType *,
                                   bool > {

    result_type operator()( first_argument_type lhs,
                            second_argument_type rhs ) const
      {
          return lhs->staminaIncMax() < rhs->staminaIncMax();
      }

};

/////////////////////////////////////////////////////////

struct RealSpeedMaxCmp
    : public std::binary_function< const rcsc::PlayerType *,
                                   const rcsc::PlayerType *,
                                   bool > {

    result_type operator()( first_argument_type lhs,
                            second_argument_type rhs ) const
      {
          if ( std::fabs( lhs->realSpeedMax() - rhs->realSpeedMax() ) < 0.005 )
          {
              return lhs->cyclesToReachMaxSpeed() < rhs->cyclesToReachMaxSpeed();
          }

          return lhs->realSpeedMax() > rhs->realSpeedMax();
      }

};

/////////////////////////////////////////////////////////

struct MaxSpeedReachStepCmp
    : public std::binary_function< const rcsc::PlayerType *,
                                   const rcsc::PlayerType *,
                                   bool > {

    result_type operator()( first_argument_type lhs,
                            second_argument_type rhs ) const
      {
          return lhs->cyclesToReachMaxSpeed() < rhs->cyclesToReachMaxSpeed();
      }

};

/*-------------------------------------------------------------------*/
/*!

*/
HelliCoach::HelliCoach()
    : rcsc::CoachAgent()
    , M_substitute_count( 0 )
{
    //
    // register audio memory & say message parsers
    //
    typedef boost::shared_ptr< rcsc::SayMessageParser > SMP;

    boost::shared_ptr< rcsc::AudioMemory > audio_memory( new rcsc::AudioMemory );

    M_worldmodel.setAudioMemory( audio_memory );

    addSayMessageParser( SMP( new rcsc::BallMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::PassMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::InterceptMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::GoalieMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::OffsideLineMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::DefenseLineMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::WaitRequestMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::PassRequestMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::DribbleMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::BallGoalieMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::OnePlayerMessageParser( audio_memory ) ) );
    addSayMessageParser( SMP( new rcsc::BallPlayerMessageParser( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 9 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 8 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 7 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 6 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 5 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 4 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 3 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 2 >( audio_memory ) ) );
//     addSayMessageParser( SMP( new rcsc::FreeMessageParser< 1 >( audio_memory ) ) );

    //
    //
    //

    for ( int i = 0; i < 11; ++i )
    {
        // init map values
        M_assigned_player_type_id[i] = rcsc::Hetero_Default;
    }

    for ( int i = 0; i < 11; ++i )
    {
        M_opponent_player_types[i] = rcsc::Hetero_Default;
    }

	for (int i = 1 ; i < 12 ; i++)
		oppPlayers[i] = true;
}

/*-------------------------------------------------------------------*/
/*!

*/
HelliCoach::~HelliCoach()
{

}

/*-------------------------------------------------------------------*/
/*!

 */
bool
HelliCoach::initImpl( rcsc::CmdLineParser & cmd_parser )
{
    //rcsc::ParamMap my_params;
    // cmd_parser.parse( my_params );

    if ( ! rcsc::CoachAgent::initImpl( cmd_parser ) )
    {
        //my_params.printHelp( std::cout );
        M_client->setServerAlive( false );
        return false;
    }

    if ( cmd_parser.failed() )
    {
        std::cerr << "coach: ***WARNING*** detected invalid options: ";
        cmd_parser.print( std::cerr );
        std::cerr << std::endl;
    }

    //////////////////////////////////////////////////////////////////
    // Add your code here.
    //////////////////////////////////////////////////////////////////

    if ( config().useTeamGraphic() )
    {
        M_team_graphic.createXpmTiles( team_logo_xpm );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
HelliCoach::actionImpl()
{
    if ( world().time().cycle() == 0
         && config().useTeamGraphic()
         && M_team_graphic.tiles().size() != teamGraphicOKSet().size() )
    {
        sendTeamGraphic();
    }

	for (list<HearMessage>::const_iterator it = audioSensor().teammateMessages().begin() ; it != audioSensor().teammateMessages().end() ; it++)
		if (it->str_.c_str()[0] == 'M' and it->str_.size() == 10 and (it->unum_ == 2 or it->unum_ == 1))
			centralSayMsg = it->str_;

//	cout << "Central Say : " << centralSayMsg << endl;


	static long sayCycle = -300;
	static bool nonPlayOnSay = false;
	static bool deathOppSay = false;
	static long nonPlayOnStartCycle = -1;
	static bool saidInThisAttack = false;

	if (world().gameMode().type() == GameMode::PlayOn)
	{
		nonPlayOnSay = false;
		deathOppSay = false;
		nonPlayOnStartCycle = -1;
	}

	if (world().gameMode().type() != GameMode::PlayOn and nonPlayOnStartCycle == -1)
		nonPlayOnStartCycle = world().time().cycle();

	// Info message say -----> PlayOn
	if ( world().time().cycle() > 0 and
		(world().ball().pos().x < -5.00 and world().ball().pos().x > -15.00) and
		!saidInThisAttack and (world().time().cycle() - sayCycle) > 300 and
		world().gameMode().type() == GameMode::PlayOn)
	{
//		generateMarkTableWithoutBackDefender();
		generatePlayOnMarkTable();
		if (sayMarkers())
		{
			sayCycle = world().time().cycle();
			saidInThisAttack = true;
			cout << "Cycle:\t" << world().time().cycle() << "\tsaying markers [INFO], PlayOn\n" ;
		}
	}
	else if (world().ball().pos().x >= -25.00)
		saidInThisAttack = false;

	// FreeForm message say -----> NonPlayOn
	if ( world().gameMode().type() != GameMode::PlayOn and world().gameMode().type() != GameMode::KickOff_ and
		!world().gameMode().isOurSetPlay(world().ourSide()) and !nonPlayOnSay and
		world().time().cycle() - nonPlayOnStartCycle > 0)
	{
		cout << "NonPlayOn Say: " << world().time().cycle() << endl;
		generateNonPlayOnMarkTable();
		cout << "jigh jigh" << endl;
		if (sayMarkers())
		{
			nonPlayOnSay = true;
			cout << "Cycle:\t" << world().time().cycle() << "\tsaying markers [FREE_FORM], NonPlayOn\n" ;
		}
	}
	else if (world().gameMode().type() != GameMode::PlayOn and world().time().cycle() > 200 and
		world().opponents().size() < 11 and !deathOppSay)
	{
		bool changed = false;
		for (int i = 1 ; i < 12 ; i++)
		{
			bool oppFound = false;
			for (uint j = 0 ; j < world().opponents().size() ; j++)
				if (world().opponents()[j]->unum() == i)
					oppFound = true;

			if (!oppFound and oppPlayers[i])
			{
				oppPlayers[i] = false;
				changed = true;
			}
		}
		if (changed)
		{
			stringstream msg("");
			msg << "F" ;
			for (int i = 1 ; i < 12 ; i++)
				msg << oppPlayers[i];

			doSayFreeform(msg.str());
			deathOppSay = true;
			cout << "Msg: " << msg.str() << endl;
			cout << "Cycle:\t" << world().time().cycle() << "\tsaying death opp players [FREE_FORM]\n" ;
		}
	}

    doSubstitute();
    sayPlayerTypes();
//     if ( world().time().cycle() > 0 )
//     {
//         M_client->setServerAlive( false );
//     }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
HelliCoach::doSubstitute()
{
    static bool S_first_substituted = false;

    if ( config().useHetero()
         && ! S_first_substituted
         && world().time().cycle() == 0
         && world().time().stopped() > 10 )
    {
        doFirstSubstitute();
        S_first_substituted = true;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
HelliCoach::doFirstSubstitute()
{
    PlayerTypePtrCont candidates;

    std::fprintf( stderr,
                  "id speed step inc  power  stam"
                  //" decay"
                  //" moment"
                  //" dprate"
                  "  karea"
                  //"  krand"
                  //" effmax effmin"
                  "\n" );

    for ( rcsc::PlayerTypeSet::PlayerTypeMap::const_iterator it
              = rcsc::PlayerTypeSet::i().playerTypeMap().begin();
          it != rcsc::PlayerTypeSet::i().playerTypeMap().end();
          ++it )
    {
        if ( it->second.id() == rcsc::Hetero_Default )
        {
            if ( rcsc::PlayerParam::i().allowMultDefaultType() )
            {
                for ( int i = 0; i <= rcsc::MAX_PLAYER; ++i )
                {
                    M_available_player_type_id.push_back( rcsc::Hetero_Default );
                    candidates.push_back( &(it->second) );
                }
            }
        }
        else
        {
            for ( int i = 0; i < rcsc::PlayerParam::i().ptMax(); ++i )
            {
                M_available_player_type_id.push_back( it->second.id() );
                candidates.push_back( &(it->second) );
            }
        }
        std::fprintf( stderr,
                      " %d %.3f  %2d  %.1f %5.1f %5.1f"
                      //" %.3f"
                      //"  %4.1f"
                      //"  %.5f"
                      "  %.3f"
                      //"  %.2f"
                      //"  %.3f  %.3f"
                      "\n",
                      it->second.id(),
                      it->second.realSpeedMax(),
                      it->second.cyclesToReachMaxSpeed(),
                      it->second.staminaIncMax(),
                      it->second.getDashPowerToKeepMaxSpeed( rcsc::ServerParam::i() ),
                      it->second.getOneStepStaminaComsumption( rcsc::ServerParam::i() ),
                      //it->second.playerDecay(),
                      //it->second.inertiaMoment(),
                      //it->second.dashPowerRate(),
                      it->second.kickableArea()
                      //it->second.kickRand(),
                      //it->second.effortMax(), it->second.effortMin()
                      );
    }


//     std::cerr << "total available types = ";
//     for ( PlayerTypePtrCont::iterator it = candidates.begin();
//           it != candidates.end();
//           ++it )
//     {
//         std::cerr << (*it)->id() << ' ';
//     }
//     std::cerr << std::endl;

    std::vector< int > unum_order;
    unum_order.reserve( 10 );

    unum_order.push_back( 11 );
    unum_order.push_back( 2 );
    unum_order.push_back( 3 );
    unum_order.push_back( 9 );
    unum_order.push_back( 10 );
    unum_order.push_back( 6 );
    unum_order.push_back( 4 );
    unum_order.push_back( 5 );
    unum_order.push_back( 7 );
    unum_order.push_back( 8 );

    for ( std::vector< int >::iterator unum = unum_order.begin();
          unum != unum_order.end();
          ++unum )
    {
        int type = getFastestType( candidates );
        if ( type != rcsc::Hetero_Unknown )
        {
            substituteTo( *unum, type );
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
HelliCoach::substituteTo( const int unum,
                           const int type )
{
    if ( world().time().cycle() > 0
         && M_substitute_count >= rcsc::PlayerParam::i().subsMax() )
    {
        std::cerr << "***Warning*** "
                  << config().teamName() << " coach: over the substitution max."
                  << " cannot change the player " << unum
                  << " to type " << type
                  << std::endl;
        return;
    }

    std::vector< int >::iterator it = std::find( M_available_player_type_id.begin(),
                                                 M_available_player_type_id.end(),
                                                 type );
    if ( it == M_available_player_type_id.end() )
    {
        std::cerr << "***ERROR*** "
                  << config().teamName() << " coach: "
                  << " cannot change the player " << unum
                  << " to type " << type
                  << std::endl;
        return;
    }

    M_available_player_type_id.erase( it );
    M_assigned_player_type_id.insert( std::make_pair( unum, type ) );
    if ( world().time().cycle() > 0 )
    {
        ++M_substitute_count;
    }


    doChangePlayerType( unum, type );

    std::cout << config().teamName() << " coach: "
              << "change player " << unum
              << " to type " << type
              << std::endl;
}

/*-------------------------------------------------------------------*/
/*!

 */
int
HelliCoach::getFastestType( PlayerTypePtrCont & candidates )
{
    if ( candidates.empty() )
    {
        return rcsc::Hetero_Unknown;
    }

    // sort by max speed
    std::sort( candidates.begin(),
               candidates.end(),
               RealSpeedMaxCmp() );

//     std::cerr << "getFastestType candidate = ";
//     for ( PlayerTypePtrCont::iterator it = candidates.begin();
//           it != candidates.end();
//           ++it )
//     {
//         std::cerr << (*it)->id() << ' ';
//     }
//     std::cerr << std::endl;

    PlayerTypePtrCont::iterator best_type = candidates.begin();
    double max_speed = (*best_type)->realSpeedMax();
    int min_cycle = 100;
    for ( PlayerTypePtrCont::iterator it = candidates.begin();
          it != candidates.end();
          ++it )
    {
        if ( (*it)->realSpeedMax() < max_speed - 0.01 )
        {
            break;
        }

        if ( (*it)->cyclesToReachMaxSpeed() < min_cycle )
        {
            best_type = it;
            max_speed = (*best_type)->realSpeedMax();
            min_cycle = (*best_type)->cyclesToReachMaxSpeed();
            continue;
        }

        if ( (*it)->cyclesToReachMaxSpeed() == min_cycle )
        {
            if ( (*it)->getOneStepStaminaComsumption( rcsc::ServerParam::i() )
                 < (*best_type)->getOneStepStaminaComsumption( rcsc::ServerParam::i()) )
            {
                best_type = it;
                max_speed = (*best_type)->realSpeedMax();
            }
        }
    }

    int id = (*best_type)->id();
    candidates.erase( best_type );
    return id;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
HelliCoach::sayPlayerTypes()
{
    /*
      format:
      "(player_types (1 0) (2 1) (3 2) (4 3) (5 4) (6 5) (7 6) (8 -1) (9 0) (10 1) (11 2))"
      ->
      (say (freeform "(player_type ...)"))
    */

    static rcsc::GameTime s_last_send_time( 0, 0 );

    if ( ! config().useFreeform() )
    {
        return;
    }

    if ( ! world().canSendFreeform() )
    {
        return;
    }

    int analyzed_count = 0;

    for ( int unum = 1; unum <= 11; ++unum )
    {
        const int id = world().heteroID( world().theirSide(), unum );

        if ( id != M_opponent_player_types[unum - 1] )
        {
            M_opponent_player_types[unum - 1] = id;

            if ( id != rcsc::Hetero_Unknown )
            {
                ++analyzed_count;
            }
        }
    }

    if ( analyzed_count == 0 )
    {
        return;
    }

    std::string msg;
    msg.reserve( 128 );

    msg = "(player_types ";

    for ( int unum = 1; unum <= 11; ++unum )
    {
        char buf[8];
        snprintf( buf, 8, "(%d %d)",
                  unum, M_opponent_player_types[unum - 1] );
        msg += buf;
    }

    msg += ")";

    doSayFreeform( msg );

    s_last_send_time = world().time();

    std::cout << config().teamName()
              << " Coach: "
              << world().time()
              << " send freeform " << msg
              << std::endl;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
HelliCoach::sendTeamGraphic()
{
    int count = 0;
    for ( rcsc::TeamGraphic::Map::const_reverse_iterator tile
              = M_team_graphic.tiles().rbegin();
          tile != M_team_graphic.tiles().rend();
          ++tile )
    {
        if ( teamGraphicOKSet().find( tile->first ) == teamGraphicOKSet().end() )
        {
            if ( ! doTeamGraphic( tile->first.first,
                                  tile->first.second,
                                  M_team_graphic ) )
            {
                break;
            }
            ++count;
        }
    }

    if ( count > 0 )
    {
        std::cout << config().teamName()
                  << " coach: "
                  << world().time()
                  << " send team_graphic " << count << " tiles"
                  << std::endl;
    }
}

void HelliCoach::generateNonPlayOnMarkTable()
{
	mMarkTable.clear();
	vector<const GlobalPlayerObject*> opp;
	// Recognizing very dangerous opponents
	for (uint i = 0 ; i < world().opponents().size() ; i++)
		if ( world().opponents()[i]->pos().x < -26.25 and !isInPlayers(world().opponents()[i]->unum(), centralSayMsg) )
			opp.push_back(world().opponents()[i]);

	// Recognizing a little dangerous opponents
	for (uint i = 0 ; i < world().opponents().size() ; i++)
		if ( world().opponents()[i]->pos().x >= -26.25 and !world().opponents()[i]->goalie()
			and !isInPlayers(world().opponents()[i]->unum(), centralSayMsg) )
			opp.push_back(world().opponents()[i]);

	// pushing back the empty opponent
	for (uint i = opp.size() ; i < 4 ; i++)
		opp.push_back(new GlobalPlayerObject);

	// find nearest opponents to ball
	for (uint i = 0 ; i < opp.size() ; i++)
		for (uint j = 0 ; j < opp.size() - 1 ; j++)
			if ( opp[j]->pos().dist(world().ball().pos()) > opp[j + 1]->pos().dist(world().ball().pos()) )
			{
				const GlobalPlayerObject* temp = opp[j + 1];
				opp[j + 1] = opp[j];
				opp[j] = temp;
			}

	vector<const GlobalPlayerObject*> dangerousOpp;
	for (uint i = 0 ; i < 4 ; i++)
		dangerousOpp.push_back(opp[i]);

	// Recognizing teammate markers
	vector<const GlobalPlayerObject*> defender;
	// Offense 69AB
	for (uint i = 0 ; i < world().teammates().size() ; i++)
		if ( isInPlayers(world().teammates()[i]->unum(), "69AB") )
		{
			defender.push_back(world().teammates()[i]);
		}


	// Weighted Matching =============================================================
	vector< vector<int> > wvec;
	for (uint i = 0 ; i < min((uint)4, defender.size()) ; i++)
	{
		wvec.push_back( vector<int>() );
		for (uint j = 0 ; j < min((uint)4, defender.size()) ; j++)
		{
			float weight = 1;

			weight = 125 - positioningTarget(defender[i]->unum()).dist(dangerousOpp[j]->pos());

			wvec[i].push_back(weight);// * 1000000);
		}
	}

	WMatching wmatch(4, wvec);
	wmatch.calculate();

	for (uint i = 0 ; i < 4 ; i++)
	{
		mMarkTable[defender[wmatch.match[i]]] = dangerousOpp[i];
	}
	// ===============================================================================

	for (int i = 1 ; i < 7 ; i++)
	{
		if (i != 4)
		{
			const GlobalPlayerObject* def = NULL;
			for (uint i = 0 ; i < world().teammates().size() ; i++)
				if (world().teammates()[i]->unum() == (signed) i + 2)
					def = world().teammates()[i];

			const GlobalPlayerObject* opp = NULL;
			for (uint i = 0 ; i < world().opponents().size() ; i++)
				if (world().opponents()[i]->unum() == centralSayMsg[i])
					opp = world().opponents()[i];

			if (def and opp)
				mMarkTable[def] = opp;
		}
	}

	for (map<const GlobalPlayerObject*, const GlobalPlayerObject*>::iterator it = mMarkTable.begin() ; it != mMarkTable.end() ; it++)
		cout << "MT:\t " << it->first->unum() << " --> " << it->second->unum() << endl;

}

void HelliCoach::generatePlayOnMarkTable()
{
	mMarkTable.clear();
	vector<const GlobalPlayerObject*> dangerousOpp;
	// Recognizing very dangerous opponents
	for (uint i = 0 ; i < world().opponents().size() ; i++)
		if ( world().opponents()[i]->pos().x < -26.25 )
			dangerousOpp.push_back(world().opponents()[i]);

	// Sorting very dangerous opponents
	for (uint i = 0 ; i < dangerousOpp.size() ; i++)
		for (uint j = 0 ; j < dangerousOpp.size() - 1 ; j++)
		{
			float opp1DistToGoal = min( min(dangerousOpp[j]->pos().dist(Vector2D(-52.5, +7.0)),
							dangerousOpp[j]->pos().dist(Vector2D(-52.5, -7.0))),
							dangerousOpp[j]->pos().dist(Vector2D(-52.5, 0.0)));
//			float opp1DistToGoal = worldModel->getFullPlayer(TID_OPPONENT, dangerousOpp[j]).pos().dist(Vector2D(-52.5, 0.0));
			float opp2DistToGoal = min( min(dangerousOpp[j + 1]->pos().dist(Vector2D(-52.5, +7.0)),
							dangerousOpp[j + 1]->pos().dist(Vector2D(-52.5, -7.0))),
							dangerousOpp[j + 1]->pos().dist(Vector2D(-52.5, 0.0)));
//			float opp2DistToGoal = worldModel->getFullPlayer(TID_OPPONENT, dangerousOpp[j + 1]).pos().dist(Vector2D(-52.5, 0.0));

			if (opp1DistToGoal > opp2DistToGoal)
			{
				const GlobalPlayerObject* temp = dangerousOpp[j + 1];
				dangerousOpp[j + 1] = dangerousOpp[j];
				dangerousOpp[j] = temp;
			}
		}

	uint veryDangerousOppCount = dangerousOpp.size();
	// Recognizing a little dangerous opponents
	for (uint i = 0 ; i < world().opponents().size() ; i++)
		if ( world().opponents()[i]->pos().x >= -26.25 and !world().opponents()[i]->goalie() )
			dangerousOpp.push_back(world().opponents()[i]);

	// Sorting a little dangerous opponents
	for (uint i = veryDangerousOppCount ; i < dangerousOpp.size() ; i++)
		for (uint j = veryDangerousOppCount ; j < dangerousOpp.size() - 1 ; j++)
			if ( dangerousOpp[j]->pos().x > dangerousOpp[j + 1]->pos().x )
			{
				const GlobalPlayerObject* temp = dangerousOpp[j + 1];
				dangerousOpp[j + 1] = dangerousOpp[j];
				dangerousOpp[j] = temp;
			}

	// pushing back the empty opponent
	for (uint i = dangerousOpp.size() ; i < 9 ; i++)
		dangerousOpp.push_back(new GlobalPlayerObject);

	if (dangerousOpp.size() > 9)
		dangerousOpp.resize(9);

	// Recognizing teammate markers
	vector<const GlobalPlayerObject*> defender;
	// Defense 34578
	for (uint i = 0 ; i < world().teammates().size() ; i++)
		if ( isInPlayers(world().teammates()[i]->unum(), "34578") )
		{
			defender.push_back(world().teammates()[i]);
		}
	uint defenseCount = defender.size();
	// Middle 6
	for (uint i = 0 ; i < world().teammates().size() ; i++)
		if ( world().teammates()[i]->unum() == 6 )
		{
			defender.push_back(world().teammates()[i]);
		}
	uint middleCount = defender.size() - defenseCount;
	// Offense 9AB
	for (uint i = 0 ; i < world().teammates().size() ; i++)
		if ( isInPlayers(world().teammates()[i]->unum(), "9AB") )
		{
			defender.push_back(world().teammates()[i]);
		}
	uint offenseCount = defender.size() - middleCount - defenseCount;

	dlog.addText(Logger::ROLE, "dang: %d\t def:%d", dangerousOpp.size(), defender.size());
	dlog.addText(Logger::ROLE, "%d %d %d", defenseCount, middleCount, offenseCount);

	dlog.flush();
	// Weighted Matching =============================================================
	vector< vector<int> > wvec;
	for (uint i = 0 ; i < min((uint)5, defender.size()) ; i++)
	{
		wvec.push_back( vector<int>() );
		for (uint j = 0 ; j < min((uint)5, defender.size()) ; j++)
		{
			float weight = 1;

			weight = 125 - defender[i]->pos().dist(dangerousOpp[j]->pos());

			wvec[i].push_back(weight);// * 1000000);
		}
	}

	WMatching wmatch(5, wvec);
	wmatch.calculate();

	for (uint i = 0 ; i < 5 ; i++)
	{
		mMarkTable[defender[wmatch.match[i]]] = dangerousOpp[i];
	}
	// ===============================================================================
}

bool HelliCoach::sayMarkers()
{
	stringstream msg("");
	for (int i = 3 ; i < 12 ; i++)
	{
		bool markTargetFound = false;
		for (map<const GlobalPlayerObject*, const GlobalPlayerObject*>::iterator it = mMarkTable.begin() ; it != mMarkTable.end() ; it++)
		{
			if (it->first->unum() == i)
			{
				if (it->second->unum() == 10)
					msg << 'A' ;
				else if (it->second->unum() == 11)
					msg << 'B' ;
				else if (it->second->unum() == -1)
					msg << '0' ;
				else
					msg << it->second->unum() ;

				markTargetFound = true;
			}
		}
		if (!markTargetFound)
			msg << '0' ;
	}

	cout << "MSG: " << msg.str() << endl;

	if ( world().gameMode().type() != GameMode::PlayOn )
	{
		cout << "Say freeform message in cycle: " << world().time().cycle() << endl;
		return doSayFreeform(msg.str());
	}
	else
	{
		stringstream infoMsg("");
		infoMsg << "(say (info (" << world().time().cycle() << " (true) " << '"' << msg.str() << '"' << ")))" ;
		cout << "info MSG:\t" << infoMsg.str() << endl;

		return ( M_client->sendMessage( infoMsg.str().c_str() ) > 0 );
	}
}

Vector2D HelliCoach::positioningTarget(int unum)
{
	Vector2D homePos;
	switch (unum)
	{
		case 1:	homePos = Vector2D(-49.0,  0.0);	break;
		case 2:	homePos = Vector2D(-36.0,  0.0);	break;
		case 3:	homePos = Vector2D(-25.0,  0.0);	break;
		case 4:	homePos = Vector2D(-25.0,-30.0);	break;
		case 5:	homePos = Vector2D(-25.0,+30.0);	break;
		case 6:	homePos = Vector2D(  0.0,  0.0);	break;
		case 7:	homePos = Vector2D(-20.0,-15.0);	break;
		case 8:	homePos = Vector2D(-20.0,+15.0);	break;
		case 9:	homePos = Vector2D(+15.0,-15.0);	break;
		case 10:homePos = Vector2D(+15.0,+15.0);	break;
		case 11:homePos = Vector2D(+15.0,  0.0);	break;
	}

	return homePos;
}
