// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef HELLIBASH_COACH_H
#define HELLIBASH_COACH_H

#include <vector>
#include <map>
#include <set>

#include <rcsc/types.h>

#include <rcsc/coach/coach_agent.h>

namespace rcsc {
class PlayerType;
}

class HelliCoach
    : public rcsc::CoachAgent {
private:
    typedef std::vector< const rcsc::PlayerType * > PlayerTypePtrCont;

    int M_substitute_count; // substitution count after kick-off
    std::vector< int > M_available_player_type_id;
    std::map< int, int > M_assigned_player_type_id;

    int M_opponent_player_types[11];

    rcsc::TeamGraphic M_team_graphic;

	std::map<const rcsc::GlobalPlayerObject*, const rcsc::GlobalPlayerObject*> mMarkTable;

	std::map<int, bool> oppPlayers;

public:

    HelliCoach();

    virtual
    ~HelliCoach();

private:

	void generateNonPlayOnMarkTable();
	void generatePlayOnMarkTable();
	bool sayMarkers();

	static std::string centralSayMsg;

	rcsc::Vector2D positioningTarget(int unum);

    /*!
      \brief substitute player.

      This methos should be overrided in the derived class
    */
    void doSubstitute();

    void doFirstSubstitute();

    void substituteTo( const int unum,
                       const int type );

    int getFastestType( PlayerTypePtrCont & candidates );

    void sayPlayerTypes();

    void sendTeamGraphic();

protected:

    /*!
      You can override this method.
      But you must call PlayerAgent::doInit() in this method.
    */
    virtual
    bool initImpl( rcsc::CmdLineParser & cmd_parser );

    //! main decision
    virtual
    void actionImpl();

};

#endif
