// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <BackDefenderGuard.h>

#include <rcsc/common/logger.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/body_go_to_point.h>
#include <body_go_to_point2008.h>

#include <cmath>

using namespace rcsc;
using namespace std;


bool BackDefenderGuard::execute(PlayerAgent* agent)
{
	dlog.addText(Logger::ROLE, "Entering BackDefenderGuard::execute");
	const WorldModel* wm = &agent->world();

	Vector2D guardPoint;
	Vector2D turnPoint;
	Vector2D ballPos = wm->ball().pos();

	bool isGuardSituation = false;

	if (ballPos.x < -28.0 and ballPos.absY() > 12.0)
	{
		guardPoint.y = ballPos.y > 0.0 ? +9.15 : -9.15;

		float goalPointY = 20.0;

		if (ballPos.y > 0.0)
			guardPoint.x = ( (guardPoint.y + goalPointY) / (ballPos.y + goalPointY) ) * (52.5 + ballPos.x) - 52.5;
		else
			guardPoint.x = ( (guardPoint.y - goalPointY) / (ballPos.y - goalPointY) ) * (52.5 + ballPos.x) - 52.5;

		guardPoint.x = max(guardPoint.x, -48.0);

		turnPoint = wm->self().pos() + Vector2D(1, 0);

		isGuardSituation = true;
	}

	if (isGuardSituation)
	{
		if (wm->self().pos().dist(guardPoint) > 6.0 and
			Body_GoToPoint(guardPoint, 2.0, dashPower(wm, guardPoint), 100, false, true, 15.0).execute(agent))
			dlog.addText(Logger::ROLE, "Going to guardPoint [FAR]");
		else if (Body_GoToPoint2008(guardPoint, 0.3, dashPower(wm, guardPoint), true, false, true).execute(agent))
			dlog.addText(Logger::ROLE, "Going to guardPoint");
		else
		{
			Body_TurnToPoint(turnPoint).execute(agent);
			dlog.addText(Logger::ROLE, "Changing the body direction");
		}

		return true;
	}

	dlog.addText(Logger::ROLE, "Exiting BackDefenderGuard::execute [FALSE]");
	return false;
}

int BackDefenderGuard::dashPower(const WorldModel* wm, Vector2D guardPoint)
{
	int dashPower = 85;

	if (guardPoint.x > wm->self().pos().x)
		dashPower = 70;
	if (guardPoint.dist(wm->self().pos()) > 10.0)
		dashPower = 100;

	return dashPower;
}
