// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <Block.h>
#include <BlockIntercept.h>
#include <StraightBlock.h>

#include <Share.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/player/intercept_table.h>
#include <rcsc/player/abstract_player_object.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/penalty_kick_state.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <Intercept.h>
#include <FastIC.h>

#include <Basics.h>

#include <SideDashGoToPoint.h>
#include <body_go_to_point2008.h>

#include <bhv_danger_area_tackle.h>
#include <bhv_basic_tackle.h>

using namespace rcsc;
using namespace std;
using namespace Basics;

//#define BLOCK_LOG

Block::Block( ): oppRouteLine(Vector2D(0,0), Vector2D(0,0))
{
}

bool
Block::execute( PlayerAgent * agent )
{
	dlog.addText( Logger::BLOCK, "Entering Block::execute"); 
    this->agent = agent;

	N = 3;
	FastIC fastIC;
	fastestOpp = fastIC.getFastestOpponent();
	if( ! fastestOpp )
	{
		dlog.addText( Logger::BLOCK," fastestOpp does not exist"); 
		if ( agent->world().gameMode().isPenaltyKickMode() )
		{
			dlog.addText( Logger::BLOCK, "so use intercept");
			Intercept().execute(agent);
			return true;
		}
		return false;
	}
//	dlog.addText( Logger::BLOCK,"before neck action"); 
    LookToBallAndPlayer(fastestOpp, agent);

    state = agent->world().penaltyKickState();
	myPos = getMyPos();
	ballPos = getBallPos();
	fastestOppPos = getFastestOppPos();

	oppRoutePoint = getOppRoute();
	oppRouteLine = Line2D(fastestOppPos, oppRoutePoint);
	oppRouteRay = Ray2D(fastestOppPos, oppRoutePoint);

	dlog.addText( Logger::BLOCK, "yes!!!");dlog.flush();
	if (agent->world().self().tackleProbability() > 0.5)
		dlog.addText( Logger::BLOCK, "tackle probability is %f", agent->world().self().tackleProbability());

	if ( Bhv_DangerAreaTackle().execute(agent) )
	{
		dlog.addText( Logger::BLOCK, "Bhv_DangerAreaTackle().execute(agent) is true, so return that");
		return true;
	}

	dlog.addText( Logger::BLOCK, "shir!");dlog.flush();
	double min_prob = 0.9;
	if ( Bhv_BasicTackle(min_prob, 70).execute(agent) )
	{
		dlog.addText( Logger::BLOCK, "Bhv_BasicTackle(%f, 10).execute(agent) is true, so return that", min_prob);
		return true;
	}

	long fastestReachCycle = fastIC.getFastestOpponentReachCycle();
	if ( fastestReachCycle > N or (fastestReachCycle == N and agent->world().time().cycle() - Share::lastInterceptBlockingCycle == 1) )
	{
		dlog.addText( Logger::BLOCK, "fastestOpponenReachCycle > N, so return BlockIntercept  (N is %d)", N);dlog.flush();
		return BlockIntercept( oppRoutePoint ).execute( agent );
	}

	dlog.addText( Logger::BLOCK, "block adi");dlog.flush();
/*	if ( myPos.x < fastestOppPos.x and myPos.dist(fastestOppPos) < 3.5 and myPos.dist(fastestOppPos) > 2)
	{
		dlog.addText( Logger::BLOCK, "JADID BlockIntercept");
		return BlockIntercept( oppRoutePoint ).execute( agent );
	}
*/
	if ( StraightBlock(fastestOppPos, oppRoutePoint).execute(agent) )
		return true;

	if ( isNearToBall( agent->world().ball().pos() ) )
		getBall();
	else
		goCloseToBall();

	Share::lastOppVel = fastestOpp->vel().r();
	Share::lastOppRouteCycle = agent->world().time().cycle();

	dlog.addText( Logger::BLOCK, "Exiting Block::execute");
	return true;
}

bool
Block::LookToBallAndPlayer(const AbstractPlayerObject * player, PlayerAgent * agent)
{
	switch (agent->world().time().cycle() % 2)
	{
		case 0:
			agent->setNeckAction( new Neck_TurnToBall() );
			break;
		case 1:
			Vector2D playerNextPos = player->pos() + player->vel();

			agent->setNeckAction( new Neck_TurnToPoint(playerNextPos) );
			break;
	}

	return false;
}

bool
Block::isNearToBall( Vector2D BallPos )
{
	dlog.addText( Logger::BLOCK, "Entering Block::isNearToBall( Vector2D ballPos )");dlog.flush();

	FastIC fastIC;
	if ( fastIC.getSelfReachCycle() < 3 or oppRouteLine.dist(myPos) < 2  )
	{
		dlog.addText( Logger::BLOCK, "isNearToBall Value is true.");
		return true;
	}
	else
	{
		dlog.addText( Logger::BLOCK, "isNearToBall Value is false"); 
		return false;
	}
}

Vector2D
Block::getMyPos()
{
	Vector2D mypos = agent->world().self().pos();
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldModel dade!! myPos.x * -1");
		return Vector2D(mypos.x * -1, mypos.y);
	}
	else
	{
		return mypos;
	}
}

Vector2D
Block::getBallPos()
{
	Vector2D ballpos = agent->world().ball().pos();
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldModel dade!! ballPos.x * -1");
		return Vector2D(ballpos.x * -1, ballpos.y);
	}
	else
	{
		return ballpos;
	}
}

Vector2D
Block::getFastestOppPos()
{
	dlog.addText( Logger::BLOCK, "Entering Block::getFastestOppPos");

	Vector2D OppPos = fastestOpp->pos();
	dlog.addText( Logger::BLOCK, "fastestOppPos(before change) is %f, %f", OppPos.x, OppPos.y);dlog.flush();
	FastIC fastIC;

	Vector2D top_left( -52.5, -34 );
	Vector2D bottom_right(52.5, 34);
	Rect2D field( top_left, bottom_right );
	
	Vector2D ballPos = agent->world().ball().pos();
	BallObject ball = agent->world().ball();
	Vector2D receivePoint;

	if (fastIC.getFastestOpponent())
	{
		int fastestReachCycle = fastIC.getFastestOpponentReachCycle();
		receivePoint = agent->world().ball().inertiaPoint(fastestReachCycle);
	}
	else
		dlog.addText( Logger::BLOCK, "sssssssssss");dlog.flush();

	if ( field.contains(receivePoint) and OppPos.dist(receivePoint) > 2 )
	{
		dlog.addText( Logger::BLOCK, "use receivePoint");dlog.flush();
		OppPos = receivePoint;
	}
//	vector<FastestPlayer> FPlayers = fastIC.getFastestPlayers();
//	Vector2D receivePoint = FPlayers[0].receivePoint;
//	if( fastestOppPos.dist(receivePoint) > 2.5 and field.contains(receivePoint)/*and ball.vel().r() > fastestOpp->vel().r()*/ )
//	{
//		dlog.addText( Logger::BLOCK, "can use receivePoint");
//		OppPos = receivePoint;
//	}

	if ( !field.contains(OppPos) and field.contains(ballPos) )
	{
		dlog.addText( Logger::BLOCK, "use BallPos");
		OppPos = agent->world().ball().pos();
	}
	else if(!field.contains(OppPos) )
	{
		dlog.addText(Logger::BLOCK, "ballPos and OppPos are not in the field, use OppPos");
		OppPos = Vector2D(OppPos.x + 1, OppPos.y);
	}

//	if( field.contains(OppPos + Vector2D(-1, 0)) and fastestOppPos.x > -30 and oppRouteLine.dist(myPos) > 4 )
//		OppPos = OppPos + Vector2D(-1, 0);

	dlog.addText( Logger::BLOCK, "final fastestOppPos is %f, %f", OppPos.x, OppPos.y);
	dlog.addText( Logger::BLOCK, "Exiting Block::getFastestOppPos");dlog.flush();

	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldModel dade!! fastestOppPos.x * -1");dlog.flush();
		return Vector2D(OppPos.x * -1, OppPos.y);
	}
	else
	{
		dlog.addText( Logger::BLOCK, "hey!!!!");dlog.flush();
		return OppPos;
	}
}

void
Block::getBall()
{
	dlog.addText( Logger::BLOCK, "Entering Block::getBall");

//	Intercept intercept;
//	intercept.execute(agent); return;

	const WorldModel & wm = agent->world();
	FastIC fastIC;

	Vector2D fastestOppVel = fastestOpp->vel();
	double hisVel = fastestOpp->vel().r();
	double myVel = agent->world().self().vel().r();
	InterceptTable IT(wm);
	const PlayerObject * ITfastestOpp;
	int kickable = 1;
	if(InterceptTable(wm).fastestOpponent())
	{
		ITfastestOpp = IT.fastestOpponent();
		if ( ITfastestOpp->isKickable() )
			kickable = 2;
		else
			kickable = 0;
	}

	dlog.addText( Logger::BLOCK, "oppRouteLine.dist(myPos) = %f    and   myPos.dist(fastestOppPos) = %f", oppRouteLine.dist(myPos), myPos.dist(fastestOppPos));
/*	double deg = Ray2D(myPos, fastestOppPos).dir().degree();
	double Deg;
	if( abs((agent->world().self().body().degree() + 90) - deg) < 90 or abs(deg - (agent->world().self().body().degree() + 90)) < 90)
		Deg = agent->world().self().body().degree() + 90;
	else
		Deg = agent->world().self().body().degree() - 90;

	dlog.addText( Logger::BLOCK, "Deg is %f   and deg is %f", Deg, deg);*/

	dlog.addText( Logger::BLOCK, "kickable = %d", kickable);
	if ( oppRouteLine.dist(myPos) < 0.8 and myPos.dist(fastestOppPos) < 2.5 and oppRoutePoint.dist(fastestOppPos) > oppRoutePoint.dist(myPos) and kickable > 0 )
	{
		dlog.addText( Logger::BLOCK,"should use sideDashGetBall");
		sideDashGetBall();
		return;
	}
	if ( oppRouteLine.dist(myPos) < 0.8 and myPos.dist(fastestOppPos) < 3.3 and oppRoutePoint.dist(fastestOppPos) > oppRoutePoint.dist(myPos) and kickable > 0 )
	{
		dlog.addText( Logger::BLOCK,"should use sideDashGetBall(bekhatere penalty)");
		sideDashGetBall();
		return;
	}
/*	else if ( (fastestOppVel.r() < min_vel or fastIC.getFastestOpponentReachCycle() > N) and oppRouteLine.dist(myPos) < 1.2 and myPos.dist(fastestOppPos) > 2.5)
	{
		dlog.addText( Logger::BLOCK, "ehtemalan dare kick mikone, sideDashGetBall");
		sideDashGetBall();
		return;
	}*/
	else if ( oppRouteLine.dist(myPos) < 0.8 and myPos.dist(fastestOppPos) > 2.5 )
	{
		dlog.addText( Logger::BLOCK, "my distance to route line is little!! but haveto go near to opponent");
//		double dash_power = 100.0;
		Vector2D thisVector2D( (oppRoutePoint - fastestOppPos) / fastestOppPos.dist(oppRoutePoint) * 2 );
		Vector2D target = fastestOppPos + thisVector2D;
		dlog.addText( Logger::BLOCK, "I want to go in    %f, %f", target.x, target.y);
//    	Body_GoToPoint2008( target, 1.0, dash_power ).execute(agent);
		goToPoint(target);
	}

	double angle = abs( oppRouteRay.dir().degree() );
	angle -= 90;
	angle = abs(angle);
	angle /= 200;

	myVel = std::max(0.35, myVel);
	double least = 400;
	Vector2D bestTarget = agent->world().ball().pos();

	if( hisVel < 0.15 and agent->world().time().cycle() - Share::lastOppRouteCycle == 1)
	{
		hisVel = Share::lastOppVel;
	}

#ifdef BLOCK_LOG
	dlog.addText( Logger::BLOCK, "myVel = %f", myVel);
	dlog.addText( Logger::BLOCK, "hisVel = %f", hisVel);
	dlog.addText( Logger::BLOCK, "before for");
#endif

	for (double x = fastestOppPos.x; x > fastestOppPos.x - angle * 20 /*and 
		 fastestOppPos.dist(myPos) < fastestOppPos.dist(Vector2D(x, oppRouteLine.getY(x)))*/; x -= angle)
	{
		double y = oppRouteLine.getY(x);
		Vector2D target(x, y);
#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "checking target(%f, %f)", target.x, target.y );
		dlog.addText( Logger::BLOCK, "least is %f", least);
#endif
		if( abs(target.dist(fastestOppPos) / hisVel - target.dist(myPos) / myVel) < least )
		{
//			double d = abs(target.dist(fastestOppPos) / hisVel - target.dist(myPos) / myVel);
//			dlog.addText( Logger::BLOCK, "abs(target.dist(fastestOppPos) / hisVel - target.dist(myPos) / myVel) = %f", d);
			bestTarget = target;
			least = abs(target.dist(fastestOppPos) / hisVel - target.dist(myPos) / myVel);
		}
	}
	
//	dlog.addText( Logger::BLOCK,"bestTarget is %f , %f", bestTarget.x, bestTarget.y);
//  TEST

/*	double x = myPos.x;
	double y = oppRouteLine.getY(x);
	bestTarget = Vector2D(x, y);
	
	if(myPos.dist(fastestOppPos) >= 2)
		bestTarget = fastestOppPos;
*/
//	if (fastestOppPos.dist(myPos) > 2 and oppRouteLine.dist(myPos) < 1.2)
	{
		dlog.addText( Logger::BLOCK, "go close to ball");
		goCloseToBall();
		return;
	}
	
//  TEST
	double dash_power = 100.0;
    Body_GoToPoint2008( bestTarget, 1.0, dash_power ).execute(agent);

	dlog.addText( Logger::BLOCK, "Exiting Block::getBall");
	return;
}

void
Block::goCloseToBall()
{
	dlog.addText( Logger::BLOCK, "Entering Block::goCloseToBall");dlog.flush();

	FastIC fastIC;
	if(  fastIC.getSelfReachCycle() - fastIC.getFastestOpponentReachCycle() < 2 )
	{
#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "fastIC.getFastestOpponentReachCycle() - fastIC.getSelfReachCycle() < 2 , so intercepting"); 
#endif
/*		Intercept intercept;
		intercept.execute(agent);
		return;*/
	} 

/*	if ( oppRouteLine.dist(myPos) > 2 and myPos.dist(fastestOppPos) > 2 )
	{
		dlog.addText( Logger::BLOCK, "JADID ----  BlockIntercept ");
		BlockIntercept( oppRoutePoint ).execute(agent);
		return;
	}
*/
	setValidPoints();
	dlog.addText( Logger::BLOCK ,"targets.size() is %d", targets.size());dlog.flush();
#ifdef BLOCK_LOG
	dlog.addText( Logger::BLOCK, "after find valid point"); 
#endif
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldmodel too penalty dade, targets ro baraks mikone");
		for (uint i = 0; i < targets.size(); i++)
		{
			targets[i] = Vector2D(targets[i].x * -1, targets[i].y);
			dlog.addText( Logger::BLOCK, "targets[%d] = %f, %f", i, targets[i].x, targets[i].y);
		}
	}

	{
		myPos = agent->world().self().pos();
		if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
			fastestOppPos = Vector2D(fastestOppPos.x * -1, fastestOppPos.y);
	}

	Vector2D myRoute;
	Ray2D Center( myPos, targets[targets.size() / 2] );
	Ray2D top( myPos, targets[0] );
	Ray2D bottom( myPos, targets[targets.size()-1] );
	double center = Center.dir().degree();
	double size = abs( top.dir().degree() - bottom.dir().degree() );
#ifdef BLOCK_LOG
	dlog.addText( Logger::BLOCK ,"before find final target point");
#endif

	Ray2D myRay( myPos, agent->world().self().body() );
	Vector2D point(myRay.intersection( oppRouteRay ) );

	Vector2D top_left( -52.5, -34 );
	Vector2D bottom_right(52.5, 34);
	Rect2D field( top_left, bottom_right );

#ifdef BLOCK_LOG
	dlog.addText( Logger::BLOCK, "center.degree() = %f", center);
	dlog.addText( Logger::BLOCK, "abs( top.dir().degree() - bottom.dir().degree() = %f", size);
	dlog.addText( Logger::BLOCK, "top.dir().degree() = %f", top.dir().degree());
	dlog.addText( Logger::BLOCK, "bottom.dir().degree() = %f", bottom.dir().degree());
#endif

	Vector2D myRay_oppRoute = myRay.intersection( oppRouteRay );
	if ( Ray2D(myPos, myRay_oppRoute).dir().degree() - agent->world().self().body().degree() > 20 )
	{
		dlog.addText( Logger::BLOCK, "barkhord myRay ba oppRouteRay posht bazikone mae");
		double a = 10;
		AngleDeg body = agent->world().self().body();
		if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
		{
			dlog.addText( Logger::BLOCK, "worldmodel dade, x Vector2D ye * -1");
			myRay_oppRoute = myPos + Vector2D(body.cos() * a/* * -1*/, body.sin() * a);
		}
		else
		{
			dlog.addText( Logger::BLOCK, "worldmodel dade, x Vector2D ye * -1");
			myRay_oppRoute = myPos + Vector2D(body.cos() * a, body.sin() * a);
		}
	}
	if ( abs(agent->world().self().body().degree() - center) < size / 2.f  and field.contains(point) )
	{
		dlog.addText( Logger::BLOCK, "my body direction is to a target point, so continue");dlog.flush();
		myRoute = myRay.intersection( oppRouteRay );
#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "origin of myRay is %f, %f and direction is %f ", myRay.origin().x, myRay.origin().y, myRay.dir().degree());
		dlog.addText( Logger::BLOCK, "origin of oppRouteRay is %f, %f and direction is %f ", oppRouteRay.origin().x, oppRouteRay.origin().y, oppRouteRay.dir().degree());
#endif
		dlog.addText( Logger::BLOCK, "myRoute is %f , %f", myRoute.x, myRoute.y);
	}
	else
	{
		dlog.addText( Logger::BLOCK, "my body direction isn't to a target point, so change body direction");dlog.flush();
		int number = targets.size() / 2;
		Ray2D raY(myPos, targets[number]);
		double myNextDir = raY.dir().degree();
		AngleDeg theta = agent->world().self().body();
		double myLastDir_myNextDir = theta.degree() - myNextDir;
		dlog.addText( Logger::BLOCK ,"myLastDir - myNextDir = %f",myLastDir_myNextDir);
		if (theta.degree() - myNextDir > 180)
			myLastDir_myNextDir -= 360;
		if (theta.degree() - myNextDir < -180)
			myLastDir_myNextDir += 360;
		myLastDir_myNextDir = abs(myLastDir_myNextDir);
		dlog.addText( Logger::BLOCK ,"[FINAL] myLastDir - myNextDir = %f",myLastDir_myNextDir);
					
		if
		(
			(myLastDir_myNextDir < 10 /*or myLastDir_myNextDir > 170*/) 
			or
			(myPos.x - fastestOppPos.x > 1 and fastestOppPos.dist(myPos) > 3 and (myLastDir_myNextDir < 15 or myLastDir_myNextDir > 165))
		)
		{
			dlog.addText( Logger::BLOCK, "majboor nisti jahateto avaz koni!");
			AngleDeg theta = agent->world().self().body();
			double a = 10.0;
			if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
			{
				dlog.addText( Logger::BLOCK, "worldmodel dade, x Vector2D ye * -1");
				myRoute = myPos + Vector2D(theta.cos() * a /** -1*/, theta.sin() * a);
			}
			else
			{
				dlog.addText( Logger::BLOCK, "worldmodel dade, x Vector2D ye * 1!!!(na -1)");
				myRoute = myPos + Vector2D(theta.cos() * a, theta.sin() * a);
			}
		}
		else
		{
			dlog.addText( Logger::BLOCK, "my Body direction == %f", agent->world().self().body().degree());
			dlog.addText( Logger::BLOCK, "myRay_oppRoute is %f, %f", myRay_oppRoute.x, myRay_oppRoute.y);
			myRoute = targets[number];
			if
			(
				fastestOppPos.dist(myPos) < 3.5 and myRay_oppRoute.dist(fastestOppPos) < 4
				and
				fastestOppPos.dist(myRay_oppRoute) > fastestOppPos.dist(myRoute)
				and
				(fastestOppPos.dist(myRoute) * 2 > fastestOppPos.dist(myRay_oppRoute) or (abs(fastestOppPos.y) > 20 and fastestOppPos.x < -35) )
			)
			{
				dlog.addText( Logger::BLOCK, "age turn kone gand mikhore");
				dlog.addText( Logger::BLOCK, "fastestOppPos.dist(myRay_oppRoute) = %f", fastestOppPos.dist(myRay_oppRoute));
				dlog.addText( Logger::BLOCK, "fastestOppPos.dist(myRoute) = %f", fastestOppPos.dist(myRoute));dlog.flush();
				myRoute = myRay_oppRoute;
			}
			else if
				 (
					agent->world().gameMode().isPenaltyKickMode() 
					and fastestOppPos.dist(myRay_oppRoute) > fastestOppPos.dist(myRoute)
					and abs(fastestOppPos.y - myPos.y) < 3.5
					and fastestOppPos.dist(myRoute) * 2.2 > fastestOppPos.dist(myRay_oppRoute)
					and myLastDir_myNextDir < 45
				 )
			{
				dlog.addText( Logger::BLOCK, "bara penalty inja turn nakone behtare");dlog.flush();
				myRoute = myRay_oppRoute;
			}
			else
				dlog.addText( Logger::BLOCK, "===");
/*			Vector2D VECTOR2D ((oppRoutePoint - fastestOppPos) / ( oppRoutePoint.dist(fastestOppPos) ));
			dlog.addText( Logger::BLOCK, "VECTOR2D.r() is %f", VECTOR2D.r());
			myRoute = myRoute + VECTOR2D;*/
//			myRoute = targets[0];
		}
		dlog.addText( Logger::BLOCK, "myRoute is %f , %f", myRoute.x, myRoute.y);
	}

/*	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldModel dade!!! target");dlog.flush();
		myRoute = Vector2D(myRoute.x * -1, myRoute.y);
	}*/

	dlog.addText( Logger::BLOCK, "[FINAL] myRoute is %f , %f", myRoute.x, myRoute.y);
	if ( myPos.dist(myRoute) < 4 or (abs(fastestOppPos.y - myPos.y) > 0.3 and abs(fastestOppPos.y - myPos.y) < 2 and myPos.x < fastestOppPos.x) )
	{
		dlog.addText( Logger::BLOCK, "use my goToPoint");
		goToPoint( myRoute );
	}
	else
	{
		dlog.addText( Logger::BLOCK, "use Body_GoToPoint2008");
		double dash_power = 100.0;
		dash_power = agent->world().self().getSafetyDashPower(dash_power);

		Body_GoToPoint2008( myRoute, 1.0, dash_power ).execute(agent);
	}

	dlog.addText( Logger::BLOCK, "Exiting Block::goCloseToBall");
}

bool
Block::goToPoint( Vector2D target_point )
{
	dlog.addText( Logger::BLOCK, "Entering Block::goToPoint");

	dlog.addText( Logger::BLOCK, "abs(Ray2D(myPos, target_point).dir().degree() - agent->world().self().body().degree() ) = %f"
	, abs(Ray2D(myPos, target_point).dir().degree() - agent->world().self().body().degree() ));
	
	AngleDeg body = agent->world().self().body();
	double myNextDir = Ray2D(myPos, target_point).dir().degree();
	double myLastDir_myNextDir = body.degree() - myNextDir;
	dlog.addText( Logger::BLOCK ,"myLastDir - myNextDir = %f",myLastDir_myNextDir);
	if (body.degree() - myNextDir > 180)
		myLastDir_myNextDir -= 360;
	if (body.degree() - myNextDir < -180)
		myLastDir_myNextDir += 360;
	myLastDir_myNextDir = abs(myLastDir_myNextDir);
	dlog.addText( Logger::BLOCK ,"[FINAL]  myLastDir - myNextDir = %f",myLastDir_myNextDir);

	if ( myLastDir_myNextDir > 5 /*and myLastDir_myNextDir < 175 */)
	{
		dlog.addText( Logger::BLOCK, "turn");
		return Body_TurnToPoint( target_point ).execute(agent);
	}

	double dash_power;
	if ( myPos.dist(target_point) < 2.487 and myPos.dist(target_point) < fastestOppPos.dist(target_point) )
	{
		dash_power = 65;
	}
	else
	{
		dash_power = 100;
	}

	dash_power = agent->world().self().getSafetyDashPower(dash_power);

	dlog.addText( Logger::BLOCK, "dash");
	dlog.addText( Logger::BLOCK, "dash_power is %f", dash_power);
	agent->doDash( dash_power );
	dlog.addText( Logger::BLOCK, "Exiting Block::goToPoint");
	return true;
}

void
Block::setValidPoints()
{
	dlog.addText( Logger::BLOCK, "Entering Block::setValidPoints"); 

	FastIC fastIC;

	double angle = abs( oppRouteRay.dir().degree() );
//	dlog.addText( Logger::BLOCK, "before change angle, angle is %f", angle);
	angle -= 90;
	angle = abs(angle);
	angle /= 200;
	if ( fastestOppPos.x < oppRoutePoint.x )
	{
/*		dlog.addText( Logger::BLOCK, "we have to change oppRoutePoint");
		double angle = abs( Ray2D(fastestOppPos, oppRoutePoint + Vector2D(-6.5, 0)).dir().degree() );
		oppRoutePoint = Vector2D(oppRoutePoint + Vector2D(-6.5, 0));
		oppRouteLine = Line2D(fastestOppPos, oppRoutePoint);
		oppRouteRay = Ray2D(fastestOppPos, oppRoutePoint);

		dlog.addText( Logger::BLOCK, "before change angle, angle is %f", angle);
		angle -= 90;
		angle = abs(angle);
		angle /= 200;
*/		angle *= -1;
	}

#ifdef BLOCK_LOG
	double cot = oppRouteRay.dir().cos() / oppRouteRay.dir().sin();
	double tan = oppRouteRay.dir().tan();
	dlog.addText( Logger::BLOCK,"now angle is %f and tan oppRouteRay is %f and cot oppRouteRay is %f", angle, tan, cot);
#endif

	dlog.addText( Logger::BLOCK, "fastestOppPos.x is %f , angle is %f", fastestOppPos.x, angle);

	double oppDistToMe = myPos.dist(fastestOppPos);
//	double oppDistToMe = oppRouteLine.dist(myPos);
	dlog.addText( Logger::BLOCK, "oppDistToMe is %f", oppDistToMe);

	dlog.addText( Logger::BLOCK, "oppRouteRay.dir().degree() is %f", oppRouteRay.dir().degree());
	Vector2D origin = oppRouteRay.origin();
	dlog.addText( Logger::BLOCK, "origin o ray is %f. %f", origin.x, origin.y);
//	> bayad beshe < , ama ba'zi vaghta!!
	if(angle >= 0)
	for ( double x = fastestOppPos.x; x > fastestOppPos.x - angle * 60; x -= angle )
	{
		double y = oppRouteLine.getY(x);
		Vector2D target(x, y);
#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "x is %f and y is %f", x, y);
#endif
		Vector2D top_left( -52.5, -34 );
		Vector2D bottom_right(52.5, 34);
		Rect2D field( top_left, bottom_right );
		double oppDistToTarget = target.dist(fastestOppPos);
		oppDistToTarget *= 1;
		double Dist2 = oppDistToTarget - target.dist(myPos)/* - oppDistToTarget*/;
		double absMyDist_OppDist = abs(oppDistToTarget - target.dist(myPos));
		if ( (absMyDist_OppDist < oppDistToMe * 2/5  /*or (0 < Dist2 and Dist2 < oppDistToMe ) */) and field.contains(target) )
		{
			if ( 0 < Dist2 and Dist2 < oppDistToMe )
				dlog.addText( Logger::BLOCK, "0 < Dist2 < oppDistToMe ");
			if ( absMyDist_OppDist < oppDistToMe * 2/5 )
				dlog.addText( Logger::BLOCK, "absMyDist_OppDist < oppDistToMe * 2/5");

			dlog.addText( Logger::BLOCK, "-------------------------- set targets[%d] , points is %f , %f", targets.size(), target.x, target.y );
			targets.push_back(target);
		}
		if ( targets.size() > 20 or fastestOppPos.dist(target) > myPos.dist(fastestOppPos) * 2)
		{
			dlog.addText( Logger::BLOCK, "break");
			break;
		}
	}
	else
	for ( double x = fastestOppPos.x; x < fastestOppPos.x - angle * 60; x -= angle )
	{
		double y = oppRouteLine.getY(x);
		Vector2D target(x, y);
#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "x is %f and y is %f", x, y);
#endif
		Vector2D top_left( -52.5, -34 );
		Vector2D bottom_right(52.5, 34);
		Rect2D field( top_left, bottom_right );
		double oppDistToTarget = target.dist(fastestOppPos);
		oppDistToTarget *= 1;
		double Dist2 = oppDistToTarget - target.dist(myPos)/* - oppDistToTarget*/;
		double absMyDist_OppDist = abs(oppDistToTarget - target.dist(myPos));
		if ( (absMyDist_OppDist < oppDistToMe * 2/5  /*or (0 < Dist2 and Dist2 < oppDistToMe ) */) and field.contains(target) )
		{
			if ( 0 < Dist2 and Dist2 < oppDistToMe )
				dlog.addText( Logger::BLOCK, "0 < Dist2 < oppDistToMe ");
			if ( absMyDist_OppDist < oppDistToMe * 2/5 )
				dlog.addText( Logger::BLOCK, "absMyDist_OppDist < oppDistToMe * 2/5");

			dlog.addText( Logger::BLOCK, "-------------------------- set targets[%d] , points is %f , %f", targets.size(), target.x, target.y );
			targets.push_back(target);
		}
		if ( targets.size() > 20 or fastestOppPos.dist(target) > myPos.dist(fastestOppPos) * 2.5)
		{
			dlog.addText( Logger::BLOCK, "break");
			break;
		}
	}

	if ( targets.size() == 0 )
	{
		dlog.addText( Logger::BLOCK, "good points are not in the field!! searching for valid points");
//		< bayad beshe >, ama ba'zi vaghta!!!
		if (angle >= 0)
		for( double x = oppRoutePoint.x - angle; x < fastestOppPos.x; x += angle )
		{
			double y = oppRouteLine.getY(x);
			Vector2D target(x, y);
			dlog.addText( Logger::BLOCK, "x is %f and y is %f", x, y);
			Vector2D top_left( -52.5, -34 );
			Vector2D bottom_right(52.5, 34);
			Rect2D field( top_left, bottom_right );
			if ( field.contains(target) )
			{
				dlog.addText( Logger::BLOCK, "-------------------------- set targets[%d] , points is %f , %f", targets.size(), target.x, target.y );
				targets.push_back(target);
			}
			if ( targets.size() > 5 )
				break;
		}
		else
		for( double x = oppRoutePoint.x; x > fastestOppPos.x; x += angle )
		{
			double y = oppRouteLine.getY(x);
			Vector2D target(x, y);
			dlog.addText( Logger::BLOCK, "x is %f and y is %f", x, y);
			Vector2D top_left( -52.5, -34 );
			Vector2D bottom_right(52.5, 34);
			Rect2D field( top_left, bottom_right );
			if ( field.contains(target) )
			{
				dlog.addText( Logger::BLOCK, "-------------------------- set targets[%d] , points is %f , %f", targets.size(), target.x, target.y );
				targets.push_back(target);
			}
			if ( targets.size() > 5 )
				break;
		}
		
		
		if( angle == 0 )
		{
			dlog.addText( Logger::BLOCK, "angle is 0 !!!!!!!!!!!!!!!!");
			double angle2 = oppRouteRay.dir().degree();
			double yp = angle2 > 0? +0.5:-0.5;
			dlog.addText( Logger::BLOCK, "yp is %f", yp);

			for( uint i = 0; i < 30; i ++)
			{
				Vector2D target(fastestOppPos.x, fastestOppPos.y + i * yp);
				dlog.addText( Logger::BLOCK, "x is %f and y is %f", target.x, target.y);
				Vector2D top_left( -52.5, -34 );
				Vector2D bottom_right(52.5, 34);
				Rect2D field( top_left, bottom_right );
				if ( abs(target.dist(fastestOppPos) - target.dist(myPos)) < 4 and field.contains(target) )
					targets.push_back(target);
			}

			if (targets.size() == 0)
			{
				dlog.addText( Logger::BLOCK, "good points is in out of field!!! and angle is 0 !!!");
				for( uint i = 30; i >= 0; i --)
				{
					Vector2D target(fastestOppPos.x, fastestOppPos.y + i * yp);
					dlog.addText( Logger::BLOCK, "x is %f and y is %f", target.x, target.y); 
					Vector2D top_left( -52.5, -34 );
					Vector2D bottom_right(52.5, 34);
					Rect2D field( top_left, bottom_right );
					if ( field.contains(target) )
						targets.push_back(target);
					if ( targets.size() > 3 )
						break;
				}
			}
		}
	}
	dlog.addText( Logger::BLOCK, "Exiting Block::setValidPoints");
}

Vector2D
Block::getOppRoute_Penalty()
{
	dlog.addText( Logger::BLOCK, "Entering Block::getOppRoute_Penalty");dlog.flush();
	vector <Vector2D> routes;
	routes.push_back( Vector2D(-52, 0 ) );
	routes.push_back( Vector2D(-52, 14) );
	routes.push_back( Vector2D(-52,-14) );
	routes.push_back( Vector2D(-52, 7 ) );
	routes.push_back( Vector2D(-52,-7 ) );

	if ( abs(fastestOppPos.y) < 16 )
		routes.push_back( Vector2D(-52, fastestOppPos.y) );

	vector <Ray2D> rays;
	double least = 400;
	AngleDeg fastestOppDir = fastestOpp->body();
	Vector2D result;

	dlog.addText( Logger::BLOCK, "oon vasata");
	for ( uint i = 0; i < routes.size(); i++ )
	{
		dlog.addText( Logger::BLOCK, "too for oomade");
		rays.push_back( Ray2D(fastestOppPos, routes[i]) );
		if ( abs(rays[i].dir().sin() - fastestOppDir.sin()) < least )
		{
			least = abs(rays[i].dir().sin() - fastestOppDir.sin());
			result = routes[i];
			dlog.addText( Logger::BLOCK, "for Vector2D(%f, %f) least is %f",routes[i].x,routes[i].y, least);
		}
	}
	dlog.addText( Logger::BLOCK, "result = %f, %f", result.x, result.y);

	Vector2D V2D(52, fastestOppPos.y);
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		V2D = Vector2D(-52, fastestOppPos.y);
	}

	double opp_right = Ray2D(fastestOppPos, V2D).dir().degree();
	double opp_dir = fastestOpp->body().degree();
	double oppdir_oppright = opp_dir - opp_right;

	if (oppdir_oppright < -180)
		oppdir_oppright += 360;
	if (oppdir_oppright > 180)
		oppdir_oppright -= 360;

	if ( oppdir_oppright < 20 or fastestOppPos.absY() < 3)
	{
		dlog.addText( Logger::BLOCK, "return -52, 0");
		return Vector2D(-52, 0);
	}

	dlog.addText( Logger::BLOCK, "Exiting Block::getOppRoute_Penalty");
	return result;
}

Vector2D
Block::getOppRoute()
{
	dlog.addText( Logger::BLOCK, "Entering Block::getOppRoute");
	vector <Ray2D> rays;
	FastIC fastIC;
	const AbstractPlayerObject * fastestOpp = fastIC.getFastestOpponent();
	Vector2D fastestOppPos = fastestOpp->pos();
	Vector2D fastestOppVel = fastestOpp->vel();
	AngleDeg fastestOppDir = fastestOpp->body();
	Rect2D dangerField(	Vector2D(-52.5, -20), Vector2D(-36, 20) );
	Vector2D result;
	double least = 400;
	Vector2D ballPos = agent->world().ball().pos();

	if ( agent->world().gameMode().isPenaltyKickMode() )
	{
		dlog.addText( Logger::BLOCK, "PM_Mode == PenaltyKickMode, so return getOppRoute_Penalty");dlog.flush();
		return getOppRoute_Penalty();
	}

	vector <FastestPlayer> FPlayers = fastIC.getFastestPlayers();
	Vector2D receivePoint;
	int fastestReachCycle = fastIC.getFastestOpponentReachCycle();
	if ( fastIC.getFastestOpponent()  )
	{
		receivePoint = agent->world().ball().inertiaPoint(fastestReachCycle);
	}
	else
		receivePoint = fastestOppPos;

	Vector2D top_left( -40, -9 );
	Vector2D bottom_right(-20, 9);
	Rect2D thisRect(top_left, bottom_right);
	if ( thisRect.contains(fastestOppPos) )
	{
		dlog.addText( Logger::BLOCK, "fastestOpp in thisRect, return Vector2D(50, fastestOppPos.y)");
		return Vector2D(-50, fastestOppPos.y);
	}

	Vector2D top_left2( -52, -7 );
	Vector2D bottom_right2(-40, 7);
	Rect2D Danger(top_left2, bottom_right2);
	if ( Danger.contains(fastestOppPos))
	{
		dlog.addText( Logger::BLOCK, "fastestOppPos.dist(Vector2D(-52, 0)) < 14, so return -52, 0");
		return Vector2D(-52, 0);
	}

	if ( ((receivePoint.x < -30 and receivePoint.y > 20) or (fastestOppPos.x < -35 and fastestOppPos.y > 20))/* and myPos.x - fastestOppPos.x > 3*/ )
	{
		dlog.addText( Logger::BLOCK, "sari' bargard aghab!!!!");
		dlog.addText( Logger::BLOCK, "oppRoutePoint(result) is %f , %f", -46.0, 9.0);
		return Vector2D(-46.0, 9.0);
	}
	else if ( ((receivePoint.x < -30 and receivePoint.y < -20) or (fastestOppPos.x < -35 and fastestOppPos.y < -20)) /*and myPos.x - fastestOppPos.x > 3*/)
	{
		dlog.addText( Logger::BLOCK, "sari' bargard aghab!!!!");
		dlog.addText( Logger::BLOCK, "oppRoutePoint(result) is %f , %f", -46.0, -9.0);
		return Vector2D(-46.0, -9.0);
	}

//	set points
	if ( dangerField.contains( agent->world().ball().pos() ) )
	{
		dlog.addText( Logger::BLOCK, "Ball is in danger field");
		setOppRoutesInDangerField();
	}
	else
	{
		dlog.addText( Logger::BLOCK, "Ball is in Other Places");
		setOppRoutesInOtherPlaces();
	}

	dlog.addText ( Logger::BLOCK, "fastestOppVel.r() is %f",fastestOppVel.r() );
	dlog.addText( Logger::BLOCK, "oppRoutePoints.size() is %d",oppRoutePoints.size());

	if (fastestOppVel.r() < min_vel or fastIC.getFastestOpponentReachCycle() > N )
	{
		for( uint i = 0; i < oppRoutePoints.size(); i++)
		{
			Vector2D top_left( -35, -20 );
			Vector2D bottom_right(-25, 20);
			Rect2D thisRect(top_left, bottom_right);
			if ( thisRect.contains(receivePoint) )
			{
				result = Vector2D(-52, 0);
				dlog.addText( Logger::BLOCK, "*********** vel < %f and receivePoint is in thisRect oppRoutePoint(result) is %f , %f", min_vel, result.x, result.y); 
				return result;
			}
			if (receivePoint.dist(oppRoutePoints[i]) < least)
			{
				result = oppRoutePoints[i];
				least = receivePoint.dist(oppRoutePoints[i]);
//				dlog.addText( Logger::BLOCK, "least is %f and this point is %f , %f", least, oppRoutePoints[i].x, oppRoutePoints[i].y);
			}
		}
		dlog.addText( Logger::BLOCK, "*********** vel < %f oppRoutePoint(result) is %f , %f", min_vel, result.x, result.y); 
		return result;
	}

	if ( abs(fastestOppVel.dir().degree()) < 90 and (fastestOpp->pos().dist(ballPos) > 1.0 or abs(fastestOppVel.dir().degree()) < 45) )
	{
		dlog.addText( Logger::BLOCK, "mire samte darvaze khodeshun, return Vector2D(-52, 0)");
		return Vector2D(-52, 0);
	}

	Ray2D lastRay(fastestOppPos, Share::lastOppRoute);
	double lastLeast = abs(lastRay.dir().degree() - fastestOppDir.degree() );

	dlog.addText( Logger::BLOCK, "fastestOppDir is %f", fastestOppDir.degree());
	for ( uint i = 0; i < oppRoutePoints.size(); i++ )
	{
		rays.push_back( Ray2D(fastestOppPos, oppRoutePoints[i]) );
/*		if ( abs(rays[i].dir().degree() - fastestOppDir.degree()) < least )
		{
			least = abs(rays[i].dir().degree() - fastestOppDir.degree());
			result = oppRoutePoints[i];
			dlog.addText( Logger::BLOCK, "for Vector2D(%f, %f) least is %f",oppRoutePoints[i].x,oppRoutePoints[i].y, least);
		}
*/		if ( abs(rays[i].dir().sin() - fastestOppDir.sin()) < least )
		{
			least = abs(rays[i].dir().sin() - fastestOppDir.sin());
			result = oppRoutePoints[i];
			dlog.addText( Logger::BLOCK, "for Vector2D(%f, %f) least is %f",oppRoutePoints[i].x,oppRoutePoints[i].y, least);
		}
	}

	Vector2D tmp = result;
	if (Share::lastOppRoute.x != result.x or Share::lastOppRoute.y != result.y) 
	{
//		Share::lastOppRoute = result;
		tmp = Share::lastOppRoute;
	}

	Share::lastOppRoute = result;
	Vector2D left_up_R(-52.5, -22), right_down_R(-33, 22)
			,left_up_r(-52.5, -18), right_down_r(-38, 18);

	Rect2D Rect(left_up_R, right_down_R);
	Rect2D rect(left_up_r, right_down_r);

	dlog.addText( Logger::BLOCK, "Share::lastOppRouteCycle = %d", Share::lastOppRouteCycle);
	dlog.addText( Logger::BLOCK, "this cycle - shareCycle = %d", agent->world().time().cycle() - Share::lastOppRouteCycle);

	if ( abs(least - lastLeast) < 10 and lastLeast != least and agent->world().time().cycle() - Share::lastOppRouteCycle == 1 
	and !(Rect.contains(fastestOppPos) and !rect.contains(fastestOppPos)) )
	{
		dlog.addText( Logger::BLOCK, "fastestOppDir do not much change from last cycle, so use lastOppRoute");
		dlog.addText( Logger::BLOCK, "lastLeast is %f and this least is %f", lastLeast, least);
		dlog.addText( Logger::BLOCK, "mikhast Vector2D(%f, %f) ro entekhab kone", result.x, result.y);
		result = tmp;
	}

	dlog.addText( Logger::BLOCK, "oppRoutePoint(result) is %f , %f", result.x, result.y);
	dlog.addText( Logger::BLOCK ,"Exiting Block::getOppRoute");
	return result;
}

void
Block::setOppRoutesInDangerField()
{
//	dlog.addText( Logger::BLOCK, "Entering Block::setOppRoutesInDangerField");

	FastIC fastIC;
	const AbstractPlayerObject * fastestOpp = fastIC.getFastestOpponent();
	Vector2D fastestOppPos = fastestOpp->pos();
	vector <Vector2D> vectors;

	vectors.push_back( Vector2D(-50, -6) );
	vectors.push_back( Vector2D(-50, 0 ) );
	vectors.push_back( Vector2D(-50, 6 ) );
	if(fastestOppPos.y < -14)
		vectors.push_back( Vector2D(-46,-9) );
	if(fastestOppPos.y > 14)
		vectors.push_back( Vector2D(-46, 9) );

//	vectors.push_back( Vector2D(fastestOppPos.x,-20) );
//	vectors.push_back( Vector2D(fastestOppPos.x, 20) );

	for ( uint i = 0; i < vectors.size(); i++ )
		if ( fastestOppPos.dist(vectors[i]) > 4 )
			oppRoutePoints.push_back(vectors[i]);

//	dlog.addText( Logger::BLOCK, "Exiting Block::setOppRoutesInDangerField");
}

void
Block::setOppRoutesInOtherPlaces()
{
//	dlog.addText( Logger::BLOCK, "Entering Block::setOppRoutesInOtherPlaces");

	FastIC fastIC;
	vector <FastestPlayer> FPlayers = fastIC.getFastestPlayers();
	Vector2D receivePoint;
	if ( FPlayers.size() == 0 )
	{
		receivePoint = fastestOppPos;
	}
	else
		FPlayers[0].receivePoint;
//	const AbstractPlayerObject * fastestOpp = fastIC.getFastestOpponent();
	vector <Vector2D> vectors;

//	if( (fastestOppPos.x > -30 and fastestOppPos.y < -18) or (fastestOppPos.y > 20 and fastestOppPos.x < -20) )
	vectors.push_back( Vector2D(-52,-16) );
	vectors.push_back( Vector2D(-52, 0 ) );
	vectors.push_back( Vector2D(-52, 16) );
//	if( (fastestOppPos.x > -30 and fastestOppPos.y > 18) or (fastestOppPos.y < -20 and fastestOppPos.x < -20) )


	if (!( myPos.x - fastestOppPos.x > 3 and ((receivePoint.x < -15 and receivePoint.y > 20) or (fastestOppPos.x < -15 and fastestOppPos.y > 20)) ))
	    if (! agent->world().gameMode().isPenaltyKickMode() )
			vectors.push_back( Vector2D(-52, 28) );

	if (!( myPos.x - fastestOppPos.x > 3 and ((receivePoint.x < -15 and receivePoint.y < -20) or (fastestOppPos.x < -15 and fastestOppPos.y < -20)) ))
	    if (! agent->world().gameMode().isPenaltyKickMode() )
			vectors.push_back( Vector2D(-52,-28) );

	if ( fastestOppPos.y > 20 and fastestOppPos.x < -25)
		vectors.push_back( Vector2D(-46, 9) );
	if ( fastestOppPos.y < -20 and fastestOppPos.x < -25)
		vectors.push_back( Vector2D(-46, -9) );

	for ( uint i = 0; i < vectors.size(); i++ )
	{
		dlog.addText( Logger::BLOCK, "checking Vector2D(%f, %f)", vectors[i].x, vectors[i].y);
		if ( (myPos.y - fastestOppPos.y > 5) and myPos.y > 0 )
		{
			dlog.addText( Logger::BLOCK, ":D");
			if (vectors[i].y > myPos.y)
			{
				dlog.addText( Logger::BLOCK,"--------------------------------1");
				dlog.addText( Logger::BLOCK,"Vector2D is %f, %f", vectors[i].x, vectors[i].y);
				continue;
			}
		}
		else if ( (myPos.y - fastestOppPos.y < -5) and myPos.y < 0 )
		{
			dlog.addText( Logger::BLOCK, ":P");
			if (vectors[i].y < myPos.y)
			{
				dlog.addText( Logger::BLOCK,"--------------------------------2");
				dlog.addText( Logger::BLOCK,"Vector2D is %f, %f", vectors[i].x, vectors[i].y);
				continue;
			}
		}

		if ( fastestOppPos.dist(vectors[i]) > 7 )
			oppRoutePoints.push_back(vectors[i]);
	}

//	dlog.addText( Logger::BLOCK, "oppRoutePoints.size() is %d", oppRoutePoints.size());
//	dlog.addText( Logger::BLOCK, "Exiting Block::setOppRoutesInOtherPlaces");
}

void
Block::sideDashGetBall( bool changeBody )
{
	dlog.addText( Logger::BLOCK, "Entering Block::sideDashGetBall");
	Vector2D ballPos = agent->world().ball().pos();
	double body = agent->world().self().body().degree();
	double nextBodyDir;

	if ( changeBody == true )
	{
		double angle = 90;
		double nextBodyDir1 = angle + Ray2D(myPos, fastestOppPos).dir().degree();
		double nextBodyDir2 = -angle + Ray2D(myPos, fastestOppPos).dir().degree();

		if (nextBodyDir1 > 180)
			nextBodyDir1 -= 360;
		if (nextBodyDir2 < -180)
			nextBodyDir2 += 360;

#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "myBodyDirection is %f", body);
		dlog.addText( Logger::BLOCK, "nextBodyDir1 is %f", nextBodyDir1);
		dlog.addText( Logger::BLOCK, "nextBodyDir2 is %f", nextBodyDir2);
#endif

		double body_nextBodyDir1 = abs(body - nextBodyDir1);
		dlog.addText( Logger::BLOCK, "body_nextBodyDir1 = %f", body_nextBodyDir1);
		if ( abs(body - nextBodyDir1) > 180 )
		{
			dlog.addText( Logger::BLOCK, "abs(body - nextBodyDir1) > 180");
			body_nextBodyDir1 = 360 - body_nextBodyDir1;
		}

		double body_nextBodyDir2 = abs(body - nextBodyDir2);
		dlog.addText( Logger::BLOCK, "body_nextBodyDir2 = %f", body_nextBodyDir2);
		if ( abs(body - nextBodyDir2) > 180 )
		{
			dlog.addText( Logger::BLOCK, "abs(body - nextBodyDir2) > 180");
			body_nextBodyDir2 = 360 - body_nextBodyDir2;
		}

		dlog.addText( Logger::BLOCK, "body_nextBodyDir1 = %f", body_nextBodyDir1);
		dlog.addText( Logger::BLOCK, "body_nextBodyDir2 = %f", body_nextBodyDir2);
		if ( body_nextBodyDir1 < body_nextBodyDir2 )
			nextBodyDir = nextBodyDir1;
		else
			nextBodyDir = nextBodyDir2;

//		dlog.addText( Logger::BLOCK, "i want change myBodyDir to target %f, %f", target.x, target.y);
		dlog.addText( Logger::BLOCK, "body - nextBodyDir = %f", body - nextBodyDir);
		if ( abs(agent->world().self().body().degree() - nextBodyDir) < 10 )
			dlog.addText( Logger::BLOCK, "dont need to change bodyDir :)");
		else
		{
			dlog.addText( Logger::BLOCK, "need to change bodyDir :(");
			dlog.addText( Logger::BLOCK, "nextBodyDir is %f", nextBodyDir);
		    Body_TurnToAngle( AngleDeg(nextBodyDir) ).execute(agent);
//		   	Body_TurnToPoint( nextBodyDir, 0) ).execute(agent);
			return;
		}
	}

/*	if ( oppRouteLine.dist(myPos) > 0.5 and agent->world().gameMode().isPenaltyKickMode() and )
	{
		dlog.addText( Logger::BLOCK, "go close to ball, without change body direction");
		AngleDeg theta = agent->world().self().body();
		double a = 10;
		bool back_mode = false;
		double dashPower = 100;
		Vector2D target;
		Vector2D target1 = myPos + Vector2D(theta.cos() * a, theta.sin() * a);
		Vector2D target2 = myPos - Vector2D(theta.cos() * a, theta.sin() * a);
		dlog.addText( Logger::BLOCK, "target1 is %f, %f", target1.x, target1.y);
		dlog.addText( Logger::BLOCK, "target2 is %f, %f", target2.x, target2.y);
		dlog.addText( Logger::BLOCK, "oppRouteLine.dist(target1) = %f", oppRouteLine.dist(target1));
		dlog.addText( Logger::BLOCK, "oppRouteLine.dist(target2) = %f", oppRouteLine.dist(target2));
		if (oppRouteLine.dist(target1) < oppRouteLine.dist(target2) )
		{
			target = target1;
//			back_mode = true;
		}
		else
		{
			target = target2;
			back_mode = true;
		}
		Line2D myRay(myPos, theta.degree());
		if ( !myRay.isParallel(oppRouteLine) )
			target = myRay.intersection(oppRouteLine);
		dlog.addText( Logger::BLOCK, "target is %f, %f", target.x, target.y);
		dashPower = agent->world().self().getSafetyDashPower(dashPower);
		dlog.addText( Logger::BLOCK, "back_mode is %d", back_mode);
		Body_GoToPoint(target, 0.3, dashPower, 100, back_mode ).execute(agent);
		return;
	}*/
	dlog.addText( Logger::BLOCK, "ballPos is %f, %f",ballPos.x, ballPos.y);
	double dash_power = 100;
	if (agent->world().self().stamina() < 2800)
		dash_power = 80;

	Vector2D sideDashTarget;
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldMdoel dade!!! target");
		sideDashTarget = Vector2D(ballPos.x * -1, ballPos.y);
	}

	SideDashGoToPoint(agent->world().ball().pos(), 1, dash_power).execute(agent);

	dlog.addText( Logger::BLOCK, "Exiting Block::sideDashGetBall");
}

void
Block::goCloseToReceivePoint()
{
	dlog.addText( Logger::BLOCK, "Entering Block::goCloseToReceivePoint");
	const SelfObject & self = agent->world().self();
	FastIC fastIC;
	double minDist = 100;
	vector<Vector2D> besttargets;

	double Theta = Ray2D(myPos, fastestOppPos).dir().degree(),
		   minTheta = Theta - 45,
		   maxTheta = Theta + 45;
	dlog.addText( Logger::BLOCK, "Theta is %f", Theta);
	vector<FastestPlayer> FPlayers = fastIC.getFastestPlayers();
	Vector2D receivePoint = FPlayers[0].receivePoint;

	double angle = maxTheta - minTheta;
	angle /= 45;
	dlog.addText( Logger::BLOCK, "angle is %f", angle);

	for ( double theta = minTheta; theta < maxTheta; theta += angle)
	{
		dlog.addText( Logger::BLOCK, "theta is %f", theta);
		Vector2D bodyPos = myPos, bodyVel = self.vel();
		double bodyStamina = self.stamina();
		double bodyAngle = self.body().degree();

		Vector2D pos;
		if ( theta > 0 )
			pos = Ray2D(myPos, theta).intersection( Line2D(Vector2D(-60,40), Vector2D(60,40)) );
		else
			pos = Ray2D(myPos, theta).intersection( Line2D(Vector2D(-60,-40),Vector2D(60,-40)) );
		if ( theta == 0 )
			pos = Ray2D(myPos, theta).intersection( Line2D(Vector2D(60,40), Vector2D(60,-40)) );
		else if(theta == 180)
			pos = Ray2D(myPos, theta).intersection( Line2D(Vector2D(-60,40),Vector2D(-60,-40)) );

		dlog.addText( Logger::BLOCK, "pos is %f, %f", pos.x, pos.y);

		// Calculate turnCycle for NonSideIntercept
		double distThread = agent->world().self().kickableArea(); // kickableDist
		// calculate turn moment threshold
		float turnThread = 180.0; // if already reached, turn is not required.
		float targetDist = pos.dist(myPos);
		if (distThread < targetDist)
			turnThread = AngleDeg::asin_deg(distThread / targetDist);

		int turnCycle = 0;
		Vector2D r = agent->world().self().vel();

		while (true)
		{
			Vector2D targetRel = pos - bodyPos;
			float turnAngle = AngleDeg(targetRel.th().degree() - bodyAngle).degree();

			if (abs(turnAngle) < turnThread)
				break;

			turnAngle = turnAngle * (1 + agent->world().self().playerType().inertiaMoment() * r.r());
			if (abs(turnAngle) > ServerParam::i().maxMoment())
				turnAngle = AngleDeg((turnAngle / abs(turnAngle)) * ServerParam::i().maxMoment()).degree();
			turnAngle = turnAngle / (1 + agent->world().self().playerType().inertiaMoment() * r.r());
			r *= agent->world().self().playerType().playerDecay();
			bodyAngle += turnAngle;

			turnCycle++;
		}
		dlog.addText( Logger::BLOCK, "turncycle is %d", turnCycle);
		int dashReachCycle = fastIC.getFastestOpponentReachCycle() - turnCycle;
		dlog.addText( Logger::BLOCK, "dashReachCycle is %d", dashReachCycle);

		for ( int i = 0; i <= dashReachCycle; i++)
		{
			double dashPower = 100;
			if (bodyStamina < 1500/*ServerParam::i().recoverDecThrValue()*/)
				dashPower = 25;

			Vector2D dashAccelerate;
			dashAccelerate.setPolar(dashPower * self.dashRate(), bodyAngle);
			bodyVel += dashAccelerate;
			if (bodyVel.r() > self.playerType().playerSpeedMax())
				bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
			bodyPos += bodyVel;
			bodyVel *= self.playerType().playerDecay();
			bodyStamina -= dashPower - self.playerType().staminaIncMax();
		}
		dlog.addText( Logger::BLOCK, "BodyPos(a point on circle) = %f, %f", bodyPos.x, bodyPos.y);

		Vector2D oppRoute = getOppRoute();
		Ray2D oppRay(receivePoint, oppRoute);
		double ang = abs( oppRay.dir().degree() );
		dlog.addText( Logger::BLOCK, "before change angle, ang is %f", ang);
		ang -= 90;
		ang = abs(ang);
		ang /= 100;
		vector <Vector2D> testtargets;

		double cot = oppRay.dir().cos() / oppRay.dir().sin();
		double tan = oppRay.dir().tan();
		dlog.addText( Logger::BLOCK,"now ang is %f and tan oppRouteRay is %f and cot oppRouteRay is %f", ang, tan, cot);

		dlog.addText( Logger::BLOCK, "receivePoint.x is %f , ang is %f", receivePoint.x, ang);
		for ( double x = receivePoint.x; x > receivePoint.x - ang * 30; x -= ang )
		{
			double y = oppRay.line().getY(x);
			Vector2D target(x, y);
			dlog.addText( Logger::BLOCK, "x is %f and y is %f", x, y);
			Vector2D top_left( -52.5, -34 );
			Vector2D bottom_right(52.5, 34);
			Rect2D field( top_left, bottom_right );
			double oppDistToTarget = target.dist(receivePoint);
			oppDistToTarget *= 1;
			double oppDistToMe = bodyPos.dist(receivePoint);
//			double Dist2 = oppDistToTarget - target.dist(bodyPos)/* - oppDistToTarget*/;
			double absMyDist_OppDist = abs(oppDistToTarget - target.dist(bodyPos));
			if ( (absMyDist_OppDist < oppDistToMe * 2/5  /*or (0 < Dist2 and Dist2 < oppDistToMe ) */) and field.contains(target) )
			{
				if ( absMyDist_OppDist < oppDistToMe * 2/5 )
					dlog.addText( Logger::BLOCK, "absMyDist_OppDist < oppDistToMe * 2/5");

				dlog.addText( Logger::BLOCK, "--------- set testtargets[%d] , points is %f , %f", testtargets.size(), target.x, target.y );
				testtargets.push_back(target);
			}
			if ( testtargets.size() > 20 or receivePoint.dist(target) > oppDistToMe * 2)
				break;
		}

		if ( testtargets.size() == 0 )
		{
			dlog.addText( Logger::BLOCK, "good points are not in the field!! searching for valid points");
			for( double x = receivePoint.x - ang * 30; x < receivePoint.x; x += ang )
			{
				double y = oppRay.line().getY(x);
				Vector2D target(x, y);
				dlog.addText( Logger::BLOCK, "x is %f and y is %f", x, y);
				Vector2D top_left( -52.5, -34 );
				Vector2D bottom_right(52.5, 34);
				Rect2D field( top_left, bottom_right );
				if ( field.contains(target) )
				{
					dlog.addText( Logger::BLOCK, "----------------- set targets[%d] , points is %f , %f", testtargets.size(), target.x, target.y );
					testtargets.push_back(target);
				}
				if ( testtargets.size() > 5 )
					break;
			}
			if( ang == 0 )
			{
				dlog.addText( Logger::BLOCK, "ang is 0 !!!!!!!!!!!!!!!!");
				double ang2 = oppRay.dir().degree();
				double yp = ang2 > 0? +1:-1;
				dlog.addText( Logger::BLOCK, "yp is %f", yp);

				for( uint i = 0; i < 30; i ++)
				{
					Vector2D target(receivePoint.x, receivePoint.y + i * yp);
					dlog.addText( Logger::BLOCK, "x is %f and y is %f", target.x, target.y);
					Vector2D top_left( -52.5, -34 );
					Vector2D bottom_right(52.5, 34);
					Rect2D field( top_left, bottom_right );
					if ( abs(target.dist(receivePoint) - target.dist(myPos)) < 4 and field.contains(target) )
						testtargets.push_back(target);
				}

				if (targets.size() == 0)
				{
					dlog.addText( Logger::BLOCK, "good points is in out of field!!! and ang is 0 !!!");
					for( uint i = 30; i >= 0; i --)
					{
						Vector2D target(receivePoint.x, receivePoint.y + i * yp);
						dlog.addText( Logger::BLOCK, "x is %f and y is %f", target.x, target.y); 
						Vector2D top_left( -52.5, -34 );
						Vector2D bottom_right(52.5, 34);
						Rect2D field( top_left, bottom_right );
						if ( field.contains(target) )
							testtargets.push_back(target);
						if ( testtargets.size() > 3 )
							break;
					}
				}
			}
		}

		if ( minDist > bodyPos.dist( testtargets[testtargets.size() / 2] ) )
		{
			dlog.addText ( Logger::BLOCK, "minDist > bodyPos.dist( testtargets[testtargets.size() / 2] )");
			besttargets = testtargets;
			minDist = bodyPos.dist( testtargets[testtargets.size() / 2] );
			dlog.addText( Logger::BLOCK, "now minDist is %f", minDist);
		}

		dlog.addText( Logger::BLOCK, "---------------------------------------");
	}

	dlog.addText( Logger::BLOCK, "Exiting Block::goCloseToReceivePoint");
}

