// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <BlockIntercept.h>

#include <Share.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/player/intercept_table.h>
#include <rcsc/player/abstract_player_object.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/penalty_kick_state.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>

#include <FastIC.h>
#include <Basics.h>

#include <body_go_to_point2008.h>

//#define BLOCK_LOG

using namespace rcsc;
using namespace std;
using namespace Basics;

BlockIntercept::BlockIntercept( Vector2D RoutePoint )
{
	oppRoutePoint = RoutePoint;
}

bool
BlockIntercept::execute( PlayerAgent * agent )
{
	dlog.addText( Logger::BLOCK, "Entering BlockIntercept::execute");dlog.flush();

	const WorldModel & wm = agent->world();
	FastIC fastIC;
	fastIC.setByWorldModel(wm);

    const rcsc::PenaltyKickState * state = agent->world().penaltyKickState();
	fastestOpp = fastIC.getFastestOpponent();
	fastestOppPos = fastestOpp->pos();
	Vector2D myPos = agent->world().self().pos();
	
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldModel dade!!!");dlog.flush();
		myPos = Vector2D(myPos.x * -1, myPos.y);
		fastestOppPos = Vector2D(fastestOppPos.x * -1, fastestOppPos.y);
	}

	int fastestReachCycle = fastIC.getFastestOpponentReachCycle();
//	vector <FastestPlayer> FPlayers = fastIC.getFastestPlayers();
	if ( !fastIC.getFastestOpponent() )
	{
		dlog.addText( Logger::BLOCK, "fastestOpponent isn't valid !!! so return false  :(");dlog.flush();
		return false;
	}
/*	Vector2D receivePoint = FPlayers[0].receivePoint;
	Vector2D ballPos = fastIC.getFastestPlayers()[0].receivePoint;
	dlog.addText( Logger::BLOCK, "fastestPlayer unum is %d", FPlayers[0].unum);*/
	Vector2D receivePoint = agent->world().ball().inertiaPoint(fastestReachCycle);
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldModel dade!!!");dlog.flush();
		receivePoint = Vector2D(receivePoint.x * -1, receivePoint.y);
	}
	dlog.addText( Logger::BLOCK, "receivePoint = %f, %f", receivePoint.x, receivePoint.y);
	int fastestOppDelay = fastIC.getFastestOpponentReachCycle();

	float minTurnThread = 0.f;
	if (wm.self().pos().dist(receivePoint) > 30.f)
		minTurnThread = 2;

//	const Vector2D & inertiaFinalPoint = calculatingBall.inertiaFinalPoint();
	const Vector2D & inertiaFinalPoint = oppRoutePoint;

	dlog.addText( Logger::BLOCK, "inertiaFinalPoint is %f, %f", inertiaFinalPoint.x, inertiaFinalPoint.y);
	dlog.addText( Logger::BLOCK, "myPos is %f, %f", myPos.x, myPos.y);dlog.flush();
	Vector2D VECTOR2D ((oppRoutePoint - receivePoint) * 3 / ( oppRoutePoint.dist(receivePoint) * 4 ));
//	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
//	{
//		dlog.addText( Logger::BLOCK, "worldModel dade!!!");
//		VECTOR2D = Vector2D(VECTOR2D.x * -1, VECTOR2D.y);
//	}
	
#ifdef BLOCK_LOG
	dlog.addText( Logger::BLOCK, "oppRoutePoint.y - fastestOppPos.y = %f",oppRoutePoint.y - fastestOppPos.y);
	dlog.addText( Logger::BLOCK, "oppRoutePoint.x - fastestOppPos.x = %f",oppRoutePoint.x - fastestOppPos.x);
#endif
	dlog.addText( Logger::BLOCK, "VECTOR2D = %f, %f", VECTOR2D.x, VECTOR2D.y);
	Vector2D interceptPoint = receivePoint;
	double fastestOppMaxSpeed = fastestOpp->playerTypePtr()->playerSpeedMax();
	dlog.addText( Logger::BLOCK, "fastestOppMaxSpeed is %f", fastestOppMaxSpeed );
	dlog.addText( Logger::BLOCK, "VECTOR2D.r() is %f", VECTOR2D.r());

	dlog.addText( Logger::BLOCK, "fastestOppDelay = %d", fastestOppDelay);dlog.flush();
	int cycle = 0;
	Vector2D curBallPos;// = calculatingBall.inertiaPoint(cycle);
	while (true)
	{
#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "---->cycle = %d", cycle);dlog.flush();
#endif
		if ( cycle >= fastestOppDelay )
		{
#ifdef BLOCK_LOG
			dlog.addText( Logger::BLOCK, "---->cycle >= fastestOppDelay");
#endif
			curBallPos = receivePoint + VECTOR2D * (cycle - fastestOppDelay);
		}
		else
		{
#ifdef BLOCK_LOG
			dlog.addText( Logger::BLOCK, "---->cycle < fastestOppDelay");
#endif
			curBallPos = receivePoint;
		}

#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "curBallPos.dist(inertiaFinalPoint) = %f", curBallPos.dist(inertiaFinalPoint));
		dlog.addText( Logger::BLOCK, "curBallPos.dist(myPos) = %f", curBallPos.dist(myPos));
		dlog.addText( Logger::BLOCK, "curBallPos is %f, %f", curBallPos.x, curBallPos.y);
#endif
		if ( curBallPos.dist(inertiaFinalPoint) < 1.5 ) // if ball stopped
		{
			dlog.addText( Logger::BLOCK, "curBallPos.dist(inertiaFinalPoint) < 1");dlog.flush();
			break;
		}

		int arriveCycle = getArriveCycle(agent, curBallPos, cycle);
		if (arriveCycle != -1)
		{
			dlog.addText( Logger::BLOCK, "arriveCycle != -1, so dashTo curBallPos = %f, %f", curBallPos.x, curBallPos.y);
			dlog.addText( Logger::BLOCK, "cycle = %d", cycle);
			break;
		}
#ifdef BLOCK_LOG
		dlog.addText( Logger::BLOCK, "---------------------------------------------------");
#endif
		cycle++;
	}

	AngleDeg body = agent->world().self().body();
	dlog.addText( Logger::BLOCK, "myBodyDirection is %f", body.degree());
//	double nextDir = Ray2D(myPos, curBallPos).dir().degree();
	double nextDir = Ray2D(Vector2D(myPos.x * -1, myPos.y), Vector2D(curBallPos.x * -1, curBallPos.y)).dir().degree();
	dlog.addText( Logger::BLOCK, "nextDir is %f", nextDir);
	dlog.addText( Logger::BLOCK, "body.degree() - nextDir = %f", body.degree() - nextDir);
	double body_nextDir = body.degree() - nextDir;
	if (body_nextDir > 180)
		body_nextDir -= 360;
	if (body_nextDir < -180)
		body_nextDir += 360;

	body_nextDir = abs(body_nextDir);
	dlog.addText( Logger::BLOCK, "body_nextDir is %f", body_nextDir);

	if
	(
		body_nextDir < 10
		or
		(myPos.x - fastestOppPos.x > 1 and fastestOppPos.dist(myPos) > 3 and body_nextDir < 15)
		or
		(fastestOppPos.dist(myPos) > 10 and body_nextDir < 20 )
	)
	{
		dlog.addText( Logger::BLOCK, "i will change body direction a little, so dont change bodyDirection");
//		curBallPos = wm.self().playerType().inertiaPoint(wm.self().pos(), wm.self().vel(), 10);
		double a = 10.0;
		if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
			curBallPos = myPos + Vector2D(body.cos() * a * -1, body.sin() * a);
		else
			curBallPos = myPos + Vector2D(body.cos() * a, body.sin() * a);
		dlog.addText( Logger::BLOCK, "final point is %f, %f",curBallPos.x, curBallPos.y);
	}

	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
		dlog.addText( Logger::BLOCK, "worldModel dade!!! target");
		curBallPos = Vector2D(curBallPos.x * -1, curBallPos.y);
	}

	double dash_power = 100.0;
	dash_power = agent->world().self().getSafetyDashPower(dash_power);

	Body_GoToPoint2008( curBallPos, 1.0, dash_power ).execute(agent);

	Share::lastInterceptBlockingCycle = agent->world().time().cycle();
	dlog.addText( Logger::BLOCK, "Exiting BlockIntercept::execute");
	return true;
}

int
BlockIntercept::getArriveCycle(const PlayerAgent *agent, const Vector2D & pos, int maxCycle)
{
//	dlog.addText( Logger::BLOCK, "Entering BlockIntercept::getArriveCycle");

    const rcsc::PenaltyKickState * state = agent->world().penaltyKickState();
	int turnCycle = 0, /*sideDashReachCycle = -1,*/ dashReachCycle = 0, minusDashReachCycle = 0, plusDashReachCycle = 0;
	int reachCycle = 0;
	const SelfObject & self = agent->world().self();
	bool isMinusDash = false;
	bool isPlusDash = false;
	bool isTurnMove;
	double turnSimAngle;

	// Calculate turnCycle for NonSideIntercept
	double distThread = self.kickableArea(); // kickableDist
	// calculate turn moment threshold
	float turnThread = 180.0; // if already reached, turn is not required.
	float targetDist = pos.dist(self.pos());
	if (distThread < targetDist)
		turnThread = AngleDeg::asin_deg(distThread / targetDist);

	double bodyAngle = self.body().degree();
	Vector2D myPos = self.pos();
	Vector2D myVel = self.vel();
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
//		dlog.addText( Logger::BLOCK ,"worldModel dade!!!");
		myPos = Vector2D(myPos.x * -1, myPos.y);
		myVel = Vector2D(myVel.x * -1, myVel.y);
	}
	Vector2D targetRel = pos - myPos;
	Vector2D myPoint = self.playerType().inertiaPoint(myPos, myVel, 1);

	int maxTurnCycle = 99;
	for (int i = 0; i < maxTurnCycle; i++)
	{
		Vector2D bodyPos = myPos, bodyVel = myVel;
		Vector2D tmpPos(0, 0), tmpVel(0, 0);
		double bodyStamina = self.stamina();
		int turnInterceptTime = 0;
		double dashDist = 0;
		while (true)
		{
			if (pos.dist(bodyPos) < dashDist + distThread)
				break;

			// One cycle past
			if (turnInterceptTime > i - 1)
			{
				double dashPower = 100;
				if (bodyStamina < 1500/*ServerParam::i().recoverDecThrValue()*/)
					dashPower = 25;
				Vector2D dashAccelerate;
				dashAccelerate.setPolar(dashPower * self.dashRate(), 0);
				tmpVel += dashAccelerate;
				if (tmpVel.r() > self.playerType().playerSpeedMax())
					tmpVel.setPolar(self.playerType().playerSpeedMax(), tmpVel.th().degree());
				tmpPos += tmpVel;
				tmpVel *= self.playerType().playerDecay();
				dashDist = tmpPos.x;
			}
			turnInterceptTime++;

			bodyPos += bodyVel;
			bodyVel *= self.playerType().playerDecay();
			bodyStamina += self.playerType().staminaIncMax();
		}
		turnCycle = 0;
		bodyAngle = self.body().degree();
		Vector2D r = myVel;

		while (true)
		{
			Vector2D targetRel = pos - bodyPos;
			float turnAngle = AngleDeg(targetRel.th().degree() - bodyAngle).degree();

			if (abs(turnAngle) < turnThread)
				break;

			turnAngle = turnAngle * (1 + self.playerType().inertiaMoment() * r.r());
			if (abs(turnAngle) > ServerParam::i().maxMoment())
				turnAngle = AngleDeg((turnAngle / abs(turnAngle)) * ServerParam::i().maxMoment()).degree();
			if (turnCycle == 0)
				turnSimAngle = turnAngle;
			turnAngle = turnAngle / (1 + self.playerType().inertiaMoment() * r.r());
			r *= self.playerType().playerDecay();
			bodyAngle += turnAngle;

			turnCycle++;
		}
		if (turnCycle <= i)
		{
			isTurnMove = (i == 0);
			break;
		}
	}

	dashReachCycle = turnCycle;
	Vector2D bodyPos = myPos, bodyVel = myVel;
	double bodyStamina = self.stamina();
	for (int i = 0; i < turnCycle; i++)
	{
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();
		bodyStamina += self.playerType().staminaIncMax();
	}
	if (bodyStamina > ServerParam::i().staminaMax())
		bodyStamina = ServerParam::i().staminaMax();

	while (dashReachCycle < maxCycle)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
			break;
		double dashPower = 100;
		if (bodyStamina < 1500/*ServerParam::i().recoverDecThrValue()*/)
			dashPower = 25;

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * self.dashRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > self.playerType().playerSpeedMax())
			bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();
		bodyStamina -= dashPower - self.playerType().staminaIncMax();
		dashReachCycle++;
	}
	if (dashReachCycle >= maxCycle)
		dashReachCycle = -1;

//	Minus Dash Calculate
	bodyPos = myPos;
	bodyVel = myVel;
	bodyStamina = self.stamina();

	minusDashReachCycle = 0;
	bool ahmaghane = false;
	bodyAngle = self.body().degree();
	while (minusDashReachCycle < maxCycle && bodyStamina > 1500 && minusDashReachCycle < 7/*ServerParam::i().recoverDecThrValue()*/)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
		{
			ahmaghane = true;
			break;
		}

		double dashPower = -100;

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * self.dashRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > self.playerType().playerSpeedMax())
			bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();
		bodyStamina -= -2 * dashPower - self.playerType().staminaIncMax();
		minusDashReachCycle++;
	}
	if (!ahmaghane)
		minusDashReachCycle = -1;

//	Plus Dash Calculate
	bodyPos = myPos;
	bodyVel = myVel;
	bodyStamina = self.stamina();

	plusDashReachCycle = 0;
	ahmaghane = false;
	while (plusDashReachCycle < maxCycle && bodyStamina > 1500 && plusDashReachCycle < 50/*ServerParam::i().recoverDecThrValue()*/)
	{
		if (bodyPos.dist(pos) < distThread) // we reached the pos
		{
			ahmaghane = true;
			break;
		}

		double dashPower = 100;

		Vector2D dashAccelerate;
		dashAccelerate.setPolar(dashPower * self.dashRate(), bodyAngle);
		bodyVel += dashAccelerate;
		if (bodyVel.r() > self.playerType().playerSpeedMax())
			bodyVel.setPolar(self.playerType().playerSpeedMax(), bodyVel.th().degree());
		bodyPos += bodyVel;
		bodyVel *= self.playerType().playerDecay();
		bodyStamina -= dashPower - self.playerType().staminaIncMax();
		plusDashReachCycle++;
	}
	if (!ahmaghane)
		plusDashReachCycle = -1;

	reachCycle = dashReachCycle;
	if (minusDashReachCycle != -1 && minusDashReachCycle < reachCycle)
	{
		isMinusDash = true;
		reachCycle = minusDashReachCycle;
	}
	if (plusDashReachCycle != -1 && plusDashReachCycle < reachCycle)
	{
		isMinusDash = false;
		isPlusDash = true;
		reachCycle = plusDashReachCycle;
	}

//	dlog.addText( Logger::BLOCK, "Exiting BlockIntercept::getArriveCycle");
	return reachCycle;
}
