// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <BlockOpponent.h>

#include <body_go_to_point2008.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>

#include <rcsc/common/logger.h>
#include <rcsc/geom/angle_deg.h>

#include <Degree.h>
#include <cmath>

using namespace rcsc;

Vector2D BlockOpponent::mBlockPoint;

void BlockOpponent::execute(PlayerAgent* agent)
{
	wm = &agent->world();

	dlog.addText(Logger::ROLE, "I'm blocking opponent %d", targetOpp->unum());
	dlog.addText(Logger::ROLE, "TargetOpp: (%f, %f)", targetOpp->pos().x, targetOpp->pos().y);
	dlog.addText(Logger::ROLE, "selfPos: (%f, %f)", wm->self().pos().x, wm->self().pos().y);
	dlog.addText(Logger::ROLE, "TargetPoint: (%f, %f)", targetPoint.x, targetPoint.y);

	Vector2D blockPoint(-52.5, 0.0);

	// Calculating the goal
	Vector2D goal(-52.5, 0.0);
	if (targetOpp->pos().x > -28.0)
	{
		goal.y = targetPoint.y;
		blockPoint.y = targetPoint.y;
	}

	dlog.addText(Logger::ROLE, "goal: (%f, %f)", goal.x, goal.y);
	float bodyOppGoal = abs(Degree::getDegree(wm->self().pos(), targetPoint, goal));

	if (wm->self().pos().dist(goal) < targetPoint.dist(goal))
	{
		Vector2D p1 = wm->self().pos() - targetPoint;
		p1 = p1.normalize();
		Vector2D p2 = (goal - targetPoint);
		p2 = p2.normalize();
		bodyOppGoal = AngleDeg(p1.th().degree() - p2.th().degree()).degree();
//		dlog.addText(Logger::ROLE, "@@: %.2f", p1.th().degree());
//		dlog.addText(Logger::ROLE, "@@: %.2f", p2.th().degree());
//		bodyOppGoal = abs((wm->self().pos() - targetOpp->pos()).th().degree() - (goal - targetOpp->pos()).th().degree());

		float dist = AngleDeg(bodyOppGoal).sin() * wm->self().pos().dist(targetPoint) / AngleDeg(180 - 2 * bodyOppGoal).sin();


		Vector2D opp2Goal = goal - targetPoint;
		blockPoint = targetPoint + opp2Goal.setLength(dist);

		dlog.addText(Logger::ROLE, "BodyOppGoal : %f", bodyOppGoal);
		dlog.addText(Logger::ROLE, "Dist : %f", dist);

	}

//////////
/*		double k = 0.8;

		double sinA = AngleDeg::sin_deg(bodyOppGoal);
		double sinB = k * sinA;
		double B = AngleDeg::asin_deg(k * sinA);
		double b = sinB * (wm->self().pos() - targetPoint).r() / AngleDeg::sin_deg(180.0 - B - bodyOppGoal);

		Vector2D f = goal - targetPoint;
		f.setLength(b);
		blockPoint = targetPoint + f;

		if (b > (goal - targetPoint).r())
			blockPoint = goal;
*/
//////////

	Line2D goal2Opp(goal, targetPoint);
	if (goal2Opp.dist(wm->self().pos()) < 1.5)
		blockPoint = targetPoint;


	mBlockPoint = blockPoint;

	double goToPointThr = 3.0;

	// Checking if opp is in penalty area
	if (targetOpp->pos().x < -34.0)
	{
//		mBlockPoint = targetOpp->pos();
		mBlockPoint = targetPoint;

		if (targetOpp->vel().r() > 0.2)
//			mBlockPoint = targetOpp->pos() + targetOpp->vel() * 10;
			mBlockPoint = targetPoint + targetOpp->vel() * 10;

		goToPointThr = 2.0;
	}

	if (wm->self().pos().dist(targetPoint) < 10.0)
		goToPointThr = 1.0;

	mBlockPoint.x = min_max(-52.5, mBlockPoint.x, +52.5);
	mBlockPoint.y = min_max(-34.0, mBlockPoint.y, +34.0);

	dlog.addText(Logger::ROLE, "BlockOpp Point : (%f, %f)", mBlockPoint.x, mBlockPoint.y);

//	Body_GoToPoint2008(mBlockPoint, goToPointThr, getDashPower(), false).execute(agent);
	Body_GoToPoint(mBlockPoint, goToPointThr, getDashPower(), 100, false, true, 20.0).execute(agent);
}

int BlockOpponent::getDashPower()
{
	Vector2D goal(-52.5, 0.0);
	if (targetOpp->pos().x > -28.0)
		goal.y = targetOpp->pos().y;

	Line2D goal2Opp(goal, targetOpp->pos());

	int dashPower = 100;
	if (goal2Opp.dist(wm->self().pos()) < 5.0)
	{
		if (targetOpp->pos().x > -30.0 and wm->self().pos().x > -36.0 and targetPoint.x > wm->self().pos().x)
		{
			dashPower = 50;
			if (goal2Opp.dist(wm->self().pos()) < 2.0)
			{
				dashPower = 30;
				if (targetPoint.x > 10.0)
					dashPower = 20;
			}
			else if (goal2Opp.dist(wm->self().pos()) < 5.0)
				dashPower = min_max(40.0, goal2Opp.dist(wm->self().pos()) * 20.0, 75.0);
		}
	}

	dashPower = wm->self().getSafetyDashPower(dashPower);

	dlog.addText(Logger::ROLE, "BlockOpp DashPower : %d", dashPower);

	return dashPower;
}
