// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <DangerAreaClear.h>

#include <rcsc/action/body_kick_one_step.h>

#include <rcsc/common/logger.h>

using namespace rcsc;

bool DangerAreaClear::execute(rcsc::PlayerAgent* agent)
{
	const rcsc::WorldModel* wm = &agent->world();

	if (wm->ball().pos().x > -32.0 or wm->ball().pos().absY() > 25.0)
	{
		dlog.addText(Logger::CLEAR, "DANGER_AREA_CLEAR: ball out of danger area [FALSE]");
		return false;
	}

	if (!wm->self().isKickable() or !wm->existKickableOpponent())
	{
		dlog.addText(Logger::CLEAR, "DANGER_AREA_CLEAR: ball out of opp or self kickable [FALSE]");
		return false;
	}

	AngleDeg kickDir = (wm->ball().pos() - Vector2D(-52.5, 0.0)).th();

	Vector2D kickTarget;
	kickTarget.setPolar(30.0, kickDir);
	kickTarget += wm->ball().pos();

	Body_KickOneStep(kickTarget, 3.0, true).execute(agent);

	dlog.addText(Logger::CLEAR, "DANGER_AREA_CLEAR: executed successfuly [TRUE] (%f, %f)", kickTarget.x, kickTarget.y);
	return true;
}
