// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <strategy.h>

#include <DefensiveKick.h>

#include "body_kick_to_corner.h"
#include "body_dribble2008.h"
#include "body_hold_ball2008.h"
#include <body_clear_ball2008.h>

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/intention_kick.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/geom/vector_2d.h>

#include <iostream>

#include "HelliBASH09Pass.h"
#include "HelliBASH09SRP.h"
#include "DribbleDecision.h"
#include "DefenceBreaker.h"
#include "Config.h"
#include "PassConfig.h"
#include "PassDecision.h"
#include "PositioningConfig.h"

using namespace std;
using namespace rcsc;

bool DefensiveKick::execute( rcsc::PlayerAgent * agent )
{
	rcsc::dlog.addText( rcsc::Logger::PASS,
						"%s:%d: Bhv_BasicDefensiveKick"
						,__FILE__, __LINE__ );

	const WorldModel & wm = agent->world();

	if (wm.ball().pos().absY() < 20.0 and wm.ball().pos().x < -36.0 and
		Body_ClearBall2008().execute(agent))
	{
		dlog.addText( Logger::PASS, "Defense clear !" );
		return true;
	}

	DribbleDecision dribble(wm.self().pos());
	
	dribble.setRiskyDribbleRange(15);
	dribble.SRPDribble().setHasForbiddenPolygon(true);
	dribble.SRPDribble().setMaxForwardMovement(10);
	dribble.SRPDribble().forbiddenPolygon().addVertex(Vector2D(36,20));
	dribble.SRPDribble().forbiddenPolygon().addVertex(Vector2D(36,-20));
	dribble.SRPDribble().forbiddenPolygon().addVertex(Vector2D(52.5,-20));
	dribble.SRPDribble().forbiddenPolygon().addVertex(Vector2D(52.5,20));
	dribble.SRPDribble().setSafetyValues(0.9,1,0.9,1.1);

	Body_HoldBall2008 holdBall(true , Vector2D( 0 , 0 ) );
	
	dlog.addText( Logger::PASS , "KHAR!!!!!!!!!!!!!!!!" );
	dlog.flush();
	
	PassDecision pass(2);
	pass.setSafetyFactors(0.9,1.2,0.9,1,4,0.9,1.2,1,0,30);

	
	Decision dd = dribble.getValue(agent);
	Decision pd = pass.getValue(agent,true);

	if ( pd == SHOULD_DO_IT )
	{
//		cerr << "PASS!!!First!!!" << endl;
		rcsc::dlog.addText ( rcsc::Logger::PASS , "First Touch PASS!!");
		pass.execute(agent , false , 2);
		agent->setNeckAction( new rcsc::Neck_ScanField() );
		return true;
	}
	if ( dd == CAN_DO_IT )
	{
		rcsc::dlog.addText ( rcsc::Logger::PASS , "Dribbling!!");
		dribble.execute(agent);
		agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
		return true;
	}
	if ( pd == CAN_DO_IT )
	{
//		cerr << "PASS!!!!!!!!!!!" << endl;
		rcsc::dlog.addText ( rcsc::Logger::PASS , "PASS!!");
		pass.execute(agent , false , 2);
		agent->setNeckAction( new rcsc::Neck_ScanField() );
		return true;
	}
	if ( dd == CAN_NOT_DO_IT )
	{
		rcsc::dlog.addText ( rcsc::Logger::PASS , "ControlBALL!!!");
		agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
		holdBall.execute(agent);
		return true;
	}
	dribble.execute(agent);
	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	return true;
}
