// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <Mark.h>

#include <Basics.h>
#include <strategy.h>

#include <body_go_to_point2008.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/audio_sensor.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/body_turn_to_angle.h>
#include <rcsc/action/view_normal.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_relative.h>

#include <SideDashGoToPoint.h>
#include <BlockOpponent.h>
#include <Share.h>
#include <vector>

#include <FastIC.h>

using namespace rcsc;
using namespace std;
using namespace Basics;

bool Mark::mLookForTarget = false;

Mark::Mark(rcsc::PlayerAgent* _agent, int targetUnum):
	markTarget(NULL),
	agent(_agent),
	wm(&_agent->world()),

	markTargetUnum(targetUnum),

	nonPlayOnMark(false),

	distBehindOpp(3.0)
{
}

bool Mark::getValue()
{
	dlog.addText(Logger::ROLE | Logger::MARK, "Entering Mark::getValue");

	dlog.addText(Logger::MARK, "MarkTarget:\t%d", markTargetUnum);

	if (markTargetUnum > 11 or markTargetUnum < 1)
	{
		dlog.addText(Logger::MARK, "Illegal markTarget unum (not between 1 and 11)");
		return false;
	}

//	dlog.addText(Logger::MARK, "OPPs:");
	for (uint i = 0 ; i < wm->opponentsFromBall().size() ; i++)
	{
//		dlog.addText(Logger::MARK, "--> %d : (%f, %f)", wm->opponentsFromBall()[i]->unum(), wm->opponentsFromBall()[i]->pos().x,
//												wm->opponentsFromBall()[i]->pos().y);
		if ( wm->opponentsFromSelf()[i]->unum() == markTargetUnum)
		{
			markTarget = wm->opponentsFromSelf()[i];
//			break;
		}
	}

	if (!markTarget)
	{
		dlog.addText(Logger::MARK, "MarkTarget Unavailable");
		return false;
	}

	if (nonPlayOnMark)
		return true;

	return true;

	dlog.addText(Logger::ROLE | Logger::MARK, "Exiting Mark::getValue");
}

void Mark::execute()
{
	dlog.addText(Logger::ROLE | Logger::MARK, "Entering Mark::execute");

	Share::isMarkExecuted = true;
	mLookForTarget = false;
	isOffsideTrapSituation = false;

	Vector2D ballPos = wm->ball().pos() + wm->ball().vel() * 2;


//	TAVAJJOH **********************************************************
	static Vector2D lastCycleOppDir = markTarget->vel();
	static Vector2D lastRealOppDir = markTarget->vel();
	static bool doubt = false;

	dlog.addText(Logger::MARK, "============================================================");
	dlog.addText(Logger::MARK, "Doubt:\t%d", doubt);
	dlog.addText(Logger::MARK, "MARK_TARGET_VEL:\t%f,%f", markTarget->vel().th().degree(), markTarget->vel().r());
	dlog.addText(Logger::MARK, "OPP_DIR:\t%f, %f", oppDir.th().degree(), oppDir.r());
	dlog.addText(Logger::MARK, "LAST_CYCLE_OPP_DIR:\t%f,%f", lastCycleOppDir.th().degree(), lastCycleOppDir.r());
	dlog.addText(Logger::MARK, "LAST_REAL_OPP_DIR:\t%f,%f", lastRealOppDir.th().degree(), lastRealOppDir.r());
	dlog.addText(Logger::MARK, "REAL_BODY: %f / R_B_COUNT: %d", markTarget->realBody().degree(), markTarget->realBodyCount());
	dlog.addText(Logger::MARK, "SELF_BODY: %f", wm->self().body().degree());
	dlog.addText(Logger::MARK, "---------------------------------------------------------");

	oppDir = markTarget->vel();

	if ( AngleDeg(markTarget->vel().th().degree() - lastRealOppDir.th().degree()).abs() > 20.0 and 
		(oppDir.r() > 0.12 or (oppDir.r() > 0.09 and Share::isLastMarkActionChasboondan)) 
		and lastRealOppDir.r() > 0.09 )
	{
		if (!doubt or AngleDeg(markTarget->vel().th().degree() - lastCycleOppDir.th().degree()).abs() > 35.0)
		{dlog.addText(Logger::MARK, "^^^^^^^^^^^^^^^^^^^666666666666666666666666666666666666");
			oppDir.setDir(lastRealOppDir.th());
		}
		doubt = true;
	}
	else
	{
		if (lastRealOppDir.r() > 0.09)
			oppDir.setDir(lastRealOppDir.th());
		doubt = false;
	}

	if (markTarget->realBodyCount() < 2)
	{
		oppDir.setDir(markTarget->realBody());

		// Recongnizing the opponent back dash [oppDir]
		if ( AngleDeg(oppDir.th() + 180 - markTarget->vel().th()).abs() < 20.0 )
		{
			oppDir.setDir(oppDir.th() + 180);
			dlog.addText(Logger::MARK, "BACK BACK BACK BACK");
		}
		// --------------------------------------------
	}

	oppDir.setLength(markTarget->vel().r());

	lastCycleOppDir = markTarget->realBodyCount() < 2 ?	Vector2D().setPolar(markTarget->vel().r(), markTarget->realBody())
							  : 	markTarget->vel();

	lastRealOppDir = oppDir;

	// Recongnizing the opponent back dash [lastCycleOppDir]
	if ( markTarget->realBodyCount() < 2 and AngleDeg(lastCycleOppDir.th() + 180 - markTarget->vel().th()).abs() < 20.0 )
	{
		lastCycleOppDir.setDir(lastCycleOppDir.th() + 180);
		dlog.addText(Logger::MARK, "BACK BACK BACK");
	}
	// -----------------------------------------------------

	dlog.addText(Logger::MARK, "Doubt:\t%d", doubt);
	dlog.addText(Logger::MARK, "MARK_TARGET_VEL:\t%f,%f", markTarget->vel().th().degree(), markTarget->vel().r());
	dlog.addText(Logger::MARK, "OPP_DIR:\t%f, %f", oppDir.th().degree(), oppDir.r());
	dlog.addText(Logger::MARK, "LAST_OPP_DIR:\t%f,%f", lastCycleOppDir.th().degree(), lastCycleOppDir.r());
	dlog.addText(Logger::MARK, "============================================================");
//	TAVAJJOH **********************************************************

	Vector2D markPoint;
	if (nonPlayOnMark)
		markPoint = getNonPlayOnMarkPoint();
	else
		markPoint = getPlayOnMarkPoint();

	dlog.addText(Logger::MARK, "MARK_POINT:\t(%.2f, %.2f)", markPoint.x, markPoint.y);

	Vector2D turnPoint = getTurnPoint();

	dlog.addText(Logger::MARK, "I'm marking opponent %d", markTarget->unum());

	dlog.addText(Logger::MARK, "MarkTarget posCount:\t%d", markTarget->posCount());
	dlog.addText(Logger::MARK, "MarkTarget velCount:\t%d", markTarget->velCount());
	dlog.addText(Logger::MARK, "is our team ball possessor %d", wm->interceptTable()->isOurTeamBallPossessor());

	// NonPlayOn Mark ======================================================================================
	if (nonPlayOnMark)
	{
		// Going to markPoint (GoToPoint)
		if ( Body_GoToPoint2008(markPoint, markPoint.x < -34.0 ? 0.3 : 0.75, getDashPower(markPoint)).execute(agent) )
		{
			dlog.addText(Logger::MARK, "Mark --> Going to MARK pos!!!");
			dlog.addText(Logger::MARK, "Exiting Mark::execute");
			return;
		}

		// Changing the body direction
		if ( Body_TurnToPoint( turnPoint ).execute(agent) )
		{
			dlog.addText(Logger::MARK, "Mark --> Turning and setting the angle!!!");
			dlog.addText(Logger::MARK, "Exiting Mark::execute");
			return;
		}

		dlog.addText(Logger::ROLE | Logger::MARK, "Exiting Mark::execute -- nothing to do");
		return;
	}
	// ==================================================================================



	dlog.addText(Logger::MARK, "IS MARKTARGET SIDE OPP: %s", isMarkTargetSideOpp() ? "yes": "no");

	dlog.addText(Logger::MARK, "self_markPoint dist: %f", wm->self().pos().dist(markPoint));
	dlog.addText(Logger::MARK, "self_markTarget dist: %f", wm->self().pos().dist(markTarget->pos()));

	dlog.addText(Logger::MARK, "isLastMarkActionCHASBOONDAN: %d", Share::isLastMarkActionChasboondan);

	bool isChasboondanSituation = (
					(wm->self().pos().dist(markPoint) < 1.5 and wm->self().pos().dist(markTarget->pos()) < 5.0)
					or
					wm->self().pos().dist(markTarget->pos()) < 4.0
					or
					( (  (wm->self().pos().dist(markPoint) < 2.5 and wm->self().pos().dist(markTarget->pos()) < 6.0)
					or wm->self().pos().dist(markTarget->pos()) < 5.0) and
					Share::isLastMarkActionChasboondan )
	);

	if (markTarget->pos().x < -36.0)
		isChasboondanSituation = (
					wm->self().pos().dist(markPoint) < 1.5 * 2
					or
					wm->self().pos().dist(markTarget->pos()) < 1.0
					or
					( (wm->self().pos().dist(markPoint) < 1.8 or
					wm->self().pos().dist(markTarget->pos()) < 2.0) and
					Share::isLastMarkActionChasboondan )
	);

	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	Vector2D nextBallPos = wm->ball().pos() + wm->ball().vel() * 2;
	if (wm->existKickableOpponent())
	{
		if (FastIC::getFastestOpponent())
			nextBallPos = FastIC::getFastestOpponent()->pos();
		else
			nextBallPos = wm->ball().pos();
	}
	dlog.addText(Logger::MARK, "NextBallPos: (%f, %f)    markPoint (%f, %f)", nextBallPos.x, nextBallPos.y, markPoint.x, markPoint.y);
	if (nextBallPos.x > -31.0 and markPoint.x <= -36.0/*and wm->ball().pos().x > markTarget->pos().x + 25.0*/)
	{
		markPoint.x = max(markPoint.x, -36.0);

		if (markPoint.x > markTarget->pos().x)
		{
			turnPoint = wm->self().pos() + Vector2D(0.0, wm->ball().pos().y - wm->self().pos().y);
		}

		isOffsideTrapSituation = true;

		dlog.addText(Logger::MARK, "OFFSIDE TRAP !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
	}
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////

	// It's just a TOF....... must be debugged
/*	float a = hypot(3.5, markTarget->pos().dist(markPoint));
	bool b = isChasboondanSituation;
	isChasboondanSituation = isChasboondanSituation or ( (wm->self().pos().dist(markPoint) + wm->self().pos().dist(markTarget->pos())) < (2 * a) );
	if (!b and isChasboondanSituation)
		dlog.addText(Logger::MARK, "BEIZIIIIIIIIIIIIIIII!!!!");
*/	//////////////////////////////////////////

	int maxCount = 4;
	if (markTarget->pos().x > -20.0)
		maxCount = 8;

	dlog.addText(Logger::MARK, "MAX_COUNT: %d", maxCount);
	dlog.addText(Logger::MARK, "POS_COUNT: %d", markTarget->posCount());
	dlog.addText(Logger::MARK, "VEL_COUNT: %d", markTarget->velCount());

	// CHESBOONDANE 1
	if
	(
		isChasboondanSituation
		and
		(oppDir.r() > 0.12 or (oppDir.r() > 0.09 and Share::isLastMarkActionChasboondan))
		and
	//////////////////////////////////////////////////////////////////////////// just for debug
		(oppDir.th().abs() > 90 or markTarget->pos().x < -34.0)
	////////////////////////////////////////////////////////////////////////////////////////////
		and
		(markTarget->posCount() < maxCount and markTarget->velCount() < maxCount)
	)
	{
		dlog.addText(Logger::MARK, "++++ chesboondane 1 (vel > x)");

		Share::isLastMarkActionChasboondan = true;

		// ------------------------------------------------------
		Vector2D oppVel = oppDir * 10;

		if 
		(
			((wm->self().pos().y * oppVel.y) > 0.0 and markTarget->pos().absY() > 20.0)
			or
			(wm->ball().pos().y > markTarget->pos().y and wm->self().pos().y < markTarget->pos().y - 0.5 and oppVel.y < 0.0)
			or
			(wm->ball().pos().y < markTarget->pos().y and wm->self().pos().y > markTarget->pos().y + 0.5 and oppVel.y > 0.0)
		)
		{
			oppVel.y = 0.0;
			dlog.addText(Logger::MARK, "Vertical velocity filtering");
		}
		// ------------------------------------------------------

		dlog.addText(Logger::MARK, "POOOOOOOOOOOOOOOOOOOOOOOOOOOOO\tMarkTargetVel: %.2f", oppDir.th().degree());
		dlog.addText(Logger::MARK, "POOOOOOOOOOOOOOOOOOOOOOOOOOOOO\tMyBodyDir: %.2f", wm->self().body().degree());

/*		if (markTarget->pos().x < -36.0 and wm->ball().pos().x < -32.0)
		{
			// -----------------
			Vector2D newMarkPoint = wm->self().pos() + oppVel / 2;

			newMarkPoint.y = min_max(-30.0, newMarkPoint.y, +30.0);
			// -----------------

			Vector2D frontVec, backVec;
			frontVec.setPolar(5, wm->self().body());
			backVec.setPolar(5, wm->self().body() - 180);

			bool backMode = false;
			if ( (backVec + wm->self().pos()).dist(newMarkPoint) < (frontVec + wm->self().pos()).dist(newMarkPoint)
				and markTarget->pos().absY() < 18.0 and wm->self().pos().dist(newMarkPoint) < 3.0 )
				backMode = true;

			if (Body_GoToPoint(newMarkPoint, 0.5, getDashPower(newMarkPoint, true), 100, backMode, true, 15.0).execute(agent))
			{
				dlog.addText(Logger::MARK, "back_mode: %d", backMode);
				dlog.addText(Logger::MARK, "Mark --> chesboondane 1 BASE_GO_TO_POINT(%f, %f)", newMarkPoint.x, newMarkPoint.y);

				return;
			}

//			if (Body_GoToPoint2008(newMarkPoint, 0.5, getDashPower(newMarkPoint, true)).execute(agent))
//			{
//				dlog.addText(Logger::MARK, "Mark --> chesboondane 1 GO_TO_POINT_2008 (%f, %f)", newMarkPoint.x, newMarkPoint.y);
//
//				return;
//			}
		}
		else
*/		{
			// -----------------

			Vector2D newMarkPoint = wm->self().pos() + oppVel;

			if (markTarget->pos().x < -36.0 and wm->ball().pos().x < -32.0)
				newMarkPoint = wm->self().pos() + oppVel / 2;

			newMarkPoint.y = min_max(-30.0, newMarkPoint.y, +30.0);

			if (isOffsideTrapSituation)
				newMarkPoint.x = max(newMarkPoint.x, -36.0);
			// -----------------

			dlog.addText(Logger::MARK, "Using mySelf GoToPoint");

			if (AngleDeg( wm->self().body() - (newMarkPoint - wm->self().pos()).th()/*oppVel.th()*/ ).abs() < 20.0)
			{
				agent->doDash(getDashPower(newMarkPoint, true));
				dlog.addText(Logger::MARK, "Mark --> chesboondane 1 DASH (%f, %f)", newMarkPoint.x, newMarkPoint.y);
				return;
			}
			else if (Body_TurnToPoint(newMarkPoint).execute(agent))
			{
				dlog.addText(Logger::MARK, "Mark --> chesboondane 1 TURN (%f, %f)", newMarkPoint.x, newMarkPoint.y);
				return;
			}
		}

	}

	Share::isLastMarkActionChasboondan = false;

	// Block Opponent ------------------------------------------------------
	if (!isChasboondanSituation /*and !isOffsideTrapSituation*/)
	{
		BlockOpponent(markTarget, markPoint).execute(agent);
		dlog.addText(Logger::MARK, "Mark --> Going to MARK pos!!! ==>> [BLOCK_OPP] ");
		return;
	}


	// CHESBOONDANE 2
	{
		dlog.addText(Logger::MARK, "++++ chesboondane 2 (vel <= x)");

		dlog.addText(Logger::MARK, "selfPos: (%f, %f)", wm->self().pos().x, wm->self().pos().y);
		dlog.addText(Logger::MARK, "dist 2 markPoint: %f", wm->self().pos().dist(markPoint));

		double error = 0.5;

		if (oppDir.r() < 0.05 and markTarget->pos().x > -34.0)
		{
			error = 0.75;
		}

		Vector2D frontVec, backVec;
		frontVec.setPolar(5, wm->self().body());
		backVec.setPolar(5, wm->self().body() - 180);

		bool backMode = false;
		if ( (backVec + wm->self().pos()).dist(markPoint) < (frontVec + wm->self().pos()).dist(markPoint)
			and markTarget->pos().x < -36.0 and markTarget->pos().absY() < 18.0 and wm->self().pos().dist(markPoint) < 2.0 )
			backMode = true;


		if (Body_GoToPoint(markPoint, error, getDashPower(markPoint), 100, backMode, true, 15.0).execute(agent))
		{
			dlog.addText(Logger::MARK, "back_mode: %d", backMode);
			dlog.addText(Logger::MARK, "Mark --> chesboondane 2 DASH_GO_TO_POINT (%f, %f)", markPoint.x, markPoint.y);
			return;
		}
/*
		if (Body_GoToPoint2008(markPoint, error, getDashPower(markPoint)).execute(agent))
		{
			dlog.addText(Logger::MARK, "Mark --> chesboondane 2 GO_TO_POINT_2008 (%f, %f)", markPoint.x, markPoint.y);

			return;
		}
*/	}

	// Changing the body direction --------------------
	if ( Body_TurnToPoint( turnPoint ).execute(agent) )
	{
		dlog.addText(Logger::MARK, "Mark --> Turning and setting the angle!!! (%f, %f)", turnPoint.x, turnPoint.y);
		dlog.addText(Logger::MARK, "Exiting Mark::execute");
		return;
	}

	dlog.addText(Logger::ROLE | Logger::MARK, "Exiting Mark::execute -- nothing to do");
}

void Mark::setNonPlayOnMark(bool nonPlayOn)
{
	nonPlayOnMark = nonPlayOn;
}

Vector2D Mark::getPlayOnMarkPoint()
{
	Vector2D markTargetPos = markTarget->pos() + oppDir * 2;
	if (markTarget->pos().x < -36.0)
		markTargetPos = markTarget->pos();

	if (wm->interceptTable()->fastestOpponent() and wm->interceptTable()->fastestOpponent()->unum() == markTarget->unum())
		markTargetPos = markTarget->pos() + oppDir * 5;

	Vector2D ballPos = wm->ball().pos() + wm->ball().vel() * 2;
	Vector2D markPoint = ballPos - markTargetPos;

	double dist2Target = 1.5;
	if (markTarget->pos().dist(wm->ball().pos()) < 10.0)
		dist2Target = 0.75;

	markPoint.setLength(min(dist2Target, ballPos.dist(markTargetPos) / 3));
	markPoint += markTargetPos;

	markPoint.x -= 0.5;

	if (wm->ball().pos().x > -36.0 and wm->ball().pos().absY() < 20.0)
	{
		markPoint = markTarget->pos() + Vector2D(-0.5, 0.0);
		if (abs(markTarget->pos().y - wm->ball().pos().y) > 7.5 and wm->ball().pos().x < markTarget->pos().x + 6.0)
			markPoint.y = markTarget->pos().y + 0.3 * ( (wm->ball().pos().y - markTarget->pos().y) > 0.0 ? +1 : -1 );

		if (markTarget->pos().absY() > 15.0)
			markPoint.y = markTarget->pos().y + 0.7 * ( markTarget->pos().y > 0.0 ? -1 : +1 );
	}

	// dist behind opp
	distBehindOpp = 3.0;
	if (isMarkTargetSideOpp())
	{
		if (markTarget->pos().x > 0.0)
			distBehindOpp = 10.0;
		else if (markTarget->pos().x > -15.0)
			distBehindOpp = 6.0;
		else
			distBehindOpp = 3.0;
	}
	else if (wm->ball().pos().x < -36.0)
	{
		distBehindOpp = 1.2;
	}
	// ---------------

	float newY = markTargetPos.y;
	if (markTargetPos.x < -34.0 and wm->ball().pos().x < markTargetPos.x)
	{
		Vector2D markPoint = ballPos - markTargetPos;

		double dist2Target = 1.2;

		markPoint.setLength(min(dist2Target, ballPos.dist(markTargetPos) / 3));
		markPoint += markTargetPos;

		newY = markPoint.y;
	}

	// ------------------------------------------
	if (markTargetPos.x > -36.0)
	{
		markPoint.x = markTargetPos.x - min(markTargetPos.x + 36.0, distBehindOpp/*3.0*/);
		markPoint.y = newY;

/*		if (markTargetPos.x < -30.0 and markTargetPos.x > -34.0)
		{
			markPoint.y = markTargetPos.y;
			markPoint.x = markTargetPos.x - 2.0;
		}
*/
		// (bala, payin)tar vaistadan az yaroo
		if (!nonPlayOnMark or markTarget->pos().y * wm->ball().pos().y < 0)
		{
			if (markTarget->pos().y > 0.0 and isMarkTargetSideOpp())
			{
				markPoint.y = markTargetPos.y - 3.0;
			}
			else if (markTarget->pos().y < 0.0 and isMarkTargetSideOpp())
			{
				markPoint.y = markTargetPos.y + 3.0;
			}
		}
		// ------------------------------------

	}


	markPoint.x = min_max(-51.0, markPoint.x, +51.0);
	markPoint.y = min_max(-30.0, markPoint.y, +30.0);

	dlog.addText(Logger::MARK, "---Ball.pos: (%f , %f)", wm->ball().pos().x, wm->ball().pos().y);
	dlog.addText(Logger::MARK, "---Ball.vel: (%f , %f), %f", wm->ball().vel().x, wm->ball().vel().y, wm->ball().vel().r());
	dlog.addText(Logger::MARK, "---MarkTarget.pos: (%f , %f)", markTarget->pos().x, markTarget->pos().y);
	dlog.addText(Logger::MARK, "---MarkTarget.vel: (%f , %f), %f", oppDir.x, oppDir.y, oppDir.r());
	dlog.addText(Logger::MARK, "-- pos + vel * 2 : (%f, %f)", markTargetPos.x, markTargetPos.y);
	dlog.addText(Logger::MARK, "---MarkPoint: (%f , %f)", markPoint.x, markPoint.y);
	dlog.addText(Logger::MARK, " && %f && ", wm->ball().pos().dist(markTarget->pos()));

	return markPoint;
}

Vector2D Mark::getNonPlayOnMarkPoint()
{
	// use playOn markPoint when markTarget is far from ball (just for defenders)
	if (markTarget->pos().dist(wm->ball().pos()) > (wm->gameMode().side() == wm->ourSide() ? 15.0 : 25.0)
		and isInPlayers(wm->self().unum(), "34578") and markTarget->pos().x < 25.0)
		return getPlayOnMarkPoint();

	Vector2D markPoint = wm->ball().pos() - markTarget->pos();

	double dist2Target = 2.0;

	// our nonPlayOn
	if (wm->gameMode().side() == wm->ourSide() and agent->world().gameMode().type() != rcsc::GameMode::FreeKick_)
	{
		dist2Target = 6.5;
	}

	markPoint.setLength(min(dist2Target, wm->ball().pos().dist(markTarget->pos()) / 3));
	markPoint += markTarget->pos();

	// Checking the forbidden area in nonPlayOn mark
	if ( markTarget->pos().dist(wm->ball().pos()) < 5.0 )
	{
		Vector2D tmp = Vector2D(-52.5, 0.0) - wm->ball().pos();
		markPoint = tmp.setLength(9.15) + wm->ball().pos();
	}
	else if ( nonPlayOnMark and markTarget->pos().dist(wm->ball().pos()) < 9.15 )
	{
		Vector2D tmp = markTarget->pos() - wm->ball().pos();
		markPoint = tmp.setLength(9.15) + wm->ball().pos();
	}

	// Checking the opponent penalty area in goalKick mark
	if ( wm->gameMode().type() == GameMode::GoalKick_ and markTarget->pos().x > 25.0)
	{
		if ( markTarget->pos().absY() < 15.0 or (markTarget->pos().x - 36.0) < (20.0 - markTarget->pos().absY()) )
			markPoint = Vector2D(35.0, markTarget->pos().y);
		else if (markTarget->pos().absY() < 21.0)
			markPoint = Vector2D(markTarget->pos().x, markTarget->pos().y > 0.0 ? 21.0 : -21.0);

		Rect2D topArea(Vector2D(36.0, -34.0), Vector2D(53.0, -20.0));
		Rect2D bottomArea(Vector2D(36.0, 20.0), Vector2D(53.0, 34.0));
		Rect2D backArea(Vector2D(-53.0, -20.0), Vector2D(36.0, 20.0));

		Rect2D topCornerArea(Vector2D(-53.0, -34.0), Vector2D(36.0, -20.0));
		Rect2D bottomCornerArea(Vector2D(-53.0, 20.0), Vector2D(36.0, 34.0));

		if (!topCornerArea.contains(wm->self().pos()) and !bottomCornerArea.contains(wm->self().pos()))
		{
			if
			(
				(topArea.contains(markPoint) and backArea.contains(wm->self().pos()))
				or
				(topArea.contains(wm->self().pos()) and backArea.contains(markPoint))
				or
				(topArea.contains(wm->self().pos()) and bottomArea.contains(markPoint))
			)
				markPoint = Vector2D(35.0, -21.0);

			else if
			(
				(bottomArea.contains(markPoint) and backArea.contains(wm->self().pos()))
				or
				(bottomArea.contains(wm->self().pos()) and backArea.contains(markPoint))
				or
				(bottomArea.contains(wm->self().pos()) and topArea.contains(markPoint))
			)
				markPoint = Vector2D(35.0, 21.0);
		}

	}

	// Changing the markPoint in opponent goalieCatch mode
	if ( wm->gameMode().type() == GameMode::GoalieCatch_ and markTarget->pos().x > 25.0)
		markPoint = markTarget->pos() + Vector2D(+0.5, 0.0);

	return markPoint;
}

Vector2D Mark::getTurnPoint()
{
	double theta = (wm->ball().pos() - markTarget->pos()).dir().degree() ;
	Vector2D turnPoint = wm->self().pos() + Vector2D::polar2vector(10.0, AngleDeg(theta + (theta > 0 ? +90 : -90)));

	if (	!nonPlayOnMark
		and
		!( wm->ball().pos().y > markTarget->pos().y and AngleDeg(oppDir.th() - (-90)).abs() < 60 )
		and
		!( wm->ball().pos().y < markTarget->pos().y and AngleDeg(oppDir.th() - (+90)).abs() < 60 )
	)
		turnPoint = wm->self().pos() + oppDir * 50;

	if (
		nonPlayOnMark and ( markTarget->pos().dist(wm->ball().pos()) < 5.0 or
		(wm->gameMode().type() == GameMode::GoalKick_ and
		markTarget->pos().x > 36.0 and markTarget->pos().absY() < 20.0) )
	)
	{
		theta = (wm->ball().pos() - wm->self().pos()).dir().degree();
		turnPoint = wm->self().pos() + Vector2D::polar2vector(10.0, AngleDeg(theta + (theta > 0 ? +90 : -90)));
	}

	if (nonPlayOnMark and markTarget->pos().dist(wm->ball().pos()) > (wm->gameMode().side() == wm->ourSide() ? 15.0 : 25.0) and
		isInPlayers(wm->self().unum(), "34578") and markTarget->pos().x > -36.0)
		turnPoint = markTarget->pos();


	if (markTarget->pos().x > -36.0 and !nonPlayOnMark)
	{
		turnPoint = wm->self().pos() + Vector2D(1.0, 0.0);
		if (wm->ball().pos().x < -30.0 and wm->self().pos().dist(markTarget->pos()) < 4.0)
			turnPoint = wm->self().pos() + Vector2D(-1.0, 0.0);

		if 
		(
			( wm->ball().pos().x < markTarget->pos().x + 20.0 and abs(markTarget->pos().y - wm->ball().pos().y) > 7.0
			and (wm->self().pos().x > wm->ball().pos().x + 3.0 or wm->ball().pos().x > -20.0) )
			or
			(wm->ball().pos().x < -30.0 and wm->self().pos().dist(markTarget->pos()) < 4.0)
		)
		{
			if (markTarget->posCount() > 3)
			{
				turnPoint = markTarget->pos();
				mLookForTarget = true;
				dlog.addText(Logger::MARK, "LookForTarget !!!!!!!!!!!");
			}
			else
				turnPoint = wm->self().pos() + Vector2D(-1.0, 0.0);
		}
	}

	dlog.addText(Logger::MARK, "TurnPoint: (%f, %f)", turnPoint.x, turnPoint.y);

	return turnPoint;
}

int Mark::getDashPower(Vector2D markPoint, bool chasboondan)
{
	int dashPower = 100;

	if (wm->self().pos().dist(markPoint) < 1.0)
		dashPower = min_max(50, (int)wm->self().pos().dist(markPoint) * 50, 100);

	if (chasboondan)
	{
		if (wm->self().pos().dist(markPoint) < 1.0)
			dashPower = min_max(50, (int)wm->self().pos().dist(markPoint) * 50, 100);
		else
			dashPower = 100;
	}
	else
	if (markPoint.x > wm->self().pos().x and markTarget->pos().x > -30.0 and wm->self().pos().x > -36.0)
	{
		if (abs(markPoint.y - wm->self().pos().y) < 2.0)
		{
			dashPower = 30;
			if (wm->self().pos().x > markPoint.x - 2.0)
				dashPower = 25;
		}
		else if (abs(markPoint.y - wm->self().pos().y) < 4.0)
			dashPower = min_max(30.0, abs(markPoint.y - wm->self().pos().y) * 25.0, 100.0);
	}


	//////////////////////////////////
	if (isOffsideTrapSituation and markPoint.x > wm->self().pos().x)
		dashPower = 100;


	dlog.addText(Logger::MARK, "Mark Dash Power: %d %d", dashPower, wm->self().getSafetyDashPower(dashPower));

	dashPower = wm->self().getSafetyDashPower(dashPower);

	return dashPower;
}

bool Mark::isMarkTargetSideOpp()
{
	double calcY = markTarget->pos().y - (wm->ball().pos().y * 0.4);

	return abs(calcY) > 20.0;
}
