// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <SayMarkTable.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/basic_client.h>
#include <rcsc/player/audio_sensor.h>

#include <Share.h>
#include <WMatching.h>
#include <Basics.h>
#include <sstream>

using namespace std;
using namespace rcsc;
using namespace Basics;

bool SayMarkTable::execute(rcsc::PlayerAgent* agent)
{
	const rcsc::WorldModel* wm = &agent->world();

	if (wm->ball().pos().x < -20.0)
	{
		dlog.addText(Logger::ROLE, "Bad ball POSITION");
		return false;
	}

	if (!generateMarkTable(agent))
	{
		dlog.addText(Logger::ROLE, "No MarkTable generated!!!");
		return false;
	}
	dlog.addText(Logger::ROLE, "MarkTable generated!!!");

	stringstream msg("");
	msg << 'M' ;
	for (int i = 3 ; i < 12 ; i++)
	{
		bool markTargetFound = false;
		for (map<const PlayerObject*, const PlayerObject*>::iterator it = mMarkTable.begin() ; it != mMarkTable.end() ; it++)
		{
			if (it->first->unum() == i)
			{
				if (it->second->unum() == 10)
					msg << 'A' ;
				else if (it->second->unum() == 11)
					msg << 'B' ;
				else if (it->second->unum() == -1)
					msg << '0' ;
				else
					msg << it->second->unum() ;

				markTargetFound = true;
				break;
			}
		}
		if (!markTargetFound)
			msg << '0' ;
	}

	bool halfOpp = false;
	for (uint i = 0 ; i < 7 ; i++)
		if (msg.str()[i] == '0' and i != 4)
			halfOpp = true;

	dlog.addText(Logger::ROLE, " @ %s  * %s", Share::markSayMsg.c_str(), msg.str().c_str());

///////////////////////////
	vector<const PlayerObject*> addedPlayers, removedPlayers;

	for (map<const PlayerObject*, const PlayerObject*>::iterator it = mMarkTable.begin() ; it != mMarkTable.end() ; it++)
		if (!isInPlayers(it->second->unum(), Share::markSayMsg))
			addedPlayers.push_back(it->second);
	for (uint i = 0 ; i < Share::markSayMsg.size() ; i++)
	{
		char unumChar = Share::markSayMsg[i];
		int oppUnum;
		if (unumChar == 'A')
			oppUnum = 10;
		else if (unumChar == 'B')
			oppUnum = 11;
		else
			oppUnum = unumChar - '0';

		if (!isInPlayers(oppUnum, msg.str()))
		{
			for (uint j = 0 ; j < wm->opponentsFromSelf().size() ; j++)
				if (wm->opponentsFromSelf()[j]->unum() == oppUnum)
					removedPlayers.push_back(wm->opponentsFromSelf()[j]);
		}
	}

	for (uint i = 0 ; i < addedPlayers.size() ; i++)
		dlog.addText(Logger::ROLE, "added player: %d", addedPlayers[i]->unum());
	for (uint i = 0 ; i < removedPlayers.size() ; i++)
		dlog.addText(Logger::ROLE, "removed player: %d", removedPlayers[i]->unum());

	if (wm->ball().pos().x > 20.0 and addedPlayers.size() == 1 and removedPlayers.size() == 1 and
		addedPlayers[0]->pos().x < 26.0 and removedPlayers[0]->pos().x > addedPlayers[0]->pos().x + 10.0)
		;
	else if (wm->ball().pos().x > 20.0 and Share::markSayCycle != -1)
		return false;	
///////////////////////////wm->ball().pos().x > -20.0 and wm->ball().pos().x < 20.0 and 

	// Checking the vision *****************************
	bool vision = true;
	for (double i = -90.0 ; i <= 90.0 ; i+= 5.0)
		if (wm->dirCount(AngleDeg(i)) > 10)
		{
			vision = false;
			dlog.addText(Logger::MARK, "dirCount BAD !!! %d (%d)", i, wm->dirCount(AngleDeg(i)));
		}
	for (map<const PlayerObject*, const PlayerObject*>::iterator it = mMarkTable.begin() ; it != mMarkTable.end() ; it++)
		if (it->second->posCount() > 7 or !Share::isOppAlive[it->second->unum() - 1])
		{
			vision = false;
			dlog.addText(Logger::MARK, "oppPosCount BAD !!! %d (%d)", it->second->unum(), it->second->posCount());
		}
	for (uint i = 1 ; i < 7 ; i++)
		if (Share::markSayMsg[i] != '0' and i != 4)
		{
			char unumChar = Share::markSayMsg[i];
			int oppUnum;
			if (unumChar == 'A')
				oppUnum = 10;
			else if (unumChar == 'B')
				oppUnum = 11;
			else
				oppUnum = unumChar - '0';

			bool oppFound = false;
			for (uint j = 0 ; j < wm->opponentsFromSelf().size() ; j++)
				if (wm->opponentsFromSelf()[j]->unum() == oppUnum)
				{
					oppFound = true;
					if (wm->opponentsFromSelf()[j]->posCount() > 7)
					{
						vision = false;
						dlog.addText(Logger::MARK, "oppPosCount BAD (LastSay)!!! %d (%d)",
							wm->opponentsFromSelf()[j]->unum(), wm->opponentsFromSelf()[j]->posCount());
					}
				}

			if (!oppFound and Share::isOppAlive[oppUnum - 1])
				vision = false;
		}

	// *************************************************

	if
	(
		(msg.str().substr(0, 4) != Share::markSayMsg.substr(0, 4) or msg.str().substr(5, 2) != Share::markSayMsg.substr(5, 2)) 
	)
	{

		dlog.addText(Logger::ROLE, "Dir(0)Count: %d", wm->dirCount(AngleDeg(0.0)));

		//////////////////////////////////////////////////////
/*		if (Share::markSayCycle != -1 and msg.str() == Share::markSayMsg and wm->time().cycle() >= Share::generatedMarkSayCycle + 20)
		{
			Share::markSayMsg = Share::generatedMarkSayMsg;
			Share::markSayCycle = Share::generatedMarkSayCycle;
		}
		else
*/		//////////////////////////////////////////////////////
		if (halfOpp and
			agent->audioSensor().infoMessage().size() == 9 and agent->audioSensor().infoTime().cycle() > Share::lastKickOffCycle and
			Share::markSayMsg == "M000000000" and !wm->self().goalie())
		{
			Share::markSayMsg = "M" + agent->audioSensor().infoMessage();
			dlog.addText(Logger::ROLE, "*** Coach Say ***");
		}
		else if (halfOpp)
		{
			dlog.addText(Logger::ROLE, "HALF OPP");
			return false;
		}
		else if (vision)
		{
			// Starting default = -1;
			// Changing default = -2;
			if (Share::markSayCycle == -1)
			{
				Share::markSayMsg = msg.str();
				Share::markSayCycle = wm->time().cycle();
				Share::generatedMarkSayCycle = -2;
			}
			else if ( Share::generatedMarkSayCycle < 0 )
			{
				Share::generatedMarkSayMsg = msg.str();
				Share::generatedMarkSayCycle = wm->time().cycle();
			}
			else
			{
				if ( ( wm->time().cycle() - Share::generatedMarkSayCycle >= 20 ) and ( Share::generatedMarkSayMsg == msg.str() ) )
				{
					Share::markSayMsg = msg.str();
					Share::markSayCycle = wm->time().cycle();
					Share::generatedMarkSayCycle = -2;
				}
				if ( wm->time().cycle() - Share::generatedMarkSayCycle >= 30 and ( Share::generatedMarkSayMsg != msg.str() ) )
				{
					Share::generatedMarkSayCycle = -2;
				}
			}

			dlog.addText(Logger::ROLE, "MarksayMsg %s %d", Share::markSayMsg.c_str(), Share::markSayCycle);
			dlog.addText ( Logger::ROLE , "generatedMarkSayMsg %s %d" , Share::generatedMarkSayMsg.c_str() , Share::generatedMarkSayCycle );
			//////////////////////////////
			if (msg.str() != Share::markSayMsg)
			{
				dlog.addText(Logger::ROLE, "markSayMsg != msg.str()");
				return false;
			}
			//////////////////////////////

			dlog.addText(Logger::ROLE, "$$$ Self Say $$$");
		}
		else
		{
			dlog.addText(Logger::ROLE, "Bad Vision :D");
			dlog.addText(Logger::ROLE, "No say available!!!");
			return false;
		}

		dlog.addText(Logger::ROLE, "MarkTable Msg: %s", Share::markSayMsg.c_str());

		stringstream sayMsg("");
		sayMsg << "(say " << Share::markSayMsg << ")" ;

		return ( agent->M_client->sendMessage( sayMsg.str().c_str() ) > 0 );
	}
	else
		return false;
}


bool SayMarkTable::generateMarkTable(rcsc::PlayerAgent* agent)
{
	const rcsc::WorldModel* wm = &agent->world();

	mMarkTable.clear();
	vector<const PlayerObject*> dangerousOpp;
	// Recognizing very dangerous opponents
	for (uint i = 0 ; i < wm->opponentsFromSelf().size() ; i++)
		if ( wm->opponentsFromSelf()[i]->pos().x < -26.25 )
			dangerousOpp.push_back(wm->opponentsFromSelf()[i]);

	// Sorting very dangerous opponents
	for (uint i = 0 ; i < dangerousOpp.size() ; i++)
		for (uint j = 0 ; j < dangerousOpp.size() - 1 ; j++)
		{
			float opp1DistToGoal = min( min(dangerousOpp[j]->pos().dist(Vector2D(-52.5, +7.0)),
							dangerousOpp[j]->pos().dist(Vector2D(-52.5, -7.0))),
							dangerousOpp[j]->pos().dist(Vector2D(-52.5, 0.0)));
//			float opp1DistToGoal = worldModel->getFullPlayer(TID_OPPONENT, dangerousOpp[j]).pos().dist(Vector2D(-52.5, 0.0));
			float opp2DistToGoal = min( min(dangerousOpp[j + 1]->pos().dist(Vector2D(-52.5, +7.0)),
							dangerousOpp[j + 1]->pos().dist(Vector2D(-52.5, -7.0))),
							dangerousOpp[j + 1]->pos().dist(Vector2D(-52.5, 0.0)));
//			float opp2DistToGoal = worldModel->getFullPlayer(TID_OPPONENT, dangerousOpp[j + 1]).pos().dist(Vector2D(-52.5, 0.0));

			if (opp1DistToGoal > opp2DistToGoal)
			{
				const PlayerObject* temp = dangerousOpp[j + 1];
				dangerousOpp[j + 1] = dangerousOpp[j];
				dangerousOpp[j] = temp;
			}
		}

	uint veryDangerousOppCount = dangerousOpp.size();
	// Recognizing a little dangerous opponents
	for (uint i = 0 ; i < wm->opponentsFromSelf().size() ; i++)
		if ( wm->opponentsFromSelf()[i]->pos().x >= -26.25 and !wm->opponentsFromSelf()[i]->goalie() )
			dangerousOpp.push_back(wm->opponentsFromSelf()[i]);

	// Sorting a little dangerous opponents
	for (uint i = veryDangerousOppCount ; i < dangerousOpp.size() ; i++)
		for (uint j = veryDangerousOppCount ; j < dangerousOpp.size() - 1 ; j++)
			if ( dangerousOpp[j]->pos().x > dangerousOpp[j + 1]->pos().x )
			{
				const PlayerObject* temp = dangerousOpp[j + 1];
				dangerousOpp[j + 1] = dangerousOpp[j];
				dangerousOpp[j] = temp;
			}

	// pushing back the empty opponent
	for (uint i = dangerousOpp.size() ; i < 9 ; i++)
		dangerousOpp.push_back(new PlayerObject);

	if (dangerousOpp.size() > 9)
		dangerousOpp.resize(9);

	// Recognizing teammate markers
	vector<const PlayerObject*> defender;
	// Defense 34578
	for (uint i = 0 ; i < wm->teammatesFromSelf().size() ; i++)
		if ( isInPlayers(wm->teammatesFromSelf()[i]->unum(), "34578") )
		{
			defender.push_back(wm->teammatesFromSelf()[i]);
		}
	uint defenseCount = defender.size();
	// Middle 6
	for (uint i = 0 ; i < wm->teammatesFromSelf().size() ; i++)
		if ( wm->teammatesFromSelf()[i]->unum() == 6 )
		{
			defender.push_back(wm->teammatesFromSelf()[i]);
		}
	uint middleCount = defender.size() - defenseCount;
	// Offense 9AB
	for (uint i = 0 ; i < wm->teammatesFromSelf().size() ; i++)
		if ( isInPlayers(wm->teammatesFromSelf()[i]->unum(), "9AB") )
		{
			defender.push_back(wm->teammatesFromSelf()[i]);
		}
	uint offenseCount = defender.size() - middleCount - defenseCount;

	dlog.addText(Logger::ROLE, "dang: %d\t def:%d", dangerousOpp.size(), defender.size());
	dlog.addText(Logger::ROLE, "%d %d %d", defenseCount, middleCount, offenseCount);

	// Checking the vision on teammate defenders
	if (defenseCount < 5)
	{
		dlog.addText(Logger::ROLE, "I haven't seen all defenders");
		return false;
	}

	dlog.flush();
	// Weighted Matching =============================================================
	vector< vector<int> > wvec;
	for (uint i = 0 ; i < min((uint)5, defender.size()) ; i++)
	{
		wvec.push_back( vector<int>() );
		for (uint j = 0 ; j < min((uint)5, defender.size()) ; j++)
		{
			float weight = 1;

			weight = 125 - defender[i]->pos().dist(dangerousOpp[j]->pos());

			wvec[i].push_back(weight);// * 1000000);
		}
	}

	dlog.addText(Logger::ROLE, "weights --------------------------------");
	dlog.addText(Logger::ROLE, "  \t  %d\t%d\t%d\t%d\t%d", defender[0]->unum(), defender[1]->unum(), defender[2]->unum(), defender[3]->unum(), defender[4]->unum());
	dlog.addText(Logger::ROLE, "----------------------------------------");
	for (uint i = 0 ; i < 5 ; i++)
		dlog.addText(Logger::ROLE, "%d\t| %d\t%d\t%d\t%d\t%d", dangerousOpp[i]->unum(), wvec[0][i], wvec[1][i], wvec[2][i], wvec[3][i], wvec[4][i]);
	dlog.flush();

	WMatching wmatch(5, wvec);
	wmatch.calculate();

	for (uint i = 0 ; i < 5 ; i++)
	{
		mMarkTable[defender[wmatch.match[i]]] = dangerousOpp[i];
	}
	// ===============================================================================
	dlog.flush();

/*	dlog.addText(Logger::ROLE, "TMMs____");
	for (uint i = 0 ; i < wm->teammatesFromSelf().size() ; i++)
		dlog.addText(Logger::ROLE, "--> %d  (%f, %f)", wm->teammatesFromSelf()[i]->unum(),
			wm->teammatesFromSelf()[i]->pos().x, wm->teammatesFromSelf()[i]->pos().y);
*/	dlog.addText(Logger::ROLE, "OPPs____");
	for (uint i = 0 ; i < dangerousOpp.size() ; i++)
		dlog.addText(Logger::ROLE, "--> %d  (%.2f, %.2f) ( %d ) [%d]", dangerousOpp[i]->unum(),
			dangerousOpp[i]->pos().x, dangerousOpp[i]->pos().y, dangerousOpp[i]->posCount(), dangerousOpp[i]->posValid());

	dlog.addText(Logger::ROLE, "mMarkTable++++++++:");dlog.flush();
	for (map<const PlayerObject*, const PlayerObject*>::iterator it = mMarkTable.begin() ; it != mMarkTable.end() ; it++)
		dlog.addText(Logger::ROLE, "tmm: %d (%.2f, %.2f) \t|  opp: %d (%.2f, %.2f)", it->first->unum(), it->first->pos().x, it->first->pos().y,
											it->second->unum(), it->second->pos().x, it->second->pos().y);
	dlog.addText(Logger::ROLE, "MSIZE: %d", mMarkTable.size());dlog.flush();

	return true;
}
