// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <StraightBlock.h>

#include <Share.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <rcsc/player/abstract_player_object.h>
#include <rcsc/player/penalty_kick_state.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/player_agent.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_point.h>

#include <FastIC.h>
#include <Basics.h>

#include <SideDashGoToPoint.h>

#include <body_go_to_point2008.h>

using namespace rcsc;
using namespace std;
using namespace Basics;

StraightBlock::StraightBlock( Vector2D fastestOppPos, Vector2D RoutePoint )
{
	oppRoutePoint = RoutePoint;
	this->fastestOppPos = fastestOppPos;
}

bool
StraightBlock::execute( PlayerAgent * agent )
{
	dlog.addText( Logger::BLOCK, "Entering StraightBlock::execute");

	const rcsc::PenaltyKickState * state = agent->world().penaltyKickState();
	Vector2D myPos = agent->world().self().pos();
//	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
//		myPos = Vector2D(myPos.x * -1, myPos.y);

	Vector2D top_left( -15, -15 );
	Vector2D bottom_right(25, 15);
	Rect2D thisRect(top_left, bottom_right);

	AngleDeg body = agent->world().self().body();
	double myPos_oppPos = Ray2D(myPos, fastestOppPos).dir().degree();
	dlog.addText( Logger::BLOCK, "abs(body.degree() - myPos_oppPos) = %f", abs(body.degree() - myPos_oppPos));

	if ( !thisRect.contains(fastestOppPos) )
	{
		dlog.addText( Logger::BLOCK, "fastestOpp isn't in thisRect");
		return false;
	}
	if ( abs(oppRoutePoint.y - fastestOppPos.y) > 10 )
	{
		dlog.addText( Logger::BLOCK, "(oppRoutePoint.y - fastestOppPos.y) > 10");
		return false;
	}
	if ( abs(fastestOppPos.y - myPos.y) > 2 or (abs(fastestOppPos.y - myPos.y) > 1 and fastestOppPos.dist(myPos) < 2) )
	{
		dlog.addText( Logger::BLOCK, "I'm not in oppRouteLine");
		return false;
	}
	if (agent->world().self().stamina() < 3500)
	{
		dlog.addText( Logger::BLOCK, "stamina is less than 3500");
		return false;
	}
	if ( myPos.dist(fastestOppPos) < 1 )
	{
		dlog.addText( Logger::BLOCK, "kheili nazdikam, khatarie!!");
		return false;
	}
	if ( myPos.x > fastestOppPos.x )
	{
		if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
			dlog.addText( Logger::BLOCK, "velesh kon!");
		else
		{
			dlog.addText( Logger::BLOCK, "myPos.x > fastestOppPos.x");
			return false;
		}
	}
	if ( myPos.x < fastestOppPos.x )
	{
		if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
		{
			dlog.addText( Logger::BLOCK, "myPos.x < fastestOppPos.x");
			return false;
		}
		else
			dlog.addText( Logger::BLOCK, "velesh kon!");
	}
	if ( abs(body.degree() - myPos_oppPos) > 45 and myPos.dist(fastestOppPos) < 2.47 )
	{
		dlog.addText( Logger::BLOCK, "badbakht dare sidedash mikone toopo begire");
		return false;
	}
	if ( abs(myPos.y - fastestOppPos.y) / abs(myPos.x - fastestOppPos.x) > 0.6 )
	{
		dlog.addText( Logger::BLOCK, "abs(myPos.y - fastestOppPos.y) / abs(myPos.x - fastestOppPos.x) > 0.6");
		return false;
	}

	dlog.addText( Logger::BLOCK, "StraightBlock value is true");
	dlog.addText( Logger::BLOCK, "abs(myPos.y - fastestOppPos.y) = %f", abs(myPos.y - fastestOppPos.y));

	Vector2D newFastestOppPos = fastestOppPos;
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
		newFastestOppPos = Vector2D(newFastestOppPos.x * -1, newFastestOppPos.y);

	if ( abs(myPos.y - fastestOppPos.y) > 0.3 and newFastestOppPos.dist(myPos) < 20 )
	{
		dlog.addText( Logger::BLOCK, "use side dash");

		Vector2D target = Vector2D(myPos.x + 10, myPos.y);
		if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
			target = Vector2D(target.x - 20, myPos.y);

		Ray2D Ray(myPos, target);
		dlog.addText( Logger::BLOCK, "Ray.dir().degree() - body.degree() = %f", Ray.dir().degree() - body.degree());
		double body_ray = Ray.dir().degree() - body.degree();
		if (body_ray > 180)
			body_ray -= 360;
		if (body_ray < -180)
			body_ray += 360;

		dlog.addText( Logger::BLOCK, "body_ray = %f", body_ray);
		if ( body_ray > 20 )
		{
			dlog.addText( Logger::BLOCK, "we have to change body direction");
			return Body_TurnToPoint(target).execute(agent);
		}

		double Y = 1.0;
		if ( abs(myPos.y - fastestOppPos.y) < 0.6 )
			Y = 0.6;

		Vector2D target_point;
		double dash_power = 90;
		if ( agent->world().self().stamina() < 2800 )
			dash_power = 70;
		if ( myPos.y > fastestOppPos.y )
			target_point = Vector2D(myPos.x, myPos.y - Y);
		else
			target_point = Vector2D(myPos.x, myPos.y + Y);

		dlog.addText( Logger::BLOCK, "target_point is %f, %f", target_point.x, target_point.y);
		SideDashGoToPoint(target_point, 1, dash_power).execute(agent);
		return true;
	}

	dlog.addText( Logger::BLOCK, "use normal dash");
	Vector2D target = Vector2D(myPos.x + 10, myPos.y);
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
		target = Vector2D(target.x - 20, myPos.y);

	Ray2D Ray(myPos, target);
	if ( abs(Ray.dir().degree() - body.degree()) < 5 )
	{
		dlog.addText( Logger::BLOCK, "continue");
		double a = fastestOppPos.dist(myPos);
		if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
			target = myPos + Vector2D(body.cos() * a/* * -1*/, body.sin() * a);
		else
			target = myPos + Vector2D(body.cos() * a, body.sin() * a);
	}
	else
		dlog.addText( Logger::BLOCK, "change body direction");

	double dash_power = 100;
	if (fastestOppPos.dist(myPos) < 1)
		dash_power = 80;
	dash_power = agent->world().self().getSafetyDashPower(dash_power);

    return Body_GoToPoint( target, 0.5, dash_power ).execute(agent);

	dlog.addText( Logger::BLOCK, "Exiting StraightBlock::execute");
	return true;
}
