// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <GoalPlan.h>

#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>
#include <rcsc/action/intention_kick.h>
#include <rcsc/action/body_stop_ball.h>
#include <rcsc/action/body_turn_to_point.h>

#include <rcsc/player/interception.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_codec.h>

#include <OffenseVariables.h>

#include <Basics.h>
#include <Degree.h>
#include <FastIC.h>

#define MaxX 50
#define MaxY 30

using namespace rcsc;
using namespace std;

GoalPlan::GoalPlan() : PreDecidedPassBase()
{
	minBallMovement = 3.f;

	for ( int i = 0; i < 11; i++ )
	{
		addedValues[i] = 0.f;
	}
	shootingWeight = 1.f;
	angleWeight = 1.f;
}

GoalPlan::~GoalPlan()
{
}

void GoalPlan::changeBallDelay(rcsc::PlayerAgent *agent)
{
	const WorldModel & wm = agent->world();
	if ( not wm.self().isKickable() )
	{
		mBall.delay = wm.interceptTable()->selfReachCycle() - wm.time().cycle() + 1;
	}
	else
	{
		mBall.delay = 1;
		FastIC fast;
		for ( unsigned int i = 0; i < oppPlayers.size(); i++ )
			if ( oppPlayers[i].posCount <= maxOppPosCount )
				fast.players().push_back( FastPlayer( oppPlayers[i].mPos , oppPlayers[i].mVel , 0 , oppPlayers[i].uniNum , oppPlayers[i].isTeamMate 
										, 1 , 2 , oppPlayers[i].playerType , &(oppPlayers[i]) ) );
		fast.setBall( mBall.mPos , Vector2D(0.f , 0.f) , 0);
		fast.calculate( wm.self().unum() );
		if ( fast.getFastestPlayers().size() )
			mBall.delay = min ( 1 , fast.getFastestPlayers()[0].reachCycle - 1);
		if ( mBall.delay < 0 )
			mBall.delay = 0;
	}
//	dlog.addText ( Logger::PASS , "GoalPlan---->ballDelay	%d" , mBall.delay);
}

float GoalPlan::getPointShootingValue( PlayerAgent * agent , Vector2D receiverPos )
{
	if ( not agent->world().getOpponentGoalie() )
		return 1.f;
	float value = 0.f;
	const PlayerObject * goalie = agent->world().getOpponentGoalie();
	Vector2D goalieVel = Vector2D ( 0.f , 0.f );
	if ( goalie->velValid() )
		goalieVel = goalie->vel();
	FastIC shootingFast;
	shootingFast.players().push_back( FastPlayer( goalie->pos() , goalieVel , 0 , 1 /* Never mind */, false	, 2.f , 1.f , goalie->playerTypePtr()->id() , NULL ) );
	Vector2D ballPos = agent->world().ball().pos();
	Vector2D shootTargets[2];
	shootTargets[0] = Vector2D( 52.5f , 7.f );
	shootTargets[0] = Vector2D( 52.5f , -7.f );
	for ( unsigned int i = 0; i < 2; i++ )
	{
		Vector2D receiverToTarget = shootTargets[i] - receiverPos;
		receiverToTarget.setLength( ServerParam::i().ballSpeedMax() );
		shootingFast.setBall( receiverPos , receiverToTarget , 0 );
		shootingFast.calculate( agent->world().self().unum() );
		if ( not shootingFast.getFastestPlayers().size() )
			value += 0.5;
	}
	return value;
}

float GoalPlan::getPointAngleValue( Vector2D receivePos )
{
	Vector2D topBar = Vector2D ( 52.5f , 7.f );
	Vector2D bottomBar = Vector2D ( 52.5f , -7.f );
	float angle = fabsf(Degree::getDegree( topBar , receivePos , bottomBar ));
//	dlog.addText ( Logger::PASS , "\t\t\t\treceivePos ( %.2f , %.2f )" , receivePos.x , receivePos.y );
//	dlog.addText ( Logger::PASS , "\t\t\t\tangle\t%f" , angle );
	return ( angle / 180.f );
}

void GoalPlan::addValues( string players , float valueDiff )
{
	for ( unsigned int i = 0; i < players.size(); i++ )
	{
		int block = AudioCodec::i().hex2unum( players[i] ) - 1;
		if ( block < 0 )
			continue;
		if ( block > 10 )
			continue;
		addedValues[ block ] += valueDiff;
	}
}

float GoalPlan::getValue( PlayerAgent* agent )
{
	if ( agent->world().self().pos().x < 36 )
		return 0;
	dlog.addText ( Logger::PASS , "GoalPlan::getValue!!" );
	changeBallDelay( agent );
	tmmDelay = 3;
	tmmControlBuff = 1;
	tmmMaxDash = 0.85;
	oppDelay = 1;
//	oppAdditionalDelay = 2;
	oppControlBuff = 1;
	oppMaxDash = 1;
	maxTmmPosCount = 3;
	maxOppPosCount = 5;
	mBall.delay = 0;
	updateByWorldModel( agent->world() );
	constructFastIC();	
	float bestValue = 0.f;
	const WorldModel & wm = agent->world();

	maxDirectionCount = 2;

	float maxDistanceToGoal = 26;

	// TODO: optimizing Out of penalty area.
	if ( wm.self().pos().y < -20 )
	{
		startAngle = 0.f;
		endAngle = 180.f;
	}
	else if ( wm.self().pos().y > 20 )
	{
		startAngle = -180.f;
		endAngle = 0.f;
	}
	else
	{
		startAngle = -180.f;
		endAngle = 180.f;
	}
	startRate = 0.f;
	endRate = ServerParam::i().ballSpeedMax();
	float bestD , bestV;
	bestV = bestD = 0.f;
	float directionStep = (endAngle - startAngle) / 40.f;
//	dlog.addText ( Logger::PASS , "directionStep\t%f" , directionStep );
//	dlog.addText ( Logger::PASS , "startAngle\t%f\tendAngle\t%f" , startAngle , endAngle );
	int c = 0;
	mBall.delay = min ( mBall.delay , 2 );
	int tmpBallDelay = mBall.delay;
	for ( float direction = startAngle; direction <= endAngle; direction += directionStep )
	{
//		dlog.addText ( Logger::PASS , "direction\t%f" , direction );
		if ( agent->world().dirCount( direction ) > maxDirectionCount )
				continue;
		int max , sum , ave;
		agent->world().dirRangeCount( direction , 30 ,&max,&sum,&ave);
		if ( ave > maxDirectionCount )
			continue;

		float maxSpeed = endRate;
		if ( mBall.delay == 0 )
			maxSpeed = Body_KickOneStep::get_max_possible_vel( direction , wm.self().kickRate() , mBall.mVel ).r() - 0.1f;
		maxSpeed = min( maxSpeed , endRate );
		if ( maxSpeed < 1.0 )
			continue;
		float rateStep = ( maxSpeed - startRate ) / 4.f;
//		dlog.addText ( Logger::PASS , "rateStep\t%f" , rateStep );
		for ( float rate = startRate; rate < maxSpeed; rate += rateStep )
		{
//			dlog.addText ( Logger::PASS , "\trate\t%f" , rate );
			mBall.delay = tmpBallDelay;
			if ( rate < rateStep / 2.f )
				continue;
			c++;
			PreDecidedPassReceiver thisDirVel = getFastestDirectionVelocity( agent , rate , direction );
			Vector2D receivePoint = thisDirVel.receivePoint;
//			dlog.addText ( Logger::PASS , "\t\tDirection	%f--->Rate	%f	--%d->%d---> receivePoint(%f,%f) after %d cycles" , direction , rate , thisDirVel.receiver.isTeamMate , thisDirVel.receiver.uniNum ,receivePoint.x , receivePoint.y , thisDirVel.reachCycle );
			if ( thisDirVel.receiver.uniNum == wm.self().unum() )
			{
				continue;
			}
			if ( thisDirVel.receiver.uniNum < 0 )
			{
				continue;
			}
			if ( not thisDirVel.receiver.isTeamMate )
			{
				continue;
			}
			if ( thisDirVel.receiver.isInOffside )
			{
				continue;
			}
			if ( fabsf( receivePoint.x ) > MaxX )
			{
				continue;
			}
			if ( fabsf( receivePoint.y ) > MaxY )
			{
				continue;
			}
			if ( receivePoint.x < 36.f )
			{
				continue;
			}
			float receiverDist = thisDirVel.receiver.mPos.dist ( wm.self().pos() );
			if ( receiverDist < 2.f )
				continue;
			int receiverUniNum = thisDirVel.receiver.uniNum;
			if ( receivePoint.dist ( Vector2D( 52.5 , 0 ) ) > maxDistanceToGoal )
				continue;
			float ballMove = receivePoint.dist ( mBall.mPos );
			if ( ballMove < minBallMovement )
				continue;
/*			dlog.addText( Logger::PASS , "direction\t%f\trate\t%f----> ( %.2f , %.2f )"  , direction , rate , receivePoint.x , receivePoint.y );
			dlog.addText( Logger::PASS , "\tshootingValue\t%f" , getPointShootingValue( agent , receivePoint ) );
			dlog.addText( Logger::PASS , "\tpointValue\t%f" , getPointAngleValue( receivePoint ) );
			dlog.addText( Logger::PASS , "\taddedValues[]\t%f" , addedValues[receiverUniNum - 1] );
			dlog.addText( Logger::PASS , "\tballMove\t%f" , ballMove );
*/			float weightSum = shootingWeight + angleWeight + addedValues[receiverUniNum - 1];
			float thisValue = 1.f;
			if ( weightSum != 0.f )
				thisValue = ( ( shootingWeight * getPointShootingValue( agent , receivePoint ) ) + ( angleWeight * getPointAngleValue( receivePoint ) ) 
							+ addedValues[receiverUniNum -1 ] ) / weightSum;
//			dlog.addText( Logger::PASS , "\t\t\tvalue\t%f" , thisValue );
			if ( thisValue > bestValue )
			{
				bestValue = thisValue;
				tmpStats.velocity = rate;
				tmpStats.releaseCycle = agent->world().time().cycle() + mBall.delay;
				tmpStats.validaty = mBall.delay + 1;
				tmpStats.decideTime = agent->world().time().cycle();
				tmpStats.receiverStats = thisDirVel;
				tmpStats.sayValidTime = agent->world().time().cycle() + mBall.delay + 1;
				tmpStats.ballPosBeforeRelease = getReleasePointDirectionVelocity( agent , rate , direction );
				tmpStats.kickTarget = kickTargetPoint[direction][rate];
				bestD = direction;
				bestV = rate;
				setThisCycle = true;
			}
		}
	}
	dlog.addText( Logger::PASS , "GoalPlan	%.2f------>%.2f	player %d receiving	in (%.2f , %.2f) after %d cycles releasing in (%.2f , %.2f )--->Value	%.2f" , bestD , bestV , tmpStats.receiverStats.receiver.uniNum ,
				tmpStats.receiverStats.receivePoint.x , tmpStats.receiverStats.receivePoint.y , tmpStats.receiverStats.reachCycle , tmpStats.ballPosBeforeRelease.x , tmpStats.ballPosBeforeRelease.y , bestValue);
	return bestValue;
}

