// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <GoalPlanDribble.h>

#include <rcsc/soccer_math.h>

#include <rcsc/player/world_model.h>

#include <rcsc/common/logger.h>

#include <body_dribble2008.h>

#include <SRPDribble.h>

#include <FastIC.h>

using namespace rcsc;

bool GoalPlanDribble::isTargetSet = false;
Vector2D GoalPlanDribble::currentTarget = Vector2D ( 52.5f , 0.f );
DribbleDirection GoalPlanDribble::currentDribbleDirection = DD_NONE;

GoalPlanDribble::GoalPlanDribble()
{
	dribbleTarget = Vector2D ( 52.5f , 0.f );
};

void GoalPlanDribble::changeTarget( const rcsc::PlayerAgent * agent )
{
}

GoalDribbleSituation GoalPlanDribble::getValue( rcsc::PlayerAgent * agent )
{
//	if ( SRPDribble::SRPDribble().execute ( agent ) )
//		return GD_SHOULD;
	const WorldModel & wm = agent->world();
	bool shouldBreak = false;
	{
		FastIC fastIC( false );
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			if ( (*it)->posCount() > 3 )
				continue;
			if ( (*it)->goalie() )
				continue;
			Vector2D hisPos = (*it)->pos();
			if ( hisPos.x < 36.f )
				continue;
			int hisType = 0;
			if ( (*it)->playerTypePtr() )
				hisType = (*it)->playerTypePtr()->id();
			fastIC.players().push_back(	FastPlayer( hisPos , Vector2D( 0.f , 0.f ) , 1 , (*it)->unum() , false , 1.f , 1.f , hisType , NULL , NULL ) );
		}
		Vector2D ballVel = Vector2D ( 52.5f , 0.f ) - wm.self().pos();
		ballVel.setLength( 0.6 );
		fastIC.setBall ( wm.ball().pos() , ballVel , 0 );
//		dlog.addText ( Logger::PASS , "Calculating!!" );
		fastIC.calculate( wm.self().unum() , 10 );
		if ( fastIC.getFastestPlayers().size() )
		{
			Vector2D ballReceivePoint = fastIC.getFastestPlayers()[0].receivePoint;
			float dist = ballReceivePoint.dist( wm.ball().pos() );
			if ( dist > 2.f )
				shouldBreak = true;
		}
		else
			shouldBreak = true;
	}
	if ( shouldBreak )
	{
		dribbleTarget = Vector2D ( 52.5f , 0.f );
		currentDribbleDirection = DD_BREAK;
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			if ( (*it)->posCount() > 5 )
				continue;
			Vector2D hisPos = (*it)->pos();
			if ( hisPos.dist( wm.ball().pos() ) <= 1.5f )
				return GD_SHOULD_NOT;
			break;
		}
		return GD_SHOULD;
	}
	else if ( fabsf ( wm.ball().pos().y ) < 20.f )
	{
		Vector2D goalCenter = Vector2D( 52.5f , 0.f );
		int partsCount = 6;
		float partsSize = 180.f / (float)partsCount;
		Vector2D meToGoal = goalCenter - wm.self().pos();
		float ourCurrentAngularPos = 90.f + meToGoal.th().degree();
		int currentPart = (int)(ourCurrentAngularPos / partsSize);
		float clockWiseNextAngle = ( currentPart + 1 ) * partsSize + 90.f + 5.f;
		float counterClockWiseNextAngle = currentPart * partsSize + 90.f - 5.f;
		float myDist = wm.self().pos().dist( goalCenter );

		Vector2D clockWiseCandidate = Vector2D::polar2vector( myDist , clockWiseNextAngle ) + goalCenter;
		if ( clockWiseCandidate.x < 38.f )
		{
			Vector2D goalToCandidate = clockWiseCandidate - goalCenter;
			goalToCandidate *= ( 52.5 - 38.f ) / ( 52.5 - clockWiseCandidate.x );
			clockWiseCandidate = goalCenter + goalToCandidate;
		}

		Vector2D counterClockWiseCandidate = Vector2D::polar2vector( myDist , counterClockWiseNextAngle ) + goalCenter;
		if ( counterClockWiseCandidate.x < 38.f )
		{
			Vector2D goalToCandidate = counterClockWiseCandidate - goalCenter;
			goalToCandidate *= ( 52.5 - 38.f ) / ( 52.5 - counterClockWiseCandidate.x );
			counterClockWiseCandidate = goalCenter + goalToCandidate;
		}

		if ( ( clockWiseCandidate.x < 52.5 ) and ( counterClockWiseCandidate.x < 52.5 ) )
		{
			if ( currentDribbleDirection == DD_CLOCK_WISE )
			{
				dribbleTarget = clockWiseCandidate;
				currentDribbleDirection = DD_CLOCK_WISE;
			}
			else if ( currentDribbleDirection == DD_COUNTER_CLOCK_WISE )
			{
				dribbleTarget = counterClockWiseCandidate;
				currentDribbleDirection = DD_COUNTER_CLOCK_WISE;
			}
			else if ( wm.self().pos().y > 0 )
			{
				dribbleTarget = clockWiseCandidate;
				currentDribbleDirection = DD_CLOCK_WISE;
			}
			else
			{
				dribbleTarget = counterClockWiseCandidate;
				currentDribbleDirection = DD_COUNTER_CLOCK_WISE;
			}
		}
		else if ( clockWiseCandidate.x < 52.5 )
		{
				dribbleTarget = clockWiseCandidate;
				currentDribbleDirection = DD_CLOCK_WISE;
		}
		else
		{
			dribbleTarget = counterClockWiseCandidate;
			currentDribbleDirection = DD_COUNTER_CLOCK_WISE;
		}
	}
	else
	{
		if ( wm.ball().pos().y < 0 )
		{
			Vector2D center ( 36.f , -20 );
			int partsCount = 2;
			float partsSize = 90.f / (float)partsCount;
			Vector2D centerToMe = wm.self().pos() - center;
			float myDist = centerToMe.r();
			float ourCurrentAngularPos = centerToMe.th().degree() + 90.f;
			int myPart = (int)( ourCurrentAngularPos / partsSize );
			float nextPartAngle = ( myPart + 1 ) * partsSize;
			nextPartAngle -= 90;
			nextPartAngle += 5.f;
			Vector2D centerToNextPosCandidate = Vector2D::polar2vector( myDist , nextPartAngle );
			Vector2D nextPosCandidate = centerToNextPosCandidate + center;
			if ( nextPosCandidate.x > 50.f )
			{
				centerToNextPosCandidate *= ( 49.f - wm.self().pos().x ) / ( nextPosCandidate.x - wm.self().pos().x );
				nextPosCandidate = centerToNextPosCandidate + center;
			}
			dribbleTarget = nextPosCandidate;
		}
		else
		{
			Vector2D center( 36.f , 20.f );
			int partsCount = 2;
			float partsSize = 90.f / (float)partsCount;
			Vector2D centerToMe = wm.self().pos() - center;
			float myDist = centerToMe.r();
			float ourCurrentAngularPos = 90.f - centerToMe.th().degree();
			int myPart = (int)( ourCurrentAngularPos / partsSize );
			float nextPartAngle = ( myPart + 1 ) * partsSize;
			nextPartAngle = 90.f - nextPartAngle;
			nextPartAngle -= 5.f;
			Vector2D centerToNextPosCandidate = Vector2D::polar2vector ( myDist , nextPartAngle );
			Vector2D nextPosCandidate = centerToNextPosCandidate + center;
			if ( nextPosCandidate.x > 50.f )
			{
				centerToNextPosCandidate *= ( 49.f - wm.self().pos().x ) / ( nextPosCandidate.x - wm.self().pos().x );
				nextPosCandidate = centerToNextPosCandidate + center;
			}
			dribbleTarget = nextPosCandidate;
		}
	}
	const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
		it != end;
		++it )
	{
		if ( (*it)->posCount() > 5 )
			continue;
		Vector2D hisPos = (*it)->pos();
		if ( hisPos.dist( wm.ball().pos() ) <= 1.5f )
			return GD_CAN_NOT;
		break;
	}
	return GD_CAN;
}

void GoalPlanDribble::execute( rcsc::PlayerAgent * agent )
{
//	if ( SRPDribble::SRPDribble().execute ( agent ) )
//		return;
	Body_Dribble2008( dribbleTarget , 0.5 , 100 , 2 , true ).execute( agent );
}
