// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "GoalPlanPositioning.h"

#include <cmath>

#include <rcsc/player/world_model.h>

#include <rcsc/common/audio_codec.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include <OffenseVariables.h>

#include <FastIC.h>
#include <Basics.h>

using namespace rcsc;
using namespace std;

PlayerRole GoalPlanPositioning::getBallOwnerRole( int ballOwnerUnum )
{
	string playerToCheck = "QQQQ";
	string linearPlayers = OffenseVariables::Instance()->linearPlayers();
	string quarterBackPlayer = OffenseVariables::Instance()->quarterBackPlayer();
	if ( ( linearPlayers.size() >= 3 ) and ( quarterBackPlayer.size() ) )
	{
		playerToCheck[0] = linearPlayers[0];
		playerToCheck[1] = quarterBackPlayer[0];
		playerToCheck[2] = linearPlayers[1];
		playerToCheck[3] = linearPlayers[2];
	}
	dlog.addText ( Logger::PASS , "playerToCheck[%s]" , playerToCheck.c_str() );
	for ( unsigned int i = 0; i < playerToCheck.size(); i++ )
		if ( AudioCodec::i().hex2unum( playerToCheck[i] ) == ballOwnerUnum )
			return (PlayerRole)i;
	return GPP_NONE;
}

PlayerRole GoalPlanPositioning::getPlayerRole ( int myUnum )
{
	string playerToCheck = "QQQQ";
	string linearPlayers = OffenseVariables::Instance()->linearPlayers();
	string quarterBackPlayer = OffenseVariables::Instance()->quarterBackPlayer();
	if ( ( linearPlayers.size() >= 3 ) and ( quarterBackPlayer.size() ) )
	{
		playerToCheck[0] = linearPlayers[0];
		playerToCheck[1] = quarterBackPlayer[0];
		playerToCheck[2] = linearPlayers[1];
		playerToCheck[3] = linearPlayers[2];
	}
	for ( unsigned int i = 0; i < playerToCheck.size(); i++ )
		if ( AudioCodec::i().hex2unum( playerToCheck[i] ) == myUnum )
			return (PlayerRole)i;
	return GPP_NONE;
	
}

PlayerRole GoalPlanPositioning::getSubstituedRole( PlayerRole realBallRole , PlayerRole fakeBallRole , PlayerRole playerRealRole )
{
	if ( playerRealRole == realBallRole )
		return fakeBallRole;
	if ( fakeBallRole == GPP_BOTTOMOUT )
	{
		if ( playerRealRole < realBallRole )
			return (PlayerRole)(playerRealRole + 1);
		else
			return playerRealRole;
	}
	if ( fakeBallRole == GPP_TOPOUT )
	{
		if ( playerRealRole > realBallRole )
			return (PlayerRole)(playerRealRole - 1);
		else
			return playerRealRole;			
	}
}

void GoalPlanPositioning::setMiddleAreaY ( PlayerRole ballOwnerRole , float ballY )
{
	if ( ballOwnerRole == GPP_NONE )
		middleAreaY = 0.f;
	else
	{
		middleAreaY = ballY - ( ( ballOwnerRole - 2 ) * ( toBeCoveredAreaWidth / 3.f ) ) - ( toBeCoveredAreaWidth / 6.f );
	}
}

float GoalPlanPositioning::getRoleY( PlayerRole ballOwnerRole , PlayerRole roleChoosed , float ballY )
{
	float playersDiffrence = toBeCoveredAreaWidth / 3.f;
	float minY = ballY - ( ballOwnerRole * playersDiffrence );
	if ( ballOwnerRole == GPP_NONE )
		minY = middleAreaY - ( toBeCoveredAreaWidth / 2.f );
//	dlog.addText ( Logger::PASS , "minY\t%f" , minY );
	return ( minY + ( roleChoosed * playersDiffrence ) );
}

float GoalPlanPositioning::getPlayerX( int unum )
{
	if ( Basics::isInPlayers ( unum , OffenseVariables::Instance()->quarterBackPlayer() ) )
		return 37.f;
	else
		return 42.f;
}

Vector2D GoalPlanPositioning::getHomePos( const WorldModel * wm )
{
	int fastestUnum = -1;
	int myUnum = wm->self().unum();
	Vector2D ballPos = wm->ball().pos();
	if ( FastIC::getFastestTeammate() )
	{
		fastestUnum = FastIC::getFastestTeammate()->unum();
		ballPos = inertia_n_step_point( ballPos , wm->ball().vel() , FastIC::getFastestTeammateReachCycle() , ServerParam::i().ballDecay() );
	}
	PlayerRole ballOwnerRole = getBallOwnerRole( fastestUnum );
	dlog.addText ( Logger::PASS , "ballOwnerRole\t%d" , ballOwnerRole);
	PlayerRole myRole = getPlayerRole( myUnum );
	dlog.addText ( Logger::PASS , "myRole\t%d" , myRole );
	if ( ( ballPos.y > 20.f ) and ( ballOwnerRole != GPP_TOPOUT ) )
	{
		myRole = getSubstituedRole( ballOwnerRole , GPP_TOPOUT , myRole );
		ballOwnerRole = GPP_TOPOUT;
	}
	if ( ( ballPos.y < -20.f ) and ( ballOwnerRole != GPP_BOTTOMOUT ) )
	{
		myRole = getSubstituedRole( ballOwnerRole , GPP_BOTTOMOUT , myRole );
		ballOwnerRole = GPP_BOTTOMOUT;
	}
	setMiddleAreaY( ballOwnerRole , ballPos.y );
	Vector2D myPos;
	myPos.x = min( getPlayerX( myUnum ) , (float)(wm->offsideLineX()-1.f) );
	myPos.y = getRoleY ( ballOwnerRole , myRole , ballPos.y );
	if ( myPos.y > 33.f )
		myPos.y = 33.f;
	if ( myPos.y < -33.f )
		myPos.y = -33.f;
	return myPos;
}
