// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <PrimaryGoalPlanPositioning.h>

#include <cmath>
#include <string>

#include <rcsc/common/audio_codec.h>
#include <rcsc/common/server_param.h>

#include <Basics.h>
#include <FastIC.h>
#include <OffenseVariables.h>

//#define PrimaryGoalPlanPositioningDEBUG

using namespace rcsc;
using namespace std;

PlayerRole PrimaryGoalPlanPositioning::getBallCurrentRole( rcsc::Vector2D ballPos )
{
	if ( ballPos.y > 0 )
	{
		if ( fabsf( ballPos.y ) > determiningQuarterWidth )
			return GPP_TOPOUT;
		else
			return GPP_TOPIN;
	}
	else
	{
		if ( fabsf( ballPos.y ) > determiningQuarterWidth )
			return GPP_BOTTOMOUT;
		else
			return GPP_BOTTOMIN;
	}
	return GPP_NONE;
}

rcsc::Vector2D PrimaryGoalPlanPositioning::getBestPositioningPoint( const rcsc::WorldModel * wm , std::vector<rcsc::Vector2D> targetPos )
{
#ifdef PrimaryGoalPlanPositioningDEBUG
	dlog.addText ( Logger::PASS , "getBestPositioningPoint()" );
#endif

	Vector2D currentPos = wm->self().pos();
	Vector2D ballPos = wm->ball().pos();

	vector<float> xDiff;
	float maxXDiff = 0.1f;
	for ( unsigned int i = 0; i < targetPos.size(); i++ )
	{
		float diff = fabsf ( targetPos[i].x - ballPos.x );
		xDiff.push_back( diff );
		if ( diff > maxXDiff )
			maxXDiff = diff;
	}
	vector<float> xDiffValue;
	for ( unsigned int i = 0; i < xDiff.size(); i++ )
	{
		xDiffValue.push_back ( 1.f - ( xDiff[i] / maxXDiff ) );
	}

	float maxForwardMovment = 0.1f;
	for ( unsigned int i = 0; i < targetPos.size(); i++ )
	{
		if ( targetPos[i].x - 36.f > maxForwardMovment )
			maxForwardMovment = targetPos[i].x - 36.f;
	}
	vector<float> forwardMovingValues;
	for ( unsigned int i = 0; i < targetPos.size(); i++ )
	{
		forwardMovingValues.push_back( ( targetPos[i].x - 36.f ) / maxForwardMovment );
	}

	vector<int> playersNearPos;
	for ( unsigned int i = 0; i < targetPos.size(); i++ )
		playersNearPos.push_back( 0 );

	float distToCount = 5.f;
	const PlayerPtrCont::const_iterator end = wm->opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm->opponentsFromSelf().begin();
		it != end;
		++it )
	{
		if ( (*it)->posCount() > 3 )
			continue;
		if ( (*it)->goalie() )
			continue;
		Vector2D hisPos = (*it)->pos();
		for ( unsigned int i = 0; i < playersNearPos.size(); i++ )
		{
			if ( hisPos.dist( targetPos[i] ) <= distToCount )
				playersNearPos[i]++;
		}
	}

	int maxPlayersNearPos = 1;
	for ( unsigned int i = 0; i < playersNearPos.size(); i++ )
	{
		if ( playersNearPos[i] > maxPlayersNearPos )
			maxPlayersNearPos = playersNearPos[i];
	}
#ifdef PrimaryGoalPlanPositioningDEBUG
	dlog.addText( Logger::PASS , "maxPlayersNearPos\t%d" , maxPlayersNearPos );
#endif

	vector<float> playersNearPosValue;
	for ( unsigned int i = 0; i < playersNearPos.size(); i++ )
	{
		playersNearPosValue.push_back ( 1.f - ( playersNearPos[i] / maxPlayersNearPos ) );
	}
#ifdef PrimaryGoalPlanPositioningDEBUG
	for ( unsigned int i = 0; i < targetPos.size(); i++ )
	{
		dlog.addText ( Logger::PASS , "candidate %d is ( %.2f , %.2f )" , i+1 , targetPos[i].x , targetPos[i].y );
		dlog.addText ( Logger::PASS , "\tforwardMovingValues\t%f" , forwardMovingValues[i] );
		dlog.addText ( Logger::PASS , "\tplayersNearPosValue\t%f" , playersNearPosValue[i] );
		dlog.addText ( Logger::PASS , "\txDiffValue\t%f" , xDiffValue[i] );
	}
#endif

	float forwardMovementWeight = 1.f;
	float playersNearPosWeight = 1.f;
	float xDiffWeight = 1.f;
	float weightSum = forwardMovementWeight + playersNearPosWeight + xDiffWeight;
	vector<float> values;
	for ( unsigned int i = 0; i < targetPos.size(); i++ )
	{
		values.push_back( ( ( forwardMovementWeight * forwardMovingValues[i] ) + ( playersNearPosWeight * playersNearPosValue[i] ) +
		 ( xDiffValue[i] * xDiffWeight ) ) 
		 / weightSum );
#ifdef PrimaryGoalPlanPositioningDEBUG
		dlog.addText ( Logger::PASS , "value[%d] = %f" , i , values[i] );
#endif
	}

	if ( targetPos.size() )
	{
		float maxValue = values[0];
		int maxBlock = 0;
		for ( unsigned int i = 0; i < targetPos.size(); i++ )
		{
			if ( values[i] > maxValue )
			{
				maxValue = values[i];
				maxBlock = i;
			}
		}
		return targetPos[maxBlock];
	}
	return Vector2D ( 0.f , 0.f );
}

rcsc::Vector2D PrimaryGoalPlanPositioning::getRolePos( PlayerRole ballRole , PlayerRole playerRole , const rcsc::WorldModel * wm)
{
	Vector2D ballPos = wm->ball().pos();
	int playerUnum = wm->self().unum();
	float changingY = toBeCoveredAreaWidth / 3.f;
	float minY = ballPos.y - ( ballRole * changingY );
	float posY = minY + ( playerRole * changingY );
	float minX = 40.f;
	if ( Basics::isInPlayers( playerUnum , OffenseVariables::Instance()->quarterBackPlayer() ) )
		minX = 38.f;
	minX = min( minX , (float)(wm->offsideLineX() - 0.5f) );
	float maxX = max ( (float)(wm->offsideLineX() - 0.5f) , 36.f );
	float xDiff = maxX - minX;
	if ( xDiff < 2.f )
		return ( Vector2D( minX , posY ) );
	float middleX = ( minX + maxX ) / 2.f;
	if ( ( ballRole == GPP_BOTTOMOUT ) and ( playerRole == GPP_BOTTOMIN ) )
			maxX = middleX;
	if ( ( ballRole == GPP_TOPOUT ) and ( playerRole == GPP_TOPIN ) )
			maxX = middleX;
	xDiff = maxX - minX;
	if ( xDiff < 2.f )
		return ( Vector2D( minX , posY ) );
	float xChange = xDiff / 3.f;
	vector< Vector2D > placesToCheck;
	placesToCheck.push_back( Vector2D ( minX , posY ) );
	placesToCheck.push_back( Vector2D ( minX + xChange , posY ) );
	placesToCheck.push_back( Vector2D ( maxX - xChange , posY) );
	placesToCheck.push_back( Vector2D ( maxX , posY ) );
	return getBestPositioningPoint( wm , placesToCheck );
}

rcsc::Vector2D PrimaryGoalPlanPositioning::getHomePos( const WorldModel * wm )
{	
	int fastestUnum = 100;
	int myUnum = wm->self().unum();
	Vector2D ballPos = wm->ball().pos();
	if ( FastIC::getFastestTeammate() )
	{
		fastestUnum = FastIC::getFastestTeammate()->unum();
		ballPos = inertia_n_step_point( ballPos , wm->ball().vel() , FastIC::getFastestTeammateReachCycle() , ServerParam::i().ballDecay() );
	}
	if ( ( fabsf( ballPos.x ) > 52.5 ) or ( fabsf( ballPos.y ) > 34.f ) )
		ballPos = wm->ball().pos();
	PlayerRole ballRole = getBallCurrentRole( ballPos );

	if ( fastestUnum == myUnum )
		return ballPos;
#ifdef PrimaryGoalPlanPositioningDEBUG
	dlog.addText ( Logger::PASS , "ballRole\t%d" , ballRole );
#endif

	string playersToFill = "QQQQ";
	string quarterBackPlayer = OffenseVariables::Instance()->quarterBackPlayer();
	string linearPlayers = OffenseVariables::Instance()->linearPlayers();
	if ( ( quarterBackPlayer.size() ) and ( linearPlayers.size() == 3 ) )
	{
		playersToFill[0] = linearPlayers[0];
		playersToFill[1] = quarterBackPlayer[0];
		playersToFill[2] = linearPlayers[1];
		playersToFill[3] = linearPlayers[2];
	}
#ifdef PrimaryGoalPlanPositioningDEBUG
	dlog.addText ( Logger::PASS , "playersToFill\t%s" , playersToFill.c_str() );
	dlog.addText ( Logger::PASS , "fastestUnum\t%d" , fastestUnum );
#endif
	string playersInPos = "QQQQ";
	unsigned int playerMarker = 0;
	unsigned int blockMarker = 0;
	while ( ( blockMarker < playersInPos.size() ) and ( playerMarker < playersToFill.size() ) )
	{
#ifdef PrimaryGoalPlanPositioningDEBUG
		dlog.addText ( Logger::PASS , "playerMarker\t%d---->blockMarker\t%d", playerMarker , blockMarker );
#endif
		if ( blockMarker == (int)ballRole )
		{
#ifdef PrimaryGoalPlanPositioningDEBUG
			dlog.addText ( Logger::PASS , "block skipped!!" );
#endif
			blockMarker++;
			continue;
		}
		if ( AudioCodec::i().hex2unum( playersToFill[playerMarker] ) == fastestUnum )
		{
#ifdef PrimaryGoalPlanPositioningDEBUG
			dlog.addText ( Logger::PASS , "player skipped!!" );
#endif
			playerMarker++;
			continue;
		}
		playersInPos[blockMarker] = playersToFill[playerMarker];
		playerMarker++;
		blockMarker++;
	}
#ifdef PrimaryGoalPlanPositioningDEBUG
	dlog.addText ( Logger::PASS ,  "playersInPos\t%s" , playersInPos.c_str() );
#endif
	for ( unsigned int i = 0; i < playersInPos.size(); i++ )
	{
		if ( playersInPos[i] == 'Q' )
			continue;
		if ( AudioCodec::i().hex2unum( playersInPos[i] ) == wm->self().unum() )
		{
			Vector2D answerPosition = getRolePos( ballRole , (PlayerRole)i , wm );
			answerPosition.x = min ( answerPosition.x , wm->offsideLineX() - 1.f );
			return answerPosition;
		}
	}
	return Vector2D( 0.f , 0.f );
}
