// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "DribbleDecision.h"
#include "Basics.h"
#include "Ellipse.h"
#include <SRPDribble.h>
#include "body_hold_ball2008.h"

#include <strategy.h>

#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_scan_field.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/geom/triangle_2d.h>
#include <rcsc/geom/polygon_2d.h>

#include <iostream>
#include <vector>

#include <FastIC.h>

using namespace rcsc;
using namespace std;

Vector2D DribbleDecision::currentDribbleTarget( 50.f , 10.f );

DribbleDecision::DribbleDecision(Vector2D myPos):
	dribble2008_target(45.0, 0.0),
	dashCount(2),
	dribbleSpeed( 0.6f )
{
	cSRP = false;
	riskyDribbleRange = 10.f;
	maxPlayersInRectangle = 1;
	currentLocalOffsideLine = 52.5;

	if ( myPos.x > 36.f )
	{
		if ( fabsf( myPos.y ) < 20.f )
		{
			riskyDribbleRange = 4.f;
			maxPlayersInRectangle = 2;
			Vector2D myAbsPos = myPos.abs();
			if ( myAbsPos.dist ( currentDribbleTarget ) <= 2 )
				changeTarget();
			mTarget = currentDribbleTarget;
		}
		else
		{
			currentDribbleTarget = Vector2D ( 50.f , 10.f );
			riskyDribbleRange = 8;
			maxPlayersInRectangle = 1;
			mTarget.x = 47.f;
			mTarget.y = 15.f;
		}
		if ( myPos.y < 0.f )
			mTarget.y *= -1;
	}
	else if ( ( myPos.x > 30.f ) and ( fabsf( myPos.y ) <= 20 ) )
	{
		currentDribbleTarget = Vector2D ( 40.f , 10.f );
		riskyDribbleRange = 4.f;
		maxPlayersInRectangle = 2;
		mTarget = currentDribbleTarget;
		if ( myPos.y < 0.f )
			mTarget.y *= -1;
	}
	else
	{
		maxPlayersInRectangle = 1;
		riskyDribbleRange = 7.5f;
		mTarget.x = 40.f;
		mTarget.y = myPos.y;
		if ( mTarget.y > 30 )
			mTarget.y = 30;
		if ( mTarget.y < -30 )
			mTarget.y = -30;
	}
//	dlog.addText ( Logger::PASS , "dribbleTarget	= (%f , %f)" , mTarget.x , mTarget.y );
//	mSRP.reconstruct(mTarget);
}

void DribbleDecision::changeTarget()
{
	if ( currentDribbleTarget.x >= 45.f )
		currentDribbleTarget.x = 40.f;
	else
		currentDribbleTarget.x = 50.f;
}

DribbleDecision::~DribbleDecision()
{
}

void DribbleDecision::setCurrentOffsideLine( const rcsc::WorldModel & wm )
{
	currentLocalOffsideLine = -52.5;
	float distToCountAsLocal = ( ( 52.5 - wm.self().pos().x ) / ( 105.f ) ) * 26.f;
	distToCountAsLocal += 7.f;
	const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
		it != end;
		++it )
	{
		Vector2D hisPos = (*it)->pos();
		int hisUNum = (*it)->unum();
		if ( hisUNum < 0 )
			continue;
		if ( (*it)->posCount() > 3 )
			continue;
		if ( wm.opponentGoalieUnum() )
			if ( hisUNum == wm.opponentGoalieUnum() )
				continue;
		if ( fabsf( hisPos.y - wm.self().pos().y ) <= distToCountAsLocal )
			if ( hisPos.x > currentLocalOffsideLine )
				currentLocalOffsideLine = hisPos.x;
	}
}

void DribbleDecision::setRiskyDribbleRange(float f)
{
	riskyDribbleRange = f;
}

void DribbleDecision::setMaxPlayersInRectangle(int i)
{
	maxPlayersInRectangle = i;
}

HelliBASH09SRP & DribbleDecision::SRPDribble()
{
	return mSRP;
}

Decision DribbleDecision::getValue(rcsc::PlayerAgent *agent)
{
//	dlog.addText ( Logger::PASS , "DD:GETVALUE!");
//	dlog.addText ( Logger::PASS , "dribbleTarget	= (%f , %f)" , mTarget.x , mTarget.y );
//	dlog.flush();
	const rcsc::WorldModel & wm = agent->world();
	Vector2D suggestedTarget = Vector2D( 52.5 , 0.f );
	bool oneOnOne = isItOneOnOneCondition(agent , suggestedTarget );
//	dlog.addText ( Logger::PASS , "oneOnOne??\t%d" , oneOnOne );
//	if ( wm.self().pos().x <= 36 )
		if ( SRPDribble::SRPDribble().execute ( agent ) )
			return SHOULD_DO_IT;
	Vector2D tmpTarget = mTarget;
	if ( oneOnOne )
	{
		tmpTarget = suggestedTarget;
	}

	vector<Vector2D> nearOppPoses;
	vector<float> oppDists;
	const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
		it != end;
		++it )
	{
		if ( (*it)->posCount() > 5 )
			continue;
		Vector2D hisPos = (*it)->pos();
		float dist = hisPos.dist( wm.ball().pos() );
		if ( dist < riskyDribbleRange )
		{
			nearOppPoses.push_back( hisPos );
			oppDists.push_back( dist );
		}
	}
	
	bool needsFurtherSort = true;
	while ( ( needsFurtherSort ) and ( nearOppPoses.size() > 1 ) )
	{
		needsFurtherSort = false;
		for ( unsigned int i = 0; i < nearOppPoses.size() - 1; i++ )
		{
			if ( oppDists[i] > oppDists[i+1] )
			{
				float helpDist = oppDists[i];
				oppDists[i] = oppDists[i+1];
				oppDists[i+1] = helpDist;

				Vector2D helpPos = nearOppPoses[i];
				nearOppPoses[i] = nearOppPoses[i+1];
				nearOppPoses[i+1] = helpPos;
			}
		}
	}
	float direction = ( tmpTarget - wm.self().pos() ).th().degree();
	unsigned int maxGo = nearOppPoses.size();
	if ( maxGo > 2 )
		maxGo = 2;
	for ( unsigned int i = 0; i < maxGo ; i++ )
	{
//		dlog.addText ( Logger::PASS , "Checking an opponent in ( %.2f , %.2f )" , nearOppPoses[i].x , nearOppPoses[i].y );
		float oppDirection = ( nearOppPoses[i] - wm.self().pos() ).th().degree();
		float deltaAngle = fabsf( Basics::getDeltaAngle( direction , oppDirection ) );
		float tackleDist = rcsc::ServerParam::i().tackleDist() + 0.5;
		if ( oneOnOne )
			tackleDist -= 1.f;
		if ( oppDists[i] < tackleDist )
		{
			return CAN_NOT_DO_IT;
		}

		if ( deltaAngle <= 30 )
		{
			if ( oppDists[i] <= riskyDribbleRange )
				return SHOULD_NOT_DO_IT;
		}

		if ( deltaAngle <= 60 )
		{
			if ( oppDists[i] <= riskyDribbleRange / 2.f )
				return SHOULD_NOT_DO_IT;
		}

		if ( deltaAngle <= 90 )
		{
			if ( oppDists[i] <= riskyDribbleRange / 4.f )
				return SHOULD_NOT_DO_IT;
		}
	}

/*	const rcsc::PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 5 , true );
	const double nearest_opp_dist = ( nearest_opp
						? nearest_opp->distFromSelf()
						: 1000.0 );
	const double nearest_opp_dist_from_ball = ( nearest_opp
						? nearest_opp->distFromBall()
						: 1000.0 );
	const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
						? nearest_opp->pos()
						: rcsc::Vector2D( -1000.0, 0.0 ) );
	float direction = ( tmpTarget - wm.self().pos() ).th().degree();
	float nearestOppDirection = ( nearest_opp_pos - wm.self().pos() ).th().degree();
	float deltaAngle = Basics::getDeltaAngle( direction , nearestOppDirection );
	float maxDist = rcsc::ServerParam::i().tackleDist() + 0.5;
	if ( oneOnOne )
		maxDist -= 1.f;
	if ( nearest_opp )
		dlog.addText ( Logger::PASS , "NearestOppNUM	%d , dist	%f , deltaAngle	%f , distBall %f", nearest_opp->unum() , nearest_opp_dist , deltaAngle , nearest_opp_dist_from_ball);
	dlog.addText ( Logger::PASS , "RiskyDribbleRange	%f" , riskyDribbleRange );
	dlog.addText ( Logger::PASS , "TackeDist++ %f" , maxDist );
	dlog.addText ( Logger::PASS , "currentLocalOffsideLine = %f" , currentLocalOffsideLine );
	if ( ( wm.self().pos().x > currentLocalOffsideLine ) and (Strategy::i().get_ball_area(wm) != Strategy::BA_Cross) )
	{
		if ( oneOnOne )
			mTarget = tmpTarget;
		return SHOULD_DO_IT;
	}
	if ( nearest_opp_dist_from_ball < maxDist )
	{
		return CAN_NOT_DO_IT;
	}

	if ( deltaAngle <= 30 ) 
	{
		if ( nearest_opp_dist <= riskyDribbleRange )
			return SHOULD_NOT_DO_IT;
	}

	if ( deltaAngle <= 60 ) 
	{
		if ( nearest_opp_dist <= riskyDribbleRange / 2.f)
			return SHOULD_NOT_DO_IT;
	}

	if ( deltaAngle <= 90 ) 
	{
		if ( nearest_opp_dist <= riskyDribbleRange / 4.f)
			return SHOULD_NOT_DO_IT;
	}
*/
	if ( oneOnOne )
	{
		mTarget = tmpTarget;
		return SHOULD_DO_IT;
	}

	bool should = shouldWeDribble ( agent );
	if ( should )
		return SHOULD_DO_IT;
	else
		return CAN_DO_IT;
}

void DribbleDecision::execute(rcsc::PlayerAgent *agent)
{
	SRPDribble::SRPDribble selfPass;
//	if ( agent->world().self().pos().x <= 36 )
		if ( selfPass.execute(agent) )
		{
			dlog.addText ( Logger::PASS , "SelfPASS!!" );
			return;
		}
//	dlog.addText ( Logger::PASS , "AdvancedDribble!!" );
//	dlog.addText ( Logger::PASS , "Target ( %f , %f )" , mTarget.x , mTarget.y );
	Body_Dribble2008( mTarget , 1.0 , rcsc::ServerParam::i().maxPower() , 2 , true ).execute( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
}

bool DribbleDecision::shouldWeDribble ( rcsc::PlayerAgent * agent)
{
	const rcsc::WorldModel & wm = agent->world();
	if ( ( Strategy::get_ball_area ( wm.ball().pos() ) == Strategy::BA_ShootChance ) or
		( Strategy::get_ball_area ( wm.ball().pos() ) == Strategy::BA_Cross ) )
		return true;
	if ( Strategy::get_ball_area ( wm.ball().pos() ) == Strategy::BA_DribbleAttack )
	{
		Vector2D meToTarget = mTarget - wm.self().pos();
		meToTarget.setLength( dribbleSpeed );
		FastIC fast(false);
		const rcsc::WorldModel & wm = agent->world();
		fast.setBall( wm.ball().pos() , meToTarget , 0);
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( hisUNum < 0 )
				continue;
			if ( (*it)->posCount() > 5 )
				continue;
			fast.players().push_back( FastPlayer( hisPos , Vector2D( 0.f , 0.f ) , 0 , hisUNum , false , 1.f , 1.f , (*it)->playerTypePtr()->id() , NULL , NULL ) );
		}
		fast.calculate( wm.self().unum() , 150 , false );
		if ( fast.getFastestPlayers().size() < 2 )
		{
			return true;
		}
		return false;
	}
	else
	{
		Vector2D ezaf;
		return isItOneOnOneCondition( agent , ezaf);
	}
}

bool DribbleDecision::isItOneOnOneCondition ( PlayerAgent *agent , Vector2D & decidedTarget )
{
	const WorldModel & wm = agent->world();
	setCurrentOffsideLine( wm );
	if ( currentLocalOffsideLine <= wm.self().pos().x )
	{
		decidedTarget = Vector2D( 52.5f , 0.f );
		if ( wm.self().pos().y < -7.f )
			decidedTarget = Vector2D( 52.5 , -7.f );
		else if ( wm.self().pos().y > 7.f )
			decidedTarget = Vector2D( 52.5 , 7.f );
		else if ( wm.getOpponentGoalie() )
		{
			if ( wm.getOpponentGoalie()->posCount() <= 5 )
			{
				Vector2D opponentGoaliePos = wm.getOpponentGoalie()->pos();
				float upperDist = opponentGoaliePos.dist ( Vector2D( 52.5 , 7.f ) );
				float lowerDist = opponentGoaliePos.dist ( Vector2D( 52.5 , -7.f ) );
				if ( upperDist > lowerDist )
					decidedTarget = Vector2D( 52.5 , 7.f );
				else
					decidedTarget = Vector2D( 52.5 , -7.f );
			}
		}
//		dlog.addText ( Logger::PASS , "here!!" );
		return true;
	}
	int maxDirectionCount = 3;
	vector< Vector2D > targets;
	targets.push_back ( Vector2D( 52.5 , 0 ) );
	if ( wm.self().pos().y > 0 )
	{
		targets.push_back ( Vector2D( 52.5 , 12 ) );
		targets.push_back ( Vector2D( 52.5 , -12 ) );
	}
	else
	{
		targets.push_back ( Vector2D( 52.5 , -12 ) );
		targets.push_back ( Vector2D( 52.5 , 12 ) );
	}
	for ( unsigned int i = 0; i < targets.size(); i++ )
	{
		Vector2D dribbleTarget = targets[i];
		if ( wm.dirCount( ( dribbleTarget - wm.ball().pos() ).th() ) > maxDirectionCount )
		{
			continue;
		}
//		dlog.addText ( Logger::PASS , "target\t( %.2f , %.2f )", dribbleTarget.x , dribbleTarget.y );
		FastIC fast(false);
		fast.setBall( wm.ball().pos() , Vector2D::polar2vector( dribbleSpeed , ( dribbleTarget - wm.ball().pos() ).th() ) , 0);
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( hisUNum < 0 )
				continue;
			if ( (*it)->posCount() > 5 )
				continue;
//			dlog.addText ( Logger::PASS , "opponent\t%d\tin ( %f, %f )\tAdded!" , hisUNum , hisPos.x , hisPos.y );
			fast.players().push_back( FastPlayer( hisPos , Vector2D( 0.f , 0.f ) , 0 , hisUNum , false , 1.f , 1.f , (*it)->playerTypePtr()->id() , NULL , NULL ) );
		}
		fast.calculate( wm.self().unum() , 150 , false );
		if ( fast.getFastestPlayers().size() < 0 )
		{
			decidedTarget = dribbleTarget;
			return true;
		}
//		dlog.addText ( Logger::PASS , "fastestOpp is %d" , fast.getFastestPlayers()[0].unum );
		if ( wm.opponentGoalieUnum() )
			if ( fast.getFastestPlayers()[0].unum == wm.opponentGoalieUnum() )
			{
				decidedTarget = dribbleTarget;
				return true;
			}
	}
	return false;
}

void DribbleDecision::doHoldBall( rcsc::PlayerAgent *agent )
{
	Vector2D myPos = agent->world().self().pos();
	Vector2D turnTarget , kickTarget;
	if ( myPos.x <= 36 )
	{
		kickTarget = mTarget;
		if ( fabsf ( mTarget.y ) <= 20 )
			turnTarget = mTarget;
		else
			turnTarget = Vector2D ( myPos.x + 10 , 0 );
	}
	else
	{
		kickTarget = Vector2D( 52.5 , 0 );
		turnTarget = Vector2D( 42.f , 0 );
	}
	Body_HoldBall2008(true , turnTarget , kickTarget).execute ( agent );
	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
}

void DribbleDecision::reset()
{
	currentDribbleTarget = Vector2D ( 50 , 10 );
}
