// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <HelliPassReceive.h>

#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_point.h>

#include <rcsc/player/world_model.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/audio_memory.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>


#include <FastIC.h>

#include <Intercept.h>

using namespace std;
using namespace rcsc;
bool HelliPassReceive::shouldIntercept = false;
bool HelliPassReceive::isItCourseSituation = false;
int HelliPassReceive::passReleaseCycle = -10;

bool HelliPassReceive::shouldExecute( PlayerAgent *agent)
{
	WorldModel & wm = agent->editableWorldModel();
	
	if ( wm.self().isKickable() )
	{
		reset();
		wm.editableAudioMemory().setPassReceived();
		return shouldIntercept;
	}
	if ( ( not wm.audioMemory().helliPass().empty() )
	&& ( wm.audioMemory().helliPass().front().receiver_ == wm.self().unum() )
	&& 
	( ( wm.time().cycle() <= wm.audioMemory().helliPass().front().releaseCycle_ ) 
		|| abs( wm.time().cycle() - wm.audioMemory().helliPass().front().releaseCycle_ ) <= 2 )
		)
	{
		shouldIntercept = true;
		isItCourseSituation = false;
		passReleaseCycle = wm.audioMemory().helliPass().front().releaseCycle_;
		dlog.addText ( Logger::PASS , "real pass!!!!" );
	}
	else if ( ( not wm.audioMemory().helliCourse().empty() )
	&& ( wm.audioMemory().helliCourse().front().receiver_ == wm.self().unum() )
	&& 
	( ( wm.time().cycle() <= wm.audioMemory().helliCourse().front().releaseCycle_ ) 
		|| abs( wm.time().cycle() - wm.audioMemory().helliCourse().front().releaseCycle_ ) <= 2 )
		)
	{
		shouldIntercept = true;
		isItCourseSituation = true;
		passReleaseCycle = wm.audioMemory().helliCourse().front().releaseCycle_;
		dlog.addText ( Logger::PASS , "course situation1!!!");
	}

	else if ( ( wm.existKickableTeammate() ) or ( wm.existKickableOpponent() ) or ( wm.self().isKickable() ) )
	{
		reset();
	}
	
	if ( shouldIntercept )
	{
		if ( isItCourseSituation )
		{
			if ( wm.audioMemory().helliCourse().empty() )
				reset();
			else
				if ( wm.time().cycle() - 30 > wm.audioMemory().helliCourse().front().releaseCycle_ ) 
					reset();
		}
		else
		{
			if ( wm.audioMemory().helliPass().empty() )
				reset();
			else
				if ( wm.time().cycle() - 30 > wm.audioMemory().helliPass().front().releaseCycle_ ) 
					reset();
		}
	}
	if (not shouldIntercept )
	{
		wm.editableAudioMemory().setPassReceived();
	}
	return shouldIntercept;
}

void HelliPassReceive::execute(PlayerAgent *agent)
{
	dlog.addText ( Logger::PASS , "Receiving !!!! ");
	const WorldModel & wm = agent->world();
	if ( wm.time().cycle() <= passReleaseCycle )
	{
		dlog.addText ( Logger::PASS , "preReceiving!!");
		Vector2D ballPos;
		Vector2D ballVel;
		int releaseCycle;
		if ( isItCourseSituation )
		{
			ballPos = wm.audioMemory().helliCourse().front().ballPos_;
			ballVel = wm.audioMemory().helliCourse().front().ballVel_;
			releaseCycle = wm.audioMemory().helliCourse().front().releaseCycle_;
		}
		else
		{
			ballPos = wm.audioMemory().helliPass().front().ballPos_;
			ballVel = wm.audioMemory().helliPass().front().ballVel_;
			releaseCycle = wm.audioMemory().helliPass().front().releaseCycle_;
		}
		int ballDelay = releaseCycle - wm.time().cycle();
		Intercept receiveMovement;
		receiveMovement.setBallDelay( ballDelay );
		receiveMovement.ball().updateByFullstate( ballPos , ballVel , wm.self().pos() );
		receiveMovement.execute ( agent , true , 0.25 );
	}
	else
	{
		if ( isItCourseSituation )
		{
			dlog.addText ( Logger::PASS , "using base's");
			Intercept().execute(agent);
			Vector2D nextPos = wm.self().pos() + wm.self().vel();
			Vector2D ballNextPos = wm.ball().pos() + wm.ball().vel();
			float dist = nextPos.dist ( ballNextPos );
			if ( dist < 3.f )
				agent->setNeckAction(new Neck_TurnToPoint( Vector2D( 52.5 , wm.ball().pos().x > 36 ? 0:wm.self().pos().y ) ) );
			else
				agent->setNeckAction( new rcsc::Neck_TurnToBall() );
		}
		else
		{
			dlog.addText ( Logger::PASS , "using our's" );
			Intercept().execute(agent);
		}
	}
}
