// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <NewPositioningSystem.h>

#include <OffenseVariables.h>
#include <Basics.h>
#include <FastIC.h>

#include <rcsc/player/world_model.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_codec.h>
#include <rcsc/common/logger.h>

#define maxPositionY 33
#define minPositionY -33

//TODO: 6 positioning should not dramatically change.
using namespace std;
using namespace rcsc;

FakePositionInLine NewPositioningSystem::lastBallFakeSide = TOPMIDDLE;
BallOwnerSideStats NewPositioningSystem::lastBallOwnerSideStats = USUAL;
QuarteBackSide NewPositioningSystem::lastQuarterBackSide = UP;

NewPositioningSystem::NewPositioningSystem()
{
	minRestrictedYForOppositeOwner = maxPositionY - OffenseVariables::Instance()->topDist() - OffenseVariables::Instance()->bottomDist() - 1.f;
	minRestrictedYForMiddleOwner = maxPositionY - OffenseVariables::Instance()->topDist() - 1.f;
	maxAbsoluteYForKeepingLastDecision = 5.f;
}

BallOwnerSideStats NewPositioningSystem::getBallOwnerSideStats( float posY )
{
	float noiseGiriHalfWidth = 1.f;
//	dlog.addText ( Logger::PASS , "lastBallOwnerSideStats\t%d" , lastBallOwnerSideStats );
	if ( lastBallOwnerSideStats != RESET )
	{
		if ( fabsf( posY ) < minRestrictedYForOppositeOwner - noiseGiriHalfWidth )
		{
//			dlog.addText ( Logger::PASS,  "USUAL!!" );
			return USUAL;
		}
		if ( ( fabsf( posY ) > minRestrictedYForOppositeOwner + noiseGiriHalfWidth ) and ( fabsf( posY ) < minRestrictedYForMiddleOwner - noiseGiriHalfWidth ) )
		{
//			dlog.addText ( Logger::PASS, "RESTRICTEDOPPOSITE" );
			return RESTRICTEDOPPOSITE;
		}
		if ( fabsf( posY ) > minRestrictedYForMiddleOwner + noiseGiriHalfWidth )
		{
//			dlog.addText ( Logger::PASS , "RESTRICTEDMIDDLE" );
			return RESTRICTEDMIDDLE;
		}
		if ( fabsf( posY ) < minRestrictedYForMiddleOwner - noiseGiriHalfWidth )
			if ( lastBallOwnerSideStats == RESTRICTEDMIDDLE )
			{
//				dlog.addText ( Logger::PASS , "CHANGED11111111" );
				lastBallOwnerSideStats = RESTRICTEDOPPOSITE;
			}
		if ( fabsf( posY ) > minRestrictedYForOppositeOwner + noiseGiriHalfWidth )
			if ( lastBallOwnerSideStats == USUAL )
			{
//				dlog.addText ( Logger::PASS , "Changed2222222" );
				lastBallOwnerSideStats = RESTRICTEDOPPOSITE;
			}
		return lastBallOwnerSideStats;
	}
	else
	{
//		dlog.addText ( Logger::PASS , "RESETING!!!" );
		if ( fabsf(posY) < minRestrictedYForOppositeOwner )
		{
			return USUAL;
		}
		else if ( fabsf(posY) < minRestrictedYForMiddleOwner )
		{
			return RESTRICTEDOPPOSITE;
		}
		else
			return RESTRICTEDMIDDLE;
		return USUAL;
	}
	
}

PositionInLine NewPositioningSystem::getDefaultSide( int ballOwner )
{
	string linearPlayers = OffenseVariables::Instance()->linearPlayers();
	string s = " ";
	if ( linearPlayers.size() < 3 )
		return NONE;
	for ( int i = 0; i < 3; i++ )
	{
		s[0] = linearPlayers[i];
		if ( Basics::isInPlayers( ballOwner , s ) )
		{
			if ( i == 0 )
			{
				return BOTTOM;
			}
			if ( i == 1 )
			{
				return MIDDLE;
			}
			if ( i == 2 )
			{
				return TOP;
			}
		}
	}
	return NONE;
}

PositionInLine NewPositioningSystem::getBallOwnerSide( rcsc::Vector2D ballPos , int ballOwner )
{
	PositionInLine defaultSide = getDefaultSide( ballOwner );
//	dlog.addText ( Logger::PASS , "ballOwner\t%d" , ballOwner );
//	dlog.addText ( Logger::PASS , "getDefaultSide\t%d" , defaultSide );
	BallOwnerSideStats ballSideStats = getBallOwnerSideStats( ballPos.y );
	lastBallOwnerSideStats = ballSideStats;
//	dlog.addText ( Logger::PASS , "ballSideStats\t%d" , ballSideStats );
	if ( ballSideStats == USUAL )
		return defaultSide;
	else if ( ballSideStats == RESTRICTEDOPPOSITE )
	{
//		dlog.addText ( Logger::PASS , "no opposite!!" );
		if ( ( ballPos.y < 0 ) and ( defaultSide == TOP ) )
		{
//			dlog.addText ( Logger::PASS , "1" );
			return MIDDLE;
		}
		if ( ( ballPos.y > 0 ) and ( defaultSide == BOTTOM ) )
		{
//			dlog.addText ( Logger::PASS , "2" );
			return MIDDLE;
		}
		return defaultSide;
	}
	else
	{
//		dlog.addText ( Logger::PASS , "no middle!!" );
		if ( ( ballPos.y < 0 ) and ( ( defaultSide == MIDDLE ) or ( defaultSide == TOP ) ) )
		{
//			dlog.addText ( Logger::PASS , "3" );
			return BOTTOM;
		}
		if ( ( ballPos.y > 0 ) and ( ( defaultSide == MIDDLE ) or ( defaultSide == BOTTOM ) ) )
		{
//			dlog.addText ( Logger::PASS , "4" );
			return TOP;
		}
		return defaultSide;
	}
	return NONE;
}

Vector2D NewPositioningSystem::getSidePosition( PositionInLine ballSide , PositionInLine playerSide , Vector2D ballPos , float offsideLine)
{
	float yChange = 0;
	float bottomDist = OffenseVariables::Instance()->bottomDist();
	float topDist = OffenseVariables::Instance()->topDist();
	dlog.addText ( Logger::PASS , "player side \t %d" , playerSide );
	if ( ballSide == BOTTOM )
	{
		if ( playerSide == MIDDLE )
		{
//			dlog.addText ( Logger::PASS , "1-1" );
			yChange = bottomDist;
		}
		else if ( playerSide == TOP )
		{
//			dlog.addText ( Logger::PASS , "1-2" );
			yChange = bottomDist + topDist;
		}
	}
	else if ( ballSide == MIDDLE )
	{
		if ( playerSide == TOP )
		{
//			dlog.addText ( Logger::PASS , "2-1" );
			yChange = topDist;
		}
		else if ( playerSide == BOTTOM )
		{
//			dlog.addText ( Logger::PASS , "2-2" );
			yChange = -bottomDist;
		}
	}
	else if ( ballSide == TOP )
	{
		if ( playerSide == MIDDLE )
		{
//			dlog.addText ( Logger::PASS , "3-1" );
			yChange = -topDist;
		}
		else if ( playerSide == BOTTOM )
		{
//			dlog.addText ( Logger::PASS , "3-2" );
			yChange = -( bottomDist + topDist );
		}
	}
	Vector2D myPos = ballPos;
	myPos.y += yChange;
	myPos.x += OffenseVariables::Instance()->ballXAddition();
//	dlog.addText ( Logger::PASS , "%f added ----> myPos.x\t%f" , OffenseVariables::Instance()->ballXAddition() , myPos.x );
	if ( myPos.x >= offsideLine )
		myPos.x = offsideLine;
	if ( myPos.y >= maxPositionY )
		myPos.y = maxPositionY;
	if ( myPos.y <= minPositionY )
		myPos.y = minPositionY;
//	dlog.addText ( Logger::PASS , "myPos.x\t%f" , myPos.x );
	return myPos;
}

FakePositionInLine NewPositioningSystem::getBallFakeSide( Vector2D ballPos )
{
//	dlog.addText ( Logger::PASS , "ballY\t%f" , fabsf( ballPos.y ) );
//	dlog.addText ( Logger::PASS , "minRestrictedYForMiddleOwner\t%f" , minRestrictedYForMiddleOwner );
	if ( abs ( ballPos.y ) >= maxAbsoluteYForKeepingLastDecision )
	{
		if ( ballPos.y <= 0 )
		{
			if ( fabsf( ballPos.y ) <= minRestrictedYForMiddleOwner )
				return BOTTOMMIDDLE;
			else
				return BOTTOMOUT;
		}
		else
		{
			if ( fabsf(ballPos.y) <= minRestrictedYForMiddleOwner )
			{
				return TOPMIDDLE;
			}
			else
			{
				return TOPOUT;
			}
		}
	}
	else
	{
		return lastBallFakeSide;
	}
	return NONEOUT;
}

Vector2D NewPositioningSystem::getSidePosition( FakePositionInLine ballFakeSide , PositionInLine playerSide , rcsc::Vector2D ballPos , float offsideLine , int receiver )
{
	float yChange = 0;
	float bottomDist = OffenseVariables::Instance()->bottomDist();
	float bottomOutDist = OffenseVariables::Instance()->bottomOutDist();
	float topDist = OffenseVariables::Instance()->topDist();
	float topOutDist = OffenseVariables::Instance()->topOutDist();
	if ( ballFakeSide == BOTTOMOUT )
	{
		if ( playerSide == BOTTOM )
			yChange = bottomOutDist;
		else if ( playerSide == MIDDLE )
			yChange = bottomOutDist + bottomDist;
		else if ( playerSide == TOP )
			yChange = bottomOutDist + bottomDist + topDist;
	}
	else if ( ballFakeSide == BOTTOMMIDDLE )
	{
		if ( playerSide == BOTTOM )
			yChange = - bottomOutDist;
		else if ( playerSide == MIDDLE )
			yChange = bottomDist;
		else if ( playerSide == TOP )
			yChange = bottomDist + topDist;
	}
	else if ( ballFakeSide == TOPMIDDLE )
	{
		if ( playerSide == BOTTOM )
			yChange = -( topDist + bottomDist );
		else if ( playerSide == MIDDLE )
			yChange = - topDist;
		else if ( playerSide == TOP )
			yChange = topOutDist;
	}
	else if ( ballFakeSide == TOPOUT )
	{
		if ( playerSide == BOTTOM )
			yChange = - ( topOutDist + topDist + bottomDist );
		else if ( playerSide == MIDDLE )
			yChange = - ( topOutDist + topDist );
		else if ( playerSide == TOP )
			yChange = - topOutDist;
	}
	Vector2D myPos = ballPos;
	myPos.y += yChange;
	if ( myPos.y >= maxPositionY )
		myPos.y = maxPositionY;
	if ( myPos.y <= minPositionY )
		myPos.y = minPositionY;
	// Determining position x
	{
		float myX = myPos.x;
		string quarterBack = OffenseVariables::Instance()->quarterBackPlayer();
		if ( Basics::isInPlayers( receiver , quarterBack ) )
		{
			myX += OffenseVariables::Instance()->ballXAddition();
		}
		else
		{
			float ballXWeight = OffenseVariables::Instance()->ballXWeight();
			float offsideLineWeight = OffenseVariables::Instance()->offsideLineWeight();
			if ( offsideLineWeight + ballXWeight >= 1.f )
				myX = ( ( offsideLineWeight * offsideLine ) + ( ballXWeight * ballPos.x ) ) / ( offsideLineWeight + ballXWeight );
		}
		myPos.x = myX;
	}
	if ( myPos.x >= offsideLine )
		myPos.x = offsideLine;
	return myPos;
}

Vector2D NewPositioningSystem::getQBPosition( PositionInLine ballSide , rcsc::Vector2D ballPos , float offsideLine )
{
	float yChange = 0.f;
	float bottomDist = OffenseVariables::Instance()->bottomDist();
	float topDist = OffenseVariables::Instance()->topDist();
	if ( ballSide == BOTTOM )
	{
		yChange = bottomDist / 2.f;
	}
	else if ( ballSide == MIDDLE )
	{
		if ( fabsf( ballPos.y ) < 3.f )
		{
			if ( lastQuarterBackSide == UP )
				yChange = topDist / 2.f;
			else
				yChange = - bottomDist / 2.f;
		}
		else
		if ( ballPos.y > 0 )
		{
			lastQuarterBackSide = DOWN;
			yChange = - bottomDist / 2.f;
		}
		else
		{
			lastQuarterBackSide = UP;
			yChange = topDist / 2.f;
		}
	}
	else if ( ballSide == TOP )
	{
		yChange = - topDist / 2.f;
	}
	Vector2D myPos = ballPos;
/*	myPos.x += OffenseVariables::Instance()->ballXAddition();
	if ( myPos.x >= offsideLine )
		myPos.x = offsideLine;
*/	myPos.x -= OffenseVariables::Instance()->quarterBackDist();
	myPos.y += yChange;
	if ( myPos.y >= maxPositionY )
		myPos.y = maxPositionY;
	if ( myPos.y <= minPositionY )
		myPos.y = minPositionY;
	if ( myPos.x >= offsideLine )
		myPos.x = offsideLine;
	return myPos;
}

Vector2D NewPositioningSystem::getQBPosition( FakePositionInLine ballFakeSide , rcsc::Vector2D ballPos , float offsideLine )
{
	float yChange = 0.f;
	float bottomDist = OffenseVariables::Instance()->bottomDist();
	float bottomOutDist = OffenseVariables::Instance()->bottomOutDist();
	float topDist = OffenseVariables::Instance()->topDist();
	float topOutDist = OffenseVariables::Instance()->topOutDist();
	if ( ballFakeSide == BOTTOMOUT )
	{
		yChange = bottomOutDist + ( bottomDist / 2.f );
	}
	else if ( ballFakeSide == BOTTOMMIDDLE )
	{
		yChange = 0.f;
	}
	else if ( ballFakeSide == TOPMIDDLE )
	{
		yChange = 0.f;
	}
	else if ( ballFakeSide == TOPOUT )
	{
		yChange = - ( topOutDist + ( topDist / 2.f ) );
	}
	Vector2D myPos = ballPos;
/*	float ballXWeight = OffenseVariables::Instance()->ballXWeight();
	float offsideLineWeight = OffenseVariables::Instance()->offsideLineWeight();
	if ( offsideLineWeight + ballXWeight >= 1.f )
		myPos.x = ( ( offsideLineWeight * offsideLine ) + ( ballXWeight * ballPos.x ) ) / ( offsideLineWeight + ballXWeight );
	if ( myPos.x >= offsideLine )
		myPos.x = offsideLine;
	myPos.x -= OffenseVariables::Instance()->quarterBackDist();
*/	myPos.y += yChange;
	if ( myPos.y >= maxPositionY )
		myPos.y = maxPositionY;
	if ( myPos.y <= minPositionY )
		myPos.y = minPositionY;
	if ( myPos.x >= offsideLine )
		myPos.x = offsideLine;
	return myPos;
}

Vector2D NewPositioningSystem::getHomePos( const WorldModel * wm )
{
	float maxXLine = wm->offsideLineX() - 1.f;
//	dlog.addText ( Logger::PASS , "maxXLine\t%f" , maxXLine );
	int fastestUnum = -1;
	int myUnum = wm->self().unum();
	FastIC fast;
	Vector2D ballPos = wm->ball().pos();
	if ( fast.getFastestTeammate() )
	{
		fastestUnum = fast.getFastestTeammate()->unum();
		ballPos = inertia_n_step_point( ballPos , wm->ball().vel() , fast.getFastestTeammateReachCycle() , ServerParam::i().ballDecay() );
	}
//	dlog.addText ( Logger::PASS , "fastestUnum\t%d" , fastestUnum );
//	dlog.addText ( Logger::PASS , "ballPos\t( %.2f , %.2f )" , ballPos.x , ballPos.y );
	lastBallOwnerSideStats = getBallOwnerSideStats( ballPos.y );
	if ( fastestUnum == myUnum )
	{
		lastBallOwnerSideStats = RESET;
		return ballPos;
	}
	PositionInLine ballSide = getBallOwnerSide( ballPos , fastestUnum );
	FakePositionInLine fakeSide = getBallFakeSide( ballPos );
	lastBallFakeSide = fakeSide;
	string quarterBack = OffenseVariables::Instance()->quarterBackPlayer();
	dlog.addText ( Logger::PASS , "linearPlayers\t%s" , OffenseVariables::Instance()->linearPlayers().c_str() );
//	dlog.addText ( Logger::PASS , "BallSide\t%d", (int)ballSide );
	if ( ballSide == NONE )
	{
//		dlog.addText ( Logger::PASS , "fakeBallSide\t%d" , (int)fakeSide );
		if ( Basics::isInPlayers( myUnum , quarterBack ) )
		{
			return getQBPosition ( fakeSide , ballPos , maxXLine );
		}
		string playersInPos = OffenseVariables::Instance()->linearPlayers();
//		dlog.addText ( Logger::PASS , "playersInPos	%s" , playersInPos.c_str() );
		// Determining Positions
		for ( int i = 0; i < 3 ; i++ )
		{
			string currentPlayer = " ";
			currentPlayer[0] = playersInPos[i];
			if ( currentPlayer[0] == 'Q' )
				continue;
			Vector2D myTarget;
			if ( Basics::isInPlayers ( myUnum , currentPlayer ) )
			{
//				dlog.addText ( Logger::PASS , "me in %d" , i );
				myTarget = getSidePosition( fakeSide , (PositionInLine)i , ballPos , maxXLine , fastestUnum );
//				dlog.addText ( Logger::PASS , "myTarget\t( %.2f , %.2f )" , myTarget.x , myTarget.y );
				if ( wm->self().pos().x >= wm->offsideLineX() )
				{
					myTarget.x = wm->offsideLineX() - 1.f;
					myTarget.y = wm->self().pos().y;
				}
				return myTarget;
			}
		}
	}
	else
	{
		// Determining players
		if ( Basics::isInPlayers( myUnum , quarterBack ) )
		{
			return getQBPosition ( ballSide , ballPos , maxXLine );
		}
		string linearPlayers = OffenseVariables::Instance()->linearPlayers();
		string playersInPos = "QQQ";
		int playerPointer , sidePointer;
		playerPointer = sidePointer = 0;
		int ballArea = ballSide;
		char fastestPlayerChar = AudioCodec::i().unum2hex( fastestPlayerChar );
		while ( ( playerPointer < 3 ) and ( sidePointer < 3 ) ) 
		{
			string currentPlayer = " ";
			currentPlayer[0] = linearPlayers[playerPointer];
			if ( Basics::isInPlayers ( fastestUnum , currentPlayer ) )
				fastestPlayerChar = currentPlayer[0];
			if ( ( sidePointer == ballArea ) and ( Basics::isInPlayers ( fastestUnum , linearPlayers ) ) )
			{
				sidePointer++;
				continue;
			}
			if ( Basics::isInPlayers ( fastestUnum , currentPlayer ) )
			{
				playerPointer++;
				continue;
			}
			playersInPos[sidePointer] = currentPlayer[0];
			playerPointer++;
			sidePointer++;
		}
		if ( Basics::isInPlayers ( fastestUnum , linearPlayers ) )
			playersInPos[ballArea] = fastestPlayerChar;
		int equalChar = -1;
		for ( unsigned int i = 0; i < playersInPos.size(); i++ )
			if ( playersInPos[i] == linearPlayers[i] )
			{
				equalChar = i;
				break;
			}
		if ( ( equalChar > -1 ) and ( equalChar != 1 ) and ( playersInPos.size() == 3 ) )
		{
			vector<Vector2D> positions;
			if ( AudioCodec::i().hex2unum( playersInPos[0] ) == wm->self().unum() )
				positions.push_back( wm->self().pos() );
			const PlayerPtrCont::const_iterator end = wm->teammatesFromSelf().end();
			for ( PlayerPtrCont::const_iterator it = wm->teammatesFromSelf().begin();
				it != end;
				++it )
			{
				if ( (*it)->unum() < 0 )
					continue;
				if ( (*it)->unum() == AudioCodec::i().hex2unum( playersInPos[1] ) )
					continue;
				if ( not Basics::isInPlayers( (*it)->unum() , playersInPos ) )
					continue;
				if ( (*it)->posCount() > 3 )
					continue;
				positions.push_back( (*it)->pos() );
			}
			if ( positions.size() == 2 )
			{
				if ( fabsf( positions[1].x - positions[2].x ) < 2.f )
				{
					dlog.addText ( Logger::PASS , "EDITED!!" );
					playersInPos = linearPlayers;
				}
			}
		}
		// Determining Positions
		for ( int i = 0; i < 3 ; i++ )
		{
			string currentPlayer = " ";
			currentPlayer[0] = playersInPos[i];
			if ( currentPlayer[0] == 'Q' )
				continue;
			Vector2D myTarget;
			if ( Basics::isInPlayers ( myUnum , currentPlayer ) )
			{
				myTarget = getSidePosition( ballSide , (PositionInLine)i , ballPos , maxXLine );
//				dlog.addText ( Logger::PASS , "myTarget = (%.2f , %.2f)" , myTarget.x , myTarget.y );
				if ( wm->self().pos().x >= wm->offsideLineX() )
				{
					myTarget.x = wm->offsideLineX() - 1.f;
					myTarget.y = wm->self().pos().y;
				}
				return myTarget;
			}
		}
	}
	return Vector2D( -1000 , -1000 );
}

bool NewPositioningSystem::getConstantPoint( const Vector2D homePos , Vector2D & answer)
{
	int gridXCount = OffenseVariables::Instance()->gridXCount();
	int gridYCount = OffenseVariables::Instance()->gridYCount();
	float gridLenght = 105.f / (float)gridXCount;
	float gridWidth = 68.f / (float)gridYCount;
	float maxDist = hypot( gridLenght , gridWidth );
	for ( int i = 0; i < gridXCount; i++ )
		for ( int j = 0; j < gridYCount; j++ )
		{
			Vector2D bottomBack(i * gridLenght,j * gridWidth);
			Vector2D topForward( (i + 1) * gridLenght , (j + 1) * gridWidth );
			bottomBack += Vector2D( -52.5 , -34.f );
			topForward += Vector2D( -52.5 , -34.f );
			if ( homePos.x < bottomBack.x )
				continue;
			if ( homePos.y < bottomBack.y )
				continue;
			if ( homePos.x > topForward.x )
				continue;
			if ( homePos.y > topForward.y )
				continue;
			Vector2D constantPoint = ( topForward + bottomBack ) / 2.f;
			if ( homePos.dist( constantPoint ) > maxDist )
			{
				answer = constantPoint;
				return true;
			}
			else
			{
				answer = homePos;
				return false;
			}
		}
	answer = homePos;
	return false;
}
