// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <OffenseVariables.h>
#include <FastIC.h>
#include <Basics.h>

#include <string>
#include <vector>

#include <rcsc/math_util.h>

#include <rcsc/geom/vector_2d.h>

#include <rcsc/common/audio_memory.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_codec.h>
#include <rcsc/common/logger.h>

#include <rcsc/player/say_message_builder.h>
#include <rcsc/player/world_model.h>

using namespace rcsc;
using namespace std;

int OffenseVariables::lastDeclarationCycle = -50;

OffenseVariables::OffenseVariables()
{
	mLinearPlayers = "9BA";
	mQuarterBackPlayer = "6";
	mLaneShift = 0.f;
	mBallXAddition = 10.f;
	mBottomDist = mTopDist = 10.f;
	mBottomOutDist = mTopOutDist = 10.f;
	mQuarterBackDist = 5.f;
	mBallXWeight = mOffsideLineWeight = 1;
	mGridXCount = mGridYCount = 10;
}

void OffenseVariables::reset()
{
	mLinearPlayers = "9BA";
	mQuarterBackPlayer = "6";
	mLaneShift = 0.f;
	mBallXAddition = 10.f;
	mBottomDist = mTopDist = 10.f;
	mBottomOutDist = mTopOutDist = 10.f;
	mQuarterBackDist = 5.f;
	mBallXWeight = mOffsideLineWeight = 1;
	mGridXCount = mGridYCount = 10;
}

void OffenseVariables::changeVariablesIfNeeded( rcsc::Vector2D ballPos )
{
	if ( ballPos.x > 30.f )
	{
		mQuarterBackDist = 5.f;
		mTopDist = mBottomDist = mTopOutDist = mBottomOutDist = 15.f;
	}
	else
	{
		mQuarterBackDist = 5.f;
		mTopDist = mBottomDist = mTopOutDist = mBottomOutDist = 10.f;
	}
}

void OffenseVariables::editLinearPlayers( rcsc::PlayerAgent * agent )
{
	dlog.addText ( Logger::PASS , "mLinearPlayers\t%s" , mLinearPlayers.c_str() );
	vector< Vector2D > positions;
	vector< int > unum;
	int fastestBlock = -1;
	int fastestUnum = -1;
	if ( FastIC::getFastestTeammate() )
		fastestUnum = FastIC::getFastestTeammate()->unum();
	int c = -1;
	const WorldModel & wm = agent->world();
	const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
		it != end;
		++it )
	{
		Vector2D hisPos = (*it)->pos();
		int hisUNum = (*it)->unum();
		if ( hisUNum < 0 )
			continue;
		if ( Basics::isInPlayers( hisUNum , mLinearPlayers ) )
		{
			positions.push_back( hisPos );
			unum.push_back( hisUNum );
			c++;
		}
		if ( fastestUnum > 0 )
			if ( hisUNum == fastestUnum )
				fastestBlock = c;
	}
	if ( Basics::isInPlayers( wm.self().unum() , mLinearPlayers ) )
	{
		positions.push_back( wm.self().pos() );
		unum.push_back( wm.self().unum() );
		c++;
		if ( fastestUnum > 0 )
			if ( wm.self().unum() == fastestUnum )
				fastestBlock = c;
	}
	if ( positions.size() != mLinearPlayers.size() )
		return;
	bool isChanged = true;
	dlog.addText ( Logger::PASS , "fastestBlocknow %d" , fastestBlock );
	while ( isChanged )
	{
		isChanged = false;
		for ( unsigned int i = 0; i < positions.size() - 1; i++ )
		{
			if ( positions[i].y > positions[i+1].y )
			{
				Vector2D help = positions[i];
				positions[i] = positions[i+1];
				positions[i+1] = help;
				int un = unum[i];
				unum[i] = unum[i+1];
				unum[i+1] = un;
				isChanged = true;
				if ( fastestBlock == i )
					fastestBlock = i+1;
				else
					if ( fastestBlock == i+1 )
						fastestBlock = i;
			}
		}
	}
	for ( unsigned int i = 0; i < unum.size(); i++ )
		dlog.addText ( Logger::PASS , "unum[%d]%d" , i , unum[i] );
	dlog.addText ( Logger::PASS , "_____________________" );
	dlog.addText ( Logger::PASS , "fastestBlock\t%d" , fastestBlock );
/*	if ( ( fastestBlock >= 0 ) and ( Basics::isInPlayers( fastestUnum , mLinearPlayers ) ) )
	{
		dlog.addText ( Logger::PASS , "1111" );
		Vector2D fastestReachPos = inertia_n_step_point( wm.ball().pos() , wm.ball().vel() , FastIC::getFastestTeammateReachCycle() , ServerParam::i().ballDecay() );
		bool shouldCheckTop = false;
		if ( ( fastestReachPos.x > 36.f ) and ( fastestReachPos.y > 20.f ) )
			shouldCheckTop = true;
		if ( ( fastestReachPos.x < 36.f ) and ( fastestReachPos.y > 34.f - mTopDist - 2.f ) )
			shouldCheckTop = true;
		if ( shouldCheckTop )
		{
			if ( fastestBlock != ((signed)positions.size() - 1) )
			{
				if ( positions.size() > 1 )
					for ( int vectorMarker = fastestBlock; vectorMarker < ((signed)positions.size() - 1); vectorMarker++ )
						unum[vectorMarker] = unum[vectorMarker + 1];
				unum[positions.size() - 1] = fastestUnum;
			}
		}
		bool shouldCheckBottom = false;
		if ( ( fastestReachPos.x > 36.f ) and ( fastestReachPos.y < -20.f ) )
			shouldCheckBottom = true;
		if ( ( fastestReachPos.x < 36.f ) and ( fastestReachPos.y < -34.f + mBottomDist + 2.f ) )
			shouldCheckBottom = true;
		if ( shouldCheckBottom )
		{
			dlog.addText ( Logger::PASS , "fastestUnum\t%d" , fastestUnum );
			if ( fastestBlock != 0 )
			{
				for ( unsigned int i = fastestBlock; i > 0; i-- )
				{
					unum[i] = unum[i-1];
					dlog.addText ( Logger::PASS , "unum[%d]=%d" , i , unum[i] );
				}
				unum[0] = fastestUnum;
			}	
		}
	}
*/	string playersToSay = "";
	for ( unsigned int i = 0; i < positions.size(); i++ )
	{
		playersToSay += AudioCodec::i().unum2hex( unum[i] );
	}
	dlog.addText ( Logger::PASS , "playersToSay %s" , playersToSay.c_str() );
	dlog.addText ( Logger::PASS , "mLinearPlayers\t%s" , mLinearPlayers.c_str() );
	if ( mLinearPlayers != playersToSay )
	{
		mLinearPlayers = playersToSay;
		lastDeclarationCycle = wm.time().cycle();
	}
	if ( wm.time().cycle() - lastDeclarationCycle <= 7 )
	{
		dlog.addText( Logger::PASS , "said!!!" );
		agent->addSayMessage( new HelliPositionMessage(playersToSay) );
	}
}

void OffenseVariables::sayLinearPlayers( rcsc::PlayerAgent* agent )
{
	if ( agent->world().time().cycle() - lastDeclarationCycle <= 7 )
		agent->addSayMessage( new HelliPositionMessage(mLinearPlayers) );
}

void OffenseVariables::updateBySay( const WorldModel & wm )
{
	dlog.addText ( Logger::PASS , "Changing!!!!!!!?????????" );
	if ( ( not wm.audioMemory().helliPosition().empty() ) and ( wm.audioMemory().helliPositionTime().cycle() == wm.time().cycle() ) )
	{
		mLinearPlayers = wm.audioMemory().helliPosition().front().playerQueue_;
	}
}

Vector2D OffenseVariables::getRunnerPosition( float offsideLine , Vector2D runnerPos )
{
	runnerPos.x = offsideLine;
	return runnerPos;
}
