// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "PassConfig.h"

#include "OffenseVariables.h"

#include <rcsc/common/logger.h>

using namespace rcsc;
using namespace std;

PassConfig::PassConfig(Expression& passConf) : shootingPolygon() , keepingPolygon() , targetPolygon()
{
	priority = 1000;
	if ( passConf.Search("Priority") )
		priority = passConf["Priority"][0][1].AsInt();
	dlog.addText( Logger::PASS , "Constructing!!!Priority %d" ,priority );
	
	minValue = 0.f;
	if ( passConf.Search("MinValue") )
		minValue = passConf["MinValue"][0][1].AsFloat();
		
	mustKickOnce = false;
	if ( passConf.Search("MustKickOnce") )
		mustKickOnce = passConf["MustKickOnce"][0][1].AsInt();
	
	receivers = "23456789AB";
	if ( passConf.Search("Receivers") )
		receivers = passConf["Receivers"][0][1].AsString();
	
	hasShootingPolygon = false;
	hasKeepingPolygon = false;
	hasTargetPolygon = false;

	distToGoalWeight = 0;
	checkDistToGoal = false;
	maxDistanceToGoal = 110.f;
	
	forwardMovementWeight = 0;
	checkForwardMovement = false;
	maxForwardMovement = 105.f;

	receiverMovementWeight = 0;
	checkReceiverMovement = false;
	maxReceiverMovement = 110.f;
	
	receiverYChangeWeight = 0;
	checkReceiverYChange = false;
	maxReceiverYChange = 34.f;

	receiverAbilityWeight = 0;
	checkReceiverAbility = false;
	maxOpponents = 11;
	
	if ( passConf.Search("Weights") )
	{
		Expression weights = passConf["Weights"][0];
		if ( weights.Search("DistToGoal") )
		{
			distToGoalWeight = weights["DistToGoal"][0]["Weight"][0][1].AsInt();
			checkDistToGoal = weights["DistToGoal"][0]["CheckIt"][0][1].AsInt();
			if ( weights["DistToGoal"][0]["Max"][0][1].AsString() != "Current" )
				maxDistanceToGoal = weights["DistToGoal"][0]["Max"][0][1].AsFloat();
			else maxDistanceToGoal = -10;
		}
		if ( weights.Search("ForwardMovement") )
		{
			forwardMovementWeight = weights["ForwardMovement"][0]["Weight"][0][1].AsInt();
			checkForwardMovement = weights["ForwardMovement"][0]["CheckIt"][0][1].AsInt();
			maxForwardMovement = weights["ForwardMovement"][0]["Max"][0][1].AsFloat();
		}
		if ( weights.Search("ReceiverYChange") )
		{
			receiverYChangeWeight = weights["ReceiverYChange"][0]["Weight"][0][1].AsInt();
			checkReceiverYChange = weights["ReceiverYChange"][0]["CheckIt"][0][1].AsInt();
			maxReceiverYChange = weights["ReceiverYChange"][0]["Max"][0][1].AsFloat();
		}
		if ( weights.Search("ReceiverMovement") )
		{
			receiverMovementWeight = weights["ReceiverMovement"][0]["Weight"][0][1].AsInt();
			checkReceiverMovement = weights["ReceiverMovement"][0]["CheckIt"][0][1].AsInt();
			maxReceiverMovement = weights["ReceiverMovement"][0]["Max"][0][1].AsFloat();
		}
		if ( weights.Search("ReceiverAbility") )
		{
			receiverAbilityWeight = weights["ReceiverAbility"][0]["Weight"][0][1].AsInt();
			checkReceiverAbility = weights["ReceiverAbility"][0]["CheckIt"][0][1].AsInt();
			maxOpponents = weights["ReceiverAbility"][0]["Max"][0][1].AsInt();
		}
	}
	if ( passConf.Search("ShootingPolygon") )
	{
		Expression polygon = passConf["ShootingPolygon"][0];
		for ( unsigned int i = 0; i < polygon["Vertex"].size(); i++)
		{
			hasShootingPolygon = true;
			float x,y;
			x = polygon["Vertex"][i]["X"][0][1].AsFloat();
			y = polygon["Vertex"][i]["Y"][0][1].AsFloat();
			shootingPolygon.addVertex( Vector2D(x,y));
		}
	}
	
	if ( passConf.Search("KeepingPolygon") )
	{
		Expression polygon = passConf["KeepingPolygon"][0];
		for ( unsigned int i = 0; i < polygon["Vertex"].size(); i++)
		{
			hasKeepingPolygon = true;
			float x,y;
			x = polygon["Vertex"][i]["X"][0][1].AsFloat();
			y = polygon["Vertex"][i]["Y"][0][1].AsFloat();
			keepingPolygon.addVertex( Vector2D(x,y));
		}
	}
	
	if ( passConf.Search("TargetPolygon") )
	{
		Expression polygon = passConf["TargetPolygon"][0];
		for ( unsigned int i = 0; i < polygon["Vertex"].size(); i++)
		{
			hasTargetPolygon = true;
			float x,y;
			x = polygon["Vertex"][i]["X"][0][1].AsFloat();
			y = polygon["Vertex"][i]["Y"][0][1].AsFloat();
			targetPolygon.addVertex( Vector2D(x,y));
		}
	}
}

void PassConfig::reconstruct(Expression & passConf)
{
	priority = 1000;
	if ( passConf.Search("Priority") )
		priority = passConf["Priority"][0][1].AsInt();
	
	minValue = 0.f;
	if ( passConf.Search("MinValue") )
		minValue = passConf["MinValue"][0][1].AsFloat();
		
	mustKickOnce = false;
	if ( passConf.Search("MustKickOnce") )
		mustKickOnce = passConf["MustKickOnce"][0][1].AsInt();
	
	receivers = "23456789AB";
	if ( passConf.Search("Receivers") )
		receivers = passConf["Receivers"][0][1].AsString();
	
	hasShootingPolygon = false;
	hasKeepingPolygon = false;
	hasTargetPolygon = false;

	distToGoalWeight = 0;
	checkDistToGoal = false;
	maxDistanceToGoal = 110.f;
	
	forwardMovementWeight = 0;
	checkForwardMovement = false;
	maxForwardMovement = 105.f;

	receiverMovementWeight = 0;
	checkReceiverMovement = false;
	maxReceiverMovement = 110.f;
	
	receiverYChangeWeight = 0;
	checkReceiverYChange = false;
	maxReceiverYChange = 34.f;

	receiverAbilityWeight = 0;
	checkReceiverAbility = false;
	maxOpponents = 11;
	
	if ( passConf.Search("Weights") )
	{
		Expression weights = passConf["Weights"][0];
		if ( weights.Search("DistToGoal") )
		{
			distToGoalWeight = weights["DistToGoal"][0]["Weight"][0][1].AsInt();
			checkDistToGoal = weights["DistToGoal"][0]["CheckIt"][0][1].AsInt();
			if ( weights["DistToGoal"][0]["Max"][0][1].AsString() != "Current" )
				maxDistanceToGoal = weights["DistToGoal"][0]["Max"][0][1].AsFloat();
			else maxDistanceToGoal = -10;
		}
		if ( weights.Search("ForwardMovement") )
		{
			forwardMovementWeight = weights["ForwardMovement"][0]["Weight"][0][1].AsInt();
			checkForwardMovement = weights["ForwardMovement"][0]["CheckIt"][0][1].AsInt();
			maxForwardMovement = weights["ForwardMovement"][0]["Max"][0][1].AsFloat();
		}
		if ( weights.Search("ReceiverYChange") )
		{
			receiverYChangeWeight = weights["ReceiverYChange"][0]["Weight"][0][1].AsInt();
			checkReceiverYChange = weights["ReceiverYChange"][0]["CheckIt"][0][1].AsInt();
			maxReceiverYChange = weights["ReceiverYChange"][0]["Max"][0][1].AsFloat();
		}
		if ( weights.Search("ReceiverMovement") )
		{
			receiverMovementWeight = weights["ReceiverMovement"][0]["Weight"][0][1].AsInt();
			checkReceiverMovement = weights["ReceiverMovement"][0]["CheckIt"][0][1].AsInt();
			maxReceiverMovement = weights["ReceiverMovement"][0]["Max"][0][1].AsFloat();
		}
		if ( weights.Search("ReceiverAbility") )
		{
			receiverAbilityWeight = weights["ReceiverAbility"][0]["Weight"][0][1].AsInt();
			checkReceiverAbility = weights["ReceiverAbility"][0]["CheckIt"][0][1].AsInt();
			maxOpponents = weights["ReceiverAbility"][0]["Max"][0][1].AsInt();
		}
	}
	if ( passConf.Search("ShootingPolygon") )
	{
		Expression polygon = passConf["ShootingPolygon"][0];
		for ( unsigned int i = 0; i < polygon["Vertex"].size(); i++)
		{
			hasShootingPolygon = true;
			float x,y;
			x = polygon["Vertex"][i]["X"][0][1].AsFloat();
			y = polygon["Vertex"][i]["Y"][0][1].AsFloat();
			shootingPolygon.addVertex( Vector2D(x,y));
		}
	}
	
	if ( passConf.Search("KeepingPolygon") )
	{
		Expression polygon = passConf["KeepingPolygon"][0];
		for ( unsigned int i = 0; i < polygon["Vertex"].size(); i++)
		{
			hasKeepingPolygon = true;
			float x,y;
			x = polygon["Vertex"][i]["X"][0][1].AsFloat();
			y = polygon["Vertex"][i]["Y"][0][1].AsFloat();
			keepingPolygon.addVertex( Vector2D(x,y));
		}
	}
	
	if ( passConf.Search("TargetPolygon") )
	{
		Expression polygon = passConf["TargetPolygon"][0];
		for ( unsigned int i = 0; i < polygon["Vertex"].size(); i++)
		{
			hasTargetPolygon = true;
			float x,y;
			x = polygon["Vertex"][i]["X"][0][1].AsFloat();
			y = polygon["Vertex"][i]["Y"][0][1].AsFloat();
			targetPolygon.addVertex( Vector2D(x,y));
		}
	}
}

PassConfig::~PassConfig()
{
}

void PassConfig::setMinValue(float f)
{
	minValue = f;
}

float PassConfig::getMinValue() const
{
	return minValue;
}

void PassConfig::setPriority(int f)
{
	priority = f;
}

int PassConfig::getPriority() const
{
	return priority;
}

bool PassConfig::mustKickItOnce()
{
	return mustKickOnce;
}

string PassConfig::convertReceivers()
{
	string answer = "";
	string linearPlayers = OffenseVariables::Instance()->linearPlayers();
	for ( unsigned int i = 0; i < receivers.size(); i++ )
	{
		if ( receivers[i] == '6' )
			answer += OffenseVariables::Instance()->quarterBackPlayer();
		if ( linearPlayers.size() == 3 )
		{
			if ( receivers[i] == 'A' )
				answer += linearPlayers[2];
			if ( receivers[i] == 'B' )
				answer += linearPlayers[1];
			if ( receivers[i] == '9' )
				answer += linearPlayers[0];
		}
		if ( receivers[i] == 'D' )
			answer += "234578";
	}
	return answer;
}

void PassConfig::setToPass( HelliBASH09Pass & pass , Vector2D cur)
{
	pass.setInsidePolygonPass(hasTargetPolygon);
//	dlog.addText ( Logger::PASS , "HasIT	%d" , hasTargetPolygon);
//	dlog.addText ( Logger::PASS , "targetPolygon.size()	%d" , targetPolygon.vertex().size() );
	pass.targetPolygon() = targetPolygon;
	pass.setOpenPlayers(receivers);
	if ( maxDistanceToGoal < 0 )
		pass.setDistToGoalParameters(distToGoalWeight,cur.dist(Vector2D(52.5,0)),checkDistToGoal);
	else
		pass.setDistToGoalParameters(distToGoalWeight,maxDistanceToGoal,checkDistToGoal);
	pass.setForwardMovementParameters(forwardMovementWeight,maxForwardMovement - cur.x ,checkForwardMovement);
	pass.setReceiverMovementParameters(receiverMovementWeight, maxReceiverMovement , checkReceiverMovement);
	pass.setReceiverYChangeParameters(receiverYChangeWeight,maxReceiverYChange,checkReceiverYChange);
	pass.setReceiverAbilityParameters(receiverAbilityWeight,maxOpponents,checkReceiverAbility);
}

void PassConfig::setToPass( PreDecidedPass & pass, rcsc::Vector2D cur)
{
	pass.setInsidePolygonPass(hasTargetPolygon);
//	dlog.addText ( Logger::PASS , "HasIT	%d" , hasTargetPolygon);
//	dlog.addText ( Logger::PASS , "targetPolygon.size()	%d" , targetPolygon.vertex().size() );
	pass.targetPolygon() = targetPolygon;
	pass.setOpenPlayers(convertReceivers());
	if ( maxDistanceToGoal < 0 )
		pass.setDistToGoalParameters(distToGoalWeight,cur.dist(Vector2D(52.5,0)),checkDistToGoal);
	else
		pass.setDistToGoalParameters(distToGoalWeight,maxDistanceToGoal,checkDistToGoal);
	pass.setForwardMovementParameters(forwardMovementWeight,maxForwardMovement - cur.x ,checkForwardMovement);
	pass.setReceiverMovementParameters(receiverMovementWeight, maxReceiverMovement , checkReceiverMovement);
	pass.setReceiverAbilityParameters(receiverAbilityWeight,maxOpponents,checkReceiverAbility);
}

Polygon2D & PassConfig::ShootingPolygon()
{
	return shootingPolygon;
}

Polygon2D & PassConfig::KeepingPolygon()
{
	return keepingPolygon;
}

Polygon2D & PassConfig::TargetPolygon()
{
	return targetPolygon;
}

bool PassConfig::HasShootingPolygon()
{
	return hasShootingPolygon;
}

bool PassConfig::HasKeepingPolygon()
{
	return hasKeepingPolygon;
}

bool PassConfig::HasTargetPolygon()
{
	return hasTargetPolygon;
}

bool operator <( const PassConfig & left , const PassConfig & right)
{
	if ( left.getPriority() < right.getPriority() )
		return true;
	else if ( left.getPriority() == right.getPriority() )
		return ( left.getMinValue() < right.getMinValue() );
	else return false;
}

bool operator >(const PassConfig & left , const PassConfig & right)
{
	if ( left.getPriority() > right.getPriority() )
		return true;
	else if ( left.getPriority() == right.getPriority() )
		return ( left.getMinValue() > right.getMinValue() );
	else return false;
}

bool operator <=(const PassConfig & left , const PassConfig & right)
{
	return ( not ( left > right ) );
}

bool operator >=(const PassConfig & left , const PassConfig & right)
{
	return ( not ( left < right ) );
}
