// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////
#ifndef _PASS_CONFIG_H_
#define _PASS_CONFIG_H_

#include <HelliBASH09Pass.h>
#include <PreDecidedPass.h>
#include <Config.h>

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/polygon_2d.h>
#include <rcsc/player/player_agent.h>

#include <string>

class PassConfig
{
private:
	int priority;
	float minValue;
	bool mustKickOnce;
	
	rcsc::Polygon2D shootingPolygon;
	bool hasShootingPolygon;
	rcsc::Polygon2D keepingPolygon;
	bool hasKeepingPolygon;
	rcsc::Polygon2D targetPolygon;
	bool hasTargetPolygon;
	
	int distToGoalWeight;
	bool checkDistToGoal;
	float maxDistanceToGoal;

	int forwardMovementWeight;
	bool checkForwardMovement;
	float maxForwardMovement;
	
	int receiverMovementWeight;
	bool checkReceiverMovement;
	float maxReceiverMovement;
	
	int receiverYChangeWeight;
	bool checkReceiverYChange;
	float maxReceiverYChange;
	
	int receiverAbilityWeight;
	bool checkReceiverAbility;
	int maxOpponents;
	
	std::string receivers;
	std::string convertReceivers();

public:
	PassConfig(Expression & passConf);
	void reconstruct(Expression & passConf);
	~PassConfig();
	void setMinValue(float f);
	float getMinValue() const;
	bool mustKickItOnce();
	void setPriority(int f);
	int getPriority() const;
	void setToPass( HelliBASH09Pass & pass, rcsc::Vector2D cur);
	void setToPass( PreDecidedPass & pass, rcsc::Vector2D cur);
	rcsc::Polygon2D & ShootingPolygon();
	bool HasShootingPolygon();
	bool HasKeepingPolygon();
	bool HasTargetPolygon();
	rcsc::Polygon2D & KeepingPolygon();
	rcsc::Polygon2D & TargetPolygon();
};

bool operator <(const PassConfig & left , const PassConfig & right);
bool operator >(const PassConfig & left , const PassConfig & right);
bool operator <=(const PassConfig & left , const PassConfig & right);
bool operator >=(const PassConfig & left , const PassConfig & right);

#endif // PassConfig
