// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "PassDecision.h"

#include <string>
#include <algorithm>

#include <rcsc/common/logger.h>
#include <Share.h>

#include <PreDecidedPass.h>

using namespace std;
using namespace rcsc;

int PassDecision::passStyle = -10;
vector<PassConfig> PassDecision::passConfs;

PassDecision::PassDecision(int style) : mPass()
{
	mustEstimateHalfs = false;
	oppDelay = 1;
	oppMaxDistanceDelay = 40;
	oppKickable = 1;
	oppCatchable = 1.2;
	oppMaxDash = 1;
	tmmDelay = 1;
	tmmMaxDistanceDelay = 40;
	tmmKickable = 1;
	tmmCatchable = 1.2;
	tmmMaxDash = 1;
	maxDirectionCount = 10;
	oppMaxPosCount = tmmMaxPosCount = 30;
	if ( passStyle != style )
	{
		passStyle = style;
		passConfs.clear();
		string addr = "PassDecision ";
		addr[addr.length() - 1] = (char)(style + 48);
		for ( unsigned int i = 0; i < CONFIG[addr]["PassConf"].size(); i++)
		{
			passConfs.push_back(PassConfig(CONFIG[addr]["PassConf"][i]));
		}
	}
//	dlog.addText ( Logger::PASS , "Constructed");
}

PassDecision::~PassDecision()
{
}

void PassDecision::setSafetyFactors(float _tmmKickable,float _tmmCatchable,float _tmmMaxDash,int _tmmDelay,int _tmmMaxPosCount,float _oppKickable,float _oppCatchable,float _oppMaxDash,int _oppDelay,int _oppMaxPosCount)
{
	oppDelay = _oppDelay;
	oppKickable = _oppKickable;
	oppCatchable = _oppCatchable;
	oppMaxDash = _oppMaxDash;
	oppMaxPosCount = _oppMaxPosCount;
	tmmDelay = _tmmDelay;
	tmmKickable = _tmmKickable;
	tmmCatchable = _tmmCatchable;
	tmmMaxDash = _tmmMaxDash;
	tmmMaxPosCount = _tmmMaxPosCount;
}

Decision PassDecision::getValue(PlayerAgent *agent, bool estimateHalfs)
{
    const rcsc::WorldModel & wm = agent->world();
	mustEstimateHalfs = estimateHalfs;
	prePass.reset();
	prePass.setDefaultSafetyFactors(3 , 0.85 , 0.85 , 1 , 1.05 , 1 , 0);
	prePass.setMaxDirectionCount( maxDirectionCount );
	tmmMaxPosCount = 2;
	oppMaxPosCount = 5;
	prePass.setMaxTmmPosCount( tmmMaxPosCount );
	prePass.setMaxOppPosCount( oppMaxPosCount );
/*	mPass.setSafetyFactors(tmmKickable,tmmCatchable,tmmMaxDash,tmmDelay,tmmMaxPosCount,oppKickable,oppCatchable,oppMaxDash,oppDelay,oppMaxPosCount);
	mPass.setDelayFactors( tmmDelay , tmmMaxDistanceDelay , oppDelay , oppMaxDistanceDelay);
	mPass.setMaxDirectionCount( maxDirectionCount );
*/	sort(passConfs.begin() , passConfs.end());
	float configValues[passConfs.size()];
	for ( unsigned int i = 0; i < passConfs.size(); i++ )
		configValues[i] = 0.f;
//	float valueChange = 1.f / (float)passConfs.size();
	float valueChange = 0.5;
	bool anyThingSelected = false;
	for ( unsigned int i = 0; i < passConfs.size();i++)
	{
		if ( ( anyThingSelected ) and ( i > 0 ) )
		{
			if ( ( passConfs[i].getPriority() >= 0 ) and ( passConfs[i-1].getPriority() < 0 ) )
				break;
			if ( ( passConfs[i].getPriority() >= 100 ) and ( passConfs[i-1].getPriority() < 100 ) )
				break;
		}
		dlog.addText ( Logger::PASS , "Testing Priority %d --- shoot %d , keep %d" , passConfs[i].getPriority() ,passConfs[i].HasShootingPolygon() , passConfs[i].HasKeepingPolygon() );
//		dlog.flush();
//		passConfs[i].setToPass(mPass , wm.ball().pos() );
		passConfs[i].setToPass(prePass , wm.ball().pos() );
		if ( passConfs[i].HasShootingPolygon() )
			if ( not passConfs[i].ShootingPolygon().contains( wm.self().pos() ) )
				continue;
		if ( passConfs[i].HasKeepingPolygon() )
			if ( passConfs[i].KeepingPolygon().contains( wm.self().pos() ) )
				continue;
//		dlog.addText ( Logger::PASS , "Can Do It");
//		dlog.flush();
//		float value	= mPass.getValue(agent,passConfs[i].mustKickItOnce(),estimateHalfs);
		float value = prePass.getValue(agent);
//		dlog.addText ( Logger::PASS , "MinValue	=	%f , Value	= %f" , passConfs[i].getMinValue() , value);
//		dlog.flush();
		if ( value < passConfs[i].getMinValue() )
			continue;
		configValues[i] = value + ( ( passConfs.size() - 1 - i ) * valueChange );
		if ( passConfs[i].getPriority() < 0 )
			configValues[i] += 1.f;
		if ( passConfs[i].getPriority() < 100 )
			configValues[i] += 1.f;
		anyThingSelected = true;
	}
	int selectedBlock = -1;
	float bestValue = 0.f;
	for ( unsigned int i = 0; i < passConfs.size(); i++ )
		if ( configValues[i] > bestValue )
		{
			bestValue = configValues[i];
			selectedBlock = i;
		}
	dlog.addText ( Logger::PASS , "selectedBlock\t%d" , selectedBlock );
	if ( selectedBlock < 0 )
		return CAN_NOT_DO_IT;
	passConfs[selectedBlock].setToPass(prePass , wm.ball().pos() );
	prePass.getValue( agent );
	if ( passConfs[selectedBlock].getPriority() < 0 )
		return SHOULD_DO_IT;
	else
//	else if ( passConfs[selectedBlock].getPriority() < 100 )
		return CAN_DO_IT;
//	else
//		return SHOULD_NOT_DO_IT;
}

void PassDecision::execute(rcsc::PlayerAgent *agent,bool isEnforced,int kickSteps)
{
//	mPass.execute(agent,isEnforced,kickSteps);
	prePass.execute(agent);
}

void PassDecision::setDelayFactors( int _tmmDelay , float _tmmMaxDistanceDelay , int _oppDelay , float _oppMaxDistanceDelay)
{
	tmmDelay = _tmmDelay;
	tmmMaxDistanceDelay = _tmmMaxDistanceDelay;
	oppDelay = _oppDelay;
	oppMaxDistanceDelay = _oppMaxDistanceDelay;
}

void PassDecision::setMaxDirectionCount(int f)
{
	maxDirectionCount = f;
}

