// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <PreDecidedBreaker.h>

#include <Basics.h>

#include <OffenseVariables.h>

#include <body_hold_ball2008.h>

#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>
#include <rcsc/action/intention_kick.h>
#include <rcsc/action/body_stop_ball.h>
#include <rcsc/action/body_turn_to_point.h>

#include <rcsc/player/interception.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/basic_client.h>
#include <rcsc/common/audio_codec.h>
#include <rcsc/common/logger.h>

#include <vector>

#define maxX 50
#define maxY 30
#define lateMovingDelay 7

using namespace rcsc;
using namespace std;

int PreDecidedBreaker::sumCount = 0;
int PreDecidedBreaker::numCount = 0;

float PreDecidedBreaker::dirCount = 40;
float PreDecidedBreaker::rateCount = 5;
int PreDecidedBreaker::countSetCycle = -1;

PreDecidedBreaker::PreDecidedBreaker( bool b ) : PreDecidedPassBase() , behindEnemyLines() , shouldCallRun( b )
{
	defenceBreakers = "69AB";
	targetLineX = -52.5;
	maxDistanceToTargetLine = 15.f;
	maxDistanceForCallingRun = 5.f;
	maxReceiverYChange = 10.f;
	minBallMovement = 5.f;
	maxPossibleRunCycles = 3;

	maxTmmPosCount = 4;
	maxOppPosCount = 20;

	maxRunCycles = 8;
	targetLineX = 0.f;
	startAngle = -90.f;
	endAngle = 90.f;
	startRate = 1.f;
	endRate = 3.f;
}

PreDecidedBreaker::~PreDecidedBreaker()
{
}

void PreDecidedBreaker::setMaxRunCycles(rcsc::PlayerAgent *agent)
{
	const WorldModel & wm = agent->world();
/*	if ( not wm.self().isKickable() )
	{
		maxRunCycles = wm.interceptTable()->selfReachCycle() - wm.time().cycle() + 1;
	}
	else
*/	{
		maxRunCycles = maxPossibleRunCycles;
		FastIC fast;
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisVel = Vector2D( 0.f , 0.f );
			if ( (*it)->posCount() > maxOppPosCount )
				continue;
			if ( (*it)->velValid() )
				hisVel = (*it)->vel();
			fast.players().push_back( FastPlayer( (*it)->pos() , hisVel , 0 , (*it)->unum() , false , 1.f , 1.5f , (*it)->playerTypePtr()->id() , NULL ) );
//			dlog.addText ( Logger::PASS , "player %d in ( %f , %f ) with count %d" , (*it)->unum() , (*it)->pos().x , (*it)->pos().y , (*it)->posCount() );
		}
		fast.setBall( wm.ball().pos() , Vector2D(0.f , 0.f) , 0);
		fast.calculate( wm.self().unum() );
		if ( fast.getFastestPlayers().size() )
		{
//			dlog.addText ( Logger::PASS , "fastetOpp is number %d reaching in %d cycle" , fast.getFastestPlayers()[0].unum , fast.getFastestPlayers()[0].reachCycle );
			maxRunCycles = max( min ( maxPossibleRunCycles , fast.getFastestPlayers()[0].reachCycle - 1) , 0 );
		}
	}
//	dlog.addText ( Logger::PASS , "maxBallDelay	%d" , maxRunCycles);
}

void PreDecidedBreaker::setBehindEnemyLines(const WorldModel & wm , float countingRadius)
{
	float xDistToChoose = 5.f;
	float maxXDistTooUniteLines = 10.f;
	targetLineX = 0.f;
	opponentLine.clear();
	behindEnemyLines.assign();
	vector<Vector2D> centers;
	{
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( hisUNum > 0 )
			{
				if ( ( Basics::isInPlayers( hisUNum , defenceBreakers ) ) and ( hisUNum != wm.self().unum() ) )
				{
//					dlog.addText ( Logger::PASS , "addded Tmm	%d--->(%f,%f)" , hisUNum , hisPos.x , hisPos.y );
					centers.push_back(hisPos);
				}
			}
		}
	}
	centers.push_back( wm.self().pos() );
	vector<Vector2D> candidateEnemies;
	vector<Vector2D> enemies;
	{
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			if ( (*it)->posCount() > 6 )
				continue;
			if ( hisUNum > 0 )
			{
				if ( wm.opponentGoalieUnum() > 0 )
					if ( hisUNum == wm.opponentGoalieUnum() )
					continue;
				bool shouldBeCount = false;
//				dlog.addText ( Logger::PASS , "testing	%d--->(%f,%f)" , hisUNum , hisPos.x , hisPos.y );
				for ( unsigned int i = 0; i < centers.size(); i++ )
				{
					if ( ( hisPos.dist(centers[i]) < countingRadius ) and ( hisPos.x >= wm.ball().pos().x ) and ( hisPos.x >= centers[i].x - 2.f ) )
					{
						shouldBeCount = true;
						break;
					}
				}
				if ( shouldBeCount )
				{
					if ( hisPos.x > targetLineX )
						targetLineX = hisPos.x;
//					dlog.addText ( Logger::PASS , "Added!!!" );
					enemies.push_back(hisPos);
				}
				else if ( hisPos.x >= wm.ball().pos().x )
				{
//					dlog.addText ( Logger::PASS , "To be checked!" );
					candidateEnemies.push_back(hisPos);
				}
			}
		}
	}
//	dlog.addText( Logger::PASS , "candidateEnemies.size()\t%d" , candidateEnemies.size() );
	for ( unsigned int i = 0; i < candidateEnemies.size(); i++ )
	{
		Vector2D candidatePos = candidateEnemies[i];
		unsigned int firstSize = enemies.size();
		for ( unsigned int j = 0; j < firstSize; j++ )
		{
			if ( fabsf( (candidatePos - enemies[j]).x ) < xDistToChoose )
			{
				enemies.push_back(candidatePos);
				if ( candidatePos.x > targetLineX )
					targetLineX = candidatePos.x;
				break;
			}
		}
	}
//	dlog.addText ( Logger::PASS , "targetLineX	%f" , targetLineX );
	vector<Vector2D> enemiesToBeSet;
	for ( unsigned int i = 0; i < enemies.size(); i++ )
	{
		if ( targetLineX - enemies[i].x >= maxXDistTooUniteLines )
			continue;
		enemiesToBeSet.push_back( enemies[i] );
	}
	enemies = enemiesToBeSet;
	{
		bool sortNeeded = true;
		while ( sortNeeded )
		{
			sortNeeded = false;
			for ( unsigned int j = 0; (signed)j < (signed)enemies.size() - 1; j++)
			{
				if ( enemies[j].y > enemies[j+1].y )
				{
					sortNeeded = true;
					Vector2D replace = enemies[j];
					enemies[j] = enemies[j+1];
					enemies[j+1] = replace;
				}
			}
		}
	}
/*	for ( unsigned int i =0; i < enemies.size(); i++ )
	{
//		dlog.addText ( Logger::PASS , "enemy[%d]-->(%f,%f)" , i ,enemies[i].x , enemies[i].y );
	}
*/	if ( enemies.size() )
	{
		Vector2D start = enemies[0];
		start.y = -34.f;
		Vector2D end = enemies[enemies.size() - 1];
		end.y = 34.f;
		behindEnemyLines.addVertex(start);
		opponentLine.push_back(start);
		for ( unsigned int i = 0; i < enemies.size(); i++ )
		{
			behindEnemyLines.addVertex(enemies[i]);
			opponentLine.push_back(enemies[i]);
		}
		behindEnemyLines.addVertex(end);
		opponentLine.push_back(end);
	}
	else
	{
		targetLineX = wm.offsideLineX();
		behindEnemyLines.addVertex ( Vector2D(targetLineX , -34.f) );
		opponentLine.push_back( Vector2D(targetLineX , -34.f) );
		behindEnemyLines.addVertex ( Vector2D(targetLineX , 34.f) );
		opponentLine.push_back( Vector2D(targetLineX , 34.f) );
	}
	behindEnemyLines.addVertex( Vector2D( 52.5 , 34.f ) );
	behindEnemyLines.addVertex( Vector2D( 52.5 , 20.f) );
	behindEnemyLines.addVertex( Vector2D( 36.f , 20.f) );
	behindEnemyLines.addVertex( Vector2D( 36.f , -20.f) );
	behindEnemyLines.addVertex( Vector2D( 52.5 , -20.f) );
	behindEnemyLines.addVertex( Vector2D( 52.5 , -34.f) );
}

float PreDecidedBreaker::getTargetX(float lineY)
{
	unsigned int i = 0;
	while ( ( i < opponentLine.size() ) and ( lineY > opponentLine[i].y ) )
		i++;
	Vector2D lowerPart = opponentLine[i - 1];
	Vector2D upperPart = opponentLine[i];
	if ( upperPart.x == lowerPart.x )
		return upperPart.x;
	Vector2D ltou = upperPart - lowerPart;
	ltou *= ( lineY - lowerPart.y ) / ( ltou.y );
	Vector2D target = lowerPart + ltou;
	return target.x;
}

int PreDecidedBreaker::getMaxDelay( PreDecidedPassPlayer receiver , bool tillTargetLine)
{
	FastIC calculator;
	calculator.players().push_back ( FastPlayer( receiver.mPos , receiver.mVel , 0 , receiver.uniNum , true , 1 , 1 , receiver.playerType , NULL) );
	Vector2D runTarget = receiver.mPos;
	if ( tillTargetLine )
		runTarget.x = targetLineX;
	else
		runTarget.x = getTargetX( receiver.mPos.y );
	calculator.setBall( runTarget , Vector2D ( 0  , 0 ) , 0 );
	calculator.calculate(15);
	if ( calculator.getFastestPlayers().size() )
		return calculator.getFastestPlayers()[0].reachCycle;
	return 100;
}

BreakerResult PreDecidedBreaker::getThisPlayerValue( const PlayerAgent *agent , PreDecidedPassPlayer receiver )
{
	float dirStep = (endAngle - startAngle) / dirCount;
	float bestValue = 0.f;
	const WorldModel & wm = agent->world();
	PreDecidedPassRelease best;
	best.receiverStats.receiver.isTeamMate = false;
	best.velocity = 0.f;
	constructFastIC();
	int c = 0;
	int tmpBallDelay = mBall.delay;
//	dlog.addText ( Logger::PASS , "isInOffside??%d" , receiver.isInOffside );
	for ( float direction = startAngle; direction <= endAngle; direction += dirStep )
	{
		if ( agent->world().dirCount( direction ) > maxDirectionCount )
			continue;
		int max , sum , ave;
		agent->world().dirRangeCount( direction , 30 ,&max,&sum,&ave);
//		dlog.addText( Logger::PASS , "direction\t%f--->dirCount\t%d\tave\t%d\tmax\t%d" , direction , agent->world().dirCount( direction ) , ave , max);
		if ( ave > maxDirectionCount )
			continue;

		float maxSpeed = endRate;
		float maxDirectionSpeed = Body_KickOneStep::get_max_possible_vel( direction , 1000, agent->world().ball().vel() ).r() - 0.1f;
		if ( mBall.delay == 0 )
		{
			maxSpeed = maxDirectionSpeed;
			maxSpeed = min( maxSpeed , endRate);
		}
		if ( maxSpeed < 1.f )
			continue;
		float rateStep = (maxSpeed - startRate)/rateCount;
//		dlog.addText ( Logger::PASS , "direction\t%f\tmaxSpeed\t%f" , direction , maxSpeed );
		for ( float rate = startRate; rate <= maxSpeed; rate += rateStep )
		{
			mBall.delay = tmpBallDelay;
			if ( rate < rateStep / 2.f )
				continue;
			c++;
//			dlog.addText ( Logger::PASS , "checking direction %f vel %f" , direction , rate );
			PreDecidedPassReceiver thisDirVel = getFastestDirectionVelocity( agent , rate , direction , true);
/*			if ( receiver.uniNum == 11 )
			{
				Vector2D releasePoint = getReleasePointDirectionVelocity( agent , rate , direction );
				dlog.addText ( Logger::PASS , "direction %.2f--->rate %.2f player %d-->%d ( isInOffside? %d ) receiving after %d cycles in (%.2f , %.2f) releasing form ( %.2f , %.2f )" , direction , rate ,
					thisDirVel.receiver.isTeamMate , thisDirVel.receiver.uniNum , thisDirVel.receiver.isInOffside , thisDirVel.reachCycle , thisDirVel.receivePoint.x , thisDirVel.receivePoint.y , releasePoint.x , releasePoint.y );
			}
*/			if ( mBall.delay == 0 )
			{
//				dlog.addText ( Logger::PASS , "111" );
				if ( rate > maxDirectionSpeed )
					continue;
			}
			if ( thisDirVel.receiver.uniNum != receiver.uniNum )
				continue;
			if ( thisDirVel.receiver.isTeamMate != receiver.isTeamMate )
				continue;
			if ( thisDirVel.receivePoint.x >= maxX )
				continue;
			if ( fabsf( thisDirVel.receivePoint.y ) >= maxY )
				continue;
			if ( not thisDirVel.receiver.isTeamMate )
				continue;
			if ( thisDirVel.receiver.uniNum < 0 )
				continue;
			if ( thisDirVel.receiver.uniNum == agent->world().self().unum() )
				continue;
			if ( thisDirVel.receiver.mPos.x > thisDirVel.receivePoint.x - 1 )
			{
//				dlog.addText ( Logger::PASS , "22222" );
				continue;
			}
			if ( thisDirVel.receivePoint.dist( agent->world().ball().pos() ) <= minBallMovement )
			{
//				dlog.addText ( Logger::PASS , "LESS than min1!!" );
				continue;
			}
			if ( thisDirVel.receiver.isInOffside )
			{
//				dlog.addText ( Logger::PASS , "is in Offside!!!" );
				continue;
			}
			if ( not Basics::isInPlayers( thisDirVel.receiver.uniNum , defenceBreakers ) )
				continue;
/*			if ( thisDirVel.receivePoint.x <= getTargetX( thisDirVel.receivePoint.y ) + 1.f )
			{
//				dlog.addText ( Logger::PASS , "33333" );
				continue;
			}
*/			if ( thisDirVel.receivePoint.x <= targetLineX )
			{
//				dlog.addText ( Logger::PASS , "Before Target!!!" );
				continue;
			}
			if ( ( thisDirVel.receivePoint.x >= 36.f ) and ( abs( thisDirVel.receivePoint.y ) <= 20.f ) )
				continue;
			if ( fabsf( thisDirVel.receivePoint.y - thisDirVel.receiver.mPos.y ) >= maxReceiverYChange )
			{				
/*				if ( receiver.uniNum == 11 )
				{
					dlog.addText ( Logger::PASS , "receivePoint.y	%f" , thisDirVel.receivePoint.y );
					dlog.addText ( Logger::PASS , "playerPos.y		%f" , thisDirVel.receiver.mPos.y );
					dlog.addText ( Logger::PASS , "YChangeProb" );
				}
*/				continue;
			}
			float thisValue = ( ( ( thisDirVel.receivePoint.x - mBall.mPos.x ) / 105.f ) + ( thisDirVel.receivePoint.dist( Vector2D(52.5 , 0 ) ) / 110.f ) ) / 2.f;
/*			if ( receiver.uniNum == 9 )
			{
//				dlog.addText ( Logger::PASS , "thisValue	%f" , thisValue );
			}
*/			if ( thisValue > bestValue )
			{
/*				if ( receiver.uniNum == 11 )
				{
					dlog.addText ( Logger::PASS , "Changed!!" );
				}				
*/				bestValue = thisValue;
				best.receiverStats = thisDirVel;
				best.velocity = rate;
				best.ballPosBeforeRelease = getReleasePointDirectionVelocity ( agent , rate , direction );
				best.kickTarget = kickTargetPoint[direction][rate];
				best.releaseCycle = wm.time().cycle() + mBall.delay;
				best.decideTime = wm.time().cycle();
				best.validaty = mBall.delay + 1;
				best.sayValidTime = agent->world().time().cycle() + mBall.delay + 1;
			}
		}
//		dlog.addText ( Logger::PASS , "\t\trateCount\t%d" , rateCount );
	}
//	dlog.addText ( Logger::PASS , "count\t%d" , c );
	BreakerResult answer;
	answer.value = bestValue;
	answer.releaseStats = best;
	return answer;
}

void PreDecidedBreaker::advantagePlayer( unsigned int tmmBlock , bool changePos , bool zeroDelay )
{
	if ( tmmBlock >= tmmPlayers.size() )
		return;
//	dlog.addText ( Logger::PASS , "advantaging player %d" , tmmPlayers[tmmBlock].uniNum );
	if ( changePos )
		tmmPlayers[tmmBlock].mPos.x = getTargetX( tmmPlayers[tmmBlock].mPos.y ) - 1.f;
	tmmPlayers[tmmBlock].delay = 2;
	if ( zeroDelay )
		tmmPlayers[tmmBlock].delay = 0;
}

void PreDecidedBreaker::disadvantageDefendingLine(float countingRadius , int myUniNum)
{
	for ( unsigned int i = 0; i < oppPlayers.size(); i++ )
	{
//		dlog.addText ( Logger::PASS , "-------------------------------------------" );
		for ( unsigned int j = 0; j < tmmPlayers.size(); j++ )
		{
			if ( tmmPlayers[j].uniNum < 0 )
				continue;
			if ( not Basics::isInPlayers( tmmPlayers[j].uniNum , defenceBreakers ) )
				continue;
			if ( tmmPlayers[j].uniNum == myUniNum )
				continue;
			if ( (oppPlayers[i].mPos.x > mBall.mPos.x ) and ( oppPlayers[i].mPos.x >= tmmPlayers[j].mPos.x - 2.f ) and ( oppPlayers[i].mPos.dist ( tmmPlayers[j].mPos ) <= countingRadius ) )
			{
//				dlog.addText ( Logger::PASS , "testing opp %d" , oppPlayers[j].uniNum );
				int defenderDelay = tmmPlayers[j].delay + 1;
//				dlog.addText ( Logger::PASS , "tmmPlayers[%d].delay++ %d" , tmmPlayers[j].uniNum , defenderDelay );
				if ( oppPlayers[i].delay < defenderDelay )
				{
//					dlog.addText ( Logger::PASS , "changed!!" );
					oppPlayers[i].delay = defenderDelay;
				}
			}
		}
//		dlog.addText ( Logger::PASS , "opponent %d delay set to %d" , oppPlayers[i].uniNum , oppPlayers[i].delay );
	}
}

void PreDecidedBreaker::disadvantageNearDefences(float countingRadius , rcsc::Vector2D ballPos )
{
	for ( unsigned int i = 0; i < oppPlayers.size(); i++ )
	{
		float hisDist = ballPos.dist ( oppPlayers[i].mPos );
		if ( hisDist <= countingRadius )
		{
//			dlog.addText ( Logger::PASS , "changing player %d" , oppPlayers[i].uniNum );
			int hisDelay = mBall.delay + 1;
			hisDelay = max ( hisDelay , oppPlayers[i].delay );
			oppPlayers[i].delay = hisDelay;
			oppPlayers[i].controlBuff = 1.f;
//			dlog.addText ( Logger::PASS , "opponent %d delay set to %d" , oppPlayers[i].uniNum , oppPlayers[i].delay );
		}
	}
}

void PreDecidedBreaker::makeStaticDefendingLine(float countingRadius , int myUniNum)
{
	unsigned int opponentSize = oppPlayers.size();
	for ( unsigned int i = 0; i < opponentSize; i++ )
	{
		if ( oppPlayers[i].mPos.dist ( mBall.mPos ) <= maxDistanceToTargetLine / 2.f )
			continue;
		PreDecidedPassPlayer sta;
		sta.mPos = oppPlayers[i].mPos;
		sta.mVel = Vector2D( 0.f , 0.f);
		sta.delay = 0;
		sta.uniNum = oppPlayers[i].uniNum;
		sta.isTeamMate = false;
		float control = 4.f;
		float ballDist = oppPlayers[i].mPos.dist ( mBall.mPos );
		if ( ballDist <= maxDistanceToTargetLine / 2.f )
			control *= 1.f / 2.f;
		else if ( ballDist <= maxDistanceToTargetLine )
			control *= 1.f;
		else if ( ballDist <= 2.f * maxDistanceToTargetLine )
			control *= 1.5;
		sta.controlBuff = 3.f;
		sta.maxDash = 0.f;
		sta.isInOffside = false;
		sta.posCount = oppPlayers[i].posCount;
		sta.playerType = oppPlayers[i].playerType;
		oppPlayers.push_back(sta);
	}
}

int PreDecidedBreaker::getFastestOppReachCycle(const rcsc::WorldModel & wm)
{
	FastIC fast;
	const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
	for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
		it != end;
		++it )
	{
		Vector2D hisVel = Vector2D( 0.f , 0.f );
		if ( (*it)->posCount() > maxOppPosCount )
			continue;
		if ( (*it)->velValid() )
			hisVel = (*it)->vel();
		fast.players().push_back( FastPlayer( (*it)->pos() , hisVel , 0 , (*it)->unum() , false , 1.f , 1.5f , (*it)->playerTypePtr()->id() , NULL ) );
	}
	fast.setBall( wm.ball().pos() , Vector2D(0.f , 0.f) , 0);
	fast.calculate( wm.self().unum() );
	if ( fast.getFastestPlayers().size() )
	{
//		dlog.addText ( Logger::PASS , "fastetOpp is number %d reaching in %d cycle" , fast.getFastestPlayers()[0].unum , fast.getFastestPlayers()[0].reachCycle );
		return max( fast.getFastestPlayers()[0].reachCycle - 1 , 0 );
	}
	return 10;
}

float PreDecidedBreaker::getValue(PlayerAgent *agent)
{
//	dlog.addText( Logger::PASS , "Breaker::getValue()" );
	deleteMemory();
	vector< PreDecidedPassPlayer > breakers;
	vector< unsigned int> places;
	vector<int> runnersCycleToReach;
	const WorldModel & wm = agent->world();
	if ( wm.time().cycle() != countSetCycle )
	{
		countSetCycle = wm.time().cycle();
		dirCount = 40.f;
		rateCount = 5.f;
		int kickCount = 0;
		{
			float directionStep = ( endAngle - startAngle ) / dirCount;
			for ( float direction = startAngle; direction < endAngle; direction  += directionStep )
			{
				if ( agent->world().dirCount( direction ) > maxDirectionCount )
					continue;
				int max , sum , ave;
				agent->world().dirRangeCount( direction , 30 ,&max,&sum,&ave);
				if ( ave > maxDirectionCount )
					continue;
				float maxSpeed = endRate;
				if ( mBall.delay == 0 )
				{
					maxSpeed = Body_KickOneStep::get_max_possible_vel( direction , 1000, agent->world().ball().vel() ).r();
					maxSpeed = min( maxSpeed , endRate);
				}
				if ( maxSpeed < 1.f )
					continue;
				float rateStep = ( maxSpeed - startRate ) / rateCount;
				for ( float rate = startRate; rate <= maxSpeed; rate += rateStep )
				{
					if ( rate < rateStep / 2.f )
						continue;
					kickCount++;
				}
			}
			int maxKickCount = 2000;
			sumCount += kickCount;
			numCount++;
//			dlog.addText ( Logger::PASS , "average till now is %d" , ( sumCount / numCount ) );
			if ( kickCount > maxKickCount )
			{
				float zarib = (float)kickCount / (float)maxKickCount;
				dirCount = (float)( (int)(dirCount / zarib));
			}
		}
	}
	setMaxRunCycles( agent );
//	dlog.addText ( Logger::PASS , " max Run Cycles	%d" , maxRunCycles );
	setBehindEnemyLines ( wm , 15.f );
//	dlog.addText ( Logger::PASS , "targetLineX\t%f" , targetLineX );
//	dlog.addText ( Logger::PASS , "Dist To Offside line	%f" , targetLineX - wm.self().pos().x );
	if ( wm.self().pos().x >= targetLineX )
	{
//		dlog.addText ( Logger::PASS , "further!!");
		return 0.f;
	}
	if ( getTargetX(wm.self().pos().y) - wm.self().pos().x >= maxDistanceToTargetLine )
	{
//		dlog.addText ( Logger::PASS , "Too far!!");
		return 0.f;
	}
	if ( getTargetX(wm.self().pos().y) - wm.self().pos().x >= maxDistanceForCallingRun )
		shouldCallRun = false;
	if ( targetLineX >= 36.f )
	{
//		dlog.addText ( Logger::PASS , "In there!!" );
		return 0.f;
	}
	if ( wm.self().pos().x >= 36.f )
	{
//		dlog.addText ( Logger::PASS , "In there me!!");
		return 0.f;
	}
//	dlog.addText ( Logger::PASS , "Down to here!!" );
	int tmpMaxRunningCycles = maxPossibleRunCycles;
	updateByWorldModel( wm );
	for ( unsigned int i = 0; i < tmmPlayers.size(); i++ )
	{
		if ( tmmPlayers[i].uniNum > 0 )
			if ( Basics::isInPlayers( tmmPlayers[i].uniNum , defenceBreakers) )
				if ( ( tmmPlayers[i].uniNum != agent->world().self().unum() ) and (tmmPlayers[i].posCount <= maxTmmPosCount) )
				{
//					dlog.addText ( Logger::PASS , "player %d added ----> posCount %d" , tmmPlayers[i].uniNum , tmmPlayers[i].posCount );
					breakers.push_back( tmmPlayers[i] );
					places.push_back(i);
					int thisPlayerReachCycleToTargetLine = getMaxDelay( tmmPlayers[i] );
					runnersCycleToReach.push_back( thisPlayerReachCycleToTargetLine );
					if ( thisPlayerReachCycleToTargetLine < tmpMaxRunningCycles )
						tmpMaxRunningCycles = thisPlayerReachCycleToTargetLine;
				}
	}
//	dlog.addText ( Logger::PASS , "down to here!!" );
//	dlog.addText ( Logger::PASS , "tmpMaxRunningCycles\t%d" , tmpMaxRunningCycles );
//	dlog.flush();
	vector<BreakerResult> results;
	bool isGoingToRelease = false;
	// Queuing releases!!!
	{
		deleteMemory();
		int maxHoldCycles = min ( getFastestOppReachCycle ( wm ) , 1 );
		maxHoldCycles = min ( tmpMaxRunningCycles , maxHoldCycles );
		maxHoldCycles = 0;
		mBall.delay = maxHoldCycles;
//		dlog.addText ( Logger::PASS , "mBall.delay\t%d" , mBall.delay );
		tmmDelay = 2;
		tmmControlBuff = 0.9;
		tmmMaxDash = 1.f;
		oppDelay = mBall.delay + 1;
		oppControlBuff = 2.f;
		oppMaxDash = 1.f;
		updateByWorldModel( wm );
		for ( unsigned int i = 0; i < places.size() ; i++ )
		{
			advantagePlayer( places[i] , false , false );
		}
		disadvantageDefendingLine( 6.f , wm.self().unum() );
		disadvantageNearDefences( 3.f , mBall.mPos );
//		dlog.addText ( Logger::PASS , "all advantages and disadvantages are done!!" );
//		dlog.flush();
		for ( unsigned int i = 0; i < breakers.size(); i++ )
		{
//			dlog.addText( Logger::PASS , "testing player %d" , breakers[i].uniNum );
			BreakerResult result = getThisPlayerValue( agent , breakers[i] );
			if ( result.value > 0 )
			{
//				dlog.addText ( Logger::PASS , "player %d release queued!!" , breakers[i].uniNum );
				result.releaseStats.isItCourse = true;
//				result.releaseStats.ballPosBeforeRelease = wm.ball().pos();
				results.push_back ( result );
				isGoingToRelease = true;
			}
		}
	}
	bool runSet = false;
	// Queuing run requests
	if ( ( not isGoingToRelease ) and ( shouldCallRun ) and ( maxRunCycles > 2 ) )
	{
		deleteMemory();
		mBall.delay = 0;
		tmmDelay = 1;
		tmmControlBuff = 0.9;
		tmmMaxDash = 0.9f;
		oppDelay = tmmDelay + 2;
		oppControlBuff = 1.f;
		oppMaxDash = 1.f;
		updateByWorldModel( wm );
//		dlog.addText ( Logger::PASS , "targetLineX\t%f" , targetLineX );
		string runnerPlayers = "";
		bool anyThingChanged = false;
		for ( unsigned int i = 0; i < tmmPlayers.size(); i++ )
		{
			if ( tmmPlayers[i].uniNum < 0 )
				continue;
			if ( not Basics::isInPlayers( tmmPlayers[i].uniNum , defenceBreakers ) )
				continue;
			if ( tmmPlayers[i].isInOffside )
				continue;
			if ( tmmPlayers[i].uniNum == wm.self().unum() )
				continue;
			if ( tmmPlayers[i].posCount > maxTmmPosCount )
				continue;
			PreDecidedPassPlayer thisPlayer = tmmPlayers[i];
			Vector2D targetPosition = OffenseVariables::Instance()->getRunnerPosition( getTargetX(thisPlayer.mPos.y) , thisPlayer.mPos );
			FastIC runningPlayerCalculator;
			runningPlayerCalculator.players().push_back(FastPlayer( thisPlayer.mPos , Vector2D( 0.f , 0.f ) , 1 , thisPlayer.uniNum , true , 1.f , 1.f , thisPlayer.playerType , NULL , NULL ) );
			runningPlayerCalculator.setBall( targetPosition , Vector2D( 0.f , 0.f ) , 0 );
			runningPlayerCalculator.calculate( wm.self().unum() , 0 , false );
			if ( not runningPlayerCalculator.getFastestPlayers().size() )
				continue;
			int hisCycleToReach = runningPlayerCalculator.getFastestPlayers()[0].reachCycle;
			int hisMovingCycle = hisCycleToReach;
/*			dlog.addText ( Logger::PASS , "player %d ----> targetPos ( %f , %f )" , thisPlayer.uniNum , targetPosition.x , targetPosition.y );
			dlog.addText ( Logger::PASS , "hisPosition is ( %f , %f ) with count %d" , thisPlayer.mPos.x , thisPlayer.mPos.y , thisPlayer.posCount );
			dlog.addText ( Logger::PASS , "reaches in %d cycles!!" , hisCycleToReach );
*/			if ( hisCycleToReach > maxPossibleRunCycles )
				continue;
			if ( hisCycleToReach > maxRunCycles )
			{
				hisMovingCycle = maxRunCycles - 1;
//				dlog.addText ( Logger::PASS , "hisMovingCycle %d" , hisMovingCycle );
				targetPosition = thisPlayer.mPos + Vector2D::polar2vector( runningPlayerCalculator.getMovingDist( thisPlayer.playerType , hisMovingCycle - 1 ) , ( targetPosition - thisPlayer.mPos ).th() );
			}
			Vector2D hisVel = Vector2D::polar2vector( runningPlayerCalculator.getResultVel( thisPlayer.playerType , hisMovingCycle - 1 ) , ( targetPosition - thisPlayer.mPos ).th() );
//			dlog.addText ( Logger::PASS , "final target ( %f , %f )" , targetPosition.x , targetPosition.y );
//			dlog.addText ( Logger::PASS , "reach  %d ----> resultVel ( %f , %f )" , hisMovingCycle , hisVel.x , hisVel.y );
			tmmPlayers[i].mPos = targetPosition;
			tmmPlayers[i].mVel = hisVel;
			tmmPlayers[i].delay = 0;
			anyThingChanged = true;
			runnerPlayers += AudioCodec::i().unum2hex( thisPlayer.uniNum );
			for ( unsigned int j = 0; j < breakers.size(); j++ )
				if ( breakers[j].uniNum == tmmPlayers[i].uniNum )
					runnersCycleToReach[j] = hisCycleToReach;
			advantagePlayer( i , false , false );
		}
//		dlog.addText ( Logger::PASS , "runnerPlayers %s" , runnerPlayers.c_str() );
		if ( anyThingChanged )
		{
			disadvantageDefendingLine( 15.f , wm.self().unum() );
			disadvantageNearDefences( 5.f , mBall.mPos );
//			dlog.addText ( Logger::PASS , "shouldTest!!" );
			string tmpString = defenceBreakers;
			defenceBreakers = runnerPlayers;
			// Checking
			for ( unsigned int i = 0; i < breakers.size(); i++ )
			{
				if ( not Basics::isInPlayers( breakers[i].uniNum , defenceBreakers ) )
					continue;
				BreakerResult result = getThisPlayerValue ( agent , breakers[i] );
				if ( result.value > 0 )
				{
					result.value -= runnersCycleToReach[i];
					result.releaseStats.releaseCycle = wm.time().cycle() + runnersCycleToReach[i] + 4;
					result.releaseStats.decideTime = wm.time().cycle();
					result.releaseStats.validaty = min ( runnersCycleToReach[i] , maxRunCycles ) - 1;
					result.releaseStats.reDecideTime = wm.time().cycle() + min( runnersCycleToReach[i] , maxRunCycles );
					result.releaseStats.targetLineReachCycle = wm.time().cycle() + runnersCycleToReach[i];
					result.releaseStats.isItCourse = true;
					result.releaseStats.ballPosBeforeRelease = wm.ball().pos();
					results.push_back( result );
//					dlog.addText ( Logger::PASS , "player %d runs valued!!" , breakers[i].uniNum );
					runSet = true;
				}
			}
			defenceBreakers = tmpString;
		}
	}
	if ( results.size() == 0 )
		return 0;
	int bestBlock = 0;
	float maxValue = results[0].value;
	for ( unsigned int i = 0; i < results.size(); i++ )
		if ( results[i].value > maxValue )
		{
			bestBlock = i;
			maxValue = results[i].value;
		}
	if ( runSet )
	{
		maxValue = maxValue - (int)maxValue;
	}
	if ( maxValue )
	{
//		bestPassStats = results[bestBlock].releaseStats;
		tmpStats = results[bestBlock].releaseStats;
//		primaryReceiverUnum = tmpStats.receiverStats.receiver.uniNum;
		setThisCycle = true;
		if ( not runSet )	
			dlog.addText ( Logger::PASS , "[COURSE!!]velocity	%f ------->ReceivePoint	( %.2f , %.2f ) -------> receiver	%d" , 	tmpStats.velocity
					, tmpStats.receiverStats.receivePoint.x , tmpStats.receiverStats.receivePoint.y , tmpStats.receiverStats.receiver.uniNum );
		else if ( shouldCallRun )
			dlog.addText ( Logger::PASS , "calling players for run to reach in cycle %d" , tmpStats.targetLineReachCycle );
	}
	return maxValue;
}

void PreDecidedBreaker::setPlayers(const rcsc::WorldModel & wm)
{
	mBall.mPos = wm.ball().pos();
	// Adding teamMates
	{
		tmmPlayers.clear();
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			PreDecidedPassPlayer teamMate;
			teamMate.mPos = hisPos;
			teamMate.mVel = Vector2D(0.f,0.f);
			if ( (*it)->velValid() )
				teamMate.mVel = (*it)->vel();
			teamMate.delay = max( mBall.delay - 1 , 0);
			if ( mBall.mPos.dist ( hisPos ) <= 3.5 )
				teamMate.delay = mBall.delay + 2;
			teamMate.uniNum = hisUNum;
			teamMate.isTeamMate = true;
			teamMate.controlBuff = tmmControlBuff;
			teamMate.maxDash = tmmMaxDash;
			if ( hisPos.x <= wm.offsideLineX() + 1.f  )
				teamMate.isInOffside = false;
			else
				teamMate.isInOffside = true;
			teamMate.posCount = (*it)->posCount();
			teamMate.playerType = (*it)->playerTypePtr()->id();
//			dlog.addText ( Logger::PASS , "Tmm %d in ( %f , %f ) velocity( %f , %f ) with count %d and delay %d added!" , hisUNum , hisPos.x , hisPos.y , teamMate.mVel.x , teamMate.mVel.y , teamMate.posCount , teamMate.delay);
			tmmPlayers.push_back(teamMate);
		}
	}
	// Adding self
	{
		PreDecidedPassPlayer self;
		self.mPos = wm.self().pos();
		self.mVel = wm.self().vel();
		self.delay = mBall.delay + 2;
		self.uniNum = wm.self().unum();
		self.isTeamMate = true;
		self.controlBuff = tmmControlBuff;
		self.maxDash = tmmMaxDash;
		self.posCount = 0;
		self.playerType = wm.self().playerType().id();
		tmmPlayers.push_back(self);
	}
	// Adding opponents
	{
		oppPlayers.clear();
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();	
			// Making static Samples
			if ( mBall.mPos.dist( hisPos ) >= 5.f )
			{
				int hisUNum = (*it)->unum();
				PreDecidedPassPlayer opponent;
				opponent.mPos = hisPos;
				opponent.delay = 0;
				opponent.mVel = Vector2D(0.f,0.f);
				if ( (*it)->velValid() )
					opponent.mVel = (*it)->vel();
				opponent.isTeamMate = false;
				opponent.controlBuff = 5.f * oppControlBuff;
				float hisDist = hisPos.dist ( mBall.mPos );
				if ( hisDist <= 10.f )
					opponent.controlBuff *= 0.8f;
				else if ( hisDist <= 20.f )
					opponent.controlBuff *= 1;
				else if ( hisDist <= 30.f )
					opponent.controlBuff *= 3.f / 2.f;
				else
					opponent.controlBuff *= 2.f;
				if ( wm.opponentGoalieUnum() > 0 )
					if ( hisUNum == wm.opponentGoalieUnum() )
						opponent.controlBuff *= 2.f;
				opponent.maxDash = 0.f;
				opponent.isInOffside = false;
				opponent.posCount = (*it)->posCount();
				opponent.uniNum = hisUNum + 100;
				opponent.playerType = (*it)->playerTypePtr()->id();
				dlog.addText ( Logger::PASS , "Opp %d in ( %f , %f ) velocity( %f , %f ) with count %d and delay %d added!---->(MaxDash %f),(controlBuff %f)" , opponent.uniNum , hisPos.x , hisPos.y , opponent.mVel.x , opponent.mVel.y , opponent.posCount , opponent.delay , opponent.maxDash , opponent.controlBuff);
				oppPlayers.push_back(opponent);
			}
			// Making late moving samples
			{
				int hisUNum = (*it)->unum();
				PreDecidedPassPlayer opponent;
				opponent.mPos = hisPos;
				opponent.delay = mBall.delay + lateMovingDelay;
				if ( mBall.mPos.dist( hisPos ) >= 5.f )
					opponent.delay = mBall.delay + 3;
				opponent.mVel =  Vector2D(0.f,0.f);
				opponent.isTeamMate = false;
				opponent.controlBuff = oppControlBuff;
				if ( wm.opponentGoalieUnum() > 0 )
					if ( hisUNum == wm.opponentGoalieUnum() )
						opponent.controlBuff *= 2.f;
				opponent.isInOffside = false;
				opponent.posCount = (*it)->posCount();
				opponent.uniNum = hisUNum;
				opponent.playerType = (*it)->playerTypePtr()->id();
//				dlog.addText ( Logger::PASS , "Opp %d in ( %f , %f ) velocity( %f , %f ) with count %d and delay %d added!---->(MaxDash %f),(controlBuff %f)" , opponent.uniNum , hisPos.x , hisPos.y , opponent.mVel.x , opponent.mVel.y , opponent.posCount , opponent.delay , opponent.maxDash , opponent.controlBuff);
				oppPlayers.push_back(opponent);
			}
		}
	}
}
