// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <PreDecidedPass.h>

#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>
#include <rcsc/action/intention_kick.h>
#include <rcsc/action/body_stop_ball.h>
#include <rcsc/action/body_turn_to_point.h>

#include <rcsc/player/interception.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include <FastIC.h>
#include <OffenseVariables.h>

#define MaxX 52
#define MaxY 30

using namespace rcsc;
using namespace std;

int PreDecidedPass::sumCount = 0;
int PreDecidedPass::numCount = 0;

float PreDecidedPass::dirCount = 40;
float PreDecidedPass::rateCount = 3;
int PreDecidedPass::countSetCycle = -1;

PreDecidedPass::PreDecidedPass() : PreDecidedPassBase()
{	
	insidePolygonPass = false;
	mTargetPolygon.assign();
	
	distToGoalWeight = 0;
	checkDistToGoal = false;
	maxDistanceToGoal = 110.f;
	
	forwardMovementWeight = 0;
	checkForwardMovement = false;
	maxForwardMovement = 105.f;

	receiverMovementWeight = 0;
	checkReceiverMovement = false;
	maxReceiverMovement = 110.f;
	
	receiverAbilityWeight = 0;
	checkReceiverAbility = false;
	maxOpponents = 11;
	countingRadius = 5.f;
	
	minBallMovement = 2.5f;
	maxDistanceForStraightPass = 3.f;
	minReceiverDist = 2.f;
	
	startRate = 0.f;
	endRate = 2.5;

	maxHoldBallCycles = 1;
}

PreDecidedPass::~PreDecidedPass()
{
}


Polygon2D & PreDecidedPass::targetPolygon()
{
	return mTargetPolygon;
}

void PreDecidedPass::setInsidePolygonPass(bool b)
{
	insidePolygonPass = b;
}

void PreDecidedPass::setDistToGoalParameters(int weight,float maxDist,bool checkIt)
{
	distToGoalWeight = weight;
	checkDistToGoal = ( checkIt or weight );
	maxDistanceToGoal = maxDist;
}

void PreDecidedPass::setForwardMovementParameters(int weight,float maxMove,bool checkIt)
{
	forwardMovementWeight = weight;
	checkForwardMovement = ( checkIt or weight );
	maxForwardMovement = maxMove;
}

void PreDecidedPass::setReceiverMovementParameters(int weight,float maxMove,bool checkIt)
{
	receiverMovementWeight = weight;
	checkReceiverMovement = ( checkIt or weight );
	maxReceiverMovement = maxMove;
}

void PreDecidedPass::setReceiverAbilityParameters(int weight,int max, bool checkIt)
{
	receiverAbilityWeight = weight;
	checkReceiverAbility = (checkIt or weight);
	maxOpponents = max;
}

void PreDecidedPass::setOpenPlayers(string players)
{
//	dlog.addText ( Logger::PASS , "Players	%s" , players.c_str() );
	for ( int i = 0; i < 11; i++ )
		isOpen[i] = false;
	for ( uint i = 0; i < players.size(); i++)
	{
		if ( players[i] == 'A' )
			isOpen[9] = true;
		else if ( players[i] == 'B' )
			isOpen[10] = true;
		else
			isOpen[players[i] - '1' ] = true;
	}
}

void PreDecidedPass::changeBallDelay(rcsc::PlayerAgent *agent)
{
	const WorldModel & wm = agent->world();
	if ( not wm.self().isKickable() )
	{
		mBall.delay = wm.interceptTable()->selfReachCycle() - wm.time().cycle() + 1;
	}
	else
	{
		mBall.delay = maxHoldBallCycles;
		FastIC fast;
		for ( unsigned int i = 0; i < oppPlayers.size(); i++ )
			if ( oppPlayers[i].posCount <= maxOppPosCount )
			{
//				dlog.addText( Logger::PASS , "player %d in ( %f , %f ) added!!!" , oppPlayers[i].uniNum , oppPlayers[i].mPos.x , oppPlayers[i].mPos.y );
				fast.players().push_back( FastPlayer( oppPlayers[i].mPos , oppPlayers[i].mVel , 0 , oppPlayers[i].uniNum , oppPlayers[i].isTeamMate , 1 , 2 , oppPlayers[i].playerType , &(oppPlayers[i]) ) );
			}
		fast.setBall( mBall.mPos , Vector2D(0.f , 0.f) , 0);
		fast.calculate( wm.self().unum() );
//		dlog.addText ( Logger::PASS , "fastest.size()\t%d" , fast.getFastestPlayers().size() );
		if ( fast.getFastestPlayers().size() )
		{
//			dlog.addText ( Logger::PASS , "fastestPlayer\t%d---->reachCycle\t%d" , fast.getFastestPlayers()[0].unum , fast.getFastestPlayers()[0].reachCycle );
			mBall.delay = min ( maxHoldBallCycles , fast.getFastestPlayers()[0].reachCycle - 1);
		}
		if ( mBall.delay < 0 )
			mBall.delay = 0;
	}
//	dlog.addText ( Logger::PASS , "ballDelay	%d" , mBall.delay);
}

float PreDecidedPass::getValue( PlayerAgent* agent )
{
	bool toffed = false;
	changeBallDelay( agent );
	updateByWorldModel( agent->world() );
	targetX = min ( agent->world().ball().pos().x + OffenseVariables::Instance()->ballXAddition() , agent->world().offsideLineX() );
	maxDist = 3.f;
	maxSituationCount = 1;
	receiversToEffect = OffenseVariables::Instance()->linearPlayers();

	constructFastIC();
	if ( agent->world().time().cycle() != countSetCycle )
	{
		dirCount = 40.f;
		rateCount = 3.f;
		int kickCount = 0;
		{
			float directionStep = ( endAngle - startAngle ) / dirCount;
			for ( float direction = startAngle; direction < endAngle; direction  += directionStep )
			{
				if ( agent->world().dirCount( direction ) > maxDirectionCount )
					continue;
				int max , sum , ave;
				agent->world().dirRangeCount( direction , 30 ,&max,&sum,&ave);
				if ( ave > maxDirectionCount )
					continue;
				float maxSpeed = endRate;
				if ( mBall.delay == 0 )
				{
					maxSpeed = Body_KickOneStep::get_max_possible_vel( direction , 1000, agent->world().ball().vel() ).r();
					maxSpeed = min( maxSpeed , endRate);
				}
				if ( maxSpeed < 1.f )
					continue;
				float rateStep = ( maxSpeed - startRate ) / rateCount;
				for ( float rate = startRate; rate <= maxSpeed; rate += rateStep )
				{
					if ( rate < rateStep / 2.f )
						continue;
					kickCount++;
				}
			}
			int maxKickCount = 80;
			sumCount += kickCount;
			numCount++;
//			dlog.addText ( Logger::PASS , "pass average till now is %d" , ( sumCount / numCount ) );
			if ( kickCount > maxKickCount )
			{
				float zarib = (float)kickCount / (float)maxKickCount;
				dirCount = (float)( (int)(dirCount / zarib));
			}		
		}
		countSetCycle = agent->world().time().cycle();
	}
	float bestValue = 0.f;
	float bestD = 0.f, bestV = 0.f;
	float directionStep = ( endAngle - startAngle ) / dirCount;
	const WorldModel & wm = agent->world();
	int c = 0;
	int tmpBallDelay = mBall.delay;
	dlog.addText( Logger::PASS , "mBall.delay\t%d" , mBall.delay );
	for ( float direction = startAngle; direction <= endAngle; direction += directionStep )
	{
//		dlog.addText( Logger::PASS , "direction\t%f--->dirCount\t%d" , direction , wm.dirCount(direction));
//		dlog.flush();
		if ( wm.dirCount( direction ) > maxDirectionCount )
			continue;
		float deltaAngle = 30.0;
		int sum,max,ave;
		wm.dirRangeCount( direction , deltaAngle ,&max,&sum,&ave);
//		dlog.addText( Logger::PASS , "direction\t%f--->ave\t%d" , direction , ave);
		if ( ave > maxDirectionCount )
			continue;
		float maxSpeed = endRate;
		Vector2D ballVel = mBall.mVel;
		if ( mBall.delay != 0 )
			ballVel = Vector2D( 0.f , 0.f );
//		dlog.addText ( Logger::PASS , "direction\t%f" , direction );
//		dlog.addText ( Logger::PASS , "ballVel\t%f" , ballVel.r() );
//		TODO: change max speed estimation.
		float directionMaxSpeed = Body_KickOneStep::get_max_possible_vel( direction , wm.self().kickRate() , ballVel).r() - 0.1f;
		if ( mBall.delay == 0 )
			maxSpeed = directionMaxSpeed;
//		dlog.addText ( Logger::PASS , "\t%f---->maxSpeed\t%f" , direction , maxSpeed );
		maxSpeed = min( maxSpeed , endRate);
//		dlog.addText ( Logger::PASS , "\tchangedMaxSpeed\t%f" , maxSpeed );
		if ( maxSpeed < 1.0 )
			continue;
		float rateStep = (maxSpeed - startRate)/rateCount;
		for ( float rate = startRate; rate <= maxSpeed; rate += rateStep )
		{
			mBall.delay = tmpBallDelay;
			if ( rate < rateStep / 2.f )
				continue;
			c++;
//			dlog.addText ( Logger::PASS , "direction	%f------->rate	%f" , direction , rate );
//			dlog.flush();
			PreDecidedPassReceiver thisDirVel = getFastestDirectionVelocity( agent , rate , direction );
			Vector2D receivePoint = thisDirVel.receivePoint;
/*			if ( ( direction > -125 ) and ( direction < -40 ) )
				dlog.addText ( Logger::PASS , "\t\tDirection	%f--->Rate	%f	--%d->%d---> receivePoint(%f,%f) after %d cycles" , direction , rate , thisDirVel.receiver.isTeamMate , thisDirVel.receiver.uniNum ,receivePoint.x , receivePoint.y , thisDirVel.reachCycle );
*///			dlog.addText ( Logger::PASS , "here!!!!!!!!!");
//			dlog.flush();
			if ( mBall.delay == 0 )
				if ( rate > directionMaxSpeed )
					continue;
			if ( thisDirVel.receiver.uniNum == wm.self().unum() )
			{
//				dlog.addText ( Logger::PASS , "self!!" );
				continue;
			}
			if ( thisDirVel.receiver.uniNum < 0 )
			{
//				dlog.addText ( Logger::PASS , "half!!!" );
				continue;
			}
			if ( not thisDirVel.receiver.isTeamMate )
			{
//				dlog.addText ( Logger::PASS , "notTeamMate!!" );
				continue;
			}
			if ( thisDirVel.receiver.isInOffside )
			{
//				dlog.addText ( Logger::PASS , "inOffside!!" );
				continue;
			}
			if ( fabsf( receivePoint.x ) > MaxX )
			{
//				dlog.addText ( Logger::PASS , "OUTBACK!!" );
				continue;
			}
			if ( fabsf( receivePoint.y ) > MaxY )
			{
//				dlog.addText ( Logger::PASS , "OUTSIDE!!" );
				continue;
			}
			if ( ( insidePolygonPass ) and ( not mTargetPolygon.contains( receivePoint ) ) )
			{
//				dlog.addText ( Logger::PASS , "! in polygon!!" );
				continue;
			}
			if ( not isOpen[ thisDirVel.receiver.uniNum - 1 ] )
			{
//				dlog.addText( Logger::PASS , "Not Open!!" );
				continue;
			}

			float receiverDistanceToMe = thisDirVel.receiver.mPos.dist( mBall.mPos );
			if ( receiverDistanceToMe < minReceiverDist )
			{
//				dlog.addText ( Logger::PASS , "skipped near receiver!!" );
				continue;
			}
			
			float ballMovement = receivePoint.dist(mBall.mPos);
//			dlog.addText ( Logger::PASS , "ballMovement	%f" , ballMovement );
			if ( ballMovement < 5.f )
				if ( rate > 1.5f )
				{
//					dlog.addText ( Logger::PASS , " 5----1.5 " );
					continue;
				}
			if ( ballMovement < 10.f )
				if ( rate > 2.f )
				{
//					dlog.addText ( Logger::PASS , " 10----2 " );
					continue;
				}
			if ( ballMovement < minBallMovement )
			{
//				dlog.addText ( Logger::PASS , "minMovement!!") ;
				continue;
			}
			int weightSum = distToGoalWeight + forwardMovementWeight + receiverMovementWeight + receiverAbilityWeight;
			float forwardMovement = receivePoint.x - wm.ball().pos().x;
			if ( ( ( forwardMovement < 0 ) or ( forwardMovement > maxForwardMovement ) ) and ( checkForwardMovement ) )
			{
//				dlog.addText ( Logger::PASS , "FORWARD PROBLEM!!!");
				continue;
			}
			float distToGoal = receivePoint.dist( Vector2D(52.5,0) );
			if ( ( distToGoal > maxDistanceToGoal ) and ( checkDistToGoal ) )
			{
//				dlog.addText ( Logger::PASS , "DISTTOGOAL PROBLEM!!!");
				continue;
			}
			float playerMovement = ( receivePoint - thisDirVel.receiver.mPos).r();
			if ( ( playerMovement > maxReceiverMovement ) and ( checkReceiverMovement ) )
			{
//				dlog.addText ( Logger::PASS , "PlayerMovement PROBLEM!!!");
				continue;
			}
			int oppsNear = 0;
			const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
			for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
				it != end;
				++it )
			{
				Vector2D hisPos = (*it)->pos();
				int hisUNum = (*it)->unum();
				if ( ( hisUNum > 0 ) or ( checkHalfs ) )
				{
					if ( receivePoint.dist(hisPos) <= countingRadius )
						oppsNear++;
				}
			}
			if ( ( oppsNear > maxOpponents ) and ( checkReceiverAbility ) )
			{
//				dlog.addText ( Logger::PASS , "MAX Opponents Problem");
				continue;
			}
			float curVal = 1.f;
			if ( weightSum > 0 )
			{
				curVal = ( forwardMovementWeight * ( forwardMovement / maxForwardMovement ) )
							+ ( distToGoalWeight * ( ( maxDistanceToGoal - distToGoal ) / maxDistanceToGoal ) )
							+ ( receiverMovementWeight * ( ( maxReceiverMovement - playerMovement ) / maxReceiverMovement ) )
							+ ( receiverAbilityWeight * ( ( maxOpponents - oppsNear ) / maxOpponents ) );
				curVal /= weightSum;
			}
			if ( curVal > bestValue )
			{
				toffed = false;
				// Distance toff
/*				{
					if ( thisDirVel.receivePoint.dist( thisDirVel.receiver.mPos ) <= maxDistanceForStraightPass )
					{
						thisDirVel.receivePoint = thisDirVel.receiver.mPos;
//						dlog.addText ( Logger::PASS , "toffed!!" );
						toffed = true;
					}
				}
*/				tmpStats.velocity = rate;
				tmpStats.releaseCycle = agent->world().time().cycle() + mBall.delay;
				tmpStats.validaty = mBall.delay + 1;
				tmpStats.decideTime = agent->world().time().cycle();
				tmpStats.receiverStats = thisDirVel;
				tmpStats.sayValidTime = agent->world().time().cycle() + mBall.delay + 1;
				tmpStats.ballPosBeforeRelease = getReleasePointDirectionVelocity ( agent , rate , direction );
				tmpStats.kickTarget = kickTargetPoint[direction][rate];
				setThisCycle = true;
				bestValue = curVal;
				bestD = direction;
				bestV = rate;
//				dlog.addText ( Logger::PASS , "CHANGED!!!" );
			}
//			dlog.addText ( Logger::PASS , "Toffed	?	%d" , toffed );
		}
	}
	dlog.addText ( Logger::PASS , "Pass	%.2f------>%.2f	player %d receiving	in (%.2f , %.2f) releasing in (%.2f , %.2f )--->Value	%.2f" , bestD , bestV , tmpStats.receiverStats.receiver.uniNum ,
				tmpStats.receiverStats.receivePoint.x , tmpStats.receiverStats.receivePoint.y , tmpStats.ballPosBeforeRelease.x , tmpStats.ballPosBeforeRelease.y , bestValue);
	dlog.addText ( Logger::PASS , "Count	= %d" , c );
	isReleasePointCalculated[bestD][bestV] = false;
	getReleasePointDirectionVelocity( agent , bestV , bestD);
	return bestValue;
}
