// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <PreDecidedPassBase.h>
#include <Basics.h>

#include <math.h>

#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>
#include <rcsc/action/intention_kick.h>
#include <rcsc/action/body_stop_ball.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_scan_field.h>

#include <rcsc/player/interception.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/common/basic_client.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include <body_hold_ball2008.h>

#include <OffenseVariables.h>

#include <FastIC.h>

using namespace std;
using namespace rcsc;

PreDecidedPassRelease PreDecidedPassBase::bestPassStats;
int PreDecidedPassBase::primaryReceiverUnum = -1;

PreDecidedPassBase::PreDecidedPassBase() : fastIC()
{
	tmmDelay = 2;
	tmmControlBuff = 0.9;
	tmmMaxDash = 0.85;
	oppDelay = 1;
	oppAdditionalDelay = 0;
	oppControlBuff = 0.9;
	oppMaxDash = 1;

	maxTmmPosCount = 3;
	maxOppPosCount = 20;

	targetX = 0;
	maxDist = 52.5;
	maxSituationCount = 1;
	receiversToEffect = "";

	startAngle = -180;
	endAngle = 180;
	startRate = 0.5;
	endRate = 3.f;

	maxDirectionCount = 5;

	checkHalfs = true;

	mBall.delay = 2;
	mBall.mPos = Vector2D ( -100 , 100 );
	
	setThisCycle = false;

	for (map<float, map<float, bool> >::iterator it = isChecked.begin();
		it != isChecked.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}

	for (map<float, map<float, bool> >::iterator it = isReleasePointCalculated.begin();
		it != isReleasePointCalculated.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}
}

PreDecidedPassBase::~PreDecidedPassBase()
{
}

void PreDecidedPassBase::reset()
{
	primaryReceiverUnum = bestPassStats.reDecideTime = -5;
	for (map<float, map<float, bool> >::iterator it = isChecked.begin();
		it != isChecked.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}

	for (map<float, map<float, bool> >::iterator it = isReleasePointCalculated.begin();
		it != isReleasePointCalculated.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}
}

void PreDecidedPassBase::checkReset( const PlayerAgent * agent )
{
	const rcsc::WorldModel & wm = agent->world();
	if ( wm.time().cycle() >= bestPassStats.reDecideTime )
		return;
	if ( not wm.self().isKickable() )
		return;
	updateByWorldModel( wm );
	for ( unsigned int i = 0; i < oppPlayers.size(); i++ )
		if ( oppPlayers[i].posCount <= 5 )
		{
			dlog.addText ( Logger::PASS , "checking opp %d in ( %.2f , %.2f )" , oppPlayers[i].uniNum , oppPlayers[i].mPos.x , oppPlayers[i].mPos.y );
			float hisDist = wm.ball().pos().dist( oppPlayers[i].mPos );
			if ( hisDist < 1.5f )
			{
				dlog.addText ( Logger::PASS , "CHECK----->RESET!!" );
				reset();
			}
		}
}

void PreDecidedPassBase::deleteMemory()
{
	for (map<float, map<float, bool> >::iterator it = isChecked.begin();
		it != isChecked.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}

	for (map<float, map<float, bool> >::iterator it = isReleasePointCalculated.begin();
		it != isReleasePointCalculated.end(); it++)
	{
		for (map<float, bool>::iterator jit = it->second.begin(); jit != it->second.end(); jit++)
		{
			jit->second = false;
		}
	}
}

// Calculatin' Core
bool PreDecidedPassBase::canDoStepsKick( const PlayerAgent *agent , float rate , float direction , int delay , Vector2D & releasePoint )
{
	Vector2D answer;
	Vector2D targetPoint = agent->world().self().pos() + Vector2D::polar2vector( 50 , direction );
	if ( delay == 2 )
	{
		answer = Body_KickMultiStep( targetPoint , rate ).ballReleasePos( agent );
	}
	else if ( delay == 1 )
	{
		answer = Body_KickTwoStep( targetPoint , rate ).ballReleasePos( agent );
	}
	else
		answer = mBall.mPos;
	releasePoint = answer;
	if ( answer == Vector2D::INVALIDATED )
		return false;
	return true;
}

Vector2D PreDecidedPassBase::getReleasePointDirectionVelocity( const rcsc::PlayerAgent *agent , float rate , float direction)
{
	if ( not isReleasePointCalculated[direction][rate] )
	{
		isReleasePointCalculated[direction][rate] = true;
		int delay = mBall.delay;
		while ( delay >= 0 )
		{
			Vector2D release;
			bool canKick = canDoStepsKick( agent , rate , direction , delay , release );
			if ( canKick )
			{
				releasePoint[direction][rate] = release;
				releaseDelay[direction][rate] = delay;
				break;
			}
			delay--;
		}
/*		if ( delay != mBall.delay )
			dlog.addText ( Logger::PASS , "%.2f --> %.2f changed!!" , direction , rate );
*/		if ( delay < 0 )
		{
			releasePoint[direction][rate] = Vector2D::INVALIDATED;
			releaseDelay[direction][rate] = 1000;
		}
	}
	return releasePoint[direction][rate];
}

PreDecidedPassReceiver PreDecidedPassBase::getFastestDirectionVelocity(const rcsc::PlayerAgent *agent , float rate , float direction , bool ignoreTheRest )
{
	if ( isChecked[direction][rate] )// Skipping dynamically!!!!!!!!
	{
//		dlog.addText ( Logger::PASS , "dynamic!!" );
		return checkedStats[direction][rate];
	}
	// Not decided yet
//	dlog.addText ( Logger::PASS ,  "delay = %d " , mBall.delay );
	Vector2D ballPos = getReleasePointDirectionVelocity( agent , rate , direction );
//		dlog.addText ( Logger::PASS , "ballPos = ( %.2f , %.2f )" , ballPos.x , ballPos.y );
/*	if ( ( direction > -74 ) and ( direction < -70 ) )
	{
		dlog.addText ( Logger::PASS , "\t\tballDelay\t%d" , mBall.delay );
		dlog.addText ( Logger::PASS , "\t\tballPos( %.2f , %.2f )" , ballPos.x , ballPos.y );
		dlog.addText ( Logger::PASS , "\t\tballVel( %.2f , %.2f )" , ballVel.x , ballVel.y );
	}
*/	
	PreDecidedPassPlayer receiverPlayer;
	receiverPlayer.isTeamMate = false;
	PreDecidedPassReceiver answer;
	answer.receiver = receiverPlayer;
	if ( releaseDelay[direction][rate] > 100 )
	{
		isChecked[direction][rate] = true;
		checkedStats[direction][rate] = answer;
//		dlog.addText ( Logger::PASS , "Fault!!" );
		return answer;
	}
	mBall.delay = releaseDelay[direction][rate];
	Vector2D ballVel = mBall.mPos + Vector2D::polar2vector(50 , direction) - ballPos;
	ballVel.setLength( rate );
	fastIC.setBall( ballPos , ballVel , mBall.delay);
	fastIC.calculate(agent->world().self().unum() , 0 , true );
	vector<FastestPlayer> receivers = fastIC.getFastestPlayers();
//	if ( ( direction >= 70 ) and ( direction <= 75 ) )
//		dlog.addText(Logger::PASS , "receivers.size()\t%d" , receivers.size() );
	if ( receivers.size() == 0 ) 
	{
		receiverPlayer.isTeamMate = false;
		answer.receiver = receiverPlayer;
	}
	else if ( ( receivers.size() == 1 ) or ( ignoreTheRest ) )
	{
		receiverPlayer = *receivers[0].playerItCameFrom;
		answer.receiver = receiverPlayer;
		answer.receivePoint = receivers[0].receivePoint;
		answer.reachCycle = receivers[0].reachCycle;
		answer.realReachCycle = agent->world().time().cycle() + answer.reachCycle;
	}
	else
	{
		int opponentPlayer = 0;
		for ( unsigned int i = 0; i < receivers.size(); i++ )
		{
			if ( receivers[i].reachCycle != receivers[0].reachCycle )
				break;
			if ( not receivers[i].isTeamMate )
			{
				opponentPlayer = i;
				break;
			}
		}
		receiverPlayer = *receivers[opponentPlayer].playerItCameFrom;
		answer.receiver = receiverPlayer;
		answer.receivePoint = receivers[0].receivePoint;
		answer.reachCycle = receivers[0].reachCycle;
		answer.realReachCycle = agent->world().time().cycle() + answer.reachCycle;
	}

	isChecked[direction][rate] = true;
	checkedStats[direction][rate] = answer;
//	if ( ( direction >= 70 ) and ( direction <= 75 ) )
//		dlog.addText ( Logger::PASS , "player %d->%d receiving in %d cycles in ( %f , %f )" , answer.receiver.isTeamMate , answer.receiver.uniNum , answer.reachCycle , answer.receivePoint.x , answer.receivePoint.y );
	return answer;
}

float PreDecidedPassBase::getValue( PlayerAgent *agent )
{
	return 0.f;
}

bool PreDecidedPassBase::shouldDecide( PlayerAgent *agent )
{
	return ( bestPassStats.decideTime + bestPassStats.validaty < agent->world().time().cycle() );
}

bool PreDecidedPassBase::mustDecide(rcsc::PlayerAgent *agent)
{
	return ( bestPassStats.reDecideTime >= agent->world().time().cycle() );
}

// Executing Functions
bool PreDecidedPassBase::shouldExecute(rcsc::PlayerAgent *agent)
{
	bool b = ( ( bestPassStats.releaseCycle >= agent->world().time().cycle() ) and ( not shouldDecide(agent) ) and ( agent->world().self().isKickable() ) );
	return b;
}

void PreDecidedPassBase::execute(rcsc::PlayerAgent *agent)
{
	const WorldModel & wm = agent->world();
	dlog.addText ( Logger::PASS , "setThisCycle	%d" , setThisCycle );
	dlog.addText ( Logger::PASS , "tmCheck	%d" , tmpStats.receiverStats.receiver.uniNum );
	if ( setThisCycle )
	{
		bestPassStats = tmpStats;
		setThisCycle = false;
	}
	int diffrence = bestPassStats.releaseCycle - wm.time().cycle();
	dlog.addText ( Logger::PASS , "diffrence	%d" , diffrence );
	dlog.addText ( Logger::PASS , "releaseCycle	%d" , bestPassStats.releaseCycle );

//	bestPassStats.velocity *= 0.6;
	if ( wm.time().cycle() < bestPassStats.reDecideTime	 )
	{
		agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
		if ( primaryReceiverUnum > 0 )
		{
			const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
			for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
				it != end;
				++it )
			{
				int hisUnum = (*it)->unum();
				if ( hisUnum == primaryReceiverUnum )
				{
					dlog.addText ( Logger::PASS , "looking to %d " , (*it)->unum() );
					agent->setNeckAction( new rcsc::Neck_TurnToBallAndPlayer ( *it , 0 ) );
					break;
				}
			}
		}
		Body_HoldBall2008( true , Vector2D( 52.5 , 0.0 ) , Vector2D( 52.5 , 0.0 ) ).execute(agent);
	}
	else
	{
		agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
		Vector2D targetPoint = bestPassStats.receiverStats.receivePoint;
		Vector2D turnTarget = wm.self().pos();
		turnTarget.x = 60.f;
		dlog.addText ( Logger::PASS , "velocity %f" , bestPassStats.velocity );
		dlog.addText ( Logger::PASS , "targetPoint ( %.2f , %.2f )" , targetPoint.x , targetPoint.y );
		dlog.addText ( Logger::PASS , "ballVel ( %.2f , %.2f )" , wm.ball().vel().x , wm.ball().vel().y );
		if ( diffrence > 2 )
		{
			Body_HoldBall2008( true , Vector2D( 52.5 , 0.0 ) , Vector2D( 52.5 , 0.0 ) ).execute(agent);
		}
		else if ( diffrence == 2 )
		{
			Body_KickMultiStep( targetPoint, bestPassStats.velocity ).completeKickExecute( agent );
		}
		else if ( diffrence == 1 )
		{
			Body_KickTwoStep( targetPoint, bestPassStats.velocity ).completeKickExecute( agent );
		}
		else
		{
			dlog.addText ( Logger::PASS , "In HERE!!!" );
			Body_KickOneStep( targetPoint , bestPassStats.velocity , false ).execute( agent );
			bestPassStats.ballPosBeforeRelease = wm.ball().pos();
		}
	}
	communicate(agent);
	agent->setNeckAction( new rcsc::Neck_ScanField() );
	if ( diffrence == 0 )
		reset();
}

bool PreDecidedPassBase::shouldCommunicate(rcsc::PlayerAgent *agent)
{
	return ( bestPassStats.sayValidTime >= agent->world().time().cycle() );
}

void PreDecidedPassBase::communicate(rcsc::PlayerAgent *agent)
{
	if ( agent->world().time().cycle() < bestPassStats.reDecideTime	 )
	{
		agent->addSayMessage( new HelliBreakerMessage( 13.5 , bestPassStats.targetLineReachCycle ) );
//		agent->addSayMessage(new  WaitRequestMessage() );
	}
	else
	{
		Vector2D ballToTarget = bestPassStats.receiverStats.receivePoint - agent->world().ball().pos();
		Vector2D veloc = Vector2D::polar2vector( bestPassStats.velocity , ballToTarget.th() );
		dlog.addText ( Logger::PASS , "Velocity	%f" , veloc.r() );
//		Vector2D ballPosToSay = agent->world().ball().pos();
//		if ( bestPassStats.releaseCycle < agent->world().time().cycle() )
		Vector2D ballPosToSay = bestPassStats.ballPosBeforeRelease;
		if ( bestPassStats.isItCourse )
			agent->addSayMessage( new HelliCourseMessage( bestPassStats.receiverStats.receiver.uniNum , bestPassStats.releaseCycle , ballPosToSay , veloc ) );
		else
			agent->addSayMessage( new HelliPassMessage( bestPassStats.receiverStats.receiver.uniNum , bestPassStats.releaseCycle , ballPosToSay , veloc ) );
		dlog.addText ( Logger::PASS , "mesageAdded!!" );
	}
}
// Editing Functions
void PreDecidedPassBase::setBallPos( rcsc::Vector2D _pos )
{
	mBall.mPos = _pos;
}

void PreDecidedPassBase::setBallDelay( int _delay )
{
	mBall.delay = _delay;
}

void PreDecidedPassBase::setBall( rcsc::Vector2D _pos , int _delay )
{
	mBall.mPos = _pos;
	mBall.delay = _delay;
}
	
vector<PreDecidedPassPlayer> & PreDecidedPassBase::TmmPlayers()
{
	return tmmPlayers;
}

vector<PreDecidedPassPlayer> & PreDecidedPassBase::OppPlayers()
{
	return oppPlayers;
}
	
void PreDecidedPassBase::setDefaultSafetyFactors(int _tmmDelay , float _tmmControlBuff , float _tmmMaxDash , int _oppDelay , float _oppControlBuff , float _oppMaxDash , int _oppAdditionalDelay )
{
	deleteMemory();
	tmmDelay = _tmmDelay;
	tmmControlBuff = _tmmControlBuff;
	tmmMaxDash = _tmmMaxDash;
	oppDelay = _oppDelay;
	oppAdditionalDelay = _oppAdditionalDelay;
	oppControlBuff = _oppControlBuff;
	oppMaxDash = _oppMaxDash;
}

void PreDecidedPassBase::setRanges(float _startAngle , float _endAngle , float _startRate , float _endRate)
{
	startAngle = _startAngle;
	endAngle = _endAngle;
	startRate = _startRate;
	endRate = _endRate;
}

void PreDecidedPassBase::setHalfsCheck(bool check)
{
	checkHalfs = check;
}

void PreDecidedPassBase::constructFastIC()
{
	fastIC.reconstruct();
	for ( unsigned int i = 0; i < tmmPlayers.size(); i++ )
	{
		int tmpMaxTmmPosCount = maxTmmPosCount;
		if ( tmmPlayers[i].uniNum > 0 )
			if ( Basics::isInPlayers ( tmmPlayers[i].uniNum , receiversToEffect ) )
				if ( fabsf( tmmPlayers[i].mPos.x - targetX ) > maxDist )
				{
					tmpMaxTmmPosCount = maxSituationCount;
				}
		if ( tmmPlayers[i].posCount <= tmpMaxTmmPosCount )
		{
		fastIC.players().push_back( FastPlayer( tmmPlayers[i].mPos , tmmPlayers[i].mVel , tmmPlayers[i].delay , tmmPlayers[i].uniNum , tmmPlayers[i].isTeamMate 
								, tmmPlayers[i].controlBuff , tmmPlayers[i].maxDash , tmmPlayers[i].playerType , &(tmmPlayers[i]) ) );
		}
	}
	for ( unsigned int i = 0; i < oppPlayers.size(); i++ )
	{
		if ( oppPlayers[i].posCount <= maxOppPosCount )
		{
		fastIC.players().push_back( FastPlayer( oppPlayers[i].mPos , oppPlayers[i].mVel , oppPlayers[i].delay , oppPlayers[i].uniNum , oppPlayers[i].isTeamMate 
									, oppPlayers[i].controlBuff , oppPlayers[i].maxDash , oppPlayers[i].playerType , &(oppPlayers[i]) ) );
		}
	}
}

void PreDecidedPassBase::updateByWorldModel(const rcsc::WorldModel & wm  , float offsideThr)
{
	mBall.mPos = wm.ball().pos();
	mBall.mVel = wm.ball().vel();
	// Adding TeamMates
	float offsideLine = wm.offsideLineX() - offsideThr;
	dlog.addText ( Logger::PASS , "realOffsideLine\t%f , offsideThr\t%f ====> OffsideLine\t%f" , wm.offsideLineX() , offsideThr , offsideLine );
	{
		tmmPlayers.clear();
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			PreDecidedPassPlayer teamMate;
			teamMate.mPos = hisPos;
			teamMate.mVel = Vector2D(0.f,0.f);
			if ( (*it)->velValid() )
				teamMate.mVel = (*it)->vel();
			teamMate.delay = tmmDelay;
			if ( mBall.mPos.dist ( hisPos ) <= 3.5 )
				teamMate.delay = max( mBall.delay + 2 , tmmDelay );
//				teamMate.delay = mBall.delay + 2;
			if ( hisUNum > 0 )
				if ( not Basics::isInPlayers ( hisUNum , OffenseVariables::Instance()->linearPlayers() + OffenseVariables::Instance()->quarterBackPlayer() ) )
					teamMate.delay = max( mBall.delay + 2 , tmmDelay + 2 );
			teamMate.uniNum = hisUNum;
			teamMate.isTeamMate = true;
			teamMate.controlBuff = tmmControlBuff;
			teamMate.maxDash = tmmMaxDash;
			if ( hisPos.x <= offsideLine  )
				teamMate.isInOffside = false;
			else
				teamMate.isInOffside = true;
			teamMate.posCount = (*it)->posCount();
			teamMate.playerType = (*it)->playerTypePtr()->id();
//			dlog.addText ( Logger::PASS , "Tmm %d in ( %f , %f ) velocity( %f , %f ) ( in offside?? %d ) with count %d and delay %d added!" , hisUNum , hisPos.x , hisPos.y , teamMate.mVel.x , teamMate.mVel.y , teamMate.isInOffside , teamMate.posCount , teamMate.delay);
			tmmPlayers.push_back(teamMate);
		}
	}
	// Adding self
	{
		PreDecidedPassPlayer self;
		self.mPos = wm.self().pos();
		self.mVel = wm.self().vel();
		self.delay = mBall.delay + 2;
		self.uniNum = wm.self().unum();
		self.isTeamMate = true;
		self.controlBuff = tmmControlBuff;
		self.maxDash = tmmMaxDash;
		self.posCount = 0;
		self.playerType = wm.self().playerType().id();
		tmmPlayers.push_back(self);
	}
//	dlog.addText ( Logger::PASS , "(((((((((((((((((()))))))))))))))))))" );
	// Adding Opponents
	{
		int oppoentsDelay = max ( oppDelay , mBall.delay + oppAdditionalDelay );
//		dlog.addText ( Logger::PASS , "oppoentsDelay\t%d" , oppoentsDelay );
		oppPlayers.clear();
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			Vector2D hisPos = (*it)->pos();
			int hisUNum = (*it)->unum();
			PreDecidedPassPlayer opponent;
			opponent.mPos = hisPos;
			opponent.delay = oppoentsDelay;
			opponent.mVel = Vector2D(0.f,0.f);
			if ( (*it)->velValid() )
				opponent.mVel = (*it)->vel();
			if ( mBall.mPos.dist ( hisPos ) <= 2.f )
			{
//				dlog.addText ( Logger::PASS , "In tu!" );
				opponent.delay = max( mBall.delay + 2 , oppoentsDelay );
			}
			opponent.uniNum = hisUNum;
			opponent.isTeamMate = false;
			opponent.controlBuff = oppControlBuff;
			opponent.maxDash = oppMaxDash;
			if ( wm.opponentGoalieUnum() > 0 )
				if ( hisUNum == wm.opponentGoalieUnum() )
				{
					opponent.controlBuff *= 2.f;
					opponent.maxDash = max ( 1.f , oppMaxDash );
				}
			opponent.isInOffside = false;
			opponent.posCount = (*it)->posCount();
			opponent.playerType = (*it)->playerTypePtr()->id();
//			dlog.addText ( Logger::PASS , "Opp %d in ( %f , %f ) velocity( %f , %f ) with count %d and delay %d and controlBuff %f added!" , hisUNum , hisPos.x , hisPos.y , opponent.mVel.x , opponent.mVel.y , opponent.posCount , opponent.delay , opponent.controlBuff );
			oppPlayers.push_back(opponent);
		}
	}
}
