// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

//Caution!!!
/*
	In this class and it's inheritants, the cycle is determined from when the process is started, not from when the ball starts moving!!
	Be carful if you use cycle parameters!
*/

#ifndef _PRE_DECIDED_PASS_BASE_H
#define _PRE_DECIDED_PASS_BASE_H

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/polygon_2d.h>
#include <rcsc/player/player_agent.h>
#include <FastIC.h>

#include <map>
#include <string>

struct PreDecidedPassPlayer
{
	rcsc::Vector2D mPos;
	rcsc::Vector2D mVel;
	int delay;
	int uniNum;
	bool isTeamMate;
	float controlBuff;
	float maxDash;
	bool isInOffside;
	int posCount;
	int playerType;
	PreDecidedPassPlayer()
	{
		mPos = mVel = rcsc::Vector2D(0.f , 0.f);
		delay = uniNum =  posCount = playerType = 0;
		isTeamMate = isInOffside = false;
		controlBuff = maxDash = 1;
	}
};

struct PreDecidedPassBall
{
	rcsc::Vector2D mPos;
	rcsc::Vector2D mVel;
	int delay;
	PreDecidedPassBall()
	{
		mPos = rcsc::Vector2D(0.f , 0.f);
		delay = 0;
	}
};

struct PreDecidedPassReceiver
{
	PreDecidedPassPlayer receiver;
	rcsc::Vector2D receivePoint;
	int reachCycle;
	int realReachCycle;
	PreDecidedPassReceiver() : receiver()
	{
		receivePoint = rcsc::Vector2D ( 0.f , 0.f );
		reachCycle = realReachCycle = -10;
	}
};

struct PreDecidedPassRelease
{
	PreDecidedPassReceiver receiverStats;
	float velocity;
	int releaseCycle;
	int validaty;
	int decideTime;
	int reDecideTime;
	int sayValidTime;
	int targetLineReachCycle;
	rcsc::Vector2D ballPosBeforeRelease;
	rcsc::Vector2D kickTarget;
	bool isItCourse;
	PreDecidedPassRelease() : receiverStats()
	{
		releaseCycle = reDecideTime = validaty = decideTime = reDecideTime = sayValidTime = -10;
		ballPosBeforeRelease = kickTarget = rcsc::Vector2D( 0.f , 0.f );
		targetLineReachCycle = -10;
		isItCourse = false;
	}
};

class PreDecidedPassBase
{
protected:
	std::vector<PreDecidedPassPlayer> tmmPlayers;
	std::vector<PreDecidedPassPlayer> oppPlayers;
	PreDecidedPassBall mBall;

	FastIC fastIC;
	
	// Default Safety Factors
		int tmmDelay;
		float tmmControlBuff;
		float tmmMaxDash;
		int oppDelay;
		int oppAdditionalDelay;
		float oppControlBuff;
		float oppMaxDash;
	// Max PosCounts
		int maxTmmPosCount;
		int maxOppPosCount;
	// Changing things a bit
		float targetX;
		float maxDist;
		int maxSituationCount;
		std::string receiversToEffect;
	// Calculating Ranges
		float startAngle;
		float endAngle;
		float startRate;
		float endRate;
	int maxDirectionCount;

	// Dynamic Parameters
		std::map<float , std::map<float,PreDecidedPassReceiver> > checkedStats;
		std::map<float , std::map<float,bool> > isChecked;
		std::map<float , std::map<float,rcsc::Vector2D > > releasePoint;
		std::map<float , std::map<float,int> > releaseDelay;
		std::map<float , std::map<float,bool> > isReleasePointCalculated;
		std::map<float , std::map<float,rcsc::Vector2D > > kickTargetPoint;

	// Check Halfs Or Not
		bool checkHalfs;

	// Best
		PreDecidedPassRelease tmpStats;
		bool setThisCycle;
		static PreDecidedPassRelease bestPassStats;	
		static int primaryReceiverUnum;

	rcsc::Vector2D getReleasePointDirectionVelocity( const rcsc::PlayerAgent *agent , float rate , float direction);
	PreDecidedPassReceiver getFastestDirectionVelocity(const rcsc::PlayerAgent *agent , float rate , float direction , bool ignoreTheRest = false);
	bool canDoStepsKick( const rcsc::PlayerAgent *agent , float rate , float direction , int delay , rcsc::Vector2D & releasePoint );
	void constructFastIC();

public:
	PreDecidedPassBase();
	virtual ~PreDecidedPassBase();
	
	void setBallPos( rcsc::Vector2D _pos );
	void setBallDelay( int _delay );
	void setBall( rcsc::Vector2D _pos , int _delay );
	
	std::vector<PreDecidedPassPlayer> & TmmPlayers();
	std::vector<PreDecidedPassPlayer> & OppPlayers();
	
	void setDefaultSafetyFactors(int _tmmDelay , float _tmmControlBuff , float _tmmMaxDash , int _oppDelay , float _oppControlBuff , float _oppMaxDash , int _oppAdditionalDelay = 0);
	void setRanges(float _startAngle , float _endAngle , float _startRate , float _endRate);
	void setHalfsCheck(bool check);

	void setMaxDirectionCount(int max)
	{
		maxDirectionCount = max;
	}
	void setMaxTmmPosCount(int max)
	{
		maxTmmPosCount = max;
	}
	void setMaxOppPosCount(int max)
	{
		maxOppPosCount = max;
	}

	void reset();
	void checkReset( const rcsc::PlayerAgent * agent );
	void deleteMemory();

	virtual float getValue(rcsc::PlayerAgent *agent);
	bool shouldDecide(rcsc::PlayerAgent *agent);
	bool mustDecide(rcsc::PlayerAgent *agent);

	void updateByWorldModel(const rcsc::WorldModel & wm , float offsideThr = 1.f);
	
	bool shouldExecute(rcsc::PlayerAgent *agent);
	void execute(rcsc::PlayerAgent *agent);
	bool shouldCommunicate(rcsc::PlayerAgent *agent);
	void communicate(rcsc::PlayerAgent *agent);
};

#endif
