// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <QBPassReceive.h>

#include <rcsc/common/audio_memory.h>

#include <rcsc/player/world_model.h>

#include <rcsc/geom/vector_2d.h>

#include <FastIC.h>

using namespace rcsc;
using namespace std;


int QBPassReceive::getBestVelocity( const rcsc::PlayerAgent * agent )
{
	const WorldModel & wm = agent->world();
	Vector2D meToBall = wm.ball().pos() - wm.self().pos();
	int maxDirectionCount = 3;
	if ( wm.dirCount( meToBall.th().degree() > maxDirectionCount ) )
		return -1.f;
	int ave , max , sum;
	agent->world().dirRangeCount( meToBall.th().degree() , 60 ,&max,&sum,&ave);
	if ( ave > maxDirectionCount )
		return -1.f;

	FastIC calculatingIC;

	{	
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
			if ( (*it)->posCount() > 3 )
				continue;
			Vector2D hisPos = (*it)->pos();
			Vector2D hisVel = Vector2D ( 0.f , 0.f );
			if ( (*it)->velValid() )
				hisVel = (*it)->vel();
			int hisType = 0;
			if ( (*it)->playerTypePtr() )
				hisType = (*it)->playerTypePtr()->id();
			calculatingIC.players().push_back( FastPlayer( hisPos , hisVel , 1 , (*it)->unum() , false , 1.f , 1.f , hisType , NULL , NULL ) );
		}
	}

	{
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
			it != end;
			++it )
		{
			if ( (*it)->posCount() > 3 )
				continue;
			Vector2D hisPos = (*it)->pos();
			Vector2D hisVel = Vector2D ( 0.f , 0.f );
			if ( (*it)->velValid() )
				hisVel = (*it)->vel();
			int hisType = 0;
			if ( (*it)->playerTypePtr() )
				hisType = (*it)->playerTypePtr()->id();
			calculatingIC.players().push_back( FastPlayer( hisPos , hisVel , 1 , (*it)->unum() , true , 1.f , 1.f , hisType , NULL , NULL ) );
		}
	}

	{
		Vector2D myPos = wm.self().pos();
		Vector2D myVel = wm.self().vel();
		int myType = wm.self().playerType().id();
		calculatingIC.players().push_back( FastPlayer( myPos , myVel , 2 , wm.self().unum() , true , 0.9 , 0.85 , myType , NULL , NULL ) );
	}

	Vector2D ballVel = - meToBall;
	int answerCount = 0;
	float bestMovement = 1000.f;
	int c = 0;
	for ( float rate = startRate; rate <= endRate ; rate += rateStep )
	{
		c++;
		ballVel.setLength( rate );
		calculatingIC.setBall( wm.ball().pos() , ballVel , 1 );
		calculatingIC.calculate( -1 , 0 , true );
		if ( calculatingIC.getFastestPlayers().size() )
			continue;
		if ( not calculatingIC.getFastestPlayers()[0].isTeamMate )
			continue;
		if ( calculatingIC.getFastestPlayers()[0].unum != wm.self().unum() )
			continue;
		float movingDist = calculatingIC.getFastestPlayers()[0].receivePoint.dist ( wm.self().pos() );
		if ( movingDist < bestMovement )
		{
			bestMovement = movingDist;
			answerCount = c;
		}
	}
	return answerCount;
}

bool executeKick( rcsc::PlayerAgent * agent )
{
	const WorldModel & wm = agent->world();
	if ( wm.audioMemory().forfeit().empty() )
		return false;
	if ( wm.audioMemory().forfeitTime() != wm.time() )
		return false;
//	if ( wm.audioMemory().forfeit().front().receivedNumber_ 
}
