// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include "SRPDribble.h"

#include <Basics.h>

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>
#include <rcsc/player/penalty_kick_state.h>
#include <body_go_to_point2008.h>
#include <rcsc/action/neck_scan_field.h>
#include <neck_offensive_intercept_neck.h>
#include <body_intercept2008.h>
#include <Intercept.h>
#include <rcsc/geom/segment_2d.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#include <cmath>

#define SRPLOG
#define MIN_VALUE -10.0
#define IDEAL_MOVEMENT 15

using namespace rcsc;
using namespace std;

// ------------------------------------------------------------SRPDribble class-----------------------------------------------------

int SRPDribble::lastSRPCycle = -10;
SRPDribble::SRPModel SRPDribble::lastModel;

bool SRPDribble::execute(PlayerAgent * agent)
{
	const WorldModel & wm = agent->world();
	setRates(agent);
	if (Target == Vector2D::INVALIDATED)
	{
		Target = Vector2D(52.5, 0);
		if (agent->world().ball().pos().x < 30)
			Target.y = agent->world().ball().pos().y;
	}
	
	if (wm.ball().pos().x < 10 and not agent->world().gameMode().isPenaltyKickMode())
	    return false;
	if (wm.self().stamina() < 3500)
	    return false;
	if (wm.self().collidesWithPlayer())
	{
#ifdef SRPLOG
	    dlog.addText(Logger::DRIBBLE, "Collision!");
#endif
	    return false;
	}
	// Added for bad vision problem
/*	{
		double checkDegree = (Target - wm.self().pos()).th().degree();
#ifndef SRPLOG
		dlog.addText(Logger::DRIBBLE, "Target(%f, %f) , Target Diff(%f, %f), Degree(%f)", Target.x, Target.y, (Target - wm.self().pos()).x, (Target - wm.self().pos()).y, checkDegree);
#endif
		if (not visionCheck(agent, checkDegree))
		{
			dlog.addText(Logger::ROLE, "Bad vision situation! Skipping SRP!");
			return false;
		}
	}*/
	if (lastSRPCycle != wm.time().cycle())
	{
	    //dlog.addText( Logger::DRIBBLE, "SRPing!!!");
	
	    if ( ! agent->world().self().isKickable() )
	    {
	    	dlog.addText( Logger::DRIBBLE,
	    		"SRPDribble::execute --> Ball is not kickable." );
	    	return false; 
	    }
	
	    findBestSRPModel(agent);
	
	    lastSRPCycle = wm.time().cycle();
	    lastModel = bestModel;
#ifdef SRPLOG
	    dlog.addText( Logger::DRIBBLE, "Best Model Value = %f!!!", bestModel.Value);
#endif
	    if (bestModel.Value < MIN_VALUE)
	    	return false;
	    
	    if (Body_KickOneStep(bestModel.SRPTarget, bestModel.KickSpeed.r()).execute(agent))
	    {	
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "SRP EXECUTED!");
#endif
			agent->setIntention(new SRPIntercept(bestModel.SRPTarget, wm.time().cycle(), bestModel.SelfDelay, bestModel.KickSpeed));
			agent->setNeckAction(new Neck_TurnToPoint(lastModel.SRPTarget));
			return true;
	    }
	    return false;
	}
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "Executing last action");
	dlog.addText( Logger::DRIBBLE, "Best Model Value = %f!!!", lastModel.Value);
#endif
	if (lastModel.Value < MIN_VALUE)
	    	return false;

	if (Body_KickOneStep(lastModel.SRPTarget, lastModel.KickSpeed.r()).execute(agent))
	{
	    lastSRPCycle = wm.time().cycle();
	    agent->setIntention(new SRPIntercept(lastModel.SRPTarget, wm.time().cycle(), lastModel.SelfDelay, lastModel.KickSpeed));
	    agent->setNeckAction(new Neck_TurnToPoint(lastModel.SRPTarget));
#ifdef SRPLOG
	    dlog.addText(Logger::DRIBBLE, "SRP EXECUTED!");
#endif
	    return true;
	}
	return false;
}

void SRPDribble::findBestSRPModel(PlayerAgent * agent)
{
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "SRPDribble::findBestSRPModel()");
#endif
/*	const rcsc::PenaltyKickState * state = agent->world().penaltyKickState();
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )*/
		
	const WorldModel & wm = agent->world();
	
	FastIC *ic = new FastIC();
	
	contrustFastIC(agent, ic);
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "\tFastIC created with %d players", ic->players().size());
#endif
	
	double minDir = agent->world().self().body().degree() - 45.0, maxDir = agent->world().self().body().degree() + 45.0;
	const rcsc::PenaltyKickState * state = agent->world().penaltyKickState();
	if ( agent->world().gameMode().isPenaltyKickMode() )
		if ( state->onfieldSide() != agent->world().ourSide() )
		{
			minDir = -90;
			maxDir = 90;
		}
		else
		{
			minDir = -270;
			maxDir = -90;
		}
	for (double Dir = minDir;Dir <= maxDir; Dir += DirectionRate)
	{
		double Direction = AngleDeg(Dir).degree();
#ifdef SRPLOG
		dlog.addText(Logger::DRIBBLE, "\tChecking direction %f", Direction);
#endif
		if (not visionCheck(agent, Direction))
		    continue;
		double dirMaxSpeed = min(2.0, Body_KickOneStep::get_max_possible_vel(AngleDeg(Direction), wm.self().kickRate(), wm.ball().vel()).r());
		double MinSpeed = dirMaxSpeed - 1.0;
		double SpeedRate = .2;
		while (MinSpeed < .8 and MinSpeed < dirMaxSpeed - 0.5)
		{
		    MinSpeed += .1;
		    SpeedRate = (dirMaxSpeed - MinSpeed) / 5.0;
		}
#ifdef SRPLOG
		dlog.addText(Logger::DRIBBLE, "\tDirection max %f min %f rate %f", dirMaxSpeed, MinSpeed, SpeedRate);
#endif
		for (double Speed = MinSpeed;Speed <= dirMaxSpeed; Speed += SpeedRate)
		{
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "\t\tChecking speed %f", Speed);
#endif
			Vector2D speedvec;
			speedvec.setPolar(Speed, AngleDeg(Direction));
			int delay = calculateSelfDelay(agent, speedvec);
			setPlayerDelays(agent, ic, speedvec);
			ic->setBall(wm.ball().pos(), speedvec, 0);
			ic->calculate(wm.self().unum(), MaxFastICCycles);
			Vector2D selfPos = wm.self().inertiaPoint(delay);
			Intercept selfIC(true);
			BallObject ICBall = wm.ball();
			Vector2D icspeed = speedvec;
			for (int i = 0;i < delay;i++)
				icspeed *= ServerParam::i().ballDecay();
			ICBall.updateOnlyVel(icspeed, Vector2D(0, 0), 0);
			ICBall.updateByFullstate(ICBall.inertiaPoint(delay), icspeed, selfPos);
			selfIC.ball() = ICBall;
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "BodyVel(%f, %f)", wm.self().vel().x, wm.self().vel().y);
			dlog.addText(Logger::DRIBBLE, "BallVel(%f, %f)", ICBall.vel().x, ICBall.vel().y);
			dlog.addText(Logger::DRIBBLE, "BallPos(%f, %f)", ICBall.pos().x, ICBall.pos().y);
			dlog.addText(Logger::DRIBBLE, "BodyPos(%f, %f) dist = %f", selfPos.x, selfPos.y, selfPos.dist(ICBall.pos()));
#endif
			Intercept::InterceptModel myModel = selfIC.finalIntercept(agent, true, 1.f, MaxFastICCycles, 0, delay);

			Vector2D opppoint(5000, 5000);
			{
			    for (unsigned i = 0;i < ic->getFastestPlayers().size();i++)
			    {
					if (not ic->getFastestPlayers()[i].isTeamMate and ic->getFastestPlayers()[i].receivePoint.x > myModel.getBallPos().x)
					{
				    	    opppoint = ic->getFastestPlayers()[i].receivePoint;
				    	    break;
					}
			    }
			}
			int OtherReach = 1000;
			unsigned OtherNum = 100;
			for (unsigned i = 0;i < ic->getFastestPlayers().size();i++)
			    if (not ic->getFastestPlayers()[i].isTeamMate)
			    {
					OtherReach = ic->getFastestPlayers()[i].reachCycle;
					OtherNum = ic->getFastestPlayers()[i].unum;
					break;
			    }
			FastestPlayer fastest = FastestPlayer( -1 , false , 1000 , Vector2D(1000.0, 1000.0) , NULL);
			if (ic->getFastestPlayers().size())
				fastest = ic->getFastestPlayers()[0];
			else
				dlog.addText(Logger::DRIBBLE, "No player! %d", ic->players().size());
			dlog.addText(Logger::DRIBBLE, "fast reach = %d", fastest.reachCycle);
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "\t\t\t\tFastest Reach Cycle = %d", OtherReach);
			dlog.addText(Logger::DRIBBLE, "\t\t\t\tFastest = %d", OtherNum);
			dlog.addText(Logger::DRIBBLE, "\t\t\t\tMy Reach Cycle = %d", myModel.getReachCycle() );
			//dlog.addText(Logger::DRIBBLE, "\t\t\t\tReceive Point = (%f, %f)", fastest.receivePoint.x, fastest.receivePoint.y);
			dlog.addText(Logger::DRIBBLE, "\t\t\t\tSRP Length = %f", (wm.ball().pos() - myModel.getBallPos()).r());
			//
#endif
			if (fastest.reachCycle <= myModel.getReachCycle() + delay or myModel.getReachCycle() == -1)
			{
#ifdef SRPLOG
				dlog.addText(Logger::DRIBBLE, "\t\t\tNot Fastest");
				dlog.addText(Logger::DRIBBLE, "\t\t\tReach cycles -> %d <= %d", fastest.reachCycle, myModel.getReachCycle() + delay);
#endif
				continue;
			}
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "\t\t\treceivePoint = (%f, %f)", myModel.getBallPos().x, myModel.getBallPos().y);
#endif
			if (not (Basics::isInField(myModel.getBallPos(), 0.8) ))	
			{
#ifdef SRPLOG
				dlog.addText(Logger::DRIBBLE, "\t\t\tNot in field");
#endif
				continue;
			}
			if ((wm.ball().pos() - myModel.getBallPos()).r() > getMaxMove(agent, myModel.getBallPos()))
			{
#ifdef SRPLOG
				dlog.addText(Logger::DRIBBLE, "\t\t\tLong SRP * limit = %f", getMaxMove(agent, myModel.getBallPos()));
#endif
				continue;
			}
			if ((wm.ball().pos() - myModel.getBallPos()).r() < MinMove or delay == 10)
			{
#ifdef SRPLOG
				dlog.addText(Logger::DRIBBLE, "\t\t\tShort SRP");
#endif
				continue;
			}
			SRPModel model;
			model.SRPTarget = myModel.getBallPos();
			model.MyReachCycle = myModel.getReachCycle() + delay;
			model.OtherReachCycle = OtherReach;
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "My reach = %d", model.MyReachCycle);
#endif
			model.SelfDelay = delay;
			model.KickSpeed = speedvec;
			model.Movement = (myModel.getBallPos() - wm.ball().pos());
			double value = setModelValue(agent, ic, &model, myModel);
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "\tSRP model created with value %f", value);
#endif
			if (model.Value > bestModel.Value)
				bestModel = model;
		}
	}
}

void SRPDribble::setLimitVars(double minAngle, double maxAngle)
{
	this->maxAngle = maxAngle;
	this->minAngle = minAngle;
}

void SRPDribble::setDirectionRate(double DirectionRate)
{
	this->DirectionRate = DirectionRate;
}

void SRPDribble::setMaxPosCount(int MaxPosCount)
{
	this->MaxPosCount = MaxPosCount;
}

void SRPDribble::setMaxDirCount(int MaxDirCount)
{
	this->MaxDirCount = MaxDirCount;
}

void SRPDribble::setMaxFastICCycles(int MaxFastICCycles)
{
	this->MaxFastICCycles = MaxFastICCycles;
}

void SRPDribble::setMaxMove(double MaxMove)
{
	this->MaxMove = MaxMove;
}

void SRPDribble::setMinMove(double MinMove)
{
	this->MinMove = MinMove;
}

void SRPDribble::contrustFastIC(PlayerAgent * agent, FastIC * ic)
{
	const WorldModel & wm = agent->world();
	if ( not agent->world().gameMode().isPenaltyKickMode())
	{
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for (PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();it != end;++it)
		{
			int playerDelay = 0;
			if ((*it)->posCount() > MaxPosCount)
				continue;
			if (not isInCheckArea(agent, (*it)->pos()))
			{
#ifdef SRPLOG
				dlog.addText(Logger::DRIBBLE, "\t\t\t\t\t\t\t!!CHECKAREA!!, %d, %d", true, (*it)->unum());
#endif
				continue;
			}
			double controlBuff = 1.0;
			Vector2D velocity = Vector2D(0.0, 0.0);
			if ((*it)->velValid())
				velocity = (*it)->vel();
//			if ((*it)->goalie())
//				controlBuff += 0.5;
			FastPlayer tmm = FastPlayer( (*it)->pos() , (*it)->vel() , playerDelay , (*it)->unum() , true , controlBuff , 1.0 , (*it)->playerTypePtr()->id() , new PreDecidedPassPlayer() );
			ic->players().push_back(tmm);
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Player pushed -> (%d,%d) -> pos = (%.2f, %.2f) count %d -> delay(%d)", true, (*it)->unum(), (*it)->pos().x , (*it)->pos().y, (*it)->posCount() , playerDelay);
#endif
		}
	}	
	{
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for (PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();it != end;++it)
		{
			int playerDelay = 0;
			if ((*it)->posCount() > MaxPosCount)
			{
#ifdef SRPLOG
				dlog.addText(Logger::DRIBBLE, "Player %d has low pos count ->%d", (*it)->unum(), (*it)->posCount());
#endif
				continue;
			}
			if (not isInCheckArea(agent, (*it)->pos()))
			{
#ifdef SRPLOG
				dlog.addText(Logger::DRIBBLE, "\t\t\t\t\t\t\t!!CHECKAREA!!, %d, %d", false, (*it)->unum());
#endif
				continue;
			} 
			double controlBuff = 1.0;
			Vector2D velocity = Vector2D(0.0, 0.0);
			if ((*it)->velValid())
				velocity = (*it)->vel();
//			if ((*it)->goalie())
//				controlBuff += 1.5;
			if ( agent->world().gameMode().isPenaltyKickMode() )
			{
				if (not (*it)->goalie())
					continue;
				controlBuff = 3.0;
				if ( (*it)->pos().dist(agent->world().self().pos()) < 4 )
				    controlBuff = 1.5;
				if ( (*it)->isTackling() )
				{
				    dlog.addText(Logger::DRIBBLE, "Tackled %d", (*it)->tackleCount());
				    int tdelay = max( 10 - (*it)->tackleCount() - 2, 1 );
				}
			}
			FastPlayer opp = FastPlayer( playerDelay, false, controlBuff , 1 , (*it) );
			ic->players().push_back(opp);
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Player pushed -> (%d,%d) -> pos = (%.2f, %.2f) count %d -> delay(%d)", false, (*it)->unum(), (*it)->pos().x , (*it)->pos().y, (*it)->posCount() , playerDelay);
#endif
		}
	}
}

int SRPDribble::calculateSelfDelay(PlayerAgent * agent, Vector2D ballVel)
{
	const WorldModel & wm = agent->world();
	BallObject ICBall = wm.ball();
	ICBall.updateOnlyVel(ballVel, Vector2D(0, 0), 0);
	int cycles = 0;
	do 
	{
		cycles++;
		if (ICBall.inertiaPoint(cycles).dist(wm.self().pos()) > wm.self().kickableArea())
			break;
	}
	while (cycles < 4);
	if (cycles == 4)
		cycles = 10;
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "Self delay = %d", cycles);
#endif
	return cycles;
}

bool SRPDribble::isInCheckArea(PlayerAgent * agent, Vector2D pos)
{
	const WorldModel & wm = agent->world();
	Circle2D chCir(wm.ball().pos(), 15);
	if (chCir.contains(pos) or fabs(((pos - wm.ball().pos()).th() - wm.self().body()).degree()) < 90 and wm.ball().pos().dist(pos))
		return true; 
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "\t\t\t\t\t\t\tTRI = %f", fabs(((pos - wm.ball().pos()).th() - wm.self().body()).degree()));
#endif
	return false;
}

bool SRPDribble::visionCheck(PlayerAgent * agent, double Direction)
{
	int max = 0;
	if (agent->world().gameMode().isPenaltyKickMode())
	    if (agent->world().getOpponentGoalie() and agent->world().getOpponentGoalie()->posCount() < 2)
		return true;
	    else
		return false;
	
#if 1
	agent->world().dirRangeCount(AngleDeg(Direction), 50, NULL, NULL, &max);
	if (max > MaxDirCount)
	{
#ifdef SRPLOG
		dlog.addText(Logger::DRIBBLE, "Vision Fail!");
#endif
		return false;
	}
#else
	for (double i = Direction - 25.0;i <= Direction + 25.0;i += 5.0)
	{
		if (agent->world().dirCount(AngleDeg(i)) > MaxDirCount)
		{
#ifdef SRPLOG
		    dlog.addText(Logger::DRIBBLE, "Vision Fail at %f -> %d", i, agent->world().dirCount(AngleDeg(i)));
#endif
		    return false;
		}
#ifdef SRPLOG
		dlog.addText(Logger::DRIBBLE, "Good Vision at %f -> %d", i, agent->world().dirCount(AngleDeg(i)));
#endif
	}
#endif
	return true;
}

double SRPDribble::getMaxMove(PlayerAgent * agent, Vector2D Target)
{
	if ( agent->world().gameMode().isPenaltyKickMode())
		return 35;
	if (Target.x - agent->world().offsideLineX() > -5.0)
		return 20;
	return 15;
}

double SRPDribble::getOneOnOneWeight(PlayerAgent * agent, FastIC * ic, AngleDeg direction, double targetLine, int ballDelay, double dashDistance, double yLimit)
{
	if (agent->world().gameMode().isPenaltyKickMode())
	    return 1.0;
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "TargetLine = %f, y = %f", targetLine, yLimit);
#endif
	ic->setCalculateBallDecay(false);
	double SRPRate = .95;
	int PathCycle = ceil(dashDistance / SRPRate);
	Vector2D speed;
	speed.setPolar(SRPRate, direction);
	///////////////////////////////////////////////////////////////////////////////////
	double max = 0;
	const std::vector<FastPlayer>::iterator end = ic->players().end();
	std::vector<FastPlayer>::iterator f = ic->players().end();
	FastPlayer fast(0, 0, 0, 0, 0);
	bool isDel = false;
	for (std::vector<FastPlayer>::iterator it = ic->players().begin();it != end;++it)
	{
	    if (it->mPos.x > max)
	    {
			max = it->mPos.x;
			f = it;
	    }
	}
	if (f != ic->players().end())
	{
	    isDel = true;
	    fast = *f;
	    ic->players().erase(f);
	}
	///////////////////////////////////////////////////////////////////////////////////
	ic->setBall(agent->world().ball().pos(), speed, 0);
	ic->calculate(agent->world().self().unum());
	Vector2D newBallPos;
	newBallPos.setPolar(dashDistance, direction);
	newBallPos += agent->world().ball().pos();
	int fullPathCycle = ceil((targetLine - newBallPos.x) / SRPRate) + PathCycle;
	if (targetLine <= newBallPos.x)
	{
	    fullPathCycle = PathCycle = ceil(((targetLine - agent->world().ball().pos().x) / speed.th().cos()) / SRPRate);
	}
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "Full Path Cycle = %d", fullPathCycle);
#endif
	if (ic->getFastestPlayers().size() and ic->getFastestPlayers()[0].reachCycle <= PathCycle)
	{
#ifdef SRPLOG
		dlog.addText(Logger::DRIBBLE, "1 - opp (%d, %d) reachs at %d in (%.2f,%.2f )", ic->getFastestPlayers()[0].isTeamMate, ic->getFastestPlayers()[0].unum, ic->getFastestPlayers()[0].reachCycle, ic->getFastestPlayers()[0].receivePoint.x, ic->getFastestPlayers()[0].receivePoint.y);
#endif
		if (isDel)
		    ic->players().push_back(fast);
		ic->setCalculateBallDecay(true);
		return ((float)ic->getFastestPlayers()[0].reachCycle) / ((float)fullPathCycle);
	}
#ifdef SRPLOG
	if (ic->getFastestPlayers().size())
	    dlog.addText(Logger::DRIBBLE, "2 - opp (%d, %d) reachs at %d in (%.2f,%.2f )", ic->getFastestPlayers()[0].isTeamMate, ic->getFastestPlayers()[0].unum, ic->getFastestPlayers()[0].reachCycle, ic->getFastestPlayers()[0].receivePoint.x, ic->getFastestPlayers()[0].receivePoint.y);
#endif
	if (targetLine <= newBallPos.x)
	{
	    ic->setCalculateBallDecay(true);
	    if (isDel)
	    {
	        ic->players().push_back(fast);
	    }	
	    return 1.0;
	}
	speed = Vector2D(0.9, 0.0);
	if (newBallPos.y > yLimit)
	    speed.setPolar(SRPRate, (Vector2D(targetLine, yLimit) - newBallPos).th());
	else if (newBallPos.y < -yLimit)
	    speed.setPolar(SRPRate, (Vector2D(targetLine, -yLimit) - newBallPos).th());
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "path from (%.2f, %.2f) angle = %f", newBallPos.x, newBallPos.y, speed.th().degree());
#endif
	ic->setBall(newBallPos, speed, PathCycle );
	ic->calculate(agent->world().self().unum());
	if (ic->getFastestPlayers().size() and ic->getFastestPlayers()[0].reachCycle <= fullPathCycle)
	{
		ic->setCalculateBallDecay(true);
#ifdef SRPLOG
		dlog.addText(Logger::DRIBBLE, "3 - opp (%d, %d) reachs at %d in (%.2f,%.2f )", ic->getFastestPlayers()[0].isTeamMate, ic->getFastestPlayers()[0].unum, ic->getFastestPlayers()[0].reachCycle, ic->getFastestPlayers()[0].receivePoint.x, ic->getFastestPlayers()[0].receivePoint.y);
#endif
		if (isDel)
		{
		    ic->players().push_back(fast);
		}
		return ((float)ic->getFastestPlayers()[0].reachCycle) / ((float)fullPathCycle);
	}
#ifdef SRPLOG
	if (ic->getFastestPlayers().size())
	    dlog.addText(Logger::DRIBBLE, "4 - opp (%d, %d) reachs at %d in (%.2f,%.2f )", ic->getFastestPlayers()[0].isTeamMate, ic->getFastestPlayers()[0].unum, ic->getFastestPlayers()[0].reachCycle , ic->getFastestPlayers()[0].receivePoint.x, ic->getFastestPlayers()[0].receivePoint.y);
#endif
	ic->setCalculateBallDecay(true);
	if (isDel)
	{
	    ic->players().push_back(fast);
	}
	return 1.0;
}

void SRPDribble::setRates(PlayerAgent * agent)
{
    XRate = 1.0;
    const rcsc::PenaltyKickState * state = agent->world().penaltyKickState();
	if ( agent->world().gameMode().isPenaltyKickMode() and state->onfieldSide() != agent->world().ourSide() )
	{
	    XRate = -1.0;
	}
    OneRate = 10.0;
    ReachRate = 1.0;
    IsTurnRate = -3.0;
    TargetRate = -.6;
    IdealRate = -1;
    AngleRate = -.1;
    HighAngleRate = -20.0;
}

double SRPDribble::setModelValue(PlayerAgent * agent, FastIC * ic, SRPDribble::SRPModel * model, Intercept::InterceptModel icmodel)
{
	/* Weight       | Rate
	*  -------------|------
	*  X Movement   |  1
	*  One On One   |  10
	*  Cycle Diff   |  1 
	*  Turn Or No   |  -3 
	*  Target angle | -.2
	*  Movement     |  -1
	*  Turn Angle   | -.1
	*/
	double targetLine = 42.0;
	double yLimit = 25;
	Segment2D line(Vector2D(targetLine, yLimit), Vector2D(targetLine, -yLimit));
	
	if (agent->world().ball().pos().x > 30)
	{
		targetLine = 52.5;
		yLimit = 9.0;
	}
	
	if (line.dist(agent->world().ball().pos()) > 15.0)
	{
		targetLine = max(agent->world().ball().pos().x + 15.0, 26.0);
		if (targetLine < 36.0)
		    yLimit = 30.0;
		else
		    yLimit = 20.0;
	}
	
	if (model->OtherReachCycle - model->MyReachCycle > 7)
	    model->OtherReachCycle = model->MyReachCycle + 7;
	double oneonone = getOneOnOneWeight(agent, ic, model->KickSpeed.th(), targetLine, model->MyReachCycle, model->Movement.r(), yLimit);
	float oneOneThr = .55;
	if (agent->world().ball().pos().x > 38)
	    oneOneThr = .6;
	if (fabs((model->KickSpeed.th() - (Target - agent->world().ball().pos()).th()).degree()) > 20.0 and agent->world().gameMode().isPenaltyKickMode())
	    return model->Value = -1000.0;
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "One on one = %f", oneonone);
	dlog.addText(Logger::DRIBBLE, "Target = (%f, %f)", Target.x , Target.y);
	dlog.addText(Logger::DRIBBLE, "Target Angle = %f", (Target - agent->world().ball().pos()).th().degree());
#endif
	if (oneonone < oneOneThr)
	    return model->Value = -1000.0;
	model->Value = model->Movement.x                                  * XRate
			+ oneonone                                        * OneRate
			+ (model->OtherReachCycle - model->MyReachCycle)  * ReachRate
			+ (icmodel.isTurnMove() ? 0:1)                    * IsTurnRate
			+ fabs((model->KickSpeed.th() - (Target - 
			    agent->world().ball().pos()).th()).degree())  * TargetRate
			+ fabs(model->Movement.r() - IDEAL_MOVEMENT)      * IdealRate;
	if (not icmodel.isTurnMove())			
	    model->Value += fabs(icmodel.getTurnAngle())          * AngleRate;
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "Weights:");
	dlog.addText(Logger::DRIBBLE, "1 - X Movement: %f -> %f", model->Movement.x, model->Movement.x * XRate);
	dlog.addText(Logger::DRIBBLE, "2 - One On One: %f -> %f", oneonone, oneonone * OneRate);
	dlog.addText(Logger::DRIBBLE, "3 - Cycle Diff: %d -> %f", (model->OtherReachCycle - model->MyReachCycle), (model->OtherReachCycle - model->MyReachCycle) * ReachRate);
	dlog.addText(Logger::DRIBBLE, "4 - Turn Or No: %d -> %f", (icmodel.isTurnMove() ? 0:1), (icmodel.isTurnMove() ? 0:1) * IsTurnRate);
	dlog.addText(Logger::DRIBBLE, "5 - Target Angle: %f -> %f", fabs((model->KickSpeed.th() - (Target - agent->world().ball().pos()).th()).degree()), fabs((model->KickSpeed.th() - (Target - agent->world().ball().pos()).th()).degree()) * TargetRate);
	if (not icmodel.isTurnMove())			
	    dlog.addText(Logger::DRIBBLE, "6 - Turn Angle: %f -> %f", fabs(icmodel.getTurnAngle()), fabs(icmodel.getTurnAngle()) * AngleRate);
	dlog.addText(Logger::DRIBBLE, "7 - Movement Ideal: %f -> %f", fabs(model->Movement.r() - IDEAL_MOVEMENT), fabs(model->Movement.r() - IDEAL_MOVEMENT) * IdealRate); 
#endif
	double angleThr = 30.0;
	if (pointOOO(agent, ic, Target) and fabs((model->KickSpeed.th() - (Target - agent->world().ball().pos()).th()).degree()) > angleThr and not agent->world().gameMode().isPenaltyKickMode())
	{
	    model->Value += HighAngleRate;
#ifdef SRPLOG
	    dlog.addText(Logger::DRIBBLE, "Target High Angle! Value Decreased %f", HighAngleRate);
#endif
	}
#ifdef SRPLOG
	dlog.addText(Logger::DRIBBLE, "SUM %f", model->Value);
#endif
	return model->Value;
}

void SRPDribble::setPlayerDelays(PlayerAgent * agent, FastIC * ic, Vector2D speed)
{
	const std::vector<FastPlayer>::iterator end = ic->players().end();
	for (std::vector<FastPlayer>::iterator it = ic->players().begin();it != end;++it)
	{
		it->delay = 0;
		// Turn delay 
		if (not (*it).shouldConsiderFace)
		{
			(*it).delay++;
			if ((*it).playerPtr and (*it).playerPtr->velCount() < 4 and (*it).playerPtr->vel().r() > ServerParam::i().defaultPlayerSpeedMax() / 2)
			{
				(*it).shouldConsiderFace = true;
				(*it).playerBodyDir = (*it).playerPtr->vel().th().degree();
				(*it).delay--;
			}
		}
		// Underestimate behind players
		if (abs((((*it).mPos - agent->world().ball().pos()).th() - speed.th()).degree()) > 70.0)
		{
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Behind, Angle %f ", abs((((*it).mPos - agent->world().ball().pos()).th() - speed.th()).degree()));
#endif
			(*it).delay++;
		}
		// Underestimate far players
		if ((*it).mPos.dist(agent->world().ball().pos()) > 10.0 and not agent->world().gameMode().isPenaltyKickMode())
		{
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Far");
#endif
			(*it).delay++;
		}
		if ((*it).mPos.dist(agent->world().ball().pos()) > 15.0 and not agent->world().gameMode().isPenaltyKickMode())
		{
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Very Far");
#endif
			(*it).delay++;
		}
		// Underestimate very near players 
		if ((*it).mPos.dist(agent->world().ball().pos()) < 2.5 and abs((((*it).mPos - agent->world().ball().pos()).th() - speed.th()).degree()) > 60.0)
		{
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Near");
#endif
			(*it).delay++;
		}
		// Underestimate offside line players
		if ( ( (*it).mPos.dist(agent->world().ball().pos()) < 15 ) and ( (*it).mPos.x >= agent->world().ball().pos().x ) and abs((*it).mPos.x - agent->world().offsideLineX()) < 10 and not agent->world().gameMode().isPenaltyKickMode())
		{
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Offside");
#endif
			(*it).delay++;
		}
		if (not it->isTeamMate and it->unum == agent->world().opponentGoalieUnum())
		{			
			if (not agent->world().gameMode().isPenaltyKickMode())
			    it->delay = 0;
#ifdef SRPLOG
			dlog.addText(Logger::DRIBBLE, "Player (%d, %d) is Goalie!", it->isTeamMate, it->unum);
#endif
		}
#ifdef SRPLOG
		dlog.addText(Logger::DRIBBLE, "Player (%d, %d) delayed %d cycles", it->isTeamMate, it->unum, it->delay);
#endif
	}
}

bool SRPDribble::pointOOO(PlayerAgent * agent, FastIC * ic, Vector2D point)
{
	ic->setCalculateBallDecay(false);
	///////////////////////////////////////////////////////////////////////////////////
	double max = 0;
	const std::vector<FastPlayer>::iterator end = ic->players().end();
	std::vector<FastPlayer>::iterator f = ic->players().end();
	FastPlayer fast(0, 0, 0, 0, 0);
	bool isDel = false;
	for (std::vector<FastPlayer>::iterator it = ic->players().begin();it != end;++it)
	{
	    if (it->mPos.x > max)
	    {
		max = it->mPos.x;
		f = it;
	    }
	}
	if (f != ic->players().end())
	{
	    isDel = true;
	    fast = *f;
	    ic->players().erase(f);
	}
	///////////////////////////////////////////////////////////////////////////////////
	double SRPRate = 0.9;
	int PathCycle = ceil((point - agent->world().ball().pos()).r() / SRPRate);
	double direction = (point - agent->world().ball().pos()).th().degree();
	Vector2D speed;
	speed.setPolar(SRPRate, direction);
	ic->setBall(agent->world().ball().pos(), speed, 0);
	ic->calculate(agent->world().self().unum());
	float OOOThr = 0.75;
	if (isDel)
		ic->players().push_back(fast);
	ic->setCalculateBallDecay(true);
	if (ic->getFastestPlayers().size() and ic->getFastestPlayers()[0].reachCycle < PathCycle * OOOThr)
		return false;
	return true;
}

void SRPDribble::setTarget(Vector2D target)
{
	Target = target;
}

// -------------------------------------------------Intention SRP intercept---------------------------------------------

bool SRPIntercept::finished(const PlayerAgent * agent)
{
	if ( cycleCount >= icDelay && agent->world().self().isKickable() )
	{
		rcsc::dlog.addText( Logger::DRIBBLE, "Intention intercept completed!" );
		return true;
	}

	if ( agent->world().existKickableOpponent() || agent->world().existKickableTeammate() )
	{
		rcsc::dlog.addText( Logger::DRIBBLE, "Intention intercept failed!" );
		return true;
	}

	return false;
}

bool SRPIntercept::execute( PlayerAgent * agent)
{
	cycleCount++;
	Vector2D toBall = target - agent->world().self().pos();
	if (cycleCount < 2 and agent->world().self().isKickable())
	{
	    Intercept ic(true, true);
	    if (agent->world().gameMode().isPenaltyKickMode())
		ic = Intercept();	
	    BallObject ICBall = agent->world().ball();
	    ICBall.updateByFullstate(ICBall.inertiaPoint(1), speed, agent->world().self().pos());
	    ic.ball() = ICBall;
	    ic.setBallDelay(1);
	    ic.execute(agent, true);
	}
	else
	    if (agent->world().gameMode().isPenaltyKickMode())
		Intercept().execute(agent);	
	    else
		Intercept(true, true).execute(agent);
	if (agent->world().gameMode().isPenaltyKickMode())
	    agent->setNeckAction(new Neck_TurnToBallAndPlayer(agent->world().getOpponentGoalie(), 0 ));
	dlog.addText( Logger::DRIBBLE, "Intention intercepting!");
	return true;
}
