// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef SRP_DRIBBLE_H
#define SRP_DRIBBLE_H

#include <body_go_to_point2008.h>
#include <rcsc/player/soccer_action.h>
#include <rcsc/common/server_param.h>
#include <rcsc/player/soccer_intention.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/world_model.h>
#include <rcsc/common/logger.h>

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/polygon_2d.h>

#include <FastIC.h>
#include <Intercept.h>

#include <PreDecidedPassBase.h>
#include <vector>

using namespace rcsc;

class SRPDribble
{
	struct SRPModel
	{
		Vector2D SRPTarget;
		Vector2D OpponentReachPoint;
		double OpponentDistance;
		int MyReachCycle;
		int OtherReachCycle;
		Vector2D KickSpeed;
		Vector2D Movement;
		int SelfDelay;
		double Value;
		/*double getValue(double special);*/
		SRPModel(): SRPTarget(0, 0), 
					MyReachCycle(-1), 
					OtherReachCycle(-1),
					KickSpeed(0, 0),
					Movement(0, 0),
					Value(-1000)
		{}
	};	
	
	/*struct SRPFastest
	{
	    Vector2D ReachPoint;
	    int ReachCycle;
	    PlayerObject * Player;
	};*/
	
	SRPModel bestModel;
	Vector2D Target;
	static SRPModel lastModel;
	double maxAngle, minAngle;
	double DirectionRate;
	int MaxPosCount, MaxDirCount;
	int MaxFastICCycles;
	double MaxMove;
	double MinMove;
	double MinOppDist;
	static int lastSRPCycle;
	
	// RATES
	double XRate, OneRate, ReachRate, IsTurnRate, TargetRate, IdealRate, AngleRate, HighAngleRate;
	
	public:
	
	SRPDribble() : bestModel()
	{
		DirectionRate = 15.0;
		MaxPosCount = 6;
		MaxDirCount = 3;
		MaxFastICCycles = 50;
		MaxMove = 20.0;
		MinMove = 3.0;
		MinOppDist = 2.3;
		setLimitVars(-90.0, 90.0);
		Target = Vector2D(52.5, 0);
	}
	
	bool execute(PlayerAgent * agent);
	
	void setDirectionRate(double DirectionRate);
	void setMaxPosCount(int MaxPosCount);
	void setMaxDirCount(int MaxDirCount);
	void setMaxFastICCycles(int MaxFastICCycles);
	void setMaxMove(double MaxMove);
	void setMinMove(double MaxMove);
	void setLimitVars(double minAngle, double maxAngle);
	
	void setTarget(Vector2D Target); 
	
	private:
	
	void contrustFastIC(PlayerAgent * agent, FastIC * ic/*, Vector2D speed*/);
	void findBestSRPModel(PlayerAgent * agent); // Algorithm 1
	int calculateSelfDelay(PlayerAgent * agent, Vector2D ballVel);
	bool isInCheckArea(PlayerAgent * agent/*, Vector2D vel*/, Vector2D pos);
	bool visionCheck(PlayerAgent * agent, double dir);
	double getMaxMove(PlayerAgent * agent, Vector2D Target);
	double getOneOnOneWeight(PlayerAgent * agent, FastIC * ic, AngleDeg direction, double targetLine, int ballDelay, double dashDistance, double yLimit);
	double setModelValue(PlayerAgent * agent, FastIC * ic, SRPModel * model, Intercept::InterceptModel icmodel);
	void setPlayerDelays(PlayerAgent * agent, FastIC * ic, Vector2D speed);
	void setRates(PlayerAgent * agent);
	bool pointOOO(PlayerAgent * agent, FastIC * ic, Vector2D point);
};

class SRPIntercept : public rcsc::SoccerIntention 
{
private:

    const rcsc::Vector2D target;
    int lastCycle;
    int cycleCount;
    int icDelay;
    Vector2D speed;

public:

    SRPIntercept( const rcsc::Vector2D & _target, const int _lastCycle, const int _icDelay, Vector2D _speed ) : target( _target ), lastCycle( _lastCycle ), cycleCount( 0 ), icDelay( _icDelay ), speed(_speed)
	{ }

    bool finished( const rcsc::PlayerAgent * agent );

    bool execute( rcsc::PlayerAgent * agent );

};


#endif

