// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <Shoot.h>

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/view_change_width.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/vector_2d.h>

#include <view_tactical.h>

#include <FastIC.h>
#include <vector>

using namespace rcsc;
using namespace std;

bool Shoot::execute(rcsc::PlayerAgent* agent)
{
	const WorldModel* wm = &agent->world();

	if (wm->ball().pos().dist(Vector2D(52.5, 0.0)) > 30.0)
	{
		dlog.addText(Logger::SHOOT, "SHOOT: Too far from goal [FALSE]");
		return false;
	}

	dlog.addText ( Logger::SHOOT , "ballVel\t%f" , wm->ball().vel().r() );
/*	int max, sum, ave;
	double topPostDir = atan( ((-9.0) - wm->ball().pos().y) / (52.5 - wm->ball().pos().x) ) * AngleDeg::RAD2DEG;
	double bottomPostDir = atan( ((+9.0) - wm->ball().pos().y) / (52.5 - wm->ball().pos().x) ) * AngleDeg::RAD2DEG;
	wm->dirRangeCount((topPostDir + bottomPostDir) / 2, bottomPostDir - topPostDir + 10, &max, &sum, &ave);

	dlog.addText ( Logger::SHOOT , "ave\t%d\tmax\t%d" , ave , max );
	if (ave > 3 or max > 5)
	{
		dlog.addText(Logger::SHOOT, "SHOOT: bad dirCount [FALSE]");
		return false;
	}
*/
	vector<double> possiblePointsY;

	FastIC fastIC;
	for (uint i = 0 ; i < wm->opponentsFromBall().size() ; i++)
	{
		if (wm->opponentsFromBall()[i]->posCount() <= 5)
		{
			fastIC.players().push_back( FastPlayer(1, false, 1.1 ,/*wm->opponentsFromBall()[i]->goalie() ? 2.0 : 1.0,*/
							 1, wm->opponentsFromBall()[i]) );

			dlog.addText(Logger::SHOOT, "Opp player added %d (%f, %f)", wm->opponentsFromBall()[i]->unum(),
						wm->opponentsFromBall()[i]->pos().x, wm->opponentsFromBall()[i]->pos().y);
		}
		else
			dlog.addText(Logger::SHOOT, "opp PLayer %d !!!!!!!!!", wm->opponentsFromBall()[i]->unum());
	}

	double minY = -6.0 , maxY = +6.0;
	if (wm->ball().pos().y < -7.0)
		maxY = +5.0;
	else if (wm->ball().pos().y > +7.0)
		minY = -5.0;

	for (double y = minY ; y <= maxY ; y += 0.25)
	{

		int max, sum, ave;
//		double topPostDir = atan( ((-9.0) - wm->ball().pos().y) / (52.5 - wm->ball().pos().x) ) * AngleDeg::RAD2DEG;
//		double bottomPostDir = atan( ((+9.0) - wm->ball().pos().y) / (52.5 - wm->ball().pos().x) ) * AngleDeg::RAD2DEG;
		double dir = atan((y - wm->ball().pos().y) / (52.5 - wm->ball().pos().x)) * AngleDeg::RAD2DEG;
		wm->dirRangeCount(dir, 40, &max, &sum, &ave);

		dlog.addText ( Logger::SHOOT , "ave\t%d\tmax\t%d  Dir: %f" , ave , max , dir);
		if (/*ave > 3 or*/ max > 5)
		{
//			dlog.addText(Logger::SHOOT, "SHOOT: bad dirCount [FALSE]");
			continue;
		}


		double kickDir = atan( (y - wm->ball().pos().y) / (52.5 - wm->ball().pos().x) ) * AngleDeg::RAD2DEG;

		dlog.addText ( Logger::SHOOT , "direction\t%f---> maxVel\t%f" , kickDir , Body_KickOneStep::get_max_possible_vel(kickDir, 100, agent->world().ball().vel()).r() - 0.1 );
		Vector2D ballVel = Vector2D().setPolar(
					Body_KickOneStep::get_max_possible_vel(kickDir, 1000, agent->world().ball().vel()).r() - 0.1, kickDir);

		fastIC.setBall(wm->ball().pos(), ballVel, 0);
		fastIC.calculate(wm->self().unum(), 0, true);

		if (fastIC.getFastestPlayers().size() == 0)
		{
			possiblePointsY.push_back(y);
			dlog.addText(Logger::SHOOT, "SHOOT: possible point Y : %.2f", y);
		}
		else
		{
			dlog.addText(Logger::SHOOT, "SHOOT: dir fastest : %d in ( %.2f , %.2f ) cycles", fastIC.getFastestPlayers()[0].unum , fastIC.getFastestPlayers()[0].receivePoint.x , fastIC.getFastestPlayers()[0].receivePoint.y );
		}
	}


	if (possiblePointsY.size() == 0)
	{
		dlog.addText(Logger::SHOOT, "SHOOT: No possible points were found [FALSE]");
		return false;
	}


	const PlayerObject* oppGoalie = NULL;
	for (uint j = 0 ; j < fastIC.players().size() ; j++)
		if (fastIC.players()[j].playerObjectPtr->goalie())
			oppGoalie = fastIC.players()[j].playerObjectPtr;

	vector<double> pointsWeight;

	for (uint i = 0 ; i < possiblePointsY.size() ; i++)
	{
		pointsWeight.push_back(1.0);
		if (oppGoalie)
			pointsWeight[i] *= fabs(possiblePointsY[i] - oppGoalie->pos().y);
		else
			pointsWeight[i] *= min(7.0 - possiblePointsY[i], 7.0 + possiblePointsY[i]);

	}


	double bestPointY = possiblePointsY[0];
	double bestPointWeight = pointsWeight[0];

	for (uint i = 1 ; i < possiblePointsY.size() ; i++)
		if (pointsWeight[i] > bestPointWeight)
		{
			bestPointY = possiblePointsY[i];
			bestPointWeight = pointsWeight[i];
		}

	dlog.addText(Logger::SHOOT, "SHOOT: targetY: %.2f", bestPointY);

	double kickDir = atan( (bestPointY - wm->ball().pos().y) / (52.5 - wm->ball().pos().x) ) * AngleDeg::RAD2DEG;

	Vector2D ballVel = Vector2D().setPolar(
				Body_KickOneStep::get_max_possible_vel(kickDir, 1000, agent->world().ball().vel()).r() - 0.1, kickDir);

	dlog.addText(Logger::SHOOT, "kickDir: %f \t ballVel: (%f, %f)", kickDir, ballVel.r(), ballVel.th().degree());

	return Body_KickOneStep(Vector2D(52.5, bestPointY), ballVel.r(), true).execute(agent);
}

void Shoot::changeLookMode(rcsc::PlayerAgent* agent)
{
	Vector2D lookPoint;
	switch(agent->world().time().cycle() % 4)
	{
		case 0:
			lookPoint = Vector2D(43.0, -25.0);
			break;
		case 1:
			lookPoint = Vector2D(52.5, -6.0);
			break;
		case 2:
			lookPoint = Vector2D(43.0, +25.0);
			break;
		case 3:
			lookPoint = Vector2D(52.5, +6.0);
			break;
	}

	if (agent->world().ball().pos().x > 36.0 and agent->world().ball().pos().absY() < 15.0 and agent->world().self().isKickable())
	{
		View_ChangeWidth(rcsc::ViewWidth::NARROW).execute(agent);
		agent->doChangeView(rcsc::ViewWidth::NARROW);
		agent->setNeckAction(new rcsc::Neck_TurnToPoint(lookPoint));
		dlog.addText(Logger::ROLE, "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA");
		agent->setViewAction( new View_Tactical() );
	}
}
