// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

//Caution!!!
/*
	In this class and it's inheritants, the cycle is determined from when the process is started, not from when the ball starts moving!!
	Be carful if you use cycle parameters!
*/

#ifndef __VIRTUAL_PASS_H_
#define __VIRTUAL_PASS_H_

#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/polygon_2d.h>
#include <rcsc/player/player_agent.h>
#include <FastIC.h>

#include <map>
#include <string>

#include <PreDecidedPassBase.h>

struct VirtualPassDecision
{
	PreDecidedPassReceiver receiverStats;
	float velocity;
	int releaseCycle;
	VirtualPassDecision() : receiverStats()
	{
		releaseCycle = -10;
	}

};
class VirtualPassBase
{
protected:
	std::vector<PreDecidedPassPlayer> tmmPlayers;
	std::vector<PreDecidedPassPlayer> oppPlayers;
	PreDecidedPassBall mBall;

	FastIC fastIC;
	
	// Default Safety Factors
		int tmmDelay;
		float tmmControlBuff;
		float tmmMaxDash;
		int oppDelay;
		float oppControlBuff;
		float oppMaxDash;
	// Max PosCounts
		int maxTmmPosCount;
		int maxOppPosCount;
	// Calculating Ranges
		float startAngle;
		float endAngle;
		float startRate;
		float endRate;
	int maxDirectionCount;

	// Dynamic Parameters
		std::map<float , std::map<float,PreDecidedPassReceiver> > checkedStats;
		std::map<float , std::map<float,bool> > isChecked;

	// Check Halfs Or Not
		bool checkHalfs;

	// Best
		VirtualPassDecision tmpStats;
		bool setThisCycle;
		static VirtualPassDecision bestPassStats;

	PreDecidedPassReceiver getFastestDirectionVelocity(const rcsc::PlayerAgent *agent , float rate , float direction);
	void constructFastIC();

public:
	VirtualPassBase();
	virtual ~VirtualPassBase();

	void setBallPos( rcsc::Vector2D _pos )
	{
		mBall.mPos = _pos;
	}

	void setBallDelay( int _delay )
	{
		mBall.delay = _delay;
	}

	void setBall( rcsc::Vector2D _pos , int _delay )
	{
		mBall.mPos = _pos;
		mBall.delay = _delay;
	}
	
	std::vector<PreDecidedPassPlayer> & TmmPlayers()
	{
		return tmmPlayers;
	}

	std::vector<PreDecidedPassPlayer> & OppPlayers()
	{
		return oppPlayers;
	}
	void setDefaultSafetyFactors(int _tmmDelay , float _tmmControlBuff , float _tmmMaxDash , int _oppDelay , float _oppControlBuff , float _oppMaxDash)
	{
		deleteMemory();
		tmmDelay = _tmmDelay;
		tmmControlBuff = _tmmControlBuff;
		tmmMaxDash = _tmmMaxDash;
		oppDelay = _oppDelay;
		oppControlBuff = _oppControlBuff;
		oppMaxDash = _oppMaxDash;
	}

	void setRanges(float _startAngle , float _endAngle , float _startRate , float _endRate)
	{
		startAngle = _startAngle;
		endAngle = _endAngle;
		startRate = _startRate;
		endRate = _endRate;
	}

	void setHalfsCheck(bool check)
	{
		checkHalfs = check;
	}
	
	void setMaxDirectionCount(int max)
	{
		maxDirectionCount = max;
	}

	void setMaxTmmPosCount(int max)
	{
		maxTmmPosCount = max;
	}

	void setMaxOppPosCount(int max)
	{
		maxOppPosCount = max;
	}

	void reset();
	void deleteMemory();
};

#endif
