// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <DefenseSystem.h>

#include <DefenseBack.h>
#include <NormalDefense.h>

#include <rcsc/player/world_model.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/audio_sensor.h>

#include <rcsc/common/basic_client.h>
#include <rcsc/common/logger.h>

#include <Share.h>
#include <Basics.h>
#include <FastIC.h>
#include <list>

using namespace rcsc;
using namespace std;

void DefenseSystem::execute(rcsc::PlayerAgent * _agent)
{
	dlog.addText(Logger::ROLE, "Entering DefenseSystem::execute");

	agent = _agent;
	wm = &agent->world();

	Share::markLookMode = false;

	// Calculating the defense line -----------------------------------------
	Share::defenseLine = min( (wm->ball().pos().x + 52.5) / 3.0 + 52.5, 0.0);

	vector<double> ballX;
	ballX.push_back(-25.0);
	ballX.push_back(-10.0);
	ballX.push_back(+05.0);
	ballX.push_back(+20.0);
	ballX.push_back(+35.0);
	ballX.push_back(+55.0);

	vector<double> defX;
	defX.push_back(-40.0);
	defX.push_back(-25.0);
	defX.push_back(-15.0);
	defX.push_back(-05.0);
	defX.push_back(-02.5);
	defX.push_back(+00.0);

	for (uint i = 0 ; i < min(ballX.size(), defX.size()) - 1 ; i++)
		if (wm->ball().pos().x >= ballX[i] and wm->ball().pos().x < ballX[i + 1])
			Share::defenseLine = (defX[i + 1] - defX[i]) * ((wm->ball().pos().x - ballX[i]) / (ballX[i + 1] - ballX[i])) + defX[i];

	dlog.addText(Logger::ROLE, "Ideal Defense Line X: %f", Share::defenseLine);
	// ----------------------------------------------------------------------

	determineRole();

	Vector2D homePos = positioningTarget(wm, mRole);

	uint markTarget = 0;

	dlog.addText(Logger::ROLE, "heared messages__");
	for (list<HearMessage>::const_iterator it = agent->audioSensor().teammateMessages().begin() ; it != agent->audioSensor().teammateMessages().end() ; it++)
		dlog.addText(Logger::ROLE, " %s [ %d ]", it->str_.c_str(), it->unum_);

	// determining the markTarget from markSayMsg
	char unumChar = Share::markSayMsg[ wm->self().unum() - 2 ];

	dlog.addText(Logger::ROLE, "Mark say: %s", Share::markSayMsg.c_str());

	if (unumChar == 'A')
		markTarget = 10;
	else if (unumChar == 'B')
		markTarget = 11;
	else
		markTarget = unumChar - '0';

	dlog.addText(Logger::ROLE, "HomePos: (%f, %f)", homePos.x, homePos.y);

	// Executing the role
	switch(mRole)
	{
		case DR_BACK:
			DefenseBack(homePos, markTarget).execute(agent);
			break;
		case DR_CENTER:
		case DR_TOP:
		case DR_BOTTOM:
		case DR_FRONT_CENTER:
		case DR_FRONT_TOP:
		case DR_FRONT_BOTTOM:
			NormalDefense(homePos, markTarget).execute(agent);
			break;
	}

	dlog.addText(Logger::ROLE, "Exiting DefenseSystem::execute");
}

void DefenseSystem::determineRole()
{
	switch(wm->self().unum())
	{
		case 2:	mRole = DR_BACK;	break;
		case 3:	mRole = DR_CENTER;	break;
		case 4:	mRole = DR_TOP;		break;
		case 5:	mRole = DR_BOTTOM;	break;
		case 6: mRole = DR_FRONT_CENTER;break;
		case 7: mRole = DR_FRONT_TOP;	break;
		case 8: mRole = DR_FRONT_BOTTOM;break;
	}
}

Vector2D DefenseSystem::positioningTarget(const WorldModel* wm, DefenseRole role)
{
	// DEBUG: Someone decide to use which of reachPos, ballPos or ballPos + ballVel here
	Vector2D ballPos = wm->ball().pos();
	// DEBUG ---------------------------

	Vector2D vec = ballPos - Vector2D(-52.5, 0.0);
	vec *= (3.0 / 4.0);
	vec.y /= 2.0;
	vec += Vector2D(-52.5, 0.0);

	// TOF 1: Defensa kheili jolo ya aghab naran
	vec.x = min_max(-45.0, vec.x, +5.0);

	double dist2Center = 15.0;
	double dist2Front = 5.0;

	Vector2D homePos;

	switch(role)
	{
		// Back defender positioning
		case DR_BACK:
			dist2Center = 1.5;
			if (vec.x < -32.0)
				dist2Center = 2.5;

			if (vec.x < -40.00)
				dist2Front = 5.0;
			else if (vec.x < -15.00)
				dist2Front = 8.0;
			else
				dist2Front = 10.0;

			homePos = Vector2D(	vec.x - dist2Front,
						vec.y + (vec.y > 0.0 ? -1 : +1) * dist2Center);

			if ( (wm->ball().pos().x > 20.0 and FastIC::isOurTeamBallPossessor()) or wm->ball().pos().x > 30.0 )
			{
				dist2Front = 7.5;
				homePos.x = max(vec.x - dist2Front, Basics::defenseLineX(wm) - dist2Front);
				dlog.addText(Logger::ROLE, "DEFENSE_LINE: %f", Basics::defenseLineX(wm));
				homePos.x = min(homePos.x, 0.0);
			}

			homePos.x = max(homePos.x, -45.5);

			break;

		// Center defender positioning
		case DR_CENTER:
			dist2Center = 1.5;
			if (vec.x < -32.0)
				dist2Center = 2.5;

			homePos = Vector2D(	vec.x,
						vec.y + (vec.y > 0.0 ? +1 : -1) * dist2Center);
			break;

		// Top defender positioning
		case DR_TOP:
			dist2Center = 15.0;
			if (vec.x > -5.0)
				dist2Center = 20.0;
			else if (vec.x < -32.0)
				dist2Center = 12.0;

			homePos = Vector2D(	vec.x,
						vec.y - dist2Center);
			break;

		// Bottom defender positioning
		case DR_BOTTOM:
			dist2Center = 15.0;
			if (vec.x > -5.0)
				dist2Center = 20.0;
			else if (vec.x < -32.0)
				dist2Center = 12.0;
			
			homePos = Vector2D(	vec.x,
						vec.y + dist2Center);
			break;

		// Center midfielder positioning
		case DR_FRONT_CENTER:
			dist2Center = 0.0;

			if (vec.x > -20.0)
				dist2Front = 7.0;
			else
				dist2Front = 4.0;

			homePos = Vector2D(	vec.x + dist2Front,
						vec.y - dist2Center);
			break;

		// Top midfielder positioning
		case DR_FRONT_TOP:
			dist2Center = 8.0;

			if (vec.x > -20.0)
				dist2Front = 7.0;
			else
				dist2Front = 4.0;

			homePos = Vector2D(	vec.x + dist2Front,
						vec.y - dist2Center);
			break;

		// Bottom midfielder positioning
		case DR_FRONT_BOTTOM:
			dist2Center = 8.0;

			if (vec.x > -20.0)
				dist2Front = 7.0;
			else
				dist2Front = 4.0;

			homePos = Vector2D(	vec.x + dist2Front,
						vec.y + dist2Center);
			break;
	}

	return homePos;
}
