// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <Goalie.h>

#include <bhv_goalie_basic_move.h>
#include <bhv_goalie_chase_ball.h>
#include <bhv_goalie_free_kick.h>

#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/intention_kick.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/view_wide.h>
	
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/world_model.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/basic_client.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/math_util.h>

#include <body_go_to_point2008.h>
#include <bhv_savior.h>

#include <Basics.h>
#include <SayMarkTable.h>
#include <Intercept.h>
#include <FastIC.h>

#include <Share.h>

using namespace std;
using namespace rcsc;

bool Goalie::isGuardExecuted = false;

void Goalie::execute(rcsc::PlayerAgent * _agent)
{
	agent = _agent;
	wm = &agent->world();


	// Setting the view and neck action 
	if (wm->ball().pos().x - wm->self().pos().x > 30.0)
	{
		agent->setViewAction( new View_Wide() );
		agent->setNeckAction( new Neck_TurnToPoint(Vector2D(52.5,0.0)) );
	}
	else
	{
		agent->setNeckAction( new Neck_TurnToBall() );
	}

	// Saying markTargets
/*	if (wm->ball().pos().x > -32.0 and wm->ball().pos().x < -22.0 and SayMarkTable().execute(agent))
		dlog.addText(Logger::ROLE, "Mark message said successfuly");
	else
	{
		dlog.addText(Logger::ROLE, "No mark message said(?)");
		agent->M_client->sendMessage(("(say " + Share::markSayMsg + ")").c_str());
		dlog.addText(Logger::ROLE, "Repeat say: %s", ("(say " + Share::markSayMsg + ")").c_str());
	}
*/	dlog.flush();


	static const Rect2D our_penalty( Vector2D(-60.0, -20 + EPS), Vector2D(-36.0 - EPS, +20 - EPS) );

	if (agent->world().time().cycle() > M_last_goalie_kick_time.cycle() + 1
		and agent->world().ball().distFromSelf() < agent->world().self().catchableArea() - 0.05 and our_penalty.contains(agent->world().ball().pos()))
	{
		rcsc::dlog.addText(rcsc::Logger::ROLE, "%s:%d: catchable. ball dist=%f, my_catchable=%f", __FILE__, __LINE__,
			agent->world().ball().distFromSelf(), agent->world().self().catchableArea());
		agent->doCatch();
	}
	else if (agent->world().self().isKickable())
	{
		doKick();
	}
	else
	{
		doMove();
	}
}

void Goalie::doKick()
{
	if (Bhv_Savior().doKick(agent))
		dlog.addText(Logger::PASS, "Savior kick");
	else
	{
		rcsc::Body_ClearBall().execute(agent);
		dlog.addText(Logger::PASS, "Clear !!!");
	}

	agent->setNeckAction(new rcsc::Neck_ScanField());
}

void Goalie::doMove()
{
	dlog.addText(Logger::ROLE, "Entering Goalie::doMove");

//	int self_cycle = agent->world().interceptTable()->selfReachCycle();
//	int teammate_cycle = agent->world().interceptTable()->teammateReachCycle();
//	int opponent_cycle = agent->world().interceptTable()->opponentReachCycle();

	Vector2D reachPoint = Intercept().finalIntercept(agent).getBallPos();

	double passingY = ( ((wm->ball().pos().x + 52.5) / (wm->ball().pos().x - reachPoint.x)) * (reachPoint.y - wm->ball().pos().y) ) + wm->ball().pos().y;

	dlog.addText(Logger::ROLE, "rp: %f, %f", Basics::getFastestReachPoint(&agent->world()).x, Basics::getFastestReachPoint(&agent->world()).y);
	dlog.addText(Logger::ROLE, "pY: %f", passingY);

	Vector2D nextBallPos = Basics::getBallPosAfterCycles(wm->ball(), 1, ServerParam::i().ballDecay());

	// TurnPoint -----------------------
	static Vector2D turnVec = Vector2D(0, 1);

	if (turnVec.y == 1 and wm->ball().pos().y > -3.0)
		turnVec.y = 1;
	else if (turnVec.y == -1 and wm->ball().pos().y < +3.0)
		turnVec.y =-1;
	else if (wm->ball().pos().y > 0.0)
		turnVec.y = 1;
	else
		turnVec.y =-1;

	Vector2D turnPoint = wm->self().pos() + turnVec;
	// ---------------------------------

	Vector2D homePos = positioningTarget();
	int dashPower = getDashPower(homePos);

	float goToPointThr = 0.5;
	if (wm->self().pos().dist(homePos) < goToPointThr)
		goToPointThr = 1.0;

	dlog.addText(Logger::ROLE, "IS SELF FASTEST:\t%d", FastIC::isSelfFastestPlayer());
	dlog.addText(Logger::ROLE, "ReachPoint: (%f, %f)", reachPoint.x, reachPoint.y);


	passingY = 100;
	if (wm->ball().vel().th().abs() > 90.0)
	{
		passingY = wm->ball().pos().y + AngleDeg(180 - wm->ball().vel().th().abs()).tan() * abs(wm->ball().pos().x - (-50.0)) *
			(wm->ball().vel().th().degree() > 90 ? +1 : -1);
	}

	// INTERCEPT -----------------------------
	if (FastIC::isSelfFastestPlayer() and reachPoint.absX() < 53.0 and reachPoint.absY() < 34.0)
	{
		dlog.addText(Logger::ROLE, "Goalie Intercepting");
		Intercept().execute(agent);
		isGuardExecuted = false;
		return;
	}
	// CHASE ---------------------------------
	else if (Bhv_GoalieChaseBall::is_ball_chase_situation(agent))
	{
		dlog.addText(Logger::ROLE, "Ball Chase");
		Bhv_GoalieChaseBall().execute(agent);
		isGuardExecuted = false;
		return;
	}

	// GUARD ---------------------------------
	if (doSideGuard())
	{
		dlog.addText(Logger::ROLE, "Side Guard");
		isGuardExecuted = true;
		return;
	}
	isGuardExecuted = false;

	Vector2D frontVec, backVec;
	frontVec.setPolar(5, wm->self().body());
	backVec.setPolar(5, wm->self().body() - 180);

	bool backMode = false;
	if ( (backVec + wm->self().pos()).dist(homePos) < (frontVec + wm->self().pos()).dist(homePos)
		and wm->ball().pos().x < -25.0 )
		backMode = true;

	// GO TO POINT ----------------------------
	if (Body_GoToPoint(homePos, goToPointThr, dashPower, 100, backMode, false).execute(agent))
	{
		dlog.addText(Logger::ROLE, "Going to target");
	}
	// TURN -----------------------------------
	else if (Body_TurnToPoint(turnPoint).execute(agent))
	{
		dlog.addText(Logger::ROLE, "Turning");
	}

	dlog.addText(Logger::ROLE, "Exiting Goalie::doMove");
}

rcsc::Vector2D Goalie::positioningTarget()
{
	Vector2D target;
	Vector2D ballPos = wm->ball().pos() + wm->ball().vel() * 2;
	if (wm->existKickableOpponent())
	{
		if (FastIC::getFastestOpponent())
			ballPos = FastIC::getFastestOpponent()->pos();
		else
			ballPos = wm->ball().pos();
	}

	double goalieLine = -50.0;
	if (wm->ball().pos().x > 0.0)
		goalieLine = -36.0;

	target.x = goalieLine;

	// Calculating target location
	float postY = 7.0;
	float d1 = (Vector2D(-52.5, +postY) - ballPos).r();
	float d2 = (Vector2D(-52.5, -postY) - ballPos).r();

	float a1 = (2 * postY * d1) / (d1 + d2);

	Vector2D dVec( ( Vector2D(-52.5, +postY) - ballPos ) + Vector2D(0, -a1)  );

	dVec *= ( (ballPos.x - target.x) / (52.5 + ballPos.x) );
	target = dVec + ballPos;

	target.x = max( min(goalieLine, wm->ball().pos().x), -51.8 );
	target.y = min_max(-7.0, target.y, +7.0);

	return target;
}

bool Goalie::doSideGuard()
{
	static double lastGuardPointX = -51.3;

	Vector2D guardPoint(-1000, -1000), turnPoint(-1000, -1000);

	Vector2D ballPos = wm->ball().pos() + wm->ball().vel() * 2;
	if (wm->existKickableOpponent())
	{
		if (FastIC::getFastestOpponent())
			ballPos = FastIC::getFastestOpponent()->pos();
		else
			ballPos = wm->ball().pos();
	}

	dlog.addText(Logger::ROLE, "ballPos: (%f, %f)", ballPos.x, ballPos.y);

	if ((ballPos.x < -46.5 or (isGuardExecuted and ballPos.x < -45.0)) and ballPos.absY() > 9.0)
	{
		guardPoint = Vector2D(-51.3, ballPos.y > 0.0 ? 7.0: -7.0);
		if (ballPos.x > -49.5 and (lastGuardPointX == -50.0 or ballPos.x > -49.0))
			guardPoint.x = -50.0;
		lastGuardPointX = guardPoint.x;

		turnPoint = wm->self().pos() + Vector2D(1, 0);
	}

	if (guardPoint.x == -1000)
		return false;

	float goToPointThr = 0.2;

	if (Body_GoToPoint2008(guardPoint, goToPointThr, 100).execute(agent))
		dlog.addText(Logger::ROLE, "Going to GuardPos");
	else if (Body_TurnToPoint(turnPoint).execute(agent))
		dlog.addText(Logger::ROLE, "Turning (guard)");

	return true;
}

int Goalie::getDashPower(Vector2D homePos)
{
	int dashPower = 100;

	if (wm->ball().pos().x > -30.0)
		dashPower = 70;

	if (homePos.x > -40.0)
		dashPower = 40;

	if (homePos.x < wm->self().pos().x - 2.0)
		dashPower = 100;

	return dashPower;
}
