// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <NormalDefense.h>
/*
*/
#include <DefenseKickAction.h>
#include <OffensiveKick.h>

#include <body_go_to_point2008.h>

#include <Basics.h>
#include <OffensiveKick.h>
#include <body_clear_ball2008.h>

#include <rcsc/action/body_turn_to_point.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_relative.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/audio_sensor.h>

#include <rcsc/common/basic_client.h>
#include <rcsc/common/logger.h>

#include <rcsc/player/intercept_table.h>

#include <LookToPlayerAndBall.h>
#include <Intercept.h>
#include <Share.h>
#include <FastIC.h>

#include <DangerAreaClear.h>
#include <BlockIntercept.h>
#include <Tackle.h>
#include <Mark.h>

using namespace std;
using namespace rcsc;
using namespace Basics;

void NormalDefense::execute(rcsc::PlayerAgent * agent)
{
	static int scan = 0;
	Share::isScanning = false;

	dlog.addText(Logger::ROLE, "Entering NormalDefense::execute -----");
	wm = &agent->world();

	for (uint i = 0 ; i < wm->teammatesFromSelf().size() ; i++)
		dlog.addText(Logger::ROLE, "TMM %d (%f, %f)", wm->teammatesFromSelf()[i]->unum(), wm->teammatesFromSelf()[i]->pos().x,
								wm->teammatesFromSelf()[i]->pos().y);

	dlog.addText(Logger::ROLE, "BallPos: (%f, %f)", wm->ball().pos().x, wm->ball().pos().y);

        agent->setNeckAction( new Neck_TurnToBallOrScan() );

	// Attention to the commander
	int commander = wm->ball().pos().x > -21.0 ? 2 : 1 ;

	if (agent->doAttentionto(agent->world().ourSide(), commander))
	{
		dlog.addText(Logger::ROLE, "Attention to our %d", commander);
	}

	// markSayMsg Jangoolers
	{
		string sayMsg = "";
		for (list<HearMessage>::const_iterator it = agent->audioSensor().teammateMessages().begin() ; it != agent->audioSensor().teammateMessages().end() ; it++)
			if (it->str_.c_str()[0] == 'M' and it->str_.size() == 10)
				sayMsg = it->str_;

		if (sayMsg != "" and sayMsg != Share::markSayMsg)
		{
			Share::markSayMsg = sayMsg;
		}

	}

	// Kick decision
	bool kickable = wm->self().isKickable();

	dlog.addText(Logger::ROLE, "KICKABLE: %d", kickable);
	if (wm->existKickableTeammate())
	{
		if (wm->teammatesFromBall().front()->distFromBall() < wm->ball().distFromSelf() - 0.5)
			kickable = false;
		if (wm->teammatesFromBall().front()->unum() < wm->self().unum())
			kickable = false;
	}

	dlog.addText(Logger::ROLE, "KICKABLE: %d", kickable);

	if (kickable)
	{
		doKick(agent);
		Share::isLastActionIntercept = false;
		return;
	}
	// -----------------------

	Mark mark(agent, markTarget);
	bool markValue = mark.getValue();

	dlog.addText(Logger::ROLE, "Ball->Vel:\t (%f, %f), %f", wm->ball().vel().x, wm->ball().vel().y, wm->ball().vel().r());

//	int selfReachCycle = wm->interceptTable()->selfReachCycle();
	Vector2D selfReachPoint = Intercept().finalIntercept(agent).getBallPos();

	dlog.addText(Logger::ROLE, "SelfReachCycle by intercept: %d", Intercept().finalIntercept(agent).getReachCycle());
	dlog.addText(Logger::ROLE, "SelfReachCycle by fastIC: %d", FastIC::getSelfReachCycle());
	dlog.addText(Logger::ROLE, "SelfReachCycle by interceptTable: %d", wm->interceptTable()->selfReachCycle());

//	if (selfReachCycle < 10)
//		selfReachCycle = Intercept().finalIntercept(agent).getReachCycle();



	if (FastIC::getFastestOpponent())
		dlog.addText(Logger::ROLE, "Fastest Opponent: %d", FastIC::getFastestOpponent()->unum());
	if (wm->interceptTable()->fastestOpponent())
		dlog.addText(Logger::ROLE, "Fastest Opponent (interceptTABle): %d", wm->interceptTable()->fastestOpponent()->unum());

	dlog.addText(Logger::ROLE, "Is kickable opp: %d", wm->existKickableOpponent());

	dlog.addText(Logger::ROLE, "isSelfFastest: %d", FastIC::isSelfFastestPlayer());
	dlog.addText(Logger::ROLE, "isCelfFastest: %d", wm->interceptTable()->isSelfFastestPlayer());

	if (FastIC::getGlobalFastest())
		dlog.addText(Logger::ROLE, "Fastest Player: %d", FastIC::getGlobalFastest()->unum());

	const PlayerObject* fastestOpp = wm->interceptTable()->fastestOpponent();
	int oppReachCycle = wm->interceptTable()->opponentReachCycle();

//	int selfReachCycle = Intercept().finalIntercept(agent).getReachCycle();
	Vector2D ballReachPos = wm->ball().inertiaPoint(FastIC::getGlobalFastestReachCycle());
	dlog.addText(Logger::ROLE, "Ball Reach Pos: (%f, %f)", ballReachPos.x, ballReachPos.y);

	int selfReachCycle = FastIC::getSelfReachCycle();
	if (ballReachPos.x < -34.0 and ballReachPos.absY() < 12.0)
		selfReachCycle = Intercept().finalIntercept(agent).getReachCycle() - 1;

	dlog.addText(Logger::ROLE, "self_reach_cycle %d *", selfReachCycle);
	dlog.addText(Logger::ROLE, "self reach cycle %d", Intercept().finalIntercept(agent).getReachCycle());
	dlog.addText(Logger::ROLE, "opp_reach_cycle %d", oppReachCycle);

	// Tackle
	if (Tackle().execute(agent))
	{
		dlog.addText(Logger::ROLE, "Tackle executed !!!");
		Share::isLastActionIntercept = false;
		return;
	}


	// Interception
	if
	(
/*		FastIC::isSelfFastestPlayer()//wm->interceptTable()->isSelfFastestPlayer()
		or
*/		(markValue and fastestOpp and
		mark.markTarget->unum() == fastestOpp->unum() and
		abs(selfReachCycle - oppReachCycle) < 3 and
		selfReachCycle > 4 and oppReachCycle > 4)
		or

		(selfReachCycle <= oppReachCycle and FastIC::isSelfFastestTeamMate())
		or

		// vaghti opp az baghala course mindaze intercept konim ===========
		(markValue and fastestOpp and
		mark.markTarget->unum() == fastestOpp->unum() and
		abs(selfReachCycle - oppReachCycle) < 4 and
		oppReachCycle >= 1 and
		mark.markTarget->pos().absY() > 15.0 and
		selfReachPoint.x < wm->self().pos().x)
		// ================================================================
	)
	{
		Share::checkInterceptSituation = true;
		agent->setNeckAction(new Neck_TurnToBallOrScan() );

		Intercept().execute(agent);

		dlog.addText(Logger::ROLE, "Defense Intercepting !!!");
		Share::isLastActionIntercept = true;
		return;
	}
	Share::isLastActionIntercept = false;

	if (FastIC::getFastestOpponent())
		dlog.addText(Logger::ROLE, "Opp fastest: %d", FastIC::getFastestOpponent()->unum());

	dlog.addText(Logger::ROLE, "F_REACH_POINT: (%f, %f)", getFastestReachPoint(wm).x, getFastestReachPoint(wm).y);

	// Blocking
	if
	(
		(fastestOpp and
		markTarget == fastestOpp->unum() and !FastIC::isOurTeamBallPossessor())
		or
		(!markValue and amIFastestDefender(agent, "23456789AB", "1"))
		or
		(amIFastestDefender(agent, "23456789AB", "1") and 
		markValue and mark.markTarget->pos().dist(getFastestReachPoint(wm)/*wm->ball().pos()*/) < 6.0 and
		abs(mark.markTarget->pos().y - wm->ball().pos().y) < 2.0)
	)
	{
		Share::checkInterceptSituation = true;

		if (Block().execute(agent))
			dlog.addText(Logger::ROLE, "Defense Blocking");
		else
		{
			Intercept().execute(agent);
			dlog.addText(Logger::ROLE, "Block false value: intercepting");
		}

		return;
	}

	Share::checkInterceptSituation = false;

	if (markValue)
	{
		mark.execute();
		LookToPlayerAndBall(mark.markTarget).execute(agent);
		scan = 0;
	}
	else if (mark.markTarget == NULL and markTarget > 0 and markTarget < 12 and scan < 15)
	{
		dlog.addText(Logger::ROLE, "Scanning");
		agent->doTurn(30.0);
		agent->doChangeView(rcsc::ViewWidth::NARROW);
		agent->setNeckAction( new Neck_TurnToRelative(90.0) );
		scan++;
		Share::isScanning = true;
	}
	else if (Body_GoToPoint2008(homePos, 1.0, dashPower(wm, homePos)).execute(agent))
	{
		dlog.addText(Logger::ROLE, "Going to HomePos (I have nothing special to do)");
	}
	else
	{
		double theta = (wm->ball().pos() - Vector2D(-52.5, 0.0)).dir().degree() ;
		Vector2D turnPoint = wm->self().pos() + Vector2D::polar2vector(10.0, AngleDeg(theta + (theta > 0 ? +90 : -90)));

		if (wm->self().pos().absY() > 20.0)
			turnPoint = wm->self().pos() + Vector2D::polar2vector(10.0, AngleDeg(theta + (theta > 0 ? -90 : +90)));

		Body_TurnToPoint(turnPoint).execute(agent);
		dlog.addText(Logger::ROLE, "Changing the body direction");
	}

	dlog.addText(Logger::ROLE, "Exiting NormalDefense::execute ------");
}

void NormalDefense::doKick(rcsc::PlayerAgent * agent)
{
	dlog.addText(Logger::ROLE, "Entering NormalDefense::doKick");

	if (DangerAreaClear().execute(agent))
	{
		dlog.addText(Logger::ROLE, "Danger Area Clear");
		return;
	}

	double * dist = new double;
	if (wm->getOpponentNearestTo(wm->ball().pos(), 5, dist) and *dist < (wm->ball().pos().x < -32.0 ? 3.0 : 2.0) and wm->ball().pos().x < 10.0)
	{
		Body_ClearBall2008().execute(agent);
		dlog.addText(Logger::ROLE, "clearing the ball!!!");
		return;
	}

	if (wm->ball().pos().x > 0.0)
		OffensiveKick().execute(agent);
	else if (wm->self().unum() == 3)
		DefenseKickAction().CenterBack(agent);
	else if (isInPlayers(wm->self().unum(), "45"))
		DefenseKickAction().SideBack(agent);
	else
		DefenseKickAction().OffensiveHalf(agent);

	dlog.addText(Logger::ROLE, "Exiting NormalDefense::doKick");
}

void NormalDefense::doMove(rcsc::PlayerAgent * agent)
{
	dlog.addText(Logger::ROLE, "Entering NormalDefense::doMove");

	dlog.addText(Logger::ROLE, "Exiting NormalDefense::doMove");
}

int NormalDefense::dashPower(const rcsc::WorldModel* wm, rcsc::Vector2D homePos)
{
	int dash_power = 100;
	if (homePos.x > wm->self().pos().x)
		dash_power = 80;

	return dash_power;
}
