// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <NormalOffense.h>
/*
*/
#include <body_go_to_point2008.h>

#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_turn_to_angle.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/audio_memory.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_codec.h>


#include <OffensiveKick.h>
#include <Intercept.h>
#include <Basics.h>
#include <NewPositioningSystem.h>
#include <BreakerMove.h>
#include <PreDecidedPassBase.h>
#include <FastIC.h>
#include <OffenseVariables.h>

#include <strategy.h>

#include <BlockIntercept.h>

#include <OffenseKickAction.h>

using namespace Basics;
using namespace std;
using namespace rcsc;

int NormalOffense::lastSelfSayCycle = -6;

void NormalOffense::execute(rcsc::PlayerAgent * agent)
{
	wm = &agent->world();

	agent->setNeckAction( new Neck_TurnToBallOrScan() );

	// Kick decision
	bool kickable = wm->self().isKickable();

	if (wm->existKickableTeammate())
	{
		if (wm->teammatesFromBall().front()->distFromBall() < wm->ball().distFromSelf() - 0.5)
			kickable = false;
		if (wm->teammatesFromBall().front()->unum() < wm->self().unum())
			kickable = false;
	}

	if (kickable)
	{
		doKick(agent);
		if ( ( OffenseVariables::Instance()->quarterBackPlayer().size() ) and ( not Basics::isInPlayers( wm->self().unum() , OffenseVariables::Instance()->quarterBackPlayer() ) ) )
			agent->doAttentionto(agent->world().ourSide(), AudioCodec::i().hex2unum( OffenseVariables::Instance()->quarterBackPlayer()[0] ) );
		return;
	}
	FastIC fast;
	// TODO: No.6 position say
	// Set attention for LA
	int talkative = AudioCodec::i().hex2unum( OffenseVariables::Instance()->quarterBackPlayer()[0] );
	Vector2D ballPos = wm->ball().pos();
	if ( FastIC::getFastestTeammate() )
	{
		if ( FastIC::getFastestTeammateReachCycle() < 2 )
			talkative = FastIC::getFastestTeammate()->unum();
		ballPos = inertia_n_step_point( ballPos , wm->ball().vel() , FastIC::getFastestTeammateReachCycle() , ServerParam::i().ballDecay() );
	}
	if ( FastIC::isSelfFastestPlayer() )
	{
		if ( not Basics::isInPlayers( wm->self().unum() , OffenseVariables::Instance()->quarterBackPlayer() ) )
		{
			if ( OffenseVariables::Instance()->quarterBackPlayer().size() )
				talkative = AudioCodec::i().hex2unum( OffenseVariables::Instance()->quarterBackPlayer()[0] );
			else
				talkative = 6;
		}
		else
		{
			talkative = 2;
		}
	}
	else if ( talkative == wm->self().unum() )
	{
		if ( Basics::isInPlayers( wm->self().unum() , OffenseVariables::Instance()->quarterBackPlayer() ) )
		{
			talkative = 11;
			if ( OffenseVariables::Instance()->linearPlayers().size() > 1 )
				talkative = AudioCodec::i().hex2unum ( OffenseVariables::Instance()->linearPlayers()[1] );
		}
		else
		{
			talkative = 6;
			if ( OffenseVariables::Instance()->quarterBackPlayer().size() )
				talkative = AudioCodec::i().hex2unum ( OffenseVariables::Instance()->quarterBackPlayer()[0] );
		}
	}
	agent->doAttentionto(agent->world().ourSide(), talkative );
	dlog.addText ( Logger::PASS , "Attentioning to %d" , talkative );
	dlog.addText ( Logger::PASS , "ballPos ( %.2f , %.2f )" , wm->ball().pos().x , wm->ball().pos().y );
	bool isBallInOffense = Basics::isInPlayers( talkative , OffenseVariables::Instance()->linearPlayers() + OffenseVariables::Instance()->quarterBackPlayer() );
	qBPositionSay( agent );
	if ( BreakerMove().shouldExecute( agent->world() ) )
	{
		BreakerMove().execute( agent );
		return;
	}

	PreDecidedPassBase passBase;
	if ( passBase.shouldCommunicate(agent) )
		passBase.communicate ( agent );
	// Interception
	if ( fast.isSelfFastestPlayer() )
	{
		OffenseVariables::Instance()->editLinearPlayers( agent );
		Intercept().execute( agent );
		return;
	}

	if ( ( not isBallInOffense ) and ( agent->world().ball().pos().x < 0.f ) )
	{
		if ( OffenseVariables::Instance()->linearPlayers().size() > 2 )
			if ( agent->world().self().unum() == AudioCodec::i().hex2unum( OffenseVariables::Instance()->linearPlayers()[1] ) )
				OffenseVariables::Instance()->editLinearPlayers( agent );
	}
	// Blocking
	if ( fast.isSelfFastestTeamMate() )
	{
		if (Block().execute(agent))
			dlog.addText(Logger::ROLE, "Defense Blocking");
		else
		{
			Intercept().execute(agent);
			dlog.addText(Logger::ROLE, "Block false value: intercepting");
		}
		return;
	}

	doPositioning( agent , homePos , isBallInOffense );
	agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
	dlog.addText(Logger::ROLE, "Going to HomePos (I have nothing special to do)");

	dlog.addText(Logger::ROLE, "Exiting NormalOffense::execute ------");
}

void NormalOffense::doKick(rcsc::PlayerAgent * agent)
{
	dlog.addText(Logger::ROLE, "Entering NormalOffense::doKick");

	if (wm->ball().pos().x < -20.0)
	{
		if (isInPlayers(wm->self().unum(), "9A"))
			OffenseKickAction().SideForward(agent);
		else if (wm->self().unum() == 11)
			OffenseKickAction().CenterForward(agent);
		else
			OffenseKickAction().DefensiveHalf(agent);
	}
	else
		OffensiveKick().execute(agent);

	dlog.addText(Logger::ROLE, "Exiting NormalOffense::doKick");
}

void NormalOffense::doMove(rcsc::PlayerAgent * agent)
{
	dlog.addText(Logger::ROLE, "Entering NormalOffense::doMove");

	dlog.addText(Logger::ROLE, "Exiting NormalOffense::doMove");
}

void NormalOffense::doPositioning(rcsc::PlayerAgent * agent , rcsc::Vector2D homePos , bool isBallInOffense )
{
	float distThr = 1.f;
	Vector2D dashTarget;
	bool changed = NewPositioningSystem::getConstantPoint( homePos , dashTarget );
	int dash_power = getDashPower();
	if ( agent->world().self().pos().x > agent->world().offsideLineX() - 1.f )
	{
		dlog.addText ( Logger::PASS , "Running out!!!" );
		dash_power = agent->world().self().getSafetyDashPower(100);
		dashTarget = Vector2D ( agent->world().offsideLineX() - 2.f , agent->world().self().pos().y);
	}
	dlog.addText ( Logger::PASS , "dashTarget ( %.2f , %.2f )" , dashTarget.x , dashTarget.y );
	if ( Body_GoToPoint( dashTarget , distThr , dash_power , 100,/*backMode*/ false, true, 15.0).execute(agent) )
		return;
	dlog.addText ( Logger::PASS , "not dashed!!!!" );
	float targetAngle = 0.f;
	const WorldModel & wm = agent->world();
	float ballAngle = ( wm.ball().pos() - wm.self().pos() ).th().degree();
	if ( FastIC::isOurTeamBallPossessor() )
	{
		if ( isBallInOffense || ( agent->world().ball().pos().x > -20.f ) )
			targetAngle = 0.f;
		else
		{
			targetAngle = ballAngle;
		}
	}
	else
	{
		targetAngle = ballAngle;
	}
	Body_TurnToAngle( targetAngle ).execute( agent );
}

void NormalOffense::doCommunicationWorks(rcsc::PlayerAgent * agent , bool isKickable)
{
}

void NormalOffense::qBPositionSay(rcsc::PlayerAgent * agent )
{
	if ( not Basics::isInPlayers( agent->world().self().unum() , OffenseVariables::Instance()->quarterBackPlayer() ) )
		return;
	vector< Vector2D > defensePositions;
	vector< int > defenseUnums;
	vector< float > defenseSayValues;
	const WorldModel & wm = agent->world();
	{
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
				it != end;
				++it )
		{
			int hisUnum = (*it)->unum();
			if ( hisUnum < 0 )
				continue;
			if ( (*it)->isKickable() )
				continue;
			if ( wm.opponentGoalieUnum() > 0 )
				if ( hisUnum == wm.opponentGoalieUnum() )
					continue;
			if ( (*it)->posCount() != 0 )
				continue;
			Vector2D hisPos = (*it)->pos();
			if ( hisPos.x < wm.self().pos().x - 10.f )
				continue;
			defensePositions.push_back ( hisPos );
			defenseUnums.push_back( hisUnum );
			defenseSayValues.push_back( 1000 - hisPos.dist( wm.ball().pos() ) );
//			dlog.addText ( Logger::PASS , "player %d in ( %.2f , %.2f )" , hisUnum , hisPos.x , hisPos.y );
		}
	}
//	dlog.addText( Logger::PASS,  "players\t%d" , defensePositions.size() );
/*	for ( unsigned int i = 0; i < defenseSayValues.size(); i++ )
		dlog.addText ( Logger::PASS , "player %d ----> value %f" , defenseUnums[i] , defenseSayValues[i] );
*/	bool changed = true;
	while ( ( changed ) and ( defenseSayValues.size() > 1 ) )
	{
		changed = false;
		for ( unsigned int i = 0; i < defenseSayValues.size() - 1; i++ )
		{
			if ( defenseSayValues[i] < defenseSayValues[i+1] )
			{
				float helpValue = defenseSayValues[i];
				defenseSayValues[i] = defenseSayValues[i+1];
				defenseSayValues[i+1] = helpValue;

				Vector2D helpPos = defensePositions[i];
				defensePositions[i] = defensePositions[i+1];
				defensePositions[i+1] = helpPos;

				int helpUnum = defenseUnums[i];
				defenseUnums[i] = defenseUnums[i+1];
				defenseUnums[i+1] = helpUnum;
				changed = true;
			}
		}
	}
//	dlog.addText ( Logger::PASS , "SORTED!!" );
/*	for ( unsigned int i = 0; i < defenseSayValues.size(); i++ )
		dlog.addText ( Logger::PASS , "player %d ----> value %f" , defenseUnums[i] , defenseSayValues[i] );
*/	vector< Vector2D > offensePositions;
	vector< int > offenseUnums;
	vector< float > offenseSayValues;
	{
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
				it != end;
				++it )
		{
			int hisUnum = (*it)->unum();
			Vector2D hisPos = (*it)->pos();
			if ( (*it)->posCount() != 0 )
				continue;
			if ( (*it)->isKickable() )
				continue;
			if ( hisUnum < 0 )
				continue;
			if  ( not Basics::isInPlayers( hisUnum , OffenseVariables::Instance()->linearPlayers() ) )
				continue;
			offenseUnums.push_back( hisUnum );
			offensePositions.push_back( hisPos );
			offenseSayValues.push_back( 1000 - hisPos.dist( wm.ball().pos() ) );
//			dlog.addText ( Logger::PASS , "player %d in ( %.2f , %.2f )" , hisUnum , hisPos.x , hisPos.y );
		}
	}
	changed = true;
	while ( ( changed ) and ( offenseSayValues.size() > 1 ) )
	{
		changed = false;
		for ( unsigned int i = 0; i < offenseSayValues.size() - 1; i++ )
		{
			if ( offenseSayValues[i] < offenseSayValues[i+1] )
			{
				float helpValue = offenseSayValues[i];
				offenseSayValues[i] = offenseSayValues[i+1];
				offenseSayValues[i+1] = helpValue;

				Vector2D helpPos = offensePositions[i];
				offensePositions[i] = offensePositions[i+1];
				offensePositions[i+1] = helpPos;

				int helpUnum = offenseUnums[i];
				offenseUnums[i] = offenseUnums[i+1];
				offenseUnums[i+1] = helpUnum;
				changed = true;
			}
		}
	}
	vector<WindowPlayerToSend> playersToBeSent;
	for ( int i = 0; i < 4; i++ )
		playersToBeSent.push_back( WindowPlayerToSend( 10 , false , Vector2D( 0.f , 0.f ) , false ) );
	int fillingBlock = 0;
	if ( wm.time().cycle() - lastSelfSayCycle >= 2 )
	{
		playersToBeSent[fillingBlock] = WindowPlayerToSend( wm.self().unum() , true , wm.self().pos() , true );
		fillingBlock ++;
		unsigned int opponentBlock = 0;
		while ( ( opponentBlock < defenseSayValues.size() ) and ( fillingBlock < 4 ) )
		{
			playersToBeSent[fillingBlock] = WindowPlayerToSend( defenseUnums[opponentBlock] , false , defensePositions[opponentBlock] , true );
			opponentBlock++;
			fillingBlock++;
		}
		unsigned int teammateBlock = 0;
		while ( ( teammateBlock < offenseSayValues.size() ) and ( fillingBlock < 4 ) )
		{
			playersToBeSent[fillingBlock] = WindowPlayerToSend( offenseUnums[teammateBlock] , true , offensePositions[teammateBlock] , true );
			teammateBlock++;
			fillingBlock++;
		}
		agent->addSayMessage( new FourPlayerMessage( playersToBeSent[0] , playersToBeSent[1] , playersToBeSent[2] , playersToBeSent[3] ) );
		for ( unsigned int i = 0; i < 4; i ++ )
			dlog.addText ( Logger::PASS , "aying %d" , playersToBeSent[i].uniNum  );
	}
	else
	{
		unsigned int teammateBlock = 0;
		while ( ( teammateBlock < offenseSayValues.size() ) and ( fillingBlock < 1 ) )
		{
			playersToBeSent[fillingBlock] = WindowPlayerToSend( offenseUnums[teammateBlock] , true , offensePositions[teammateBlock] , true );
			teammateBlock++;
			fillingBlock++;
		}
		unsigned int opponentBlock = 0;
		while ( ( opponentBlock < defenseSayValues.size() ) and ( fillingBlock < 4 ) )
		{
			playersToBeSent[fillingBlock] = WindowPlayerToSend( defenseUnums[opponentBlock] , false , defensePositions[opponentBlock] , true );
			opponentBlock++;
			fillingBlock++;
		}
		while ( ( teammateBlock < offenseSayValues.size() ) and ( fillingBlock < 4 ) )
		{
			playersToBeSent[fillingBlock] = WindowPlayerToSend( offenseUnums[teammateBlock] , true , offensePositions[teammateBlock] , true );
			teammateBlock++;
			fillingBlock++;	
		}
		agent->addSayMessage( new FourPlayerMessage( playersToBeSent[0] , playersToBeSent[1] , playersToBeSent[2] , playersToBeSent[3] ) );
		for ( unsigned int i = 0; i < 4; i ++ )
			dlog.addText ( Logger::PASS , "saying %d" , playersToBeSent[i].uniNum  );
	}
}

int NormalOffense::getDashPower()
{
	int dashPower = 100;

	if (wm->self().pos().dist(homePos) < 4.0)
		dashPower = (int)min_max(50.0, wm->self().pos().dist(homePos) * 25, 100.0);

	if (homePos.x > wm->self().pos().x + 2.0)
		dashPower = 100;

	if (homePos.x < wm->self().pos().x - 2.0 and wm->self().pos().x < wm->offsideLineX())
		dashPower = 30;

	if (homePos.x > 36.0)
		dashPower = 100;

	return wm->self().getSafetyDashPower(dashPower);
}
