// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <OffenseSystem.h>
#include <GlobalPositioning.h>

#include <body_go_to_point2008.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/common/logger.h>

#include <PenaltyAreaPositioning.h>
#include <NormalOffense.h>
#include <NewPositioningSystem.h>
#include <OffenseVariables.h>
#include <strategy.h>

#include <GoalPlanPositioning.h>
#include <PrimaryGoalPlanPositioning.h>
#include <FastIC.h>

using namespace std;
using namespace rcsc;


void OffenseSystem::execute(PlayerAgent* _agent)
{
	dlog.addText(Logger::ROLE, "Entering OffenseSystem::execute");

	agent = _agent;
	wm = &agent->world();

	determineRole();

	OffenseVariables::Instance()->updateBySay( agent->world() );
//	OffenseVariables::Instance()->changeVariablesIfNeeded( agent->world().ball().pos() );

	Vector2D homePos = positioningTarget(wm, mRole);

	dlog.addText(Logger::ROLE, "HomePos: (%f, %f)", homePos.x, homePos.y);

	// Executing the role
	switch(mRole)
	{
		case OR_CENTER:
		case OR_TOP:
		case OR_BOTTOM:
		case OR_BACK:
			NormalOffense(homePos).execute(agent);
//			Body_GoToPoint2008(homePos, 1.0, 100).execute(agent);
			break;
	}

	dlog.addText(Logger::ROLE, "Exiting OffenseSystem::execute");
}

void OffenseSystem::determineRole()
{
	switch(wm->self().unum())
	{
		case 6:	mRole = OR_BACK;	break;
		case 9:	mRole = OR_TOP;		break;
		case 10:mRole = OR_BOTTOM;	break;
		case 11:mRole = OR_CENTER;	break;
	}
}

Vector2D OffenseSystem::positioningTarget(const WorldModel* wm, OffenseRole role)
{
	bool shouldUseGoalPlanPositioning = false;
	Vector2D ballPos = wm->ball().pos();
	if ( FastIC::getGlobalFastest() )
		ballPos = inertia_n_step_point( ballPos , wm->ball().vel() , FastIC::getGlobalFastestReachCycle() , ServerParam::i().ballDecay() );
	if ( ballPos.x > 36.f )
		shouldUseGoalPlanPositioning = true;
	if ( shouldUseGoalPlanPositioning )
	{
		string players = "96BA";

		dlog.addText(Logger::ROLE, "using PenaltyAreaPositioning");

		PrimaryGoalPlanPositioning positioning;
		return positioning.getHomePos( wm );		
	}

	NewPositioningSystem pos;
	return pos.getHomePos(wm);
}
