/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Ahmad Boorghany
 *  and is modified by: Mohammad Salehe, Sassan Haradji, ESKILAS Team
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#include <Basics.h>
#include <cmath>
#include <cassert>
#include <stdio.h>
#include <string.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/player/intercept_table.h>

#include <Expression.h>
#include <Intercept.h>
#include <Share.h>

using namespace std;
using namespace rcsc;

namespace Basics
{
	bool isInPlayers(const unsigned num, std::string str)
	{
		char charNum = num + '0';
		if (num >= 10) charNum = num - 10 + 'A';
		for (unsigned i = 0; i < str.size(); i++)
			if (str[i] == charNum) return 1;
		return 0;
	}

	unsigned charToUniNum(const char str)
	{
		unsigned uniNum = str - '0';
		if (str >= 'A') uniNum = str + 10 - 'A';
		return uniNum;
	}

	char uniNumToChar(const unsigned num)
	{
		char charNum = num + '0';
		if (num >= 10) charNum = num - 10 + 'A';
		return charNum;
	}
	
	unsigned firstPlayerInList(const string list, const string blackList)
	{
		for (unsigned i = 0; i < list.size(); i++)
			if (blackList.find(list[i]) == string::npos)
				return charToUniNum(list[i]);
		return 0;
	}

	float reRate(float value, float a, float b, float c,
			float d, ReRateMode mode)
	{
		if (a == b)
			return c;

		if (mode == RRM_DIRECT)
		{
			if (value < a)
				return c;
			if (value > b)
				return d;
			return (value - a) / (b - a) * (d - c) + c;
		}
		else
		{
			if (value < a)
				return d;
			if (value > b)
				return c;
			return (value - a) / (b - a) * (c - d) + d;
		}
	}

	float sigmoid(float x)
	{
		return (float)(1 / (1 + exp(-x)));
	}

	unsigned changeIntBase(int number, int base, string &str)
	{
		unsigned i = 0;

		for (i = 0; number >= base; i++)
		{
			str += (char)(number % base) ;
			number = (int)((float)number / base);
		}

		str += (char)number ;
		return i + 1;
	}

	unsigned baghimande(int number, unsigned base)
	{
		if (number < (int)base)
		{
			while (number <= (int)base)
				number += base;
			return (unsigned)(number - base);
		}

		while (number >= (int)base)
			number -= base;
		return (unsigned)number;
	}
	
	int sign(float number)
	{
		if (number > 0) return 1;
		else if (number < 0) return -1;
		return 0;
	}

	unsigned pairPlayer(unsigned number, unsigned side)
	{
		switch (number)
		{
		case 6:
			if (side == 0)
				return 9;
			else if (side == 1)
				return 6;
			else if (side == 2)
				return 9;
			assert(1);
		case 9:
			if (side == 0)
				return 6;
			else if (side == 1)
				return 6;
			else if (side == 2)
				return 9;
			assert(1);
		case 7:
			if (side == 0)
				return 8;
			else if (side == 1)
				return 7;
			else if (side == 2)
				return 8;
			assert(1);
		case 8:
			if (side == 0)
				return 7;
			else if (side == 1)
				return 7;
			else if (side == 2)
				return 8;
			assert(1);
		case 2:
			if (side == 0)
				return 4;
			else if (side == 1)
				return 2;
			else if (side == 2)
				return 4;
			assert(1);
		case 4:
			if (side == 0)
				return 2;
			else if (side == 1)
				return 2;
			else if (side == 2)
				return 4;
			assert(1);
		case 10:
			if (side == 0)
				return 11;
			else if (side == 1)
				return 10;
			else if (side == 2)
				return 11;
			assert(1);
		case 11:
			if (side == 0)
				return 10;
			else if (side == 1)
				return 10;
			else if (side == 2)
				return 11;
			assert(1);
		}
		return number;
	}

	string pairPlayers(string numbers, unsigned side)
	{
		string pairNumbers("");
		for (unsigned i = 0; i < numbers.size(); i++)
			pairNumbers += uniNumToChar(pairPlayer(charToUniNum(numbers[i])));
		return pairNumbers;
	}
	
	rcsc::Vector2D getBallPosAfterCycles(const rcsc::Vector2D initPos ,float rate ,float direction ,float cycles ,float ballDecay)
	{
		float qN = pow(ballDecay , cycles);
		float dist = (rate * (1 - qN) ) / ( 1 - ballDecay );
		rcsc::Vector2D ballRPos = rcsc::Vector2D::polar2vector( dist , direction );
		rcsc::Vector2D ballRes = initPos + ballRPos;
		return ballRes;
	}

	rcsc::Vector2D getBallPosAfterCycles(const rcsc::BallObject& ball ,float cycles ,float ballDecay)
	{
		return getBallPosAfterCycles(ball.pos(), (float)ball.vel().r(), ball.vel().dir().degree(), cycles, ballDecay);
	}

	rcsc::Vector2D getFastestReachPoint(const rcsc::WorldModel* wm)
	{
		return getBallPosAfterCycles(wm->ball(),
			min(wm->interceptTable()->teammateReachCycle(), wm->interceptTable()->opponentReachCycle()),
			rcsc::ServerParam::i().ballDecay());
	}

	float getDeltaAngle(float ang1,float ang2)
	{
		while ( ang1 < 0 )
			ang1 += 360.f;
		while ( ang2 < 0 )
			ang2 += 360.f;
		while ( ang1 > 360.f )
			ang1 -= 360.f;
		while ( ang2 > 360.f )
			ang2 -= 360.f;
		float delta = fabs(ang1 - ang2);
		if ( delta > 180.f )
			delta = 360.f - delta;
		return delta;
	}

	bool amIFastestDefender(rcsc::PlayerAgent * agent, string list, string blackList)
	{
		const rcsc::WorldModel* wm = &agent->world();

		if (isInPlayers(wm->self().unum(), blackList))
			return false;

		rcsc::Vector2D reachPoint = Basics::getFastestReachPoint(wm);
//		dlog.addText(Logger::PASS, "Reach point: (%f , %f)", reachPoint.x, reachPoint.y);
		double nearestDist = 1000;
		rcsc::PlayerObject* nearestDef = NULL;

		for (uint i = 0 ; i < wm->teammatesFromBall().size() ; i++)
		{
//			dlog.addText(Logger::PASS, "dist %d : %f", wm->teammatesFromBall()[i]->unum(), wm->teammatesFromBall()[i]->pos().dist(reachPoint));
			if (wm->teammatesFromBall()[i]->pos().dist(reachPoint) < nearestDist and isInPlayers(wm->teammatesFromBall()[i]->unum(), list))
			{
				nearestDist = wm->teammatesFromBall()[i]->pos().dist(reachPoint);
				nearestDef = wm->teammatesFromBall()[i];
			}
		}

		if (wm->self().pos().dist(reachPoint) <= nearestDist)
		{
			nearestDef = (rcsc::PlayerObject*)&wm->self();
			nearestDist = wm->self().pos().dist(reachPoint);
		}

		dlog.addText(Logger::ROLE, "fastest defender:\t%d", nearestDef->unum());

		return wm->self().unum() == nearestDef->unum();
	}

	bool isHeFastestDefender(rcsc::PlayerAgent * agent, string list, string blackList, int unum)
	{
		const rcsc::WorldModel* wm = &agent->world();

		if (isInPlayers(unum, blackList))
			return false;

		rcsc::Vector2D reachPoint = Basics::getFastestReachPoint(wm);
//		dlog.addText(Logger::PASS, "Reach point: (%f , %f)", reachPoint.x, reachPoint.y);
		double nearestDist = 1000;
		rcsc::PlayerObject* nearestDef = NULL;

		for (uint i = 0 ; i < wm->teammatesFromBall().size() ; i++)
		{
//			dlog.addText(Logger::PASS, "dist %d : %f", wm->teammatesFromBall()[i]->unum(), wm->teammatesFromBall()[i]->pos().dist(reachPoint));
			if (wm->teammatesFromBall()[i]->pos().dist(reachPoint) < nearestDist and isInPlayers(wm->teammatesFromBall()[i]->unum(), list))
			{
				nearestDist = wm->teammatesFromBall()[i]->pos().dist(reachPoint);
				nearestDef = wm->teammatesFromBall()[i];
			}
		}


		if (wm->self().pos().dist(reachPoint) <= nearestDist)
		{
			nearestDef = (rcsc::PlayerObject*)&wm->self();
			nearestDist = wm->self().pos().dist(reachPoint);
		}

		dlog.addText(Logger::ROLE, "fastest defender:\t%d", nearestDef->unum());

		return unum == nearestDef->unum();
	}

	bool amINearestDefender(rcsc::PlayerAgent * agent, string list, string blackList)
	{
		const rcsc::WorldModel* wm = &agent->world();

		if (isInPlayers(wm->self().unum(), blackList))
			return false;

		rcsc::Vector2D reachPoint = wm->ball().pos();
//		dlog.addText(Logger::PASS, "Reach point: (%f , %f)", reachPoint.x, reachPoint.y);
		double nearestDist = 1000;
		rcsc::PlayerObject* nearestDef = NULL;

		for (uint i = 0 ; i < wm->teammatesFromBall().size() ; i++)
		{
//			dlog.addText(Logger::PASS, "dist %d : %f", wm->teammatesFromBall()[i]->unum(), wm->teammatesFromBall()[i]->pos().dist(reachPoint));
			if (wm->teammatesFromBall()[i]->pos().dist(reachPoint) < nearestDist)
			{
				nearestDist = wm->teammatesFromBall()[i]->pos().dist(reachPoint);
				nearestDef = wm->teammatesFromBall()[i];
			}
		}

		if (wm->self().pos().dist(reachPoint) <= nearestDist)
		{
			nearestDef = (rcsc::PlayerObject*)&wm->self();
			nearestDist = wm->self().pos().dist(reachPoint);
		}

		dlog.addText(Logger::PASS, "fastest defender:\t%d", nearestDef->unum());

		return wm->self().unum() == nearestDef->unum();
	}

	rcsc::PlayerObject* fastestDefender(rcsc::PlayerAgent* agent, std::string list)
	{
		const rcsc::WorldModel* wm = &agent->world();

		rcsc::Vector2D reachPoint = Basics::getFastestReachPoint(wm);
//		dlog.addText(Logger::PASS, "Reach point: (%f , %f)", reachPoint.x, reachPoint.y);
		double nearestDist = 1000;
		rcsc::PlayerObject* nearestDef = NULL;

		for (uint i = 0 ; i < wm->teammatesFromBall().size() ; i++)
		{
//			dlog.addText(Logger::PASS, "dist %d : %f", wm->teammatesFromBall()[i]->unum(), wm->teammatesFromBall()[i]->pos().dist(reachPoint));
			if (wm->teammatesFromBall()[i]->pos().dist(reachPoint) < nearestDist and isInPlayers(wm->teammatesFromBall()[i]->unum(), list))
			{
				nearestDist = wm->teammatesFromBall()[i]->pos().dist(reachPoint);
				nearestDef = wm->teammatesFromBall()[i];
			}
		}

		if (wm->self().pos().dist(reachPoint) <= nearestDist and isInPlayers(wm->self().unum(), list))
		{
			nearestDef = (rcsc::PlayerObject*)&wm->self();
			nearestDist = wm->self().pos().dist(reachPoint);
		}

		dlog.addText(Logger::ROLE, "fastest defender:\t%d", nearestDef->unum());

		return nearestDef;
	}

	int getMarkTargetFromDefineMsg(std::string msg, int unum)
	{
		Expression expr = Expression(msg)[3];
		for (int i = 1 ; i < expr.GetChildrenCount() ; i++)
		{
			string tmmUnumStr = expr[i][2].AsString();
			int tmmUnum = Expression(tmmUnumStr.substr(1, tmmUnumStr.size() - 2)).AsInt();
			if (tmmUnum == unum)
				return expr[i][3][1].AsInt();
		}
		return 0;
	}

	bool isInField(rcsc::Vector2D pos)
	{
	    return (pos.absX() < 52.5 and pos.absY() < 34);
	}

	bool isInterceptSituation(rcsc::PlayerAgent* agent)
	{
		const rcsc::WorldModel* wm = &agent->world();

		int selfReachCycle = Intercept().finalIntercept(agent).getReachCycle();
		int oppReachCycle = wm->interceptTable()->opponentReachCycle();

		if (abs(oppReachCycle - selfReachCycle) > 2)
			Share::isInterceptSituation = oppReachCycle > selfReachCycle + 2;

		return Share::isInterceptSituation;
	}

	double defenseLineX(const rcsc::WorldModel* wm)
	{
		double lastX = 52.5;
		for (uint i = 0 ; i < wm->teammatesFromSelf().size() ; i++)
		{
			if (isInPlayers(wm->teammatesFromSelf()[i]->unum(), "34578") and
				wm->teammatesFromSelf()[i]->pos().x < lastX)
				lastX = wm->teammatesFromSelf()[i]->pos().x;
		}
		return lastX;
	}

	bool isInField(rcsc::Vector2D pos, double error)
	{
	    return (pos.absX() < (52.5 - error) and pos.absY() < (34.0 - error));
	}
}
