/*
 *  Copyright 2002-2005, Mersad Team, Allameh Helli High School (NODET).
 *
 *  This program is free software, you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Library General Public License for more details.
 *
 *  This file is created by: Ahmad Boorghany
 *  and is modified by: Mohammad Salehe, Sassan Haradji, ESKILAS Team
 *
 *  Released on Monday 1 August 2005, 10 Mordad 1384 by Mersad RoboCup Team.
 *  For more information please read README file.
*/

#ifndef __BASICS_H
#define __BASICS_H


#include <string>

#include <rcsc/geom/vector_2d.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/player_agent.h>


enum ReRateMode
{
	RRM_DIRECT,
	RRM_REVERSE
};

enum Decision
{
	SHOULD_DO_IT,
	CAN_DO_IT,
	SHOULD_NOT_DO_IT,
	CAN_NOT_DO_IT,
};

namespace Basics
{
	extern bool isInPlayers(const unsigned num, std::string);
	extern unsigned charToUniNum(const char str);
	extern char uniNumToChar(const unsigned num);
	extern unsigned firstPlayerInList(const std::string list,
			const std::string blackList);
	extern float reRate(float value, float a, float b, float c, float d,
			ReRateMode mode = RRM_DIRECT);
	extern float sigmoid(float x);
	extern unsigned changeIntBase(int number, int base, std::string &str);
	extern unsigned baghimande(int number, unsigned base);
	extern int sign(float number);
	extern unsigned pairPlayer(unsigned, unsigned = 0);
	extern std::string pairPlayers(std::string, unsigned = 0);

	rcsc::Vector2D getBallPosAfterCycles(const rcsc::Vector2D, float, float, float, float);
	rcsc::Vector2D getBallPosAfterCycles(const rcsc::BallObject&, float, float);
	rcsc::Vector2D getFastestReachPoint(const rcsc::WorldModel*);

	float getDeltaAngle(float, float);

	bool amIFastestDefender(rcsc::PlayerAgent* agent, std::string list, std::string blackList);
	bool isHeFastestDefender(rcsc::PlayerAgent* agent, std::string list, std::string blackList, int unum);
	bool amINearestDefender(rcsc::PlayerAgent* agent, std::string list, std::string blackList);
	rcsc::PlayerObject* fastestDefender(rcsc::PlayerAgent* agent, std::string list);

	int getMarkTargetFromDefineMsg(std::string msg, int unum);

	bool isInField(rcsc::Vector2D pos, double error = 0.0);

	bool isInterceptSituation(rcsc::PlayerAgent* agent);

	double defenseLineX(const rcsc::WorldModel* wm);
}

#endif // __CONFIG_H
