// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <Config.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cassert>
#include <stdlib.h>

using namespace std;

// Config CONSTRUCTOR
Config::Config():
	mLoadFlag(false)
{
}

// Config CONSTRUCTOR
Config::Config(string address):
	mLoadFlag(false)
{
	Load(address);
}

// Config DESTRUCTOR
Config::~Config()
{
	delete mConfigExpr;
}

// Config::Load ============
void Config::Load(string address)
{
	ifstream configFile;
	configFile.open(address.c_str(), ios::in);

	stringstream configStr("");

	string s;
	while (!configFile.eof())
	{
		configFile >> s;
		configStr << s << " ";
	}

	mConfigExpr = new Expression ("(" + configStr.str());
	mLoadFlag = true ;
}

// operator [] (string) =========
vector<Expression> Config::operator [] (string data)
{
	if (mLoadFlag)
		return (*mConfigExpr)[data];
	else
		perror("Config file has not been loaded : Config::operator [] ");
	exit(0);
}

// Config::Search ========
bool Config::Search(string data)
{
	return mConfigExpr->Search(data);
}

// Config::AsString =======
string Config::AsString() const
{
	return mConfigExpr->AsString();
}

// ConfigList CONSTRUCTOR
ConfigList::ConfigList():
	mLoadFlag(false)
{
}

// ConfigList CONSTRUCTOR
ConfigList::ConfigList(string address):
	mLoadFlag(false)
{
	Load(address);
}

// ConfigList DESTRUCTOR
ConfigList::~ConfigList()
{
}

// ConfigList::Instance =======
ConfigList* ConfigList::Instance()
{
	static ConfigList *configList=NULL;
	if (!configList)
		configList = new ConfigList ("./Configs/ConfigList.conf");
	return configList;
}

// ConfigList::Load ========
void ConfigList::Load(string address)
{
	Config list(address);

	WarnLoading(" Loading Config List: " + address);

	for (unsigned int i=0 ; i<list["Config"].size() ; i++)
	{
		string file = list["Config"][i]["File"][0][1].AsString() ;

		WarnLoading(" Loading Config File: " + file);

		mList[ list["Config"][i]["Topic"][0][1].AsString() ] = new Config ( list["Config"][i]["File"][0][1].AsString() ) ;
	}
	mLoadFlag = true ;
}

// ConfigList::LoadConfig =======
void ConfigList::LoadConfig(string topic, string address)
{
	mList[topic] = new Config (address);
}

// ConfigList::RemoveConfig ========
void ConfigList::RemoveConfig(string topic)
{
	delete mList[topic];
	mList.erase(topic);
}

// operator [] (string) =======
Config& ConfigList::operator [] (string topic)
{
	if (mLoadFlag)
		return *mList[topic] ;
	else
		perror("Config list has not been loaded : ConfigList::operator [] ");
	exit(0);
}

// ConfigList::WarnLoading ======
void ConfigList::WarnLoading(string str)
{
	system(("echo -n ''"+str).c_str());
	for (uint i=0 ; i<(70-str.size()) ; i++)
		system("echo -n ' '");
	system ("echo -n '['\\\\033[0\\;32m' OK '\\\\033[0\\;39m']\n'");
}
