// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <Expression.h>
#include <string>
#include <iostream>
#include <stdlib.h>

using namespace std;

// Expression CONSTRUCTOR
Expression::Expression (string data):
	Atomic(data)
{
	if (!CheckData(data))
	{
		cerr << data << endl ;
		perror ("Incorrect input string");
		exit(0);
	}
	else
	{
		bool flag = false;
		int pCloseCounter = 0;
		int pChk = 0;
		bool atomicChk = false;
		int startPosition = 0;

		for (uint i=0;i<data.size();i++)
		{
			if (data[i]==')') pCloseCounter++;
		}


		for (uint i=0;i<data.size();i++)
		{

			// Checking Atomic Start Position==========
			if (flag && !atomicChk && !pChk && data[i]!=' ' && data[i]!='\t' && data[i]!='\n' && data[i]!='(' && data[i]!=')')
			{
				atomicChk=true;
				startPosition=i;
			}

			// Checking Atomic End Position==========
			if (atomicChk && ( data[i]==' ' || data[i]=='\t' || data[i]=='\n' || data[i]=='(' || data[i]==')' ))
			{
				mChild.push_back( Expression (data.substr(startPosition,i-startPosition)));

				atomicChk=false;
			}

			// Checking (============
			if (flag && data[i]=='(') 
			{
				if (!pChk) startPosition=i;
				pChk++;
			}

			// Checking )=============
			if (pCloseCounter>1 && data[i]==')')
			{
				pChk--;
				pCloseCounter--;

				if (!pChk)
				{
					mChild.push_back( Expression(data.substr(startPosition,i-startPosition+1)));
				}
			}

			// Checking Main ( ==========
			if (data[i]=='(') flag=true;

		}
	}
}

// Expression DESTRUCTOR
Expression::~Expression()
{
//	for (int i=0;i<GetChildrenCount();i++)
//		delete mChild[i];
}

// Expression::Operator [int] ===============
Expression& Expression::operator [] (int index)
{
	if ( index<GetChildrenCount() )
		return mChild[index];
	else
	{
		perror("Index out of children range");
		exit(0);
	}
}

// Expression::Operator [string] SEARCHING FOR DATA =============
vector<Expression> Expression::operator [] (string data)
{
	vector<Expression> exprVec;

	for (int i=0;i<GetChildrenCount();i++)
		if (mChild[i].GetChildrenCount())
			if (mChild[i].NodeAt(0).AsString()==data) exprVec.push_back(mChild[i]);

	if (exprVec.size()) return exprVec;

	perror((data + " not found: Expression::operator[string] ").c_str());
	exit(0);
}

// Expression::Search ==========
bool Expression::Search(string data)
{
	for (int i=0;i<GetChildrenCount();i++)
		if (mChild[i].GetChildrenCount())
			if (mChild[i].NodeAt(0).AsString()==data) return true;

	return false;
}

// Expression::NodeAt ===============
Expression& Expression::NodeAt(int index)
{
	if ( index<GetChildrenCount() )
		return mChild[index];
	else
	{
		perror("Index out of children range");
		exit(0);
	}
}

// Expression::GetChildrenCount =========
int Expression::GetChildrenCount() const
{
	return mChild.size();
}

// Expression::CheckData ===========
bool Expression::CheckData(string data)
{
	int pChk=0;
	for (uint i=0;i<data.size();i++)
	{
		if (data[i]=='(') pChk++;
		if (data[i]==')') pChk--;
		if (pChk<0) return false;
	}

	if (pChk) return false;
	return true;
}

// Operator << (Expression)========================
ostream& operator << (ostream& str,const Expression& expr)
{
	str << expr.AsString();
	return str;
}
