// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <FastIC.h>

#include <rcsc/common/audio_memory.h>
#include <rcsc/common/player_type.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

#define MAXDASH 100
#define MaxFieldX 53
#define MaxFieldY 35

using namespace std;
using namespace rcsc;

bool FastIC::isSet = false;
float FastIC::dashDistance[18][151];
float FastIC::resultVelocity[18][151];
float FastIC::kickAbleArea[18];
float FastIC::decay[18];
turnModel FastIC::turnSimulator[18];

int FastIC::setCycle = -10;
const AbstractPlayerObject * FastIC::fastestTeamMate = NULL;
const AbstractPlayerObject * FastIC::fastestOpponent = NULL;
const AbstractPlayerObject * FastIC::globalFastest = NULL;
int FastIC::fastestTeamMateReachCycle = -10;
int FastIC::fastestOpponentReachCycle = -10;
int FastIC::globalFastestReachCycle = -10;
int FastIC::selfUnum = -1;
int FastIC::selfReachCycle = 1000;

int turnModel::getCyclesToTurn( float velocity , float angleChange )
{
	int velocityBlock = min ( (int)( velocity / ( maxVelocity / (float)turn_velCount ) ) , turn_velCount - 1);
	int angleBlock = min ( (int)( angleChange / ( maxAngleChange / (float)turn_angleCount ) ) , turn_angleCount - 1);
	return cyclesToTurn[velocityBlock][angleBlock];
}

FastIC::FastIC( bool _calculateBallDecay )
{
	calculateBallDecay = _calculateBallDecay;
	if ( not isSet )
	{
		isSet = true;
		for ( unsigned int i = 0; i < 18; i++ )
		{
			const PlayerType * thisPlayer = PlayerTypeSet::i().get(i);
			const float playerMaxSpeed = thisPlayer->playerSpeedMax();
			const float playerDecay = thisPlayer->playerDecay();
			const float playerDashPowerRate = thisPlayer->dashPowerRate();
			const float playerInertiaMoment = thisPlayer->inertiaMoment();
			// Determining dashDistances
			kickAbleArea[i] = thisPlayer->kickableMargin() + thisPlayer->playerSize() + ServerParam::i().ballSize();
			decay[i] = playerDecay;
			float dashDist = 0.f;
			float playerVel = 0.f;
			for ( unsigned int j = 0; j < 151; j++ )
			{
//				dlog.addText ( Logger::PASS , "%d-->%d %f" , i , j , dashDist );
				dashDistance[i][j] = dashDist;
				playerVel *= playerDecay;
				playerVel += MAXDASH * playerDashPowerRate;
				playerVel = min( playerVel , playerMaxSpeed);
				resultVelocity[i][j] = playerVel;
				dashDist += playerVel;
			}
			// Determining turnCycles
			float rateStep  = playerMaxSpeed / (float)turn_velCount;
			turnSimulator[i].maxVelocity = playerMaxSpeed;
			for ( unsigned int rateCount = 0; rateCount < turn_velCount; rateCount++ )
			{
				float thisRate = rateCount * rateStep;
				float angleStep = 180.f / (float)turn_angleCount;
				for ( unsigned int angleCount = 0; angleCount < turn_angleCount; angleCount++ )
				{
					float threshold = 5.f;
					float thisAngle = angleCount * angleStep;
					int turnCycles = 0;
					float calculatingRate = thisRate;
					while (true)
					{
						float turnAngle = thisAngle;
						if ( abs(turnAngle) < threshold )
							break;
						turnAngle = turnAngle * (1 + playerInertiaMoment * calculatingRate );
						if ( abs(turnAngle) > ServerParam::i().maxMoment() )
							turnAngle = AngleDeg( (turnAngle / abs(turnAngle) ) * ServerParam::i().maxMoment() ).degree();
						turnAngle = turnAngle / (1 + playerInertiaMoment * calculatingRate );
						calculatingRate *= playerDecay;
						thisAngle -= turnAngle;
						turnCycles++;
					}
					turnSimulator[i].cyclesToTurn[rateCount][angleCount] = turnCycles;
				}
			}
		}
		isSet = true;
	}
	mPlayers.clear();
	mFastestPlayers.clear();
}

void FastIC::calculate(int myUnum , int cyclesToPostCalculate , bool startLater , bool isItForReal)
{
//	dlog.addText ( Logger::PASS , "mPlayers.size() = %d" , mPlayers.size() );
	mFastestPlayers.clear();
	int firstReachCycle = 1000;
	bool shouldContinue = true;
	int cyclesElpased = 0;
	vector<FastPlayer> tmpPlayers = mPlayers;
	for ( unsigned int i = 0; i < tmpPlayers.size(); i++ )
	{
		tmpPlayers[i].controlBuff *= kickAbleArea[tmpPlayers[i].playerType];
		tmpPlayers[i].isSelected = false;
	}
	Vector2D tmpBallPos = ballPos;
	Vector2D tmpBallVel = ballVel;
	bool ballMoved = isItForReal;
	if ( startLater )
	{
		cyclesElpased = ballDelay + 1;
		tmpBallPos += tmpBallVel;
		tmpBallVel *= ServerParam::i().ballDecay();
		ballMoved = true;
	}
	while ( shouldContinue )
	{
		for ( unsigned int i = 0; i < tmpPlayers.size(); i++ )
		{
			int thisPlayerType = tmpPlayers[i].playerType;
			if ( tmpPlayers[i].isSelected )
				continue;
			int cyclesMoved = cyclesElpased - tmpPlayers[i].delay;
			if ( cyclesMoved < 0 )
				cyclesMoved = 0;
			int cyclesDashed = cyclesMoved;
			if ( tmpPlayers[i].shouldConsiderFace )
			{
				float hisStartingVel = mPlayers[i].mVel.r();
				float targetDir = ( tmpBallPos - mPlayers[i].mPos ).th().degree();
				float turnAngle = fabsf( targetDir - tmpPlayers[i].playerBodyDir );
				while ( turnAngle > 180.f )
					turnAngle -= 360.f;
				turnAngle = fabsf( turnAngle );
				if ( turnAngle > 90.f )
					turnAngle = 180.f - turnAngle;
				if ( ( ballVel.r() > 2.8 ) and ( ballVel.r() < 3.0 ) )
				if ( ( ballVel.th().degree() > -56.f ) and ( ballVel.th().degree() < -55.f ) )
				if ( ( not tmpPlayers[i].isTeamMate ) and ( tmpPlayers[i].unum == 1 ) )
				{
					dlog.addText ( Logger::PASS , "turnAngle\t%f" , turnAngle );
					dlog.addText ( Logger::PASS , "vel\t%f" , hisStartingVel );
				}
				if ( turnAngle > 20.f )
					cyclesDashed -= turnSimulator[thisPlayerType].getCyclesToTurn( hisStartingVel , turnAngle );
			}
			if ( cyclesDashed < 0 )
				cyclesDashed = 0;
			float distToCatch = dashDistance[thisPlayerType][cyclesDashed] * tmpPlayers[i].dashMultiplier;
			float controlBuff = tmpPlayers[i].controlBuff;
			if ( ( tmpBallPos.x > 36.f ) and ( fabsf( tmpBallPos.y ) < 20.f ) )
			{
				if ( tmpPlayers[i].penaltyAreaBuff > 0.f )
					controlBuff = tmpPlayers[i].penaltyAreaBuff * tmpPlayers[i].controlBuff;
			}
			distToCatch += controlBuff;
			float distToBall = tmpBallPos.dist( tmpPlayers[i].mPos );
//			dlog.addText ( Logger::PASS , "direction\t%f" , ballVel.th().degree() );
			if ( ( ballVel.r() > 2.1 ) and ( ballVel.r() < 2.2 ) )
			if ( ( ballVel.th().degree() > 26.f ) and ( ballVel.th().degree() < 28.f ) )
			if ( ( not tmpPlayers[i].isTeamMate ) and ( tmpPlayers[i].unum == 5 ) )
			{
				dlog.addText ( Logger::PASS , "-----------------------------");
				dlog.addText ( Logger::PASS , "cyclesElpased\t%d" , cyclesElpased );
				dlog.addText ( Logger::PASS , "cyclesMoved\t%d" , cyclesDashed );
				dlog.addText ( Logger::PASS , "ballPos(%.2f,%.2f)" , tmpBallPos.x , tmpBallPos.y );
				dlog.addText ( Logger::PASS , "ballVel(%.2f,%.2f)" , tmpBallVel.x , tmpBallVel.y );
				dlog.addText ( Logger::PASS , "myPos(%.2f,%.2f)" , tmpPlayers[i].mPos.x , tmpPlayers[i].mPos.y );
//				dlog.addText ( Logger::PASS , "ballMoved\t%d" , ballMoved );
				dlog.addText ( Logger::PASS , "MyMovement	%f" , distToCatch - tmpPlayers[i].controlBuff );
				dlog.addText ( Logger::PASS , "distToCatch %f" , distToCatch );
				dlog.addText ( Logger::PASS , "neededDist %f" , distToBall );
				dlog.addText ( Logger::PASS , "-----------------------------");
			}
			bool canCatch = true;
/*			if ( ( tmpPlayers[i].unum == myUnum ) and ( tmpPlayers[i].isTeamMate ) and ( cyclesDashed == 0 ) )
				canCatch = false;
*/			if ( ( distToBall <= distToCatch ) and ( cyclesElpased >= ballDelay ) and ( ballMoved ) and ( canCatch ) )
			{
				mFastestPlayers.push_back( FastestPlayer( tmpPlayers[i].unum , tmpPlayers[i].isTeamMate , cyclesElpased , tmpBallPos , tmpPlayers[i].playerItCameFrom , tmpPlayers[i].playerPtr ) );
				tmpPlayers[i].isSelected = true;
//				dlog.addText ( Logger::PASS , "cycle %d elpased--->Terminated!!" , cyclesElpased );
				if ( firstReachCycle == 1000 )
					firstReachCycle = cyclesElpased;
/*			if ( ( ballVel.r() > 2.1 ) and ( ballVel.r() < 2.2 ) )
			if ( ( ballVel.th().degree() > 26.f ) and ( ballVel.th().degree() < 28.f ) )
				dlog.addText ( Logger::PASS , "player %d--->%d reaches in %d cycles" , tmpPlayers[i].isTeamMate , tmpPlayers[i].unum , cyclesElpased );
*/			}
			// Simulating Player
			tmpPlayers[i].mPos += tmpPlayers[i].mVel;
			tmpPlayers[i].mVel *= decay[ thisPlayerType ];
		}
		// Simulating Ball
		if ( cyclesElpased >= ballDelay )
		{
			if ( calculateBallDecay )
				tmpBallVel *= ServerParam::i().ballDecay();
			tmpBallPos += tmpBallVel;
		}
		ballMoved = true;
		if ( cyclesElpased >= firstReachCycle + cyclesToPostCalculate )
			shouldContinue = false;
		if ( cyclesElpased >= 150 )
			shouldContinue = false;
		if ( ( abs( tmpBallPos.x ) >= MaxFieldX ) or ( abs( tmpBallPos.y ) >= MaxFieldY ) )
			shouldContinue = false;
		cyclesElpased++;
	}
}

void FastIC::setByWorldModel( const rcsc::WorldModel & wm )
{
	selfUnum = wm.self().unum();
	ballPos = wm.ball().pos();
	ballVel = wm.ball().vel();
//	dlog.addText ( Logger::PASS , "ballPos ( %.2f , %.2f )" , ballPos.x , ballPos.y );
//	dlog.addText ( Logger::PASS , "ballVel ( %.2f , %.2f )" , ballVel.x , ballVel.y );
	ballDelay = 0;
	if ( wm.time().cycle() == setCycle )
		return;
 	fastestOpponent = fastestTeamMate = globalFastest = NULL;
	setCycle = wm.time().cycle();
	//Adding teamMates
	{
		const PlayerPtrCont::const_iterator end = wm.teammatesFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.teammatesFromSelf().begin();
			it != end;
			++it )
		{
//			dlog.addText ( Logger::PASS , "tmm %d count %d" , (*it)->unum() , (*it)->posCount() );
			if ( (*it)->posCount() > 5 )
				continue;
			// -----------------------------
			if ((*it)->unum() == wm.teammateGoalieUnum())
				continue;
			// -----------------------------

			if ( (*it)->isKickable() )
			{
				// TODO : Tarjih ba kodume??
				globalFastest = fastestTeamMate = (*it);
				globalFastestReachCycle = fastestTeamMateReachCycle = 0;
//				dlog.addText ( Logger::PASS , "Teammates skipped due to kickable" );
				break;
			}	
			Vector2D hisVel = Vector2D( 0.f , 0.f );
			if ( (*it)->velValid() )
				hisVel = (*it)->vel();
			mPlayers.push_back(	FastPlayer( (*it)->pos() , hisVel , 0 , (*it)->unum() , true , 1 , 1 , (*it)->playerTypePtr()->id() , NULL , (*it) ) );
		}
	}
	//Adding self
		if ( wm.self().isKickable() )
		{
			globalFastest = fastestTeamMate = &(wm.self());
			globalFastestReachCycle = fastestTeamMateReachCycle = 0;	
//			dlog.addText ( Logger::PASS , "Self skipped due to kickable" );
		}
		else
			mPlayers.push_back(	FastPlayer( wm.self().pos() , wm.self().vel() , 0 , wm.self().unum() , true , 1 , 1 , wm.self().playerType().id() , NULL , &( wm.self() ) ) );
	//Adding opponents
	{
		const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();
		for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
			it != end;
			++it )
		{
//			dlog.addText ( Logger::PASS , "opp %d count %d in ( %.2f , %.2f )" , (*it)->unum() , (*it)->posCount() , (*it)->pos().x , (*it)->pos().y );
			if ( (*it)->posCount() > 5 )
				continue;
			if ( (*it)->isKickable() )
			{
				fastestOpponent = (*it);
				fastestOpponentReachCycle = 0;
				if ( not globalFastest )
				{
					globalFastest = (*it);
					globalFastestReachCycle = 0;
				}
//				dlog.addText ( Logger::PASS , "Opponents skipped due to kickable" );
				break;
			}	
			Vector2D hisVel = Vector2D( 0.f , 0.f );
			if ( (*it)->velValid() )
				hisVel = (*it)->vel();
			mPlayers.push_back(	FastPlayer( (*it)->pos() , hisVel , 0 , (*it)->unum() , false , 1 , 1 , (*it)->playerTypePtr()->id() , NULL , (*it) ) );
		}
	}
	calculate( wm.self().unum() , 150 , false , true);
	unsigned int courseReceiverPlace = 10000;
/*	if ( not wm.audioMemory().helliCourse().empty() )
		if ( not wm.audioMemory().helliCourse().front().isReceived )
		{
			int receiverUnum = wm.audioMemory().helliCourse().front().receiver_;
			for ( unsigned int i = 0; i < mFastestPlayers.size(); i++ )
			{
				if ( not mFastestPlayers[i].isTeamMate )
					continue;
				if ( mFastestPlayers[i].unum != receiverUnum )
					continue;
				courseReceiverPlace = i;
				break;
			}
		}
*/	for ( unsigned int player = 0; player < mFastestPlayers.size(); player++ )
	{
		if ( not globalFastest )
		{
			if ( ( not mFastestPlayers[player].isTeamMate ) or ( courseReceiverPlace >= 1000 ) or ( courseReceiverPlace == player ) )
			{
				globalFastest = mFastestPlayers[player].playerPtr;
				globalFastestReachCycle = mFastestPlayers[player].reachCycle;
			}
		}
		if ( ( not fastestTeamMate ) and ( mFastestPlayers[player].isTeamMate ) )
		{
			if ( ( courseReceiverPlace >= 1000 ) or ( courseReceiverPlace == player ) )
			{
				fastestTeamMate = mFastestPlayers[player].playerPtr;
				fastestTeamMateReachCycle = mFastestPlayers[player].reachCycle;
			}
		}
		if ( ( not fastestOpponent ) and ( not mFastestPlayers[player].isTeamMate ) )
		{
			fastestOpponent = mFastestPlayers[player].playerPtr;
			fastestOpponentReachCycle = mFastestPlayers[player].reachCycle;
		}
		if ( ( mFastestPlayers[player].isTeamMate ) and ( mFastestPlayers[player].unum == wm.self().unum() ) )
			selfReachCycle = mFastestPlayers[player].reachCycle;
	}
}

float FastIC::getResultVel( int playerType , int cycle )
{
	if ( playerType < 0 )
		playerType = 0;
	if ( cycle < 0 )
		cycle = 0;
	if ( cycle > 150 )
		cycle = 150;
	if ( playerType > 17 )
		playerType = 17;
	return resultVelocity[playerType][cycle];
}

float FastIC::getMovingDist( int playerType , int cycle )
{
	if ( playerType < 0 )
		playerType = 0;
	if ( cycle < 0 )
		cycle = 0;
	if ( cycle > 150 )
		cycle = 150;
	if ( playerType > 17 )
		playerType = 17;
	return dashDistance[playerType][cycle];
}

