// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef __FAST_IC_H_
#define __FAST_IC_H_

#include <rcsc/geom/vector_2d.h>
#include <rcsc/player/world_model.h>

#include <rcsc/common/server_param.h>
#include <rcsc/common/player_type.h>
#include <vector>

#include <rcsc/common/logger.h>

#define turn_velCount 10
#define turn_angleCount 18

struct PreDecidedPassPlayer;

struct turnModel
{
	int cyclesToTurn[turn_velCount][turn_angleCount];
	float maxVelocity;
	float maxAngleChange;

	int getCyclesToTurn( float velocity , float angleChange );

	turnModel()
	{
		maxVelocity = 1.f;
		maxAngleChange = 180.f;
	}
};

struct FastPlayer
{
	rcsc::Vector2D mPos;
	rcsc::Vector2D mVel;
	int delay;
	int unum;
	bool isTeamMate;
	float controlBuff;
	float dashMultiplier;
	float penaltyAreaBuff;
	int playerType;
	bool isSelected;
	PreDecidedPassPlayer* playerItCameFrom;
	const rcsc::AbstractPlayerObject* playerPtr;
	const rcsc::PlayerObject* playerObjectPtr;
	bool shouldConsiderFace;
	float playerBodyDir;

	FastPlayer( rcsc::Vector2D _mPos , rcsc::Vector2D _mVel , int _delay , int _unum , bool _isTeamMate , float _controlBuff , float _dashMultiplier 
			, int _playerType , PreDecidedPassPlayer* _playerItCameFrom, const rcsc::AbstractPlayerObject* _playerPtr = NULL)
	{
		mPos = _mPos;
		mVel = _mVel;
		delay = _delay * 2;
		unum = _unum;
		isTeamMate = _isTeamMate;
		controlBuff = _controlBuff * 0.8;
		penaltyAreaBuff = -1.f;
		dashMultiplier = _dashMultiplier;
		if ( ( _playerType > 17 ) or ( _playerType < 0 ) )
			playerType = 0;
		else
			playerType = _playerType;
		isSelected = false;
		shouldConsiderFace = false;
		playerBodyDir = 0.f;
		playerItCameFrom = _playerItCameFrom;
		playerPtr = _playerPtr;
	}
	// Mona should write a constructor here with ( delay , controlBuff , dashMultiplier , AbstractPlayerObject ); I am a bit "shad" on this one!
	FastPlayer( int _delay , bool _isTeamMate, float _controlBuff , float _dashMultiplier ,const rcsc::PlayerObject* _playerPtr )
	{
	    if (_playerPtr)
		mPos = _playerPtr->pos();
	    rcsc::Vector2D speed(0, 0);
	    if (_playerPtr and _playerPtr->velValid())
			speed = _playerPtr->vel();
	    mVel = speed;
	    delay = _delay;
	    if (_playerPtr)
		unum = _playerPtr->unum();
	    isTeamMate = _isTeamMate;
	    controlBuff = _controlBuff;
		penaltyAreaBuff = -1.f;
		if ( _playerPtr )
			if ( _playerPtr->goalie() )
				penaltyAreaBuff = rcsc::ServerParam::i().catchableArea();;
	    dashMultiplier = _dashMultiplier;
	    int _playerType = -1;
	    if (_playerPtr)
	    _playerType = _playerPtr->type();
	    if ( ( _playerType > 17 ) or ( _playerType < 0 ) )
			playerType = 0;
	    else
			playerType = _playerType;
	    isSelected = false;
	    shouldConsiderFace = false;
	    playerBodyDir = 0.f;
	    if (_playerPtr and _playerPtr->bodyCount() == 0 )
	    {
	    	rcsc::dlog.addText ( rcsc::Logger::PASS , "playerBody\t%f" , _playerPtr->body().degree() );
	    	shouldConsiderFace = true;
	    	playerBodyDir = _playerPtr->body().degree();
	    }
	    playerItCameFrom = NULL;
	    playerPtr = playerObjectPtr = _playerPtr;
	}

};


struct FastestPlayer
{
	int unum;
	bool isTeamMate;
	int reachCycle;
	rcsc::Vector2D receivePoint;
	PreDecidedPassPlayer* playerItCameFrom;
	const rcsc::AbstractPlayerObject* playerPtr;

	FastestPlayer( int _unum , bool _isTeamMate , int _reachCycle , rcsc::Vector2D _receivePoint , PreDecidedPassPlayer* _playerItCameFrom,
			const rcsc::AbstractPlayerObject* _playerPtr = NULL )
	{
		unum = _unum;
		isTeamMate = _isTeamMate;
		reachCycle = _reachCycle;
		receivePoint = _receivePoint;
		playerItCameFrom = _playerItCameFrom;
		playerPtr = _playerPtr;
	}
};

class FastIC
{
private:
	static bool isSet;
	static float dashDistance[18][151];
	static float resultVelocity[18][151];
	static float kickAbleArea[18];
	static float decay[18];
	static turnModel turnSimulator[18];

	rcsc::Vector2D ballPos;
	rcsc::Vector2D ballVel;
	int ballDelay;

	std::vector<FastPlayer> mPlayers;
	std::vector<FastestPlayer> mFastestPlayers;

	bool calculateBallDecay;

	static int setCycle;
	static const rcsc::AbstractPlayerObject * fastestTeamMate;
	static const rcsc::AbstractPlayerObject * fastestOpponent;
	static const rcsc::AbstractPlayerObject * globalFastest;
	static int fastestTeamMateReachCycle;
	static int fastestOpponentReachCycle;
	static int globalFastestReachCycle;
	static int selfUnum;
	static int selfReachCycle;


public:
	FastIC( bool _calculateBallDecay = true);
	void reconstruct()
	{
		mPlayers.clear();
		mFastestPlayers.clear();
	}

	void calculate(int myUnum , int cyclesToPostCalculate = 0 , bool startLater = false , bool isItForReal = false);

	std::vector<FastPlayer> & players()
	{
		return mPlayers;
	}

	const std::vector<FastestPlayer> & getFastestPlayers()
	{
		return mFastestPlayers;
	}

	void setBall(rcsc::Vector2D _ballPos , rcsc::Vector2D _ballVel, int _ballDelay )
	{
		ballPos = _ballPos;
		ballVel = _ballVel;
		ballDelay = _ballDelay;
	}

	void setCalculateBallDecay( bool calculate )
	{
		calculateBallDecay = calculate;
	}

	void resetPlayers()
	{
		mPlayers.clear();
	}

	void setByWorldModel( const rcsc::WorldModel & wm );
	float getResultVel( int playerType , int cycle );
	float getMovingDist( int playerType , int cycle );

	static const rcsc::AbstractPlayerObject* getFastestTeammate()
	{
		return fastestTeamMate;
	}

	static const rcsc::AbstractPlayerObject* getFastestOpponent()
	{
		return fastestOpponent;
	}

	static const rcsc::AbstractPlayerObject* getGlobalFastest()
	{
		return globalFastest;
	}
	
	static int getFastestTeammateReachCycle()
	{
		return fastestTeamMateReachCycle;
	}

	static int getFastestOpponentReachCycle()
	{
		return fastestOpponentReachCycle;
	}

	static int getGlobalFastestReachCycle()
	{
		return globalFastestReachCycle;
	}

	static bool isSelfFastestPlayer()
	{
		if ( globalFastest )
			return ( globalFastest->isSelf() );
		return false;
	}

	static bool isSelfFastestTeamMate()
	{
		if ( fastestTeamMate )
			return ( fastestTeamMate->unum() == selfUnum );
		return false;
	}

	static bool isOurTeamBallPossessor()
	{
		if ( ( fastestTeamMate ) and ( globalFastest ) )
			return ( fastestTeamMate == globalFastest );
		return false;
	}

	static int getSelfReachCycle()
	{
		return selfReachCycle;
	}

	static float getDashDistance(int type, int cycle)
	{
		return dashDistance[type][cycle];
	}

};

#endif // __FAST_IC_H_
