// -*-c++-*-

/*
 *Copyright:

 Copyright (C) ESKILAS TEAM

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#include <WMatching.h>

#include <iostream>
#include <cstring>

#include <cmath>

using namespace std;

WMatching::WMatching(int _n, vector< vector<int> > wvec):
	n(_n)
{
	for (uint i = 0 ; i < wvec.size() ; i++)
		for (uint j = 0 ; j < wvec.size() ; j++)
			w[i][j] = wvec[i][j];
}

void WMatching::read()
{
	memset(u, 0, sizeof u);
	memset(v, 0, sizeof v);
	for (int i = 0 ; i < n ; ++i)
	{
		u[i]=-inf;
		for (int j = 0 ; j < n ; ++j)
			u[i]=max(u[i],w[i][j]);
	}
}

bool WMatching::dfs(int vv)
{
	mark[vv]=true;
	for (int i = 0 ; i < n ; ++i)
		if (g[vv][i] && (match[i]==-1 || !mark[match[i]] && dfs(match[i])))
			return (match[i]=vv),1;
	return false;
}

bool WMatching::find()
{
	memset(mark, 0, sizeof(mark));
	for (int i = 0 ; i < n ; ++i)
		if (!mark[i] && !matched[i] && dfs(i))
			return (matched[i]=true);
	return false;
}

void WMatching::increase_matching()
{
	for (int l = 0 ; l < n ; ++l)
	{
		for (int i = 0 ; i < n ; ++i)
			for (int j = 0 ; j < n ; ++j)
				g[i][j]=(u[i] + v[j] == w[i][j]);
		if (find())
			return ;
		int eps=inf;
		for (int i = 0 ; i < n ; ++i)
			for (int j = 0 ; j < n ; ++j)
				if (mark[i] && (match[j]==-1 || !mark[match[j]]))
					eps=min(eps, u[i] + v[j] - w[i][j]);
		for (int i = 0 ; i < n ; ++i)
		{
			if (mark[i])
				u[i]-=eps;
			if (match[i]!=-1 && mark[match[i]])
				v[i]+=eps;
		}
	}
}

void WMatching::calculate()
{
	read();
	memset(match,-1,sizeof(match));
	memset(matched,0,sizeof matched);
	for (int i = 0 ; i < n ; ++i)
		increase_matching();


//	for (int i = 0 ; i < n ; ++i)
//		cout << match[i] << ' ';
//	cout << endl;
}
