#!/bin/bash

make

LIBPATH=/usr/local/lib
if [ x"$LIBPATH" != x ]; then
  if [ x"$LD_LIBRARY_PATH" = x ]; then
    LD_LIBRARY_PATH=$LIBPATH
  else
    LD_LIBRARY_PATH=$LIBPATH:$LD_LIBRARY_PATH
  fi
  export LD_LIBRARY_PATH
fi

DIR=`dirname $0`

player="${DIR}/EskilasPlayer"
teamname="ESKILAS"
host="localhost"

config="${DIR}/Configs/Player.conf"
config_dir="${DIR}/Formations"

coach_config="${DIR}/Configs/Coach.conf"

number=1
usecoach="false"

sleepprog=sleep
goaliesleep=1
sleeptime=0

debugopt=""
coachdebug=""

usage()
{
  (echo "Usage: $0 [options]"
   echo "Possible options are:"
   echo "      --help                print this"
   echo "  -h, --host HOST           specifies server host"
   echo "  -t, --teamname TEAMNAME   specifies team name"
   echo "  -n, --number NUMBER       specifies the number of players"
   echo "  -c, --with-coach          specifies to run the coach"
   echo "  -C, --without-coach       specifies not to run the coach"
   echo "  --uva                     use UvA type formation"
   echo "  --bpn                     use BPN formation"
   echo "  --debug                   create debug log"
   echo "  --debug-coach             write debug log (coach)"
   echo "  --debug-connect           connect to Soccer Viewer"
   echo "  --debug-write             create Soccer Viewer log") 1>&2
}

while [ $# -gt 0 ]
	do
	case $1 in

    --help)
      usage
      exit 0
      ;;

    -h|--host)
      if [ $# -lt 2 ]; then
        usage
        exit 1
      fi
      host=$2
      shift 1
      ;;

    -t|--teamname)
      if [ $# -lt 2 ]; then
        usage
        exit 1
      fi
      teamname=$2
      shift 1
      ;;

    -n|--number)
      if [ $# -lt 2 ]; then
        usage
        exit 1
      fi
      number=$2
      shift 1
      ;;

    -c|--with-coach)
      usecoach="true"
      ;;

    -C|--without-coach)
      usecoach="false"
      ;;

    --uva)
      config_dir="${DIR}/formations-uva"
      ;;

    --bpn)
      config_dir="${DIR}/formations-bpn"
      ;;

    --debug)
      debugopt="${debugopt} --debug"
      ;;

    --debug-coach)
      coachdebug="${coachdebug} --debug"
      ;;

    --debug-connect)
      debugopt="${debugopt} --debug_connect"
      ;;

    --debug-write)
      debugopt="${debugopt} --debug_write"
      ;;

    *)
      usage
      exit 1
      ;;
  esac

  shift 1
done

opt="--player-config ${config} --config_dir ${config_dir} --offline"

opt="${opt} -h ${host} -t ${teamname}"
opt="${opt} ${debugopt}"

if [ $number -gt 0 ]; then
  $player ${opt} -g &
  $sleepprog $goaliesleep
fi

i=2
while [ $i -le ${number} ] ; do
  $player ${opt} &
  $sleepprog $sleeptime
  i=`expr $i + 1`
done

if [ "${usecoach}" = "true" ]; then
  coachopt="--coach-config ${coach_config}"
  coachopt="${coachopt} -h ${host} -t ${teamname}"
  coachopt="${coachopt} ${coachdebug}"
  $coach  ${coachopt} &
fi
